/*
 * ReadTopologyFileAction.java
 *
 * Created on December 20, 2006, 8:09 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package GUIActions;

import Utility.JNWBitSet ;


/**ReadEMailTextFileAndPutIntoSourceNodeBufferAction reads several
 *emails from a text file.  It converts the e-mail String to a
 * JNWBitSet.  It then addes either the UDP or TCP header to
 * the data.  It then adds the IP header to the UDP packet or TCP packets.
 *<p>It then puts the JNWBitSet in the buffer of the source node.  It
 * then adds a RoutePacketAtNodeOrProcessAtDestination Event
 * to the DiscreteEventSimulator.
 *<p>One can read several e-mails into the model before
 * starting the simulation.  Each of the e-mails could have different
 * sources or different destinations.  The model starts running
 * when the RunModelAction is executed.
 *
 * @version 1.202
 * @author dixonw
 * @modified BZ May 2010 for JTRN3
 * @modified JMP Jan 2012 for JDLC3 HDLC
 */
public class ReadEMailTextFilesAndPutIntoSourceNodeBufferAction
   extends MainGUIActionsBaseClass implements GUIConstants.General
{
   /** The one and only source node for TCP email
    *  (JNW TCP does not support multiple connections;
    *  the way we limit email to a single connection
    *  is to send all from one source a serialize) */
   public static int tcpEmailSourceNode = -1;

   /** File chooser to find file*/
   javax.swing.JFileChooser jFileChooser = null ;

   /**The input file defininng the network topology*/
   java.io.File eMailFile = null ;

   /** File read to read individual lines*/
   java.io.BufferedReader fileReader = null ;

   /**The electronic mail to be sent. This is a StringBuilder*/
   java.lang.StringBuilder inputEMailAsStringBuilder = null ;

   /**A singleton vector containing all the Nodes*/
   static ModelData.JNWSingletonVectorOfNodes<ModelPhysical.Node> vectorOfNodes
      = ModelData.JNWSingletonVectorOfNodes.getInstance (  ) ;

   /**A singleton vector containing all the links*/
   ModelData.JNWSingletonVectorOfLinks<ModelPhysical.Link> vectorOfLinks
      = ModelData.JNWSingletonVectorOfLinks.getInstance (  ) ;

   /**UDP packet to be put into node*/
   Utility.JNWBitSet UDPPacket ;

   /**Individual lines in the e-mail file.  inputString cannot be
    * trimmed since it is the individual lines of the email messages, and
    * it may contain carriage control characters.*/
   String inputString = null ;

   /**Individual lines in the e-mail file trimmed to remove leading
    * and trailing white space.  This string must be trimmed since it
    * may represent the TO, FROM, and end of message.*/
   String trimmedInputString = null ;

   /**The source node for the e-mail*/
   int sourceNodeNumber = MINUS_ONE ;

   /**The destination node for the e-mail*/
   int destinationNodeNumber = MINUS_ONE ;

   /**The transmission protocol.  Presently this is limited
    *to TCP and UDP.*/
   public static ModelData.GlobalVariables.transmissionProtocols
      transmissionProtocol ;

   /**Prefix for the source node*/
   String sourceNodePrefix = "FROM:" ;

   /**Prefix for the destination node*/
   String destinationNodePrefix = "TO:" ;

   /**Termination String for e-mail.*/
   String termination = "MESSSAGE-END:" ;

   /**Prefix for the protocol*/
   String protocolPrefix = "PROTOCOL:" ;

   /**The JNW random number generator.*/
   static Utility.JNWRandom randomNumberGernerator
      = Utility.JNWRandom .getInstance ( ) ;

   /** Constructor calls AssignToAction ( ) to add the
    * accelerator key, the long description, the short description, the
    * mnemonic key, and the small icon to the action.*/
   public ReadEMailTextFilesAndPutIntoSourceNodeBufferAction ()
   {  // Call AssignToAction to add all the attributes to the action
      AssignToAction (
         ReadEMailFilesAction_ACCELERATOR_KEY
         ,ReadEMailFilesAction_LONG_DESCRIPTION
         ,ReadEMailFilesAction_MNEMONIC_KEY
         ,ReadEMailFilesAction_NAME
         ,ReadEMailFilesAction_SHORT_DESCRIPTION
         ,ReadEMailFilesAction_SMALL_ICON) ;
   }  // end of public ReadEMailTextFilesAndPutIntoSourceNodeBufferAction


   /** When the toolbar button is clicked on the menubar, reads the email file.  
    * It converts the e-mail String to a JNWBitSet by first adding an IP
    * or TCP headers to the message JNWBitSet and then an IP header to that
    * result.  It then puts the JNWBitSet in the buffer to the source node.  It
    * then addes a StartEMail Event to the DiscreteEventSimulator.*/
   public void actionPerformed (java.awt.event.ActionEvent event)
   {
      if ( ModelData.JNWSingletonVectorOfNodes.getInstance ( ) . size ( )  == 0  )
      {
         Utility.err.println  ( "ERROR IN GUIActions.ReadEMailText"
            + "FilesAndPutIntoSourceNodeBufferAction:"
            + "\n The vectorOfNodes is empty indicating that you have not "
            + "yet read in the topology.\n Read the topology file first. " ) ;
         return ;
      }
      // Find the topology input file
      jFileChooser = new javax.swing.JFileChooser ( fileNameForJFileChooser  ) ;
      jFileChooser . setDialogTitle ( "READ E-MAIL FILES" ) ;
      try
      {
         int returnVal = jFileChooser .showOpenDialog ( null );
         if (returnVal == jFileChooser . APPROVE_OPTION)
         {
            eMailFile = jFileChooser . getSelectedFile ( );
         } // end of if (returnVal == jFileChooser . APPROVE_OPTION)
         fileReader = new java.io.BufferedReader (
            new java.io.FileReader ( eMailFile ) ) ;
         // update the fileNameForJFileChooser for later access to the same
         // folder
         fileNameForJFileChooser  = jFileChooser . getSelectedFile( ) ;
         inputEMailAsStringBuilder = new java.lang.StringBuilder ( ) ;
         // read the lines until the end-of-file is found.
         // read the file line by line and remove the white space before 
         // and after the characters
         readData: while ( fileReader . ready ( ) )
         {  // read one line of input
            inputString = fileReader . readLine ( )  ;
            // readLine reads one line of input and removes the carrage control
            // Restore the carraige return that the readLine() removed
            inputString = inputString + CARRIAGE_RETURN ;   
            trimmedInputString = inputString . trim ( ) ;
            // If it is a comment skip or the string is empty, skip it.
            if ( ! trimmedInputString . startsWith ( COMMENT_DELIMITER  ) &
               ! trimmedInputString .equals ( EMPTY_STRING ) )
            {  // Is this a destination node number?
               if ( trimmedInputString  . startsWith ( destinationNodePrefix ))
               {  // Set the destination node number
                  destinationNodeNumber = FindColonAndConvertStringToInt ( trimmedInputString ) ;
               }  // end of if ( inputStringTrimmed  . startsWith ( destinationNodePrefix ))
               else if ( trimmedInputString  . startsWith ( sourceNodePrefix ) )
               {  // Set the source node number
                  sourceNodeNumber = FindColonAndConvertStringToInt ( trimmedInputString ) ;
               }  // end of else if ( inputStringTrimmed  . startsWith ( sourceNodePrefix ) )
               else if ( trimmedInputString  . startsWith ( protocolPrefix ) )
               {  // Set the transmission protocol
                  // Get the String after the ':'
                  String protocolString = FindColonAndConvertRemainderToTrimmedSubString ( 
                     trimmedInputString ) ;
                  transmissionProtocol = GetProtocol ( protocolString) ;
               }  // end of else if ( inputStringTrimmed  . startsWith ( sourceNodePrefix ) )
               else if ( trimmedInputString  . startsWith ( termination ) )
               {  // The messsages is complete so process it
                  int numberOfNodes
                     = ModelData.JNWSingletonVectorOfNodes.getInstance (  ) . size ( ) ;
                  if ( tcpEmailSourceNode < 0)tcpEmailSourceNode = sourceNodeNumber;
                  if(transmissionProtocol.equals(ModelData.GlobalVariables.
                         transmissionProtocols.TCP))
                  {
                      if ( sourceNodeNumber != tcpEmailSourceNode)
                      {
                          Utility.out.println(
                              "\nERROR IN GUIActions . ReadEMailTextFiles...\n" +
                              "TCP email files can have only one source node;" +
                              " file entered has both " + tcpEmailSourceNode + " and " +
                              sourceNodeNumber+"\n");
                          break readData;
                      }
                  }
                  if ( sourceNodeNumber < 0  |  sourceNodeNumber >= numberOfNodes
                     | destinationNodeNumber < 0 | destinationNodeNumber >= numberOfNodes )
                  {
                     Utility.err.println  ("ERROR IN GUIActions . ReadEMailTextFilesAndPut"
                        + "IntoSourceNodeBufferAction():\nSource node number " 
                        + sourceNodeNumber + " or destination node number " 
                        + destinationNodeNumber + " must be between 0 and " 
                        + ( numberOfNodes - 1 ) ) ;
                     break readData ;
                  }  // end of if ( sourceNodeNumber == MINUS_ONE  | destinationNodeNumber == MINUS_ONE )
                  else
                  {  // The message is complete
                     // First remove the last carriage return that was appended at
                     // line 145
                     if ( inputEMailAsStringBuilder .charAt ( 
                        inputEMailAsStringBuilder .length () -1 )
                     == '\n' )
                     { // The last character should be a carraige return
                        inputEMailAsStringBuilder . deleteCharAt ( 
                           inputEMailAsStringBuilder .length () - 1 ) ;
                     }
                     if ( ModelData.PrintPermissions.printEmailInputs == true )
                     {
                        Utility.out.println ("\nBuffering message:  sourceNode = "
                           + sourceNodeNumber + " destinationNode = "
                           + destinationNodeNumber + " email message =\n "
                           + new String ( inputEMailAsStringBuilder  ) ) ;
                     } // end of if ( ModelDatas.PrintPermissions.printEmailInputs == true )
 
                     // Create the IPPacket(s)
                     ConvertEMailToIPPacketAndBuffer (
                        new String ( inputEMailAsStringBuilder  ) ) ;

                     // Rest the source and destination node numbers
                     sourceNodeNumber = MINUS_ONE  ;
                     destinationNodeNumber = MINUS_ONE ;                   
                  }  // end of else if ( inputStringTrimmed  . startsWith ( termination ) )
               }  // end of else if ( inputStringTrimmed  . startsWith ( termination ) )
               else
               { // Otherwise add the line to the message
                  inputEMailAsStringBuilder.append ( inputString ) ;
               }  // end of else

            }  // end of if ( ! inputString . startsWith ( COMMENT_DELIMITER ) &
         }  // end of while (fileReader . ready ( ) )
         fileReader . close ( ) ;
      }  // end of try
      catch ( java.lang.Exception exception )
      {
         Utility.err.println ( "ERROR IN MainGUI_Actions."
            + "ReadEMailTextFilesAndPutIntoSourceNodeBufferAction.actionPerformed:"
            + "\n Exception "+ exception. getClass ( ) . getName ( ) + " thrown "
            + "while trying to open EMail file because " + exception . getCause ( )
            + "\n NOTE: A java.lang.ArrayIndexOutOfBoundsException may indicate "
            + "that you tried to read emails before reading the topology." ) ;
      } // end of catch

      // load the first email
      if(tcpEmailSourceNode >= 0) {
          ModelPhysical.Node mpn = vectorOfNodes.get(tcpEmailSourceNode);
          mpn.loadNextEmail();
      }
      else if(transmissionProtocol ==
            ModelData.GlobalVariables.transmissionProtocols.TCP)
          Utility.out.println("ERROR - NO EMAIL FOUND TO LOAD");

   }  // end of public void actionPerformed(java.awt.event.ActionEvent event)


   /**A method to convert the e-mail message to a UDP/IP packet or TCP/IP
    * packets.  It then adds the converted email message to the buffer
    * of the source node.*/
   public void ConvertEMailToIPPacketAndBuffer ( String EMailIn )
   {  //Convert the String to a JNWBitSet;
      JNWBitSet eMailBitSet = ConvertEMailStringToBitSetOfCorrectSize ( EMailIn ) ;
      int emailLength = eMailBitSet.getJNWLength()/8;
   
      // Add the UDP or TCP header
      //Get the source node
      ModelPhysical.Node sourceNode = vectorOfNodes . get ( sourceNodeNumber ) ;

      ModelPhysical.Node destinationNode = vectorOfNodes . get ( destinationNodeNumber ) ; // BZ

      switch ( transmissionProtocol )
      {
         case UDP :
            UDPPacket
               = ModelTransport.UDPUtilities.PrependUDPHeaderToData ( eMailBitSet 
               , sourceNodeNumber , destinationNodeNumber ) ;
            //Add the packet to the buffer of the node
            UDPPacket
               =  ModelTransport.IPUtilities.PrependIPHeaderToMessage (  UDPPacket
               , sourceNodeNumber , destinationNodeNumber , transmissionProtocol )  ;
            // For some strange reason I must use the clone constructor
            // or all the messsages in the node buffer will be identical
            sourceNode . nodeBuffer . add ( UDPPacket . clone ( ) ) ;
            // Put event to send packets to interfaces onto the discrete 
            // event simulator
            new ModelEvents . RoutePacketAtNodeOrProcessAtDestination ( 0
               ,  sourceNodeNumber, UDPPacket ) ;
            break ;
            
         case TCP :
            // Add the email message buffer to email queue in node BZ
            sourceNode.addMessageBuffer(
                    eMailBitSet,
                    emailLength,
                    destinationNodeNumber);
            break ;
      }  // end of switch
      // Clear the email
      inputEMailAsStringBuilder = new java.lang.StringBuilder ( ) ;
   }  // end of ConvertEMailToIPPacketAndPutIntoSourceNode

   /**Searches a String for the first ':' and trims the remainder of
    * the String.*/
   private static String FindColonAndConvertRemainderToTrimmedSubString (
      String stringIn )
   {  // find first ':'
      int firstColon = stringIn.indexOf ( ':' ) ;
      return  ( stringIn . substring ( firstColon + 1 ) ) . trim ( ) ;
   }  // end of FindColonAndConvertStringToInt


   /**Searches a String for the first ':' and converts the remainder
    * of the String to an int.*/
   private static int FindColonAndConvertStringToInt ( String stringIn )
   {  // find first ':'
      int firstColon = stringIn.indexOf ( ':' ) ;
      String string = ( stringIn . substring ( firstColon + 1 ) ) . trim ( ) ;
      // Convert String to int and return
      return ( new Integer ( string ) ) . intValue ( ) ;
   }  // end of FindColonAndConvertStringToInt


   /**Converts the String to a JNWBitSet and rounds the email messages length
    * to two bytes.  This is necessary since a two-byte ( 16 bits ) XOR is
    * performed over the UDP and TCP messages and headers.*/
   private static JNWBitSet ConvertEMailStringToBitSetOfCorrectSize ( String EMailIn )
   {  //Convert the String to a JNWBitSet;
      JNWBitSet bitSet = new JNWBitSet ( ) ;
      bitSet . setJNWBitSetToString ( EMailIn ) ;
      // Pad the message clone to 16 bits  The message String is NOT padded
      // to 16 bits, but the message clone, which is used in computing the
      // 16-bit checksum, is padded so that the CRC can be computed.
      int remainderBits = bitSet . getJNWLength ( )  % 16 ;
      if ( ( remainderBits != 0 ) )
      {  // Append additional bits
         bitSet  = bitSet . shift ( 16 - remainderBits )  ;
      }
      return bitSet ;
   }  // end of ConvertEMailStringToBitSetOfCorrectSize


   /** Converts the input protocol string to an enumeration of the transmission
    * protocol.*/
   private static ModelData.GlobalVariables.transmissionProtocols
      GetProtocol ( String string )
   {  // The enumeration of the possible transmission modes
      java . lang . Enum  enumerationValue ;
      /**What type of transmission protocol is this?*/
      java.util.Iterator iterator  = java.util.EnumSet . range (
         ModelData.GlobalVariables.transmissionProtocols.UDP
         , ModelData.GlobalVariables.transmissionProtocols.TCP )
         . iterator () ;
      while ( iterator . hasNext ( ) )
      {  // Get the enumerated value
         enumerationValue = ( java . lang . Enum ) iterator.next () ;
         if ( enumerationValue . toString () . equals ( string ) )
         {
            return ( ModelData . GlobalVariables . transmissionProtocols  )
            enumerationValue ;
         }  // end of if ( enumerationValue . toString () . equals ( string ) )
      }  // end of while.  If no enumerated value equals
      Utility.err.println  ("ERROR in GUIActions."
         + " ReadEMailTextFilesAndPutIntoSourceNodeBufferActionGUIDatas."
         +  "getProtocol:\n The stringRepresentation, " + string
         + ", does not equal any enumeration value.  Returning UDP." );
      return ModelData.GlobalVariables.transmissionProtocols.UDP ;
   }


   /**For testing only.*/
   public static void main_Testing_Only (String [ ] args)
   {
      String test = "abc" ;
      ModelData.GlobalVariables.transmissionProtocols output ;
      output = GetProtocol ( test ) ;
      System.out.println ( " Ths protocol is : " + output.toString ( ) ) ;
      System.exit ( 0 ) ;
      String email = "A:  00015305 " ;
      ////  String string = "b" ;
      for ( int i = 0 ; i < 1 ; i++ )
      {
         int source2 = FindColonAndConvertStringToInt ( email )  ;
         System.out.println ( "int for String " + email + " is " + source2 ) ;
      }
      System.exit (  1 ) ;
      System.out.println ("STARTING  ReadEMailTextFileAndPutIntoSourceNodeBufferAction_1") ;
      new ReadEMailTextFilesAndPutIntoSourceNodeBufferAction () ;
   }  // end of public static void main_Testing_Only

}  // end of class
