/*
 * ReadTopologyFileAction.java
 *
 * Created on December 20, 2006, 8:09 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package GUIActions;

import ModelPhysical.Node;
import ModelPhysical.Link;

/** ReadTopologyFileAction reads the network topology from a text and creates
 * an array containing the network description.  ReadTopologyFileAction
 * also creates a Vector, vectorOfNodes, containing
 * all the Nodes and a Vector, vectorOfLinks, containing all the Links.
 *
 *<p><b>Versions:</b>
 *<li><b>Version 1.101</b> of 20071221 reset  wholeTopologyInputText
 *= new java.lang.StringBuilder ( 1000 ) to clear out the
 *topology after printing it.  Otherwise the old topology
 *will also be printed.</li>
 *<li><b>Version 1.102</b> of 20080128 read the networkType from the
 * topology file rather than input from the Globle variables GUI.</li>
 * @version 1.202
 * @author dixonw
 * @modified JMP 18Nov2011 for JTRN3
 */
public class ReadTopologyFileAction extends MainGUIActionsBaseClass
   implements GUIConstants.General {

   /** File chooser to find file*/
   javax.swing.JFileChooser jFileChooser = null;

   /**The input file defining the network topology*/
   java.io.File inputTopologyFile = null;

   /** Buffered file read to read individual lines*/
   java.io.BufferedReader fileReader = null;

   /**Individual lines in the topology file*/
   String inputString = null;

   /**The tokenizer for the topology input file*/
   java.util.StringTokenizer stringTokenizer = null;

   /**A singleton vector containing all the Nodes*/
   static ModelData.JNWSingletonVectorOfNodes<ModelPhysical.Node> vectorOfNodes 
      = ModelData.JNWSingletonVectorOfNodes.getInstance ();

   /**A vector for temporarily storing the Nodes.  The nodes are put onto this
    *temporary Vector in the order that they appear in the topology
    *difinition file, which in the sequential Node number, the order in which
    * they were created.  The Nodes are then moved to the vectorOfNodes in the order
    *of their node numbers.  This means that the nodes will be displayed in
    *their used-defined order.*/
   private static java.util.Vector<ModelPhysical.Node> temporaryNodeVector 
      = null;

   /**A singleton vector containing all the links*/
   private static ModelData.JNWSingletonVectorOfLinks<ModelPhysical.Link> vectorOfLinks 
      = ModelData.JNWSingletonVectorOfLinks.getInstance ();

   /**A singleton vector containing all the interfaces.*/
   private static ModelData.JNWSingletonVectorOfInterfaces<ModelDataLink.InterfaceBaseClass> 
      vectorOfInterfaces = ModelData.JNWSingletonVectorOfInterfaces.getInstance ();

   /**Number of the first node read from the topology file.*/
   int firstNodeNumber;

   /**Number of the last read from the topology file.*/
   int lastNodeNumber;

   /**The whole topology file.  Used only to display the input
    *tolology file if ModelDatas . PrintPermissions . displayInputTopology
    *is set to true.*/
   java.lang.StringBuilder wholeTopologyInputText = null ;

   /**Prefix for the newtork type*/
   static String networkTypePrefix = null ;

   /**Prefix for the node position*/
   static String nodePositionPrefix = null  ;

   /**The position where the node is to be displayed on the graphical 
    * display.*/
   private static int XPos,  YPos ;
   
   /**A triple of node number, X position, and Y position.*/
   private static int [ ] triple = new int [ 3 ] ; 
   
   /**A vector of node number and X and Y position.*/
  private static java.util.Vector <int [ ] > nodes 
     = new java.util.Vector ( ) ;

   /** Constructor calls AssignToAction ( ) to add the
    * accelerator key, the long description, the short description, the
    * mnemonic key, and the small icon to the action.*/
   public ReadTopologyFileAction () {
      // Call AssignToAction to add all the attributes to the action
      AssignToAction (
         ReadTopologyFileAction_ACCELERATOR_KEY ,
         ReadTopologyFileAction_LONG_DESCRIPTION ,
         ReadTopologyFileAction_MNEMONIC_KEY , ReadTopologyFileAction_NAME ,
         ReadTopologyFileAction_SHORT_DESCRIPTION ,
         ReadTopologyFileAction_SMALL_ICON );
      // Make sure the random number generator exists
      // This will make seedForJNWRandomNumberGenerator available to
      // ModelPhysical.Link.
      Utility.JNWRandom useless = Utility.JNWRandom.getInstance ();
      nodePositionPrefix = "Position of Node" ;
      networkTypePrefix = "NETWORK-TYPE:" ;
   }

   /** When the toolbar button is clicked on the menubar, reads the topology
    * file and creates
    *a vector of Nodes and a vector of Links and it creates the interface matrix.   */
   public void actionPerformed ( java.awt.event.ActionEvent event ) {  // Find the topology input file
      jFileChooser = new javax.swing.JFileChooser ( fileNameForJFileChooser );
      jFileChooser.setDialogTitle ( "READ TOPOLOGY FILE" );
      try {
         int returnVal = jFileChooser.showOpenDialog ( null );
         if (returnVal == jFileChooser.APPROVE_OPTION) {
            inputTopologyFile = jFileChooser.getSelectedFile ();
         } // end of if
         fileReader = new java.io.BufferedReader (
            new java.io.FileReader ( inputTopologyFile ) );
         // update the fileNameForJFileChooser for later access to the same
         // folder
         fileNameForJFileChooser = jFileChooser.getSelectedFile ();
      }
      catch ( java.lang.Exception exception ) {
         Utility.err.println ( " ERROR IN MainGUI_Actions.ReadTopologyFileAction."
            + "actionPerformed:\nException " + exception.getClass ().getName () 
            + " thrown while trying to open topology file because "
            + exception.getCause ( ) );
      }
      // Remove any nodes, links, or interfaces from the Vectors.  This could happen if the user
      // opens a second topology file.  Also reset the static numbers so that the links
      // and interfaces are numbered from 0
      vectorOfNodes.removeAllElements ();
      vectorOfLinks.removeAllElements ();
      ModelPhysical.Link.staticLinkNumber = 0;
      vectorOfInterfaces.removeAllElements ();
      ModelDataLink.InterfaceBaseClass.staticInterfaceNumber = 0;
      ModelData.GlobalVariables.networkType = null;
      wholeTopologyInputText = new java.lang.StringBuilder ( 1000 );
      try {
         // read the file line by line and remove the white space before and after the characters
         while (fileReader.ready ()) {  // Read a single line
            inputString = fileReader.readLine ().trim ();
            // readLine ( ) will strip off the carraige return, so restore it fro printing
            wholeTopologyInputText.append ( inputString + CARRIAGE_RETURN );
            // Is this the network type?
            if (inputString.startsWith ( networkTypePrefix )) {  // Get the network type
               String networkTypeAsString =
                  FindColonAndConvertStringToTrimmedSubString ( inputString );
               try {
                  ModelData.GlobalVariables.networkType =
                     ModelData.GlobalVariables.networkTypes.valueOf ( 
                     networkTypeAsString );
               // Make sure you found an acceptable network type
               }
               catch ( Exception exception ) {
                  Utility.err.println ( "ERROR in ReadTopologyFileAction:exception" 
                     + exception.getClass ().getName () + " thrown " + networkTypeAsString
                     + " not found in enumeration" );
               } // end of if ( interfaceTypeFound == false )
            }  // end of if ( inputStringTrimmed  . startsWith ( destinationNodePrefix ))
            // If it is a comment skip or the string is empty, skip it.
            else {
               if (inputString.startsWith ( nodePositionPrefix )) {
                  try {
                     String stringToParse = new String ( inputString )  ; 
                     int length = nodePositionPrefix.length () ;
                     stringToParse = stringToParse . substring ( length  ) ;                 
                     // Find the "=" sign
                     length = stringToParse . indexOf ( "=" ) ;
                     String string = stringToParse . substring ( 0 , length ) ;                
                     int nodeNumber = new Integer ( string.trim () ).intValue ();
                     stringToParse = stringToParse . substring ( length + 1   ) ;                   
                     length = stringToParse . indexOf ( "(" ) ;
                     stringToParse = stringToParse . substring ( length + 1   ) ;                 
                     length = stringToParse . indexOf ( "," ) ;
                     string = stringToParse . substring ( 0 , length     ) ;                  
                     int XPos = new Integer ( string.trim () ).intValue ();
                     stringToParse = stringToParse . substring ( length + 1   ) ;
                     length = stringToParse . indexOf ( ")" ) ; 
                        string = stringToParse . substring ( 0 , length  ) ;                   
                     int YPos = new Integer ( string.trim () ).intValue ();
                     // Save the node number and position in a triple
                     triple [ 0 ]  = nodeNumber   ;
                     triple [ 1 ] = XPos ;
                     triple [ 2 ] = YPos ;
                     // Add the triple to the Vector nodes
                     nodes.add (  triple . clone ( ) ); 
                     }  // end of try
                     catch ( Exception e ) {
                        Utility.err.println ( " ERROR IN MainGUI_Actions.ReadTopologyFileAction"
                           + ".actionPerformed:\nException " + e.getClass ().getName ()
                           + " thrown while trying to read line \""
                           + inputString + "\" in input topology file"
                           + "\n    Data should be in format\n\"" + nodePositionPrefix 
                           + " int = ( XPos , YPos )"); }
               }  // end of if (inputString.startsWith ( nodePositionPrefix ))
               else
               {
                  if ( ! inputString.startsWith ( COMMENT_DELIMITER  ) &
                      ! inputString.equals ( EMPTY_STRING )) {
                     stringTokenizer = new java.util.StringTokenizer ( 
                        inputString , COMMA , false  );
                     int sourceNodeNumber = 0;
                     int destinationNodeNumber = 0;
                     int dataRate = 0;
                     float bitErrorRate = ( float ) 0.0;
                     String bitErrorRateAsString = null;
                     float transmissionDelay = ( float ) 0.0;
                     XPos = MINUS_ONE;
                     YPos = MINUS_ONE;
                     String transmissionDelayAsString = null;
                     String interfaceTypeAsString = null;
                     firstNodeNumber = 0;
                     lastNodeNumber = 0;
                     ModelData.GlobalVariables.interfaceTypes enumerationOfInterfaceType =
                        null;
                     try {  // You MUST enter a source, destination, bandwidth, and a type
                        sourceNodeNumber = new Integer ( stringTokenizer.
                           nextToken ().trim () ).intValue ();
                        destinationNodeNumber = new Integer ( stringTokenizer.
                           nextToken ().trim () ).intValue ();
                        dataRate = new Integer ( stringTokenizer.nextToken ().
                           trim () ).intValue ();
                        interfaceTypeAsString = stringTokenizer.nextToken ().
                           trim ();
                        // Define the interface type
                        for ( ModelData.GlobalVariables.interfaceTypes tempInterfaceType : ModelData.GlobalVariables.interfaceTypes.
                           values () ) {
                           if (interfaceTypeAsString.equals ( tempInterfaceType.
                              toString () )) {  // Set the interface type
                              enumerationOfInterfaceType = tempInterfaceType;
                           }  // end of if ( interfaceTypeAsString . equals ( tempInterfaceTypeType . toString ( ) )
                        }  // end of for ( int i = 0 ; i < ModelDatas.Enumerations.networkTypeNames.length ; i++ )
                        if (enumerationOfInterfaceType == null) {
                           Utility.err.println ( "ERROR in ReadTopologyFileAction: interfaceType "
                              + interfaceTypeAsString 
                              + " not found in enumeration" );
                        } // end of if ( interfaceTypeFound == false )
                        try {  // Find the bit error rate
                           bitErrorRateAsString = stringTokenizer.nextToken ().
                              trim ();
                           // The bit error rate may be blank, if so, it will be interpreted as 0.0
                           if ( bitErrorRateAsString.length () > 0 ) {
                              bitErrorRate = new Float ( bitErrorRateAsString ).
                                 floatValue ();
                           }  // end of if ( bitErrorRateAsString.length () > 0)
                        }  // end of try 
                        catch ( java.util.NoSuchElementException e ) {
                           //If there is no String ignore exception and set bitErrorRate = 0.0
                        }
                        transmissionDelayAsString = stringTokenizer.nextToken ().
                           trim ();
                        transmissionDelay = new Float ( 
                           transmissionDelayAsString ).floatValue ();
                        //If there is no String ignore exception and set transmissionDelay = 0.0
                        // Find the first and last nodes
                        firstNodeNumber = java.lang.Math.min ( firstNodeNumber ,
                           java.lang.Math.min ( sourceNodeNumber ,
                           destinationNodeNumber ) );
                        lastNodeNumber = java.lang.Math.max ( lastNodeNumber ,
                           java.lang.Math.max ( sourceNodeNumber ,
                           destinationNodeNumber ) );
                        // Test whether the Node is already on the vectorOfAllNodes, create the Node which
                        // add it to the vectorOfAllNodes in the constructor for Node
                        if (vectorOfNodes.nodeAleadyInVector ( sourceNodeNumber ) ==
                           false) {  // Create the new Node and automatically add it to the vectorOfNodes
                           new Node ( sourceNodeNumber );
                        }
                        if (vectorOfNodes.nodeAleadyInVector ( 
                           destinationNodeNumber ) == false) {  // Create the new Node and automatically add it to the vectorOfNodes
                           new Node ( destinationNodeNumber );
                        }
                        if (vectorOfLinks.linkAleadyInVector ( sourceNodeNumber ,
                           destinationNodeNumber , dataRate ) == false) {  // Create the new Link and automatically add it to the vectorOfLinks
                           // temporarily set the sequential Node numbers to 0
                           // Update the sequential node numbers in AddLinkNumbersToNodes ( )
                           new Link ( sourceNodeNumber , destinationNodeNumber /*, 0  , 0  */ ,
                              dataRate , bitErrorRate , transmissionDelay ,
                              enumerationOfInterfaceType );
                        }  // end of if if ( vectorOfLinks . linkAleadyInVector ( sourceNodeNumber  , destinationNodeNumber  , dataRate )
                     }  // end of try
                     catch ( Exception e ) {
                        Utility.err.println ( " ERROR IN MainGUI_Actions.ReadTopologyFileAction"
                           + ".actionPerformed:\nException " + e.getClass ().getName () 
                           + " thrown while trying to " + " read line \"" 
                           + inputString + "\" in input topology file" 
                           + "\n    Data should be in format\n\"SOURCE NODE , DESTINATION NODE , " 
                           + "DATA RATE, INTERFACE TYPE, BIT ERROR RATE , TRANSMISSION DELAY IN MICROSECONDS" );
                     }  // end of catch
                  } // end of if ( inputString . startsWith ( COMMENT_DELIMITER ) )
               }  // end of else
            }  // end of else
         }  // end of while (fileReader . ready ( ) )       
         // Check for errors
         if (ModelData.GlobalVariables.networkType == null) {
            {
               Utility.err.println ( "ERROR in GUI_Actions."
                  + "ReadTopologyFileAction constructor: " 
                  + "The network type has not been specified. " );
               return ;
            }  // end of if ( ( firstNodeNumber != 0 ) |
         }  // end of if

         // Test to ensure that the node numbers are contiguous-That none are missing
         if (( firstNodeNumber != 0 ) |
            lastNodeNumber != vectorOfNodes.size () - 1) {
            Utility.err.println ( "ERROR in GUI_Actions."
               + "ReadTopologyFileAction constructor:"
               + "\nFirst node number = " + firstNodeNumber 
               + " last node number = " + lastNodeNumber 
               + " but vector of nodes has size " + vectorOfNodes.size () );
            return ;
         }  // end of if ( ( firstNodeNumber != 0 ) |
      }  // end of try
      catch ( java.io.IOException exception ) {
         Utility.err.println ( " ERROR IN MainGUI_Actions."
            + "ReadTopologyFileAction constructor:" 
            + " Error in reading topology file" + exception.getCause () );
      } // end of catch
      // Call the method to
      // reorder the nodes in the Vector by the node number
      ReorderNodes ();
      AddLinkNumbersToNodes ();
      // out input topology or Nodes, Links, and Interfaces
      if (ModelData.PrintPermissions.displayInputTopology == true) {
         Utility.out.print ( "\nNETWORK TYPE = " +
            ModelData.GlobalVariables.networkType.toString () + CARRIAGE_RETURN +
            "\nInput topology = " + CARRIAGE_RETURN + new String ( 
            wholeTopologyInputText ) );
      }  // end of if (ModelData.PrintPermissions.displayInputTopology == true)
      if (ModelData.PrintPermissions.displayNodesLinksAndInterfaces == true) {
         Utility.out.print ( "\nIN ReadTopologyFileAction:\nNodes =" +
            vectorOfNodes.toString () + "\n Links = " +
            vectorOfLinks.toString () + "\n Interfaces = " + vectorOfInterfaces.
            toString () );
      }  // end of if (ModelData.PrintPermissions.displayNodesLinksAndInterfaces == true)
      // Create the routing matrices for the LAN or create the topology for the
      // Ethernet bus
      switch ( ModelData.GlobalVariables.networkType ) {
         case Router_Network:
            // Create the routing tables in each of the nodes
            for ( int i = 0 ; i < vectorOfNodes.size () ; i ++ ) {
               vectorOfNodes.get ( i ).InitializeRoutingMatrix ();
            }  // end of for
            new ModelNetwork.Routing ().ComputeRoutingMatrices ();
            break;
         case Ethernet_Bus:
            new ModelPhysical.EthernetBus ();
            break;
         case Token_Ring:
            Utility.err.println ( "ERROR IN ReadTopologyFileAction.actionPerformed: Token Ring " +
               "not implemented yet." );
      }  // end of switch
      try {
         fileReader.close ();
      }  // end of try
      catch ( java.lang.Exception exception ) {
         Utility.err.println ( " ERROR IN MainGUI_Actions.ReadTopologyFileAction.ReadTopologyFileAction" +
            " constructor:\nException " + exception.getClass ().getName () +
            " thrown while trying to close topology file because " + exception.
            getCause () );
      }  // end of catch
   } // end actionPerformed(java.awt.event.ActionEvent event)

   /**Adds the link numbers to the Vector in each Node.  An iterator steps through
    * all the Nodes. For each Node, a java.util.Iterator searches all the Links
    * to determine whether that Link originates or terminates at the
    * Node.  If so, the Link is added to the vectorOfLinksForThisNode for the Node.*/
   public static void AddLinkNumbersToNodes () {
      java.util.Iterator nodeIterator = vectorOfNodes.iterator ();
      // Iterate thgough the nodes
      while (nodeIterator.hasNext ()) {
         Node node = ( Node ) nodeIterator.next ();
         // Iterate throught the links
         java.util.Iterator linkIterator = vectorOfLinks.iterator ();
         while (linkIterator.hasNext ()) {
            Link link = ( Link ) linkIterator.next ();
            // If the Link starts or ends at this Node
            if (( node.thisNodeNumber == link.destinationNodeNumber ) ||
               ( node.thisNodeNumber == link.sourceNodeNumber )) {  // Set the sequential numbers in the node
               if (node.thisNodeNumber == link.destinationNodeNumber) {
                  link.destinationNodeNumber = node.thisNodeNumber;
               }
               if (node.thisNodeNumber == link.sourceNodeNumber) {
                  link.sourceNodeNumber = node.thisNodeNumber;
               }
               // Add it to the Node's Vector of Links'
               node.addLink ( link );
            }  // end of if ( ( node.thisNodeNumber == link . destinationNodeNumber ) ||
         } // end of while (linkIterator.hasNext( ) )
      } // end of while (nodeIterator.hasNext())
   }  // end of AddLinkNumbersToNodes ( )

   /**Reorders the Nodes on the vectorOfNodes by sequential node number.
    * This makes it easier for the user to understand the node numbers.  It 
    * then adds the Xpos and YPos to the Nodes for use in the graphical
    * displays.*/
   private static void ReorderNodes () {  // Create a temporary Vector
      int numberOfNodes = vectorOfNodes.size ();
      temporaryNodeVector = new java.util.Vector ( numberOfNodes );
      // Add the Nodes to the termporary vector
      for ( int i = 0 ; i < numberOfNodes ; i ++ ) {
         temporaryNodeVector.add ( vectorOfNodes.get ( i ) );
      }  // end of for ( int i = 0 ; i < numberOfNodes ; i++ )
      // Remove all the nodes from VectorOfNodes
      vectorOfNodes.removeAllElements ();
      for ( int lookingForNodeNumber = 0 ; lookingForNodeNumber < numberOfNodes ;
         lookingForNodeNumber ++ ) {
         for ( int location = 0 ; location < temporaryNodeVector.size () ;
            location ++ ) {
            if (temporaryNodeVector.get ( location ).thisNodeNumber ==
               lookingForNodeNumber) {  // Add the Node back onto the vectorOfNodes
               vectorOfNodes.add ( temporaryNodeVector.get ( location ) );
               // Delete the Node from the temporaryNodeVector
               temporaryNodeVector.remove ( location );
               break;
            }  // end of if ( temporaryNodeVector .
         }  // end of for ( int location = 0 ;
      }  // end of for ( int lookingForNodeNumber = 0 ;...
      // Copy the Node positions
      while ( nodes.isEmpty ( ) ==  false ) {        
         triple = nodes.firstElement () ;
         nodes. remove ( 0 ) ; 
         int nodeNumber = triple [ 0 ] ;
         Node node = vectorOfNodes . get ( nodeNumber ) ;
         node . XPos = triple [ 1 ] ;
         node . YPos = triple [ 2 ] ;        
      }  // end of while ( nodes.isEmpty ( ) ==  false )
   }  // end of ReorderNodes

   /**Searches a String for the first ':' and converts the remainder
    * to a trimmed String.*/
   private static String FindColonAndConvertStringToTrimmedSubString ( 
      String stringIn ) {  // find first ':'
      int firstColon = stringIn.indexOf ( ':' );
      return ( stringIn.substring ( firstColon + 1 ) ).trim ( );
////200080825////      return string2;
   }  // end of FindColonAndConvertStringToInt

   /**For testing only.*/
   public static void main_Testing_Only ( String[] args ) {
      String inputString = "a,a,z. ,.t,.d";
      String comma = ".,";
      java.util.StringTokenizer stringTokenizer =
         new java.util.StringTokenizer ( inputString , comma , false );
      while (stringTokenizer.hasMoreElements ()) {
         System.out.println ( " testing " );
         String data = stringTokenizer.nextToken ().trim ();
         System.out.println ( " data = " + data );

      }
      System.out.println ( "FINISHED" );
      System.exit ( 0 );
      ModelData.GlobalVariables.networkTypes network =
         ModelData.GlobalVariables.networkTypes.Token_Ring;
      System.out.println ( "network type = " + network );
      System.out.println ( "network type position = " + network.ordinal () );
      System.out.println ( "IN TESTONLY: THE NODES ARE:\n" + vectorOfNodes.
         toString () );
      ModelNetwork.Routing routing = new ModelNetwork.Routing ();
   }  // end of main_Testing_Only 

}  // end of class

