package GUIGraphical;


/**This class saves 
* graphical characteristics needed to display the GraphicalLink
*  on the Topology Display.  This is also used in 
* GUIData.Link_Characteristics to get and put the input data into
* the correct GraphicalLink.*/
public class GraphicalLink implements GUIConstants . General {
   //These fields are related only the graphical
   // display of the data.

   /**The starting Node number.*/
   public int startNodeNumber;

   /**The ending Node number.*/
   public int endNodeNumber;

   /**The channel capacity in Kbps.*/
   public int dataRate;

   /**The interface type.*/
   public ModelData.GlobalVariables.interfaceTypes interfaceType =
      ModelData.GlobalVariables.interfaceTypes.Ethernet;

   /**The bit error rate.*/
   public float bitErrorRate;

   /**The transmission delay in microseconds.*/
   public float transmissionDelay;

   // The square of the denominator used in computing
   // the distance distance to an arbitrary point ( X , Y )
   // See http://mathworld.wolfram.com/Point-LineDistance2-Dimensional.html
   private long DSquare;

   /**The minimum of the X coordinates.  Used
    * to determine if a mouse click is on a Link.*/
   public int XMin;
   
   /**The maximum of the X coordinates.  Used
    * to determine if a mouse click is on a Link.*/
   public int XMax;
   
   /**The minimaum of the Y coordinates.  Used
    * to determine if a mouse click is on a Link.*/
   public int YMin ;
   
   /**The maximum of the Y coordinates.  Used
    * to determine if a mouse click is on a Link.*/
   public int YMax;

   /**DeltaX used in computing the distance between point ( X,Y )
    * and the Link*/
   public int deltaX ;
   
   /**DeltaY used in computing the distance between point ( X,Y )
    * and the Link*/
   public int deltaY ;
   
   /**The default constructor to start the GraphicalLink.*/
   public GraphicalLink () {
   startNodeNumber = MINUS_ONE ;
   endNodeNumber = MINUS_ONE ;
   dataRate = ZERO ;
   interfaceType =
      ModelData.GlobalVariables.interfaceTypes.Ethernet;
   bitErrorRate = ZERO_FLOAT ;
   transmissionDelay = ZERO_FLOAT ;      
   }

   public GraphicalLink ( 
        int sourceNodeNumberIn 
      , int destinationNodeNumberIn 
      , int dataRateIn
      , ModelData.GlobalVariables.interfaceTypes interfaceTypeIn 
      , float bitErrorRateIn 
      , float transmissionDelayIn ) {     
      this.startNodeNumber = sourceNodeNumberIn;
      this.endNodeNumber = destinationNodeNumberIn;
      this.dataRate = dataRateIn;
      this.interfaceType = interfaceTypeIn;
      this.bitErrorRate = bitErrorRateIn;
      this.transmissionDelay = transmissionDelayIn;
      // UpdateGraphicalVariables the graphical variables
      UpdateGraphicalVariables ();     
   }

   /**A constructor to start the GraphicalLink.*/
   public GraphicalLink ( int startnodeNumberIn ) {
      this ( ) ;   
      startNodeNumber = startnodeNumberIn;
   }

   public void setStartNodeNumber ( int startNodeNumberIn ) {
      startNodeNumber = startNodeNumberIn;
      UpdateGraphicalVariables (); 
   }

   public void setEndNodeNumber ( int endNodeNumberIn ) {
      endNodeNumber = endNodeNumberIn;
      UpdateGraphicalVariables (); 
   }
   /**This method updates both the node numbers simultaneously.  This is necessary
    * since the UpdateGraphicalVariables method is called.  If both node numbers have changed,
    * a null pointer exception may be thrown by calling setStartNodeNumber
    * and setEndNodeNumber in succession.
    * @param startNodeNumberIn
    * @param endNodeNumberIn*/
   public void setStartNodeNumberAndEndNodeNumber ( int startNodeNumberIn
      , int endNodeNumberIn ) {
       startNodeNumber = startNodeNumberIn;
       endNodeNumber = endNodeNumberIn;
      // UpdateGraphicalVariables the graphical variables
      UpdateGraphicalVariables (); 
   }
   
   /**Getter for the startNodeNumber.*/
   public int getStartNodeNumber () {
      return startNodeNumber;
   }

   /**Getter for the endNodeNumber.*/
   public int getEndNodeNumber () {
      return endNodeNumber;
   }

   /**Getter for DSquare.*/   
   public long getDSquare () {
      return DSquare;
   }


   /**Updates and stores the denominator used in FindLink.*/
   public void UpdateGraphicalVariables () {
      int[] position = new int[ 2 ];   
      position = GUIGraphical.TopologyDisplay_Static.nodeVector.get ( 
        /* sourceNodeNumber */ startNodeNumber );
      int startXPos = position[ 0];
      int startYPos = position[ 1];
      position = GUIGraphical.TopologyDisplay_Static.nodeVector.get (  endNodeNumber );
      int endXPos = position[ 0];
      int endYPos = position[ 1];
      int diffX = endXPos - startXPos;
      int diffY = endYPos - startYPos;
      DSquare = diffX * diffX + diffY * diffY;
      deltaX = endXPos - startXPos ;
      deltaY = endYPos - startYPos ;
      XMax = java.lang.Math.max ( startXPos , endXPos );
      XMin = java.lang.Math.min ( startXPos , endXPos );
      YMax = java.lang.Math.max ( startYPos , endYPos );
      YMin = java.lang.Math.min ( startYPos , endYPos );    
   }  // end of public void UpdateGraphicalVariables ()
   
   /**The soString method.*/
   public String toString () {
      String string  ;
       int [ ] position =  GUIGraphical.TopologyDisplay_Static.nodeVector.get ( 
        /* sourceNodeNumber */ startNodeNumber );
      string = " Starting at Node " + startNodeNumber  + " at postion (" + position [ 0 ] 
         + ","+ position [ 1 ] + ") and \n  ending at Node " + endNodeNumber ;
      position =  GUIGraphical.TopologyDisplay_Static.nodeVector.get ( 
        /* sourceNodeNumber */ endNodeNumber );     
      return string + " at postion (" + position [ 0 ] 
         + "," + position [ 1 ]
         + ")\n with XMax, YMax = " + XMax + ", " + YMax 
         + " and XMim, YMim = " + XMin + ", " + YMin 
         + "\n  deltaX = " + deltaX + " deltaY = " + deltaY 
         + "\n  DSquare = " + DSquare  ;
   }  // end of public String toString ()

}  // end of class GraphicalLink
