/*
 * Print_Permissions.java
 *
 * Created on January 11, 2007, 10:38 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package GUIGraphical;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;

/**Displays the network topology.  Also permits the user to edit the topology.
 * This display does NOT update in real time.
 *<li><b>Version 1.201</b> of 20081106 makes changes in actionPerformed so that
 * the GUI will not close if any Nodes do not have Links.</li>
 *<li><b>Version 1.202</b> of 20081107 adds method boolean MoreThanTwoLinks (  ).
 *   The method tells whether any Node has more than two links attached.  If more
 * than two links are attached, the topology must be a router network.</li>
 * @version 1.201
 * @author dixonw*/
public class TopologyDisplay_Static extends GUIs.StandardJFrame {

   /**A singleton vector containing all the Nodes*/
   private static ModelData.JNWSingletonVectorOfNodes<ModelPhysical.Node> vectorOfNodes
      = ModelData.JNWSingletonVectorOfNodes.getInstance ();

   /**A singleton vector containing all the links*/
   private static ModelData.JNWSingletonVectorOfLinks<ModelPhysical.Link> vectorOfLinks
      = ModelData.JNWSingletonVectorOfLinks.getInstance ();

     /**A singleton vector containing all the Interfaces*/
   private static ModelData.JNWSingletonVectorOfInterfaces<ModelDataLink.InterfaceBaseClass> 
      vectorOfInterfaces =
      ModelData.JNWSingletonVectorOfInterfaces.getInstance ();
   
   /**The random number generator*/
   private static Utility.JNWRandom randomNumberGenerator
      = Utility . JNWRandom . getInstance ();

   /**A node position. The position [0] is the X position, xPos, and
    * position [1] is the Y position, xPos. The [0 , 0 ] position in
   java is at the left upper corner of the screen.  The positive X
   direction is to the right, and the positive Y direction is down.*/
   private static int[] position = new int[ 2 ];

   /**A vector of node positions.  A Node is simply the X and Y coordinates.*/
   public static java.util.Vector<int[]> nodeVector
      = new java.util.Vector ( 20 );

   /**A vector of GUIGraphical.GraphicalLinks. */
   public static java.util.Vector<GUIGraphical.GraphicalLink> linkVector
      = new java.util.Vector ( 50 );

   /**The nodeRadius of the Nodes in pixels.*/
   private static int nodeRadius = 15;

   /**The maximum distance between the mouse pointer and a
    * GraphicalLink for the pointer to be considered to be
    * on the GraphicalLink.*/
   private static int maximumDistance = 5;

   /**The maximum distance SQUARED between the mouse pointer and a
    * GraphicalLink for the pointer to be considered to be on the GraphicalLink.  If
    * the mouse pointer is within sqrt ( maximumDistanceSquared ) of the
    * GraphicalLink, then it is on the GraphicalLink.  See
    * MouseClickAdapter.FindLinkWherePressed ( )*/
   private static int maximumDistanceSquared = maximumDistance * maximumDistance;

   /**The node on which the mouse was pressed.  This is the number of the Node
    * on which the mouse is pressed.  This is computed by
    * MouseClickAdapter.FindNodeWherePressed ( ) .  If the mouse is not
    * pressed on a Node, the value is MINUS_ONE.*/
   private static int nodeWherePressed;

   /**The node on which the mouse was released.*/
   private static int nodeWhereReleased;

   /**The link on which the mouse was pressed.  This is the number of the GraphicalLink
    * on which the mouse is pressed.  This is computed by
    * MouseClickAdapter.FindLinkWherePressed ( ) .  If the mouse is not
    * pressed on a GraphicalLink, the value is MINUS_ONE.
   <p>A mouse is on the GraphicalLink if the distance between the GraphicalLink and the
   mouse click is less than sqrt ( maximumDistanceSquared ) .  This is also
   used in GUIData.Link_Characteristics to get and put the input data into
   the correct GraphicalLink. */
   public static int linkWherePressed;

   /**Is dragging taking place? This is set in
    * MouseMotionAdapter.mouseDragged.*/
   private static boolean isDraggingTakingPlace;

   /**Is the right button clicked? */
   private static boolean isRightButtonClicked;

   /**A temporary GraphicalLink.*/
   private static GraphicalLink temporaryLink = null;

   /**The X and Y mouse positions.*/
   private static int XPos , YPos;

   /**The title of the "CLEAR" button.*/
   private static String clearButtonTitle = "CLEAR" ;

   /**The "ENTER/UPDATE" button.*/
   private static javax.swing.JButton enterUpdateButton;

   /**The "CLEAR" button.*/
   private static javax.swing.JButton clearButton;

   /**The panel where the network will be drawn.*/
   private static JPanel drawingPanel = new JPanel ();

   /**The singleton instance of the class InputDataSingleton. <b><u>DO NOT
    * MOVE</u></b> the location of this field in the class.  It <b><u>MUST
    * </u></b>be the last field in  this class. It will destroy the
    * toSting() method and cause the model to go into an infinite loop.*/
   private static final TopologyDisplay_Static topologyDisplay =
      new TopologyDisplay_Static ();

   /** Creates a new instance of TopologyDisplay_Static */
   private TopologyDisplay_Static () {
      super ();
   }  // end of constractor



   static {
      // Set the title
      topologyDisplay.setTitle ( jFrameBaseTitle + " Topology Display" );
      topologyDisplay.setSize ( mainGUIInitialWidthInPixels ,
         mainGUIInitialHeightInPixels );
      topologyDisplay.setResizable ( true );
      topologyDisplay.setDefaultCloseOperation (
         javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE );
      // Set the upper left corner of the GUI to a specific location
      topologyDisplay.setLocation ( XLocationOfInputGUI , YLocationOfInputGUI );
      // Add the listeners in the static initilizer
      // They CANNOT be added in the getInstance () method or you
      // will have multiple listeners.
      // Add a mouse listener to the drawing panel
      topologyDisplay.addMouseListener ( new MouseClickAdapter () );
      // Add a mouse listener to the drawing panel
      topologyDisplay.addMouseMotionListener ( new MouseMotionAdapter () );
      // Define the standard responses to Window Events
      topologyDisplay.addComponentListener ( new MyComponentAdapter () );
      // Define the standard responses to Window Events
      topologyDisplay.addWindowListener ( new MyGraphicalWindowAdapter () );
      // Add in the JButtons
      ////     buttonPanel = new JPanel ();
      enterUpdateButton = new javax.swing.JButton ( inputGUIButtonTitle );
      // make a border for the buttons
      Border border = BorderFactory.createBevelBorder ( BevelBorder.RAISED );
      enterUpdateButton.setBackground ( standardBlueButtonColor );
      enterUpdateButton.setForeground ( standardButtonLetteringColor );
      enterUpdateButton.setFont ( standardButtonBoldFont );
      enterUpdateButton.setBorder ( border );
      enterUpdateButton.addActionListener ( new java.awt.event.ActionListener () {
         // override the actionPerformed method
         public void actionPerformed ( java.awt.event.ActionEvent e ) {
            // If any node does not have a Link then the
            // routing algorithms will explode
            if ( DoesEveryNodeHaveALink () == false ) {
               Utility.err.println ( "IN TopologyDisplay_Static.actionPerformed: Not"
                  + " all the Nodes have Links.  Display will not close.") ;
               return ;
            }
            vectorOfNodes.clear ( );
            vectorOfLinks.clear ( );
            // Call ModelPhysical.Link.RESET to restart the
            // Link numbering with 0.
            ModelPhysical.Link.RESET ();
            ModelDataLink.InterfaceBaseClass.RESET ();
            // Create the Vector of Nodes
            for ( int i = 0 ; i < nodeVector.size () ; i ++ ) {
               new ModelPhysical.Node ( i );
               // Save the current ( X , Y ) position of the Node
               ModelPhysical.Node node = vectorOfNodes.get ( i );
               position = nodeVector.get ( i );
               node.XPos = position[ 0 ];
               node.YPos = position[ 1 ];
            }
            // Create the Vector of Links
            for ( int i = 0 ; i < linkVector.size () ; i ++ ) {
               temporaryLink = linkVector.get ( i );
               new ModelPhysical.Link (
                  temporaryLink.startNodeNumber , temporaryLink.endNodeNumber ,
                  temporaryLink.dataRate , temporaryLink.bitErrorRate ,
                  temporaryLink.transmissionDelay , temporaryLink.interfaceType );
            }
            // If any node does not have a Link then the
            // routing algorithms will explode
////            if ( DoesEveryNodeHaveALink () == false ) {
////               Utility.err.println ( "IN TopologyDisplay_Static.actionPerformed: Not"
////                  + " all the Nodes have interfaces.  Display will not close.") ;
////               topologyDisplay.setVisible ( true );
////               return ;
////            }
               GUIActions . ReadTopologyFileAction . AddLinkNumbersToNodes ( );
               // Create the routing matrices for the LAN or create the topology for the
               // Ethernet bus
//// System.out.println ( " network type = " + ModelData.GlobalVariables.networkType )  ;
               switch ( ModelData.GlobalVariables.networkType ) {
                  case Router_Network:
                     // Create the routing tables in each of the nodes
                     for ( int i = 0 ; i < vectorOfNodes.size () ; i ++ ) {
                        vectorOfNodes.get ( i ).InitializeRoutingMatrix ();
                     }
                     new ModelNetwork.Routing ().ComputeRoutingMatrices ();
////            CreateInterfacesMatrixForPullen ( ) ;
                     break;
                  case Ethernet_Bus:
////            Utility.out.println ("IN ReadTopologyFileAction().actionPerformed: running EthernetBus ")   ;
                     new ModelPhysical.EthernetBus ();
                     break;
                  case Token_Ring:
                     Utility.err.println ( "ERROR IN ReadTopologyFileAction"
                        + ".actionPerformed: Token Ring not implemented yet." );
                   case Undefined:
 ////                   Utility.out.println ( "IN ReadTopologyFileAction"
 ////                       + ".actionPerformed: Network type Undefined." );
                     // Use nodeVector.size () and linkVector.size() to
                     // determine what kind of network this is.
                     if ( nodeVector.size () < linkVector.size ()
                         |  MoreThanTwoLinks (  ) == true ) {
                        // This is a router network
                        ModelData.GlobalVariables.networkType
                           = ModelData.GlobalVariables.networkTypes.Router_Network ;
                        new ModelNetwork.Routing ().ComputeRoutingMatrices ();
                     }  // end of if ( nodeVector.size () < linkVector.size () )
                      else if ( nodeVector.size () - 1 == linkVector.size () ) {
                        ModelData.GlobalVariables.networkType
                           = ModelData.GlobalVariables.networkTypes.Ethernet_Bus ;
                         // This is an Ethernet
                        new ModelPhysical.EthernetBus ();
                     }  // end of if ( nodeVector.size () - 1 == linkVector.size () )
                     else if ( nodeVector.size () == linkVector.size () ) {
                        ModelData.GlobalVariables.networkType
                           = ModelData.GlobalVariables.networkTypes.Token_Ring ;
                        Utility.err.println ( "IN TopologyDisplay_Static: "
                           + "nodeVector.size () == linkVector.size () which "
                           + "means the input is a token ring-which is not "
                           + "yet implemented.") ;
                     }  // end of else if ( nodeVector.size () == linkVector.size () )
/*                     if ( nodeVector.size () < linkVector.size ()
                         |  MoreThanTwoLinks (  ) == true ) {
                        // This is a router network
                        ModelData.GlobalVariables.networkType
                           = ModelData.GlobalVariables.networkTypes.Router_Network ;
                        new ModelNetwork.Routing ().ComputeRoutingMatrices ();
                     }  // end of if ( nodeVector.size () < linkVector.size () )
*/
                     else {
                        ModelData.GlobalVariables.networkType
                           = ModelData.GlobalVariables.networkTypes.Undefined ;
                        Utility.err.println ( "IN TopologyDisplay_Static: "
                           + "The network type is undefined." ) ;
                     }  // end of else
               }  // end of switch
 ////           }  // end of if ( DoesEveryNodeHaveALink ( ) == true )
 ////           else {
////               Utility.err.println ( "IN TopologyDisplay_Static.actionPerformed: Not"
 ////                 + " all the Nodes have interfaces.") ;
 ////           }
            nodeVector.clear ();
            linkVector.clear ();
            topologyDisplay.setVisible ( false );
         }  // end of actionPerformed
      } // After adding action listener to
         ); // end of enterUpdateButton . addActionListener
      clearButton = new javax.swing.JButton ( clearButtonTitle );
      // make a border for the buttons
////      Border border = BorderFactory.createBevelBorder ( BevelBorder.RAISED );
      clearButton.setBackground ( standardBlueButtonColor );
      clearButton.setForeground ( standardButtonLetteringColor );
      clearButton.setFont ( standardButtonBoldFont );
      clearButton.setBorder ( border );
      clearButton.addActionListener ( new java.awt.event.ActionListener () {
         // override the actionPerformed method
         public void actionPerformed ( java.awt.event.ActionEvent e ) {
////System.out.println ("IN TopologyDisplay clearing display: vectorOfNodes.size ()  = "
////   + vectorOfNodes. size ( ) + " vectorOfLinks.size ()  = " + vectorOfLinks.size () ) ;
            vectorOfNodes.clear ( );
            vectorOfLinks.clear ( );
            vectorOfInterfaces.clear ( ) ;
////System.out.println ("IN TopologyDisplay clearing display: vectorOfNodes.size ()  = "
////   + vectorOfNodes. size ( ) + " vectorOfLinks.size ()  = " + vectorOfLinks.size () ) ;            
            nodeVector . clear ( ) ;
            linkVector . clear ( ) ;
//// System.out.println ("IN TopologyDisplay clearing display: vectorOfNodes.size ()  = "
////   + vectorOfNodes. size ( ) + " vectorOfLinks.size ()  = " + vectorOfLinks.size () ) ;
            ModelData.GlobalVariables.networkType
                = ModelData.GlobalVariables.networkTypes.Undefined ;
            topologyDisplay.repaint () ;
         }  // end of actionPerformed
      } // After adding action listener
         ); // end of clearButton . addActionListener
   }  // end of static initializer

   /**Returns the one instance of an input GUI.  getinstance( ) will delete
    * all the components from the center panels and reconstruct the
    * new center panel.*/
   public static TopologyDisplay_Static getInstance () {
////System.out.println ("IN TopologyDisplay_Static.getInstance:\nNodes = \n"
////   + vectorOfNodes.toString () + "\n Links = \n" + vectorOfLinks.toString () )  ;
////System.out.println ("IN TopologyDisplay_Static.getInstance:\n network type = "
////   + ModelData.GlobalVariables.networkType ) ;
      // Initialize the arrays for display and update
      InitializeVectors ();
      // Clear out the old components
      topologyDisplay.getContentPane ().removeAll ();
      // Set the default close operation
////      topologyDisplay.setDefaultCloseOperation (
////         javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE );
      // Set the upper left corner of the GUI to a specific location
////      topologyDisplay.setLocation ( XLocationOfInputGUI , YLocationOfInputGUI );
      // Add in the JButtons
      JPanel buttonPanel = new JPanel ();
      java.awt.GridLayout grid = new java.awt.GridLayout ( 2 , 1 );
      buttonPanel.add ( clearButton );
      buttonPanel.add ( enterUpdateButton );
      topologyDisplay.add ( drawingPanel , java.awt.BorderLayout.NORTH );
      topologyDisplay.add ( buttonPanel , java.awt.BorderLayout.SOUTH );
      ////     AddNodesToGraphic ();
      topologyDisplay.repaint ();
      topologyDisplay.setVisible ( true );
      return topologyDisplay;
   }  // end of topologyDisplay ( )

   /** Initializes the arrays of Nodes and Links from
    * the input data.  It reads vectorOfNodes and vectorOfLinks and
    * creates nodeVector, linkVector, and interfaceVector.*/
   public static void InitializeVectors () {
      if (vectorOfNodes.isEmpty ()) {
////        System.err.println ( "IN AddNodeAndLinks: no nodes" );
         return; }
////System.err.println (" mainGUIInitialWidthInPixels = " + mainGUIInitialWidthInPixels
////   + " mainGUIInitialHeightInPixels = " + mainGUIInitialHeightInPixels  ) ;
      // Get the maximum X and Y positions that can be displayed
      // in the network display
      java.awt.Dimension displayDimension = topologyDisplay.getSize ( ) ;
      // Add Nodes
      for ( int i = 0 ; i < vectorOfNodes.size () ; i ++ ) {
         ModelPhysical.Node node = vectorOfNodes.get ( i );
         XPos = node.getXPos ();
         YPos = node.getYPos ();
//// System.out.println ( " XPos = " + XPos + " YPos = " + YPos + " displayDimension . height = "
////   + displayDimension . height + " displayDimension . width = "
/////   + displayDimension . width )  ;
         if ( XPos < 0 | YPos < 0 | XPos > displayDimension . width
            | YPos > displayDimension . height ) {
            // The upper left corner of the Input GUI has coordinates (2,23); 
            // therefore, it was necessary to apply a bias
            XPos = randomNumberGenerator.nextInt ( mainGUIInitialWidthInPixels -
               5 ) + 5;
            YPos =
               randomNumberGenerator.nextInt ( mainGUIInitialHeightInPixels - 23 ) +
               23;
            node.setPosition ( XPos , YPos );
////System.err.println (" i = " + i + " XPos = " + XPos + " YPos = " + YPos ) ;
         }  // end of if
         position[ 0 ] = XPos;
         position[ 1 ] = YPos;
         nodeVector.add ( position.clone () );
      }  // end of for ( int i = 0 ; i < vectorOfNodes.size ( ) ; i++ )
      if ( ModelData.GlobalVariables.networkType ==
         ModelData.GlobalVariables.networkTypes.Router_Network ) {
         for ( int i = 0 ; i < vectorOfLinks.size () ; i ++ ) {
            ModelPhysical.Link link = vectorOfLinks.get ( i );
            GraphicalLink graphicalLink =
               new GraphicalLink ( link.sourceNodeNumber ,
            link.destinationNodeNumber , link.bitRateSourceToDestination ,
            link.interfaceType , link.bitErrorRate , link.transmissionDelay );
            linkVector.add ( graphicalLink );
         }  // end of for ( int i = 0 ; i < vectorOfNodes.size ( ) ; i++ )
      }   // end of if ( ModelData.GlobalVariables.networkType ==
      else if ( ModelData.GlobalVariables.networkType ==
         ModelData.GlobalVariables.networkTypes.Ethernet_Bus ) {
         // For an Ethernet bus, create one link betwnne each successive
         // node, assign sequential link numbers, and give them all
         // the characteristics of link 0.
         // Get the Ethernet Bus*/
         ModelPhysical.EthernetBus ethernetBus
         = ( ModelPhysical.EthernetBus ) vectorOfLinks.get ( 0 );
////         ModelPhysical.Link link = vectorOfLinks.get ( 0 );
////System.out.println ( "IN TopologyDisplay_Static.InitializeVectors "
////   + "\n link.bitRateSourceToDestination = "
////   + ethernetBus .bitRateSourceToDestination
////   + " link.interfaceType = " + ethernetBus . interfaceType
////   + " link.bitErrorRate = " + ethernetBus .bitErrorRate
////   + " link.transmissionDelay = " + ethernetBus .transmissionDelay ) ;
/*        java.lang.StringBuilder stringBuilder
         = new java.lang.StringBuilder ( 2000 ) ;
      stringBuilder . append (
         "IN TopologyDisplay_Static.InitializeVectors "
         + " Ethernet bus has rate = "
         + ethernetBus . bitRateSourceToDestination + " with type  = "
         + ethernetBus . interfaceType + " with bit error rate "
         + ethernetBus . bitErrorRate + "\n and transmission delay in microseconds\n" ) ;
////         + this . transmissionDelay + " microseconds with interfaces\n" );
////            stringBuilder . append ( "\nCOST MATRIX TO GO TO GO FROM NODE I TO NODE J USING THE ROUTING TABLE" + CARRIAGE_RETURN ) ;
      for ( int i = 0 ; i  < ethernetBus . transmissionDelay.length ; i++ )
      {
         for ( int j = 0 ; j  < ethernetBus . transmissionDelay [0].length ; j++ )
         {
            stringBuilder . append ( TAB + ethernetBus . transmissionDelay  [ i ] [ j ] ) ;
         }  // end of for ( int j = 0 ; j  < numberOfNodes ; j++ )
         stringBuilder . append ( CARRIAGE_RETURN ) ;
      }  // end of  for ( int i = 0 ; i  < numberOfNodes ; i++ )
      String string =   new String ( stringBuilder ) ;
System.out.println ( string ) ;
*/
////            new GraphicalLink ( link.sourceNodeNumber ,
////            link.destinationNodeNumber , link.bitRateSourceToDestination ,
////            link.interfaceType , link.bitErrorRate , link.transmissionDelay );
         for ( int i = 0 ; i < vectorOfNodes.size () - 1 ; i++ ) {
            GraphicalLink graphicalLink =
               new GraphicalLink ( i , i + 1
                  , ethernetBus .bitRateEthernetBus
                  , ethernetBus .interfaceType , ethernetBus . bitErrorRate
                  , ethernetBus .transmissionDelay [ i ] [ i + 1 ]);
         linkVector.add ( graphicalLink );
         }  // end of for ( int i = 0 ; i < vectorOfNodes.size () - 2 ; i++ )
      }  // end of else if ( ModelData.GlobalVariables.networkType ==


   }  // end of InitializeVectors

   /** This determines if every Node has a Link.  When inputting data
    * from a text file, a Node is defined by the Link between the Node
    * and another Node.  But with a graphical input it is possible
    * to create Nodes without Links.  This method return a boolean
    * which says whether every Node has at least one Link.*/
   public static boolean DoesEveryNodeHaveALink () {
      if (nodeVector.isEmpty () == true | linkVector.isEmpty () == true) {
         return false;
      }
      boolean[] connectedViaLink = new boolean[ nodeVector.size () ];
      for ( int i = 0 ; i < linkVector.size () ; i ++ ) {
         GUIGraphical.GraphicalLink link = linkVector.get ( i );
         connectedViaLink[link.endNodeNumber] = true;
         connectedViaLink[link.startNodeNumber] = true;
      }
      for ( int i = 0 ; i < nodeVector.size () ; i ++ ) {
         if (connectedViaLink[i] == false) {
            return false;
         }
      }
      return true;
   }


      /**Checks the nodeVector and link Vector to determine if any Node has
       * more that two Links attached.  If that is true, the topology is
       * a routed network, not an Ethernet bus or a token ring.*/
      private static boolean MoreThanTwoLinks (  ) {
          int N = nodeVector.size () ;
          // create an array
          int linkCount [ ] = new int [ N  ] ;
          for ( int i = 0 ; i < N ; i++ ) {
             linkCount [ i ] = 0 ;
          }  // end of for
          for ( int i = 0 ; i < linkVector.size ( ) ; i++ ) {
             temporaryLink = linkVector.get ( i ) ;
             linkCount [ temporaryLink.startNodeNumber ]
                = linkCount [ temporaryLink.startNodeNumber ] + 1 ;
             linkCount [ temporaryLink.endNodeNumber ]
                = linkCount [ temporaryLink.endNodeNumber ] + 1 ;
             if ( linkCount [ temporaryLink.startNodeNumber ] > 2
                | linkCount [ temporaryLink.endNodeNumber ] > 2 ) {
                return true ;
             }  // end of if
          }  // end of for
          return false ;
       } // end of method private boolean MoreThanTwoLinks (  )


   /**Paint the Nodes and Links*/
   public void paint ( Graphics g ) {
      super.paint ( g );
      java.awt.Graphics2D g2d = ( java.awt.Graphics2D ) g;
////      java.awt.BasicStroke basicStroke = ( java.awt.BasicStroke ) g2d.getStroke ( ) ;
////System.out.println (" Stroke width = " + ( basicStroke . getLineWidth() ) )  ;
      g2d.setColor ( Color.BLUE );
      g2d.drawString ( "Network Type::"
         + ModelData.GlobalVariables.networkType.toString ( ) , 10 , 40  ) ;
      for ( int i = 0 ; i < nodeVector.size () ; i ++ ) {
////System.err.println ( "In paint 1 " ) ;
         position = nodeVector.get ( i );
         int xPos = position[ 0];
         int yPos = position[ 1];
         //System.err.println ("Drawing Node of radius " + nodeRadius ) ;
         g2d.fillOval ( xPos - nodeRadius , yPos - nodeRadius , 2 * nodeRadius ,
            2 * nodeRadius );
         g2d.drawString ( "Node " + i , xPos + nodeRadius , yPos - nodeRadius );
      }  // end of for ( int i = 0 ; i < vectorOfNodes.size ( ) ; i++ )
      // Add the links
      g2d.setColor ( Color.BLACK );
      g2d.setStroke ( new java.awt.BasicStroke ( ( float ) 5.0 ) );
      for ( int i = 0 ; i < linkVector.size () ; i ++ ) {
//// System.err.println ( "In paint 2 " ) ;
         temporaryLink =  linkVector.get ( i );
         int sourceNodeNumber = temporaryLink.getStartNodeNumber ();
         int destinationNodeNumber = temporaryLink.getEndNodeNumber ();
         position =  nodeVector.get ( sourceNodeNumber );
         int startXPos = position[ 0];
         int startYPos = position[ 1];
         position =  nodeVector.get ( destinationNodeNumber );
         int endXPos = position[ 0];
         int endYPos = position[ 1];
         g2d.drawLine ( startXPos , startYPos , endXPos , endYPos );
      // Make the line thicker based on the slope
      }  // end of for ( int i = 0 ; i < vectorOfNodes.size ( ) ; i++ )
   }  // end of method paint


   /**A private class to handle mouse clicks.  This class inherits
    * java.awt.event.MouseAdapter which implements the standard means of
    * handling mouse clicks.  This class overrides the methods specific
    * to this GUI.
   <p>This method only records the point's X and Y position, which Node or
   Link the click occurred on, if any, and whether the right button
   was clicked.  If a right-click occurs on a Node, a graphicalLink is
   created, though the Link may or not be added to the topology.*/
   private static class MouseClickAdapter extends java.awt.event.MouseAdapter {
      // Handle a mouse pressed
      public void mousePressed ( java.awt.event.MouseEvent event ) {
////System.err.println (" The mouse has been pressed."  ) ;
         RESET ();
         XPos = event.getX ();
         YPos = event.getY ();
////System.err.println (" Before FindNodeWherePressed ."  ) ;
         nodeWherePressed = FindNodeWherePressed ( XPos , YPos );
////System.err.println (" Before FindLinkWherePressed ."  ) ;
         linkWherePressed = FindLinkWherePressed ( XPos , YPos );
////System.err.println (" mainGUIInitialWidthInPixels = " + mainGUIInitialWidthInPixels
////   + " mainGUIInitialHeightInPixels = " + mainGUIInitialHeightInPixels  ) ;
         if (event.isMetaDown () == true) {
            isRightButtonClicked = true;
         }  // end of if
         // Create a temporary GraphicalLink if this begins on a node
         // and the right button is clicked. This only creates the
         // starting position of the GraphicalLink: The end is added in
         // mouseReleased
         if (nodeWherePressed != MINUS_ONE & isRightButtonClicked == true) {
            temporaryLink = new GraphicalLink ( nodeWherePressed );
         }  // end of if
//// System.err.println ( "RESULTS: The mouse has been pressed ENTERED DATA: XPos = " +
////XPos + " YPos = " + YPos + " nodeWherePressed = " + nodeWherePressed +
//// " linkWherePressed = " + linkWherePressed +
////" isRightButtonClicked = " + isRightButtonClicked );
      }  // end of mousePressed

      public void mouseReleased ( java.awt.event.MouseEvent event ) {
////System.err.println ( "The mouse has been released." );
         // If the mouse was NOT dragged or the mouse dragging
         // did NOT start on a Node, return and do nothing
         if (isDraggingTakingPlace == true & nodeWherePressed != MINUS_ONE) {
            // DO NOT RESET since this is a mouseReleased and mouseClick
            // is to be serviced after this
/////System.err.println ( "in mouseReleased: Should not do anything." );
            // Get the present position where the mouse was released
            XPos = event.getX ();
            YPos = event.getY ();
//// System.err.println ( "in mouseReleased: Before calling FindNodeWherePressed." );
            int nodeWhereReleased = FindNodeWherePressed ( XPos , YPos );
//// System.err.println ( "in mouseReleased: After calling FindNodeWherePressed." );
            //If the Node was dragged and released on free space
            //move the Node
//// System.err.println ( "in mouseReleased: nodeWhereReleased = " + nodeWhereReleased
////    + " isRightButtonClicked = " + isRightButtonClicked );
            if (( nodeWhereReleased == MINUS_ONE | nodeWhereReleased ==
               nodeWherePressed ) & isRightButtonClicked == false) {
               // The Node is being moved
               position = new int[ 2 ];
               position[ 0 ] = XPos;
               position[ 1 ] = YPos;
               // Move the position of the Node to data
////System.err.println ( " Setting Node at  XPos = " + XPos + " , " + YPos  );
               nodeVector.setElementAt ( position , nodeWherePressed );
               // UpdateGraphicalVariables the data in the Links attached to the Node
               for ( int i = 0 ; i < linkVector.size () ; i ++ ) {
////System.err.println ( "In mouseReleased 1 " ) ;
                  temporaryLink = linkVector.get ( i );
                  // update he value of the denominator squared
                  temporaryLink.UpdateGraphicalVariables ();
               }  // end of for ( int i = 0 ; i < linkVector.size () ; i ++ )
            }  // end of if ( nodeWhereReleased == MINUS_ONE )
            else {  // The drag began on a Node so this creates a GraphicalLink
               // if the right button was clicked AND the GraphicalLink ends on
               // another Node
               if (isRightButtonClicked == true & nodeWhereReleased != MINUS_ONE) {
                  // Set where the GraphicalLink ends
                  temporaryLink.setEndNodeNumber ( nodeWhereReleased );
                  // Check whether the GraphicalLink already exists
                  if (LinkAlreadyInLinkVector ( temporaryLink ) == false) {
                     linkVector.add ( temporaryLink );
                  }  // end of if
               }  // end of if ( isRightButtonClicked == true )
            }  // end of else
         }  // end of if (isDraggingTakingPlace == true & nodeWherePressed != MINUS_ONE)
         topologyDisplay.repaint ();
      } // end of public void mouseReleased

      // Handle a mouse click
      public void mouseClicked ( java.awt.event.MouseEvent event ) {
         int clickCount = event.getClickCount ();
////System.err.println ( "IN mouseClicked: servicing  clickCount = " + event.getClickCount () );
         if (clickCount == 1) {  // For one click
            // If this is NOT on a Node and it is NOT on a GraphicalLink
            // and it is NOT a right click add a Node.  A single-click on
            // a Link WILL NOT create a new Node.
            if (nodeWherePressed == MINUS_ONE & linkWherePressed == MINUS_ONE &
               isRightButtonClicked == false) {
               // The click is for a new Node
               position = new int[ 2 ];
               position[ 0] = XPos;
               position[ 1] = YPos;
////System.err.println ( " adding Node " + nodeVector.size () +
////" at position " + XPos + " , " + YPos );
               nodeVector.add ( position );
            }  // end of if (nodeWherePressed == MINUS_ONE & isRightButtonClicked == false)
            else if (linkWherePressed != MINUS_ONE & isRightButtonClicked ==
               true) {
////               System.err.println ( "OPEN GraphicalLink " + linkWherePressed +
////                  " display for input" );
////System.err.println ( "In mouseClicked linkWherePressed = " + linkWherePressed
////   + " sRightButtonClicked = " + isRightButtonClicked ) ;
////System.err.println ( "In mouseClicked 1 " ) ;
               GraphicalLink link = linkVector.get ( linkWherePressed );
               GUIs.InputGUI inputGUI = GUIs.InputGUI.getInstance (
                  new GUIData.Link_Characteristics () );
            }  // end of if-else
            else { // Otherwise the click is on free-space, on a GraphicalLink or
               // is a right-click, so ignore
               return;
            }  // end of else
         }  // end of if (clickCount == 1)
         else if (clickCount == 2) { // click count = 2
            // Ignore multiple right-clicks
            if (isRightButtonClicked == true) {
               return;
            }
            // For second click, remove a Node or a GraphicalLink
            // First check if to remove a Node
            if (nodeWherePressed != MINUS_ONE) {
               ////System.err.println ("In mouseClicked: removing all Links attached to node = " + nodeWherePressed ) ;
               // Remove the Links attached to Node nodeWherePressed before
               // removing the Node.  This is necessary since the Nodes will be
               // renumbered when the Node is removed.
////System.err.println ("\n\nDELETING NODE " + nodeWherePressed  ) ;
               DeleteNodeAndLinksAttachedToNode ( nodeWherePressed );
               ////             System.err.println ("\n\nIn mouseClicked: REMOVING NODE = " + nodeWherePressed ) ;
 ////              // Now remove the Node which will renumber the Nodes
 ////              nodeVector.remove ( nodeWherePressed );
////  System.err.println ("In mouseClicked: After removing all Links attached to node = " + nodeWherePressed
////     + "\n number of nodes = " + nodeVector . size ( ) + " number of links = " + linkVector . size ( )) ;
            }  // end of if-else
            else if (linkWherePressed != MINUS_ONE) {
               linkVector.remove ( linkWherePressed );
            }  // end of if
         // Second check if to remove a Node
         } // end of click count = 2
         else {  // more that 2 clicks
            return;
         }
         topologyDisplay.repaint () ;
      }  // end of public void mouseClicked


      /**Finds a Node, if any, that covers position xPosIn, xPosIn.  The method
       * returns the <u>first</u> Node that is under the click, not all the
       * Nodes.  If no Node is found, the method returns a MINUS_ONE.*/
      public static int FindNodeWherePressed ( int XPosIn , int YPosIn ) {
         for ( int i = 0 ; i < nodeVector.size ( ) ; i ++ ) {
            position = nodeVector.get ( i );
            int xDel = XPosIn - position[ 0 ];
            int yDel = YPosIn - position[ 1 ];
            // Is the click on this Node?
            if (xDel * xDel + yDel * yDel <= nodeRadius * nodeRadius) {
               return i;
            }  // end of if ( xDel * xDel + yDel * yDel <= nodeRadius * nodeRadius )
         }  // end of for ( int i = 0 ; i < nodeVector.size () ; i ++ )
         // There is no such Node
         return MINUS_ONE;
      }  // end of FindNodeWherePressed


      /**Searches the Vector of Links to find the GraphicalLink which is
       * within distance sqrt( maximumDistanceSquared ) of the mouse
       * click.  It returns a int which is the number of the
       * <u>first</u> GraphicalLink meeting the criteria.  If there
       * is no GraphicalLink within distance sqrt ( maximumDistanceSquared ),
       * the method returns MINUS_ONE.
       <p>FindLinkWherePressed first puts an imaginary rectangle around
       * the Link with the Link as the diagonal.  It first tests whether
       * the mouse click is within the box.  If if it, FindLinkWherePressed
       * computes the distance between the mouse click and the Link.  If the
       * distance is less than sqrt ( maximumDistanceSquared ), FindLinkWherePressed
       * return the Link number.<b>NOTE</b> that FindLinkWherePressed
       * returns the link number of the <u>first</u> GraphicalLink that
       * is within distance.  It is EXTREMELY unlikely that two Links will be
       * sufficiently close to the mouse click: This can ONLY happen at a Node.*/
      public int FindLinkWherePressed ( int XPosIn , int YPosIn ) {
////System.err.println ( "In FindLinkWherePressed  1 " ) ;
////System.err.println ( "\nIn FindLinkWherePressed at: XPos = " + XPosIn  + " YPos = " + YPosIn
////   + " linkVector.size () = " + linkVector.size ()) ;
////         int distance = Integer.MAX_VALUE;
////         int savedIndex = MINUS_ONE;
////System.err.println ( "In FindLinkWherePressed listing Links "  ) ;
////         for ( int i = 0 ; i < linkVector.size () ; i ++ ) {
////            temporaryLink = ( GraphicalLink ) linkVector.get ( i );
////System.err.println ( " Link " + i + "\n" + temporaryLink.toString ( )  ) ;
////System.err.println ( "In FindLinkWherePressed: linkVector.size () = " + linkVector.size () ) ;
         for ( int i = 0 ; i < linkVector.size () ; i ++ ) {
            temporaryLink = linkVector . get ( i );
////System.err.println ( "In FindLinkWherePressed i = " + i + /*"  XPos = " +  XPosIn + " YPos = "
////+  YPosIn + */" from node " + temporaryLink.getStartNodeNumber () + " to node "
////   + temporaryLink.getEndNodeNumber () + " XMin = " + temporaryLink.XMin
////+ " XMax = " + temporaryLink.XMax  + " YMin = " + temporaryLink.YMin + " YMax = " + temporaryLink.YMax) ;
            if ( // Is ( XPos , YPos ) inside the square formed
               // by the Link as the diagonal of the rectangle
                 XPosIn <= temporaryLink.XMax + maximumDistance
               & YPosIn <= temporaryLink.YMax + maximumDistance
               & XPosIn >= temporaryLink.XMin - maximumDistance
               & YPosIn >= temporaryLink.YMin - maximumDistance ) {
               position =  nodeVector.get ( temporaryLink.getStartNodeNumber ( ) );
               int startXPos = position [ 0 ];
               int startYPos = position [ 1 ];
               long numerator = temporaryLink . deltaX * ( startYPos -
                  YPosIn ) - temporaryLink . deltaY * ( startXPos - XPosIn );
 ////              long numeratorSquare = numerator * numerator;
               long distSquared = numerator * numerator / temporaryLink.getDSquare ();
////float realDistance =  ( float ) java.lang.Math.sqrt ( ( double )distSquared ) ;
////System.err.println ("BEFORE TEST: temporaryLink.deltaX= " + temporaryLink . deltaX
////+ " temporaryLink.deltaY  = " + temporaryLink . deltaY  + " numeratorSquare = " + numeratorSquare
////   + " temporaryLink.getDSquare () = " + temporaryLink.getDSquare ()
////   + " saved distance = " + distance + " distSquared = " + distSquared
////   + " realDistance = " + realDistance )  ;
               // If the distance squared < maximumDistanceSquared return i
               if ( distSquared < maximumDistanceSquared ) {
                  return i ;
               }  // end of if
            }  // end of if ( XPosIn <= temporaryLink.XMax + maximumDistance ...
         }  // end of for ( int i = 0 ; i < linkVector . size ( ) ; i++ )
         // A GraphicalLink has been found close enough
         return MINUS_ONE ;
      }


      /**Checks whether the GraphicalLink is already in the linkVector by
       * comparing the start and end node numbers.*/
      private boolean LinkAlreadyInLinkVector ( GraphicalLink linkIn ) {
         int inDestinationNode = linkIn.getEndNodeNumber ();
         int inSourceNode = linkIn.getStartNodeNumber ();
         for ( int i = 0 ; i < linkVector.size () ; i ++ ) {
            GraphicalLink link = linkVector.get ( i );
            int destinationNode = link.getEndNodeNumber ();
            int sourceNode = link.getStartNodeNumber ();
            if (( inDestinationNode == destinationNode & inSourceNode ==
               sourceNode ) | ( inDestinationNode == sourceNode &
               inSourceNode == destinationNode )) {
               return true;
            }  // end of if
         }  // end of for ( int i = 0 ; i < linkVector . size ( ) ; i++ )
         return false;
      }  // end of method LinkAlreadyInLinkVector

      /**Deletes the Links attached to the node defined by
       * nodeWherePressedIn.  It then rededfines the node numbers
       * so that the Links point to the correct Node.*/
      private void DeleteNodeAndLinksAttachedToNode ( int nodeWherePressedIn ) {
////System.err.println ( "\n\nBEGINNING DeleteNodeAndLinksAttachedToNode listing " + linkVector . size ( ) + " links "  ) ;
////          for ( int i = 0 ; i < linkVector.size () ; i ++ ) {
////             temporaryLink = ( GraphicalLink ) linkVector.get ( i );
//// System.err.println ( "\nLink " + i + "\n" + temporaryLink.toString ( )  ) ;   }
         if (nodeWherePressedIn == MINUS_ONE) {
            return;
         }
/* System.err.println ( "" ) ;
for ( int i = 0 ; i < linkVector.size () ; i++ ) {
temporaryLink =  linkVector.get ( i );
System.err.println ( "In DeleteLinksAttachedToNode i = " + i + " from node " + temporaryLink.getStartNodeNumber () + " to node "
+ temporaryLink.getEndNodeNumber () + " XMin = " + temporaryLink.XMin
+ " XMax = " + temporaryLink.XMax  + " YMin = " + temporaryLink.YMin + " YMax = " + temporaryLink.YMax) ;
} */
         for ( int i = linkVector.size () - 1 ; 0 <= i ; i -- ) {
            ////         for ( int i = 0 ; i <  linkVector.size () ; i++ ) {
            ////System.err.println ( "In DeleteNodeAndLinksAttachedToNode  1 " ) ;
            temporaryLink =  linkVector.get ( i );
            if (temporaryLink.getEndNodeNumber () == nodeWherePressedIn |
               temporaryLink.getStartNodeNumber () == nodeWherePressedIn) {
////System.err.println ( "Deleting link between Node  " + temporaryLink.getEndNodeNumber () + " and Node " + temporaryLink.getStartNodeNumber () );
               // Remove this GraphicalLink
               linkVector.remove ( i );
            }  // end of if
         }  // end of for ( int i = linkVector . size ( ) - 1 ; 0 <= i ; i-- ) {
         // Remove the Node
         nodeVector.remove ( nodeWherePressedIn ) ;
/*System.err.println ( "" ) ;
for ( int i = 0 ; i < linkVector.size () ; i++ ) {
temporaryLink =  linkVector.get ( i );
System.err.println ( "In DeleteLinksAttachedToNode i = " + i + " from node " + temporaryLink.getStartNodeNumber () + " to node "
+ temporaryLink.getEndNodeNumber () + " XMin = " + temporaryLink.XMin
+ " XMax = " + temporaryLink.XMax  + " YMin = " + temporaryLink.YMin + " YMax = " + temporaryLink.YMax) ;
}  */
         // Since the Node nodeWherePressedIn has been removed, all the nodes
         // with node number > nodeWherePressedIn must have their node numbers
         // decreased by 1.  The
         // nodeVector automatically does this for the Nodes at
         // nodeVector.remove ( nodeWherePressedIn ) above.  But it is
         // necessary to decrease the node numbers saved in all the Links.
         // BOTH the start and end number MUST be updated simultaneously.  If
         // not, using setStartNodeNumber or setEndNodeNumber may result in
         // a null pointer exception.
         for ( int i = 0 ; i < linkVector.size () ; i ++ ) {
            ////System.err.println ( "In DeleteNodeAndLinksAttachedToNode  2 " ) ;
            temporaryLink =  linkVector.get ( i );
            boolean changed = false ;
            int startNumber = temporaryLink.getStartNodeNumber ();
            int endNumber = temporaryLink.getEndNodeNumber ();
            if ( startNumber >= nodeWherePressedIn) {
               startNumber = startNumber - 1 ;
               changed = true ;
            }  // end of if ( startNumber >= nodeWherePressedIn)
            if ( endNumber >= nodeWherePressedIn) {
               endNumber = endNumber - 1 ;
               changed = true ;
            }  // end of if ( endNumber >= nodeWherePressedIn)
         if ( changed == true ) { temporaryLink
            . setStartNodeNumberAndEndNodeNumber ( startNumber , endNumber );
            }  // end of if ( changed == true )
         }  // end of for ( int j = 0 ; j < linkVector. size ( ) ; j++ )
/* System.err.println ( "" ) ;
 for ( int i = 0 ; i < linkVector.size () ; i++ ) {
temporaryLink =  linkVector.get ( i );
System.err.println ( "In DeleteLinksAttachedToNode i = " + i + " from node " + temporaryLink.getStartNodeNumber () + " to node "
+ temporaryLink.getEndNodeNumber () + " XMin = " + temporaryLink.XMin
+ " XMax = " + temporaryLink.XMax  + " YMin = " + temporaryLink.YMin + " YMax = " + temporaryLink.YMax) ;
} */
      }// end of DeleteNodeAndLinksAttachedToNode
   }  // end of private class MouseClickAdapter


   /**A private class to handle mouse motion.  This class inherits
    * java.awt.event.MouseAdapter which implements the standard means of
    * handling mouse clicks.  this class overrides the methods specific
    * to this GUI.*/
   private static class MouseMotionAdapter
      extends java.awt.event.MouseMotionAdapter {
      public void mouseDragged ( java.awt.event.MouseEvent event ) {
////System.err.println ( "In mouse is being dragged. "  );
         // If dragging has alread started, ignore the continuation
         // until the mouse is released
         isDraggingTakingPlace = true;
      }  // end of mouseDragged
   }  // end of private class MouseMotionAdapter


   /**Resets the mouse variables and booleans*/
   public static void RESET () {
      nodeWherePressed = nodeWhereReleased = MINUS_ONE;
      linkWherePressed = MINUS_ONE;
      XPos = YPos = MINUS_ONE;
      isDraggingTakingPlace = false;
      isRightButtonClicked = false;
      temporaryLink = null;
      position = null;
   }


   /**A private class to reate a CommonCloseWindowOperation( ) for the
    * topology and network windows.  It will simply makes the display
    * window invisible and deletes any topology.
    * @version 1.100
    * @author dixonw*/
   private static class MyGraphicalWindowAdapter
      extends java.awt.event.WindowAdapter {
      public void windowClosing ( java.awt.event.WindowEvent e ) {
         topologyDisplay.setVisible ( false );
         nodeVector.clear ();
         linkVector.clear ();
      }  // end of windowClosing
   } // end of private class MyWindowAdapter

   /**A private class to create a Component Adapter to redraw the
    * GUI when it is resized.
    * @version 1.100
    * @author dixonw*/
   private static class MyComponentAdapter
      extends java.awt.event.ComponentAdapter {
      // When the window is resized, repaint it.
      public void componentResized ( java.awt.event.ComponentEvent e ) {
         topologyDisplay.repaint ();
      }  // end of method componentResized
   }  // end of class MyComponentAdapter

}  // end of class

