/*
 * DiscreteEventSimulator.java
 *
 * Created on February 6, 2007, 4:55 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package ModelDES;

import Utility.JNWRandom;
import Utility.out;

/**
 * The discrete event simulator is a double linked list of events.  Events are stored firstly
 * by the time of the event and secondly, for events occurring at the same time, by the order
 * that they were added to the simulator.
 *
 * <p>Users can add events to the simulator using the method addEvent.  The event will be
 * added to the vectorInTime
 * linked list according to the event time.  It there are no events in the simulator at the
 * time of the new event, a new ModelDatas.ListOfEvents is created.  Each ModelDatas.ListOfEvents
 * maintains a list of Events, all of which occur
 * at time ModelDatas.ListOfEvents.timeOfEventsList.
 *
 * The StartModel method will execute the events in temporal and then entry order.  JNW
 * ends when all the events have been executed.
 *
 * <p><b>Versions:</b>
 * <li><b>Version 1.001</b> of 20071113 adds the method
 * reorderEvent (int eventSequenceNumber  , long newEventTime )</li>
 * <li><b>Version 1.002</b> of 20071118 adds the
 * method RestoreEventAtNewEventTime ( long newEventTime
 * , ModelEvents.ModelEventsBaseClass event  ) which
 * restores the Event to the DES at the newEventTime.  I had forgotten
 * that when you call reorderEvent from
 * within ExecuteEvent() the Event it is no long in the DES, so reorderEvent
 * will fail.  The Event is no longer in the DES since it had to be removed to
 * execute ExecuteEvent()</li>
 * <li><b>Version 1.200</b> of 20081019 added return of boolean from
 * deleteEvent.  The boolean is used in
 * ModelTransport.TCPState.DeleteTransmittedPacketInformation.</li>
 * @author dixonw
 * @version 1.202
 * @modified JMP 28Oct11 for HDLC
 */
public class DiscreteEventSimulator implements GUIConstants.General {

   /** Vector of events in time.  Each entry corresponds to a different
    * simulation time.  The vector is maintained in increasing
    * order of time with vectorInTime [0] being the most recent
    * time.  The Vector will contain ListOfEvents.    ListOfEvents
    *contains the time of theModelDatas.Events and a list
    * of the ModelDatas.Events*/
   static//???
      java.util.Vector<ModelDES.ListOfEvents> vectorInTime =
      new java.util.Vector<ModelDES.ListOfEvents> ( 100 );

   public static int getVectorSize(){return vectorInTime.size();}//???

   /**A singleton vector containing all the Nodes
      Doesn't belong in ModelDES but JNW has no way to
      input a sequence of emails to the simulator so
      I'm forced to provide reference to the nodes here*/
   static ModelData.JNWSingletonVectorOfNodes<ModelPhysical.Node> vectorOfNodes
       = ModelData.JNWSingletonVectorOfNodes.getInstance (  ) ;

   /**The singleton of a DiscreteEventSimulator*/
   private static final DiscreteEventSimulator singletonDiscreteEventSimulator =
      new DiscreteEventSimulator ();


   /**Get a reference to the network display.*/
   private static GUIGraphical.NetworkDisplay_RealTime networkDisplay
      = GUIGraphical.NetworkDisplay_RealTime.getInstance ( ) ;

   /**The clock time at the start of the simulation.  Used to computer the
    * time neede by the computer to perform the calculations.*/
         long startSimulationTime = 0;


    /** Decimal format for printing floating point numbers.  The format
     * is X.XXXX.  This field was originally instantiated in class
     * GUIConstants.DecimalFormats instantiated here because the class was
     causing the JavaDoc generator to terminate with a null pointer
     exception. */
         java.text.DecimalFormat decimalFormatFourDigits
      = new java.text.DecimalFormat ( "0.0000" ) ;

   /**Creates a new instance of DiscreteEventSimulator.  Starts by putting one
    *dummy event into the DES.  This way, it is not necessary to test each
    *Event to determine if it is the first Event.*/
   private DiscreteEventSimulator () {
      //  Get a referenced to the network display
////      networkDisplay = GUIGraphical.NetworkDisplay_RealTime.getInstance ( ) ;
      // But make it invisible
////      networkDisplay. setVisible ( false );
   }

   /**Adds an event to the discrete event simulator.  First, it searches the
    *vectorInTime.  If a link list is found which has timeOfEventsList the same
    *as the eventTime of the event being added, the event is simply added to the
    *existing linked list.  If the eventTime of the event being added is not
    * one of the timeOfEventsLists, then a new ListOfEvents is added to the
    * vectorInTime in its proper location, ordered by time.
    *<p>If eventTime >= ModelDatas.GlobalVariables.maximumDESTicks the
    * event will not be added to the events list and the model will stop when
    * the list is exhausted.*/
   public void addEvent ( ModelEvents.ModelEventsBaseClass event ) {
      long eventTime = event.getEventStartTime ();
      if (eventTime > ModelData.GlobalVariables.maximumDESTicks)  
         return;
      // Is the vector empty?
      if (vectorInTime.size () == 0) {
         vectorInTime.add ( new ModelDES.ListOfEvents ( event ) );
      }
      else {  // else 456
         long listTime = 0;  // Get the first events list
         ModelDES.ListOfEvents timeEventsList = null;
         // loop over the times in the Vector
         synchronized(vectorInTime) {
             for ( int i = 0 ; i < vectorInTime.size () ; i ++ ) {
                 timeEventsList = vectorInTime.get ( i );
                 listTime = timeEventsList.getTimeOfEventsList ();
                 if (eventTime < listTime) {
                     // Add a new events list
                     vectorInTime.add (
                         i ,
                         new ModelDES.ListOfEvents ( event ) );
                     break;
                 }  // end of if
                 else if (eventTime == listTime) {
                     timeEventsList.add ( event );  // just add an event to an existing list
                     break;
                 } // end of else if
             }  // end of for ( int i = 0 ; i < vectorInTime . size ( ) ; i++ )
             if (eventTime > listTime) {
                // Add a new ModelDatas.ListOfEvents to the vectorInTime
                vectorInTime.add ( new ModelDES.ListOfEvents ( event ) );
             } // end of if ( eventTime > listTime )
         }//end synchronized
      }  // end of else 456
      /**Permission to print data errors*/
      if (ModelData.PrintPermissions.printEventsInDES == true) {
         Utility.out.println ( "IN DiscreteEventSimulator added : " + event.
            toString () + CARRIAGE_RETURN + "\nComplete DES = :\n" + this.
            toString () );
      } // end of if ;
   }  // end of addEvent ( ModelDatas.ModelEventsBaseClass event )


   /**Deletes the event with number event sequence number from the
    * vectorInTime.  Searches each in the vectorInTime until it finds
    * the sequence number.  The event with this sequence number is deleted
    * from the appropriate eventsList.  If the events list is emptied
    * by removing the event from the list, the list is removed from the
    * vectorInTime.  Therefore, no empty lists will remain in
    * the vectorInTime.  If the sequence number is not found in the
    * vectorInTime, an error message is printed.*/
   public void deleteEvent ( int eventSequenceNumber ) {  
      boolean deleted = false;
      synchronized(vectorInTime) {
          for ( int i = 0 ; i < vectorInTime.size () ; i ++ ) {
              // Get the ith eventsList
              ModelDES.ListOfEvents eventsList = vectorInTime.elementAt ( i );
              // If possible, remove the event
              if (eventsList.deleteEvent ( eventSequenceNumber ) == true) {
                  deleted = true;
                  if (eventsList.getListSize () == 0) {  // Is the list now empty?
                  // If so, remove this linked list  from the vectorInTime
                      vectorInTime.remove ( i );
                  }  // end of if ( eventsList . getListSize ( ) == 0 )
                  break;
              }  // end of if ( eventsList . deleteEvent ( eventSequenceNumber ) == true )
          }  // end of  for ( int i = 0 ; i < vectorInTime . size ( ) ; i++ )
      }// end synchronized
      if (deleted == false) {
         Utility.war.print ( "WARNING in Model.DiscreteEventSimulator.deleteEvent( ): " +
            "Unable to remove event number " + eventSequenceNumber +
            CARRIAGE_RETURN );
      }  // end of if
   }  // end of DeleteEvent ( int eventSequenceNumber )


   /**Returns simulation time of currently executing event*/// JMP
   public static long getSimulationTime() {
       if(vectorInTime.size() > 0)
         return vectorInTime.firstElement().get(0).getEventStartTime();
       return ModelData.GlobalVariables.simulationTime;
   }


   /**This starts the model.  This gets and saves
    * the system time the system time.  It prints the date and time, and it calls
    * ExecuteEvents (), which ectually executes the Events.*/
   public void StartModel () {
      Utility.out.print ( "  STARTING Network Workbench ExecuteEvents ( ) "
         + ( new java.util.Date () ).toString () + CARRIAGE_RETURN
         + "      with random number seed "
         + JNWRandom.getInstance ().seedForJNWRandomNumberGenerator
         + DOUBLE_CARRIAGE_RETURN );
      ModelPhysical.Node sourceNode =
          vectorOfNodes.get(
              GUIActions.ReadEMailTextFilesAndPutIntoSourceNodeBufferAction.
                  tcpEmailSourceNode);
      startSimulationTime = System.currentTimeMillis ();

      // Endless loop
      while ( vectorInTime.size () > 0 ) {
         // If the model has been paused, stop
         if ( ModelData.GlobalVariables.modelPaused == true ) {return ; }
         ModelDES.ListOfEvents timeEventList;
         ModelEvents.ModelEventsBaseClass event;
         synchronized(vectorInTime) {
             timeEventList = vectorInTime.firstElement ();
             // Remove the head
             event = timeEventList.getHead ();
             if (timeEventList.isEmpty ( ) ) {
                 vectorInTime.remove ( 0 );
             }  // end of if ( timeEventList . isEmpty ( ))
         }//end synchronized
         if (ModelData.PrintPermissions.printEventsExecution == true) {
            Utility.out.println ( "IN DiscreteEventSimulator.StartModel executing" +
               " time " + ModelData.GlobalVariables.simulationTime + " " +
               event.toString () );
         } // end of if
         // Execute the event using polymorphism
         event.ExecuteEvent ();
         Thread.yield();//BZ

         // check for an email completed; start a new one
         if(sourceNode.thisTCPEmailIsDone) {
             vectorInTime =
                 new java.util.Vector<ModelDES.ListOfEvents> ( 100 );
             sourceNode.loadNextEmail();
             sourceNode.thisTCPEmailIsDone = false;
         }
         if ( networkDisplay.isNetworkDisplayCurrentlyOpen == true ) {
////            networkDisplay.repaint ( ) ;
         }
      }  // end of while ( vectorInTime .size() > 0 )
      double executionTime = ( ( double ) System.currentTimeMillis () -
         startSimulationTime ) / 1000.0;
      Utility.out.print ( "\nSIMULATION ENDS at simulation time = " +
         ModelData.GlobalVariables.simulationTime + "\nModel execution time " +
         decimalFormatFourDigits.format ( executionTime ) + " seconds." +
         DOUBLE_CARRIAGE_RETURN );
      // If the model has run to the finish reset the pauseAction
      //  and resumeAction buttons
      // Create a pseudo MainGUI, so that one can reference the public
      // static fields pauseAction and resumeAction.
      // This ia NOT a good way to do this, but it works.
      // If the model is paused, DO NOT reset the fields
      GUIs.MainGUI mainGUI = new GUIs.MainGUI ( 1 );
      mainGUI . pauseAction . setEnabled ( false );
      mainGUI . resumeAction . setEnabled ( false  );
      mainGUI . runModelAction . setEnabled ( true  );
      // Release the memory
      mainGUI = null ;
      // Reset the modelPaused field
      ModelData.GlobalVariables.modelPaused = false ;
   }  // end of StartModel ( )



   /**Resume executing the events*/
    @SuppressWarnings("static-access")
   public void ExecuteEvents () {
      // Endless loop
      while ( vectorInTime.size () > 0 ) {
         // If the model has been paused, stop
         if ( ModelData.GlobalVariables.modelPaused == true ) { return ; }
         ModelDES.ListOfEvents timeEventList;
         ModelEvents.ModelEventsBaseClass event;
         synchronized(vectorInTime) {
             timeEventList = vectorInTime.firstElement ();
             // Remove the head
             event = timeEventList.getHead ();
             if (timeEventList.isEmpty ( ) ) {
                 vectorInTime.remove ( 0 );
             }  // end of if ( timeEventList . isEmpty ( ))
         }//end synchronized
         if (ModelData.PrintPermissions.printEventsExecution == true) {
            Utility.out.println ( "IN DiscreteEventSimulator.ExecuteEvents executing : " +
            " time " + ModelData.GlobalVariables.simulationTime + " " +
                event.toString () );
         } // end of if
         // Execute the event using polymorphism
         event.ExecuteEvent ();
         Thread.yield();//BZ
         if ( networkDisplay.isNetworkDisplayCurrentlyOpen == true ) {
            networkDisplay.repaint ( ) ;
         }
      }  // end of while ( vectorInTime .size() > 0 )
      double executionTime = ( ( double ) System.currentTimeMillis () -
         startSimulationTime ) / 1000.0;
      Utility.out.print ( "\nSIMULATION ENDS at simulation time = " +
         ModelData.GlobalVariables.simulationTime + "\nModel execution time " +
         decimalFormatFourDigits.format ( executionTime ) + " seconds." +
         DOUBLE_CARRIAGE_RETURN );
      // If the model has run to the finish reset the pauseAction
      //  and resumeAction buttons
      // Create a pseudo MainGUI, so that one can reference the public
      // static fields pauseAction and resumeAction.
      // This ia NOT a good way to do this, but it works.
      // If the model is paused, DO NOT reset the fields
////      if ( ModelData.GlobalVariables.modelPaused == false ) {
      GUIs.MainGUI mainGUI = new GUIs.MainGUI ( 1 );
      mainGUI . pauseAction . setEnabled ( false );
      mainGUI . resumeAction . setEnabled ( false  );
      mainGUI . runModelAction . setEnabled ( true  );
      // Release the memory
      mainGUI = null ;
      // Reset the modelPaused field
      ModelData.GlobalVariables.modelPaused = false ;
   }  // end of StartModel ( )

   /**Executes one event only.*/
   public void ExecuteOneEvent () {
      // Is this the start of the simulation?
      if ( startSimulationTime <= ( long ) 0 ) {
      Utility.out.print ( "     STARTING Network Workbench ExecuteEvents ( ) " +
         ( new java.util.Date () ).toString () + CARRIAGE_RETURN +
         "      with random number seed " +
         JNWRandom.getInstance ().seedForJNWRandomNumberGenerator +
         DOUBLE_CARRIAGE_RETURN );
         startSimulationTime = System.currentTimeMillis ();
      }

      // If you push the single step button and there are no more Events,
      // catch the Exception and just return
      try {
          ModelDES.ListOfEvents timeEventList;
          ModelEvents.ModelEventsBaseClass event;
          synchronized(vectorInTime) {
              timeEventList = vectorInTime.firstElement ();
              // Remove the head
              event = timeEventList.getHead ();
              if (timeEventList.isEmpty ( ) ) {
                  vectorInTime.remove ( 0 );
              }  // end of if ( timeEventList . isEmpty ( ))
          }// end synchronized
          // Execute the event using polymorphism
          event.ExecuteEvent ();
          Thread.yield();//BZ
          if ( networkDisplay.isNetworkDisplayCurrentlyOpen == true ) {
              networkDisplay.repaint ( ) ;
          }
          if (ModelData.PrintPermissions.printEventsExecution == true) {
              Utility.out.println ( "IN DiscreteEventSimulator.ExecuteEvents : " +
                  " time " + ModelData.GlobalVariables.simulationTime + " " +
                  event.toString () );
          } // end of if
      } catch ( java.util.NoSuchElementException e ) { return ; }
      double executionTime = ( ( double ) System.currentTimeMillis () -
         startSimulationTime ) / 1000.0;
      if ( vectorInTime .size( ) == 0 ) Utility.out.print ( "SIMULATION ENDS at simulation time = " +
         ModelData.GlobalVariables.simulationTime + "\nModel execution time " +
         decimalFormatFourDigits.format ( executionTime ) + " seconds." +
         DOUBLE_CARRIAGE_RETURN );
   }  // end of StartModel ( )


   /**A toString method for the DiscreteEventSimulator */
   public String toString () {
      String string = "VECTOR OF LISTS contains " + vectorInTime.size () +
         " lists as follows:\n";
      for ( int i = 0 ; i < vectorInTime.size () ; i ++ ) {
         ModelDES.ListOfEvents eventInList = vectorInTime.elementAt ( i );
         string = string + eventInList.toString ();
      }
      return string;
   }  // end of toString

   /** Returns the one instance of the DiscreteEventSimulator. */
   public static DiscreteEventSimulator getInstance () {
      return singletonDiscreteEventSimulator;
   }

   /**A toString method for the DiscreteEventSimulator */
   public String toStringSequenceNumbers () {
      String string = "VECTOR OF LISTS contains " + vectorInTime.size () +
         " lists as follows:\n";
      for ( int i = 0 ; i < vectorInTime.size () ; i ++ ) {
         ModelDES.ListOfEvents eventInList = vectorInTime.elementAt ( i );
         string = string + eventInList.toStringSequenceNumbers  ();
      }
      return string;
   }  // end of toString

   /**A main method for testing the DiscreteEventSimulator
    * @param args the command line arguments
    */
   public static void main_For_Testing_Only ( String[] args ) {
      GUIData.InputDataSingleton inputData = GUIData.InputDataSingleton.
         getInstance ();
      JNWRandom random = JNWRandom.getInstance ();
      ModelDES.DiscreteEventSimulator discreteEventSimulator =
         new ModelDES.DiscreteEventSimulator ();
      ModelEvents.ModelEventsBaseClass event = null;
      int time2 = - 50;
      for ( int i = 0 ; i < 100 ; i ++ ) {
         time2 = random.nextInt ( 100000 );
         event = new ModelEvents.ModelEventsBaseClass ( time2 );
         discreteEventSimulator.addEvent ( event );
      }
      System.out.println ( "The DES is now: " +
         discreteEventSimulator.toString () );
      for ( int i = 0 ; i < 110 ; i ++ ) {
////         discreteEventSimulator.reorderEvent ( i , 10000 + 100 * i /*random.nextInt ( 100000 ) */ );
      }
      System.out.println ( "The DES is now: " +
         discreteEventSimulator.toString () );

      System.exit ( 0 );
      int bias = 0;
      int stop = 0;
      int startDelete = 0;
      int stopDelete = 0;
      for ( int i = 0 ; i < stop ; i ++ ) {
         long time = random.nextInt ( 50 );
         ModelEvents.ModelEventsBaseClass event2 =
            new ModelEvents.ModelEventsBaseClass ( 5 );
         event2.setEventStartTime ( time );
         discreteEventSimulator.addEvent ( event2 );
      }
      for ( int i = startDelete ; i < stopDelete ; i ++ ) {
         discreteEventSimulator.deleteEvent ( i );
      }
   }  // end of Main

}  // end of class DiscreteEventSimulator

////SAVED METHODS 20080727
////SAVED METHODS 20080727
////SAVED METHODS 20080727
   /**Resets the event time of event eventSequenceNumber to the time
    * newEventTime and puts the Event into the DES in order of the
    * new event time.  The Event is found, the eventTime of the
    * Event is changed, the Event is removed from the DES, and
    * finally it is added back into the DES.
    * <p>This method does <b>NOT</b> change the event
    * sequence number.  It <b>ONLY</b> changes the event time of
    * an Event and reorders the Event in the double linked list.
    * <p><b>THIS METHOD IS NOT USED ANYWHERE AS OF 20080608.  POSSIBLY IT
    * SHOULD BE REMOVED!</b>*/
/*   public void reorderEvent ( int eventSequenceNumber , long newEventTime ) {
      boolean reordered = false;
      for ( int i = 0 ; i < vectorInTime.size () ; i ++ ) {
         // Get the ith eventsList
         ModelDES.ListOfEvents eventsList = vectorInTime.elementAt ( i );
         // Get the linked list in the events list
         java.util.LinkedList<ModelEvents.ModelEventsBaseClass> linkedList =
            eventsList.getLinkedList ();
outer_Loop: for ( int j = 0 ; j < eventsList.getListSize () ; j ++ ) {
               ModelEvents.ModelEventsBaseClass eventToChange =
                  eventsList.get ( j );
               long sequenceNumber = eventToChange.getEventSequenceNumber ();
               if (sequenceNumber == eventSequenceNumber) {
                  eventToChange.setEventStartTime ( newEventTime );
                  eventsList.deleteEvent ( eventSequenceNumber );
                  if (eventsList.getListSize () == 0) {  // Is the list now empty?
                     // If so, remove this linked list  from the vectorInTime
                     vectorInTime.remove ( i );
                  }  // end of if ( eventsList . getListSize ( ) == 0 )
                  reordered = true;
                  // Add the event back on the DES in the new order
                  this.addEvent ( eventToChange );
                  break outer_Loop;
               }  // end of if ( sequenceNumber == eventSequenceNumber )
            }  // end of for ( int i = 0 ; i < linkedList.getListSize ( ) ; i++ )
      }  // end of if ( eventsList . deleteEvent ( eventSequenceNumber ) == true )
      if (reordered == false) {
         Utility.err.println ( "ERROR in Model.DiscreteEventSimulator.reorderEvent( ): " +
            "Unable to reorder event number " + eventSequenceNumber );
      }  // end of if
   }  // end of reorderEvent
*/
   /**Resets the event time of event to the time newEventTime
    * and adds the Event into the DES in order of the new event time.  This
    * method is intended for use with repeated occurring Events
    * like GenerateMessageLoad that are polled from the discrete event
    * simulator and then added back into the DES at a new time.
    * This method can be called from within the ExecuteEvent() method
    * of the event as
    * <li> discreteEventSimulator.RestoreEventAtNewEventTime ( newEventTime , this )
    * ;</li><p>This will cause the event to be reentered in the DES at the new time.
    * The eventSequenceNumber is unchanged by this method.
    * <p>ERROR:  This method is duplicative of
    * ModelEvents.ModelEventsBaseClass.RestoreThisEventAtNewEventTime
    * , and one or the other should be eliminated.</p>
    * <p><b>THIS METHOD IS NOT USED ANYWHERE AS OF 20080608.  POSSIBLY IT
    * SHOULD BE REMOVED!</b>
   @param newEventTime new time at which event is to be stored in simulator
   @param the Event to be stored*/
/*   public void RestoreEventAtNewEventTime ( long newEventTime ,
      ModelEvents.ModelEventsBaseClass event ) {
      event.setEventStartTime ( newEventTime );
      this.addEvent ( event );
   }  // end of reorderEvent
*/