/*
 * JNWVector.java
 *
 * Created on December 21, 2006, 6:34 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelData;

import ModelPhysical.Node ;
import ModelPhysical.Link ;

/**JNWVector overrides the toString() methods of java.util.Vector so 
 * that the Vector can be printed to show the number of elements 
 * in the Vector and the elements themselves.  JNWVector also 
 * provides methods to test whether the Vector already 
 * contains a particular Node or Link.
 *<p><b>Versions:</b>
 *<li><b>Version 1.001</b> of 20071202 removed the checks for the 
 * vector.size()==0 since the iterator.hasNext() which immediately
 * follows accomplishes the same thing.</li>
 * @version 1.100
 * @author dixonw*/
public class JNWVector<E> extends java.util.Vector<E>
   implements GUIConstants.General
{
   
   /**Creates a new instance of JNWVector.*/
   public JNWVector ()
   {
      super ( ) ;
   }
   
   
   /**Return boolean whether the node number is already in the vector*/
   public boolean nodeAleadyInVector ( int nodeNumberIn )
   {  // Create an iterator
      java.util.Iterator iterator = this. iterator ( ) ;
      // If the Vector is empty
      // While the Vector has more elements
      while ( iterator . hasNext ( ) )
      {
         Node node =  ( Node ) iterator . next ( ) ;
         if ( node . thisNodeNumber == nodeNumberIn )
         {
            return true ;
         }
      } // end of  while ( iterator.hasNext ( ))
      // The Node is not in the Vector
      return false ;
   } // end of nodeAleadyInVector
   
   
   /**Return boolean whether the link  is already in the vector.  The
    * method accepts the source and destination node numbers and searches the
    * vectorOfLinks to determine if the Link is already in the Vector.  A Link
    * is already in the Vector if both its source and destination Nodes are equal.
    * <p>If the link is already in the vector, the source 
    *and destination are compared.
    * If the source and destination are reversed the
    * bitRateDestinationToSource is replaced by the new rate.*/
   public boolean linkAleadyInVector ( int sourceNodeNumber 
      , int destinationNodeNumber , int rate )
   {  // Create an iterator
      java.util.Iterator iterator = this. iterator ( ) ;
      // If the Vector is empty
      // While the Vector has more elements
      while ( iterator . hasNext ( ) )
      {
         Link link =  ( Link ) iterator . next ( ) ;
         if (   // If both the source and destinations are the same
            link.destinationNodeNumber == destinationNodeNumber &
            link.sourceNodeNumber == sourceNodeNumber )
         {
            return true ; }
         else if ( // If both the source and destinations are reversed
            link.destinationNodeNumber == sourceNodeNumber &
            link.sourceNodeNumber == destinationNodeNumber   )
         {
            // Overwrite the bitRateDestinationToSource with the new rate
            // This will only happen if the link is entered a second time
            // with probably a different data rate
            link . bitRateDestinationToSource = rate ;
            return true ;
         }
      } // end of  while ( iterator.hasNext ( ))
      // The Link is not in the Vector
      return false ;
   } // end of nodeAleadyInVector
   
   
   /** This toString method returns the Vector as a String with 
    * each object in the vector returned as a separate line.*/
   public String toString ( )
   {  // a generic object
      Object object = null ;
      // Create an iterator
      java.util.Iterator iterator = this. iterator ( ) ;
      // Create a stringBuffer
      java.lang.StringBuffer stringBuffer = new java.lang.StringBuffer ( 1000 ) ;
      stringBuffer . append ( "\nJNWVector " + this . getClass ( ) .getName ( )
      + " has " + this . size ( ) + " entries" ) ;
      // If the Vector is empty
      if ( this.size ( ) == 0 )
      {  // return the almost empty StringBuffer
         return new String ( stringBuffer . append ( CARRIAGE_RETURN )  ) ;
      }
      else
      {  // Find the class of the elements
         object = iterator . next ( ) ;
         stringBuffer . append ( " of class " + object . getClass ( ) . getName ( ) 
            + CARRIAGE_RETURN ) . append ( object . toString ( ) ) ;
  ////       stringBuffer . append ( object . toString ( ) ) ;
      }
      // If the Vector has more elements
      while ( iterator . hasNext ( ) )
      {
         object = iterator . next ( ) ;
         stringBuffer . append ( object . toString ( ) ) ;
      } // end of  while ( iterator.hasNext ( ) )
      return new String ( stringBuffer )  ;
   }  // end of method toString
   

   /**For testing only.*/
   public static void main_Testing_Only ( String [ ] args ) {
      JNWVector vector = new JNWVector ( ) ;
      System.out.println ( " Is the node already in the vector " + vector .nodeAleadyInVector ( 5 ) ) ;
   }
}
