/*
 * Ethernet.java
 *
 * Created on June 8, 2007, 7:59 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelDataLink;

import Utility.*;

/**The Ethernet interface.
 * @version 1.100
 * @author dixonw
 *@see <a href="http://en.wikipedia.org/wiki/Ethernet">Ethernet</a>
 *@see <a href="http://www.cisco.com/univercd/cc/td/doc/cisintwk/ito_doc/ethernet.htm">Ethernet in Excruciation Detail</a>
 *@see <a href="http://www.ieee802.org/3/">IEEE 802.3 ETHERNET WORKING GROUP</a>
 *@see <a href="http://www.erg.abdn.ac.uk/users/gorry/eg3561/lan-pages/csma-cd.html">Carrier Sense Multiple Access with Collision Detection (CSMA/CD)</a>
 *@see <a href="http://www.geocities.com/SiliconValley/Haven/4824/ethernet.html">Ethernet Page</a>
 */
public class Ethernet extends EthernetUtilities {

    /** Transmission attempt*/
    int transmissionAttemptNumber;

    /** Time delay.*/
    long timeDelay;

    /**The Ethernet Bus*/
    ModelPhysical.EthernetBus ethernetBus;
    
    /** Creates a new instance of Ethernet */
    public Ethernet( int nodeNumber ) {
        super (  ) ;
        // Save to number of the node to which this interface is attached.
        this . nodeToWhichAttached = nodeNumber ;     
    }
    
   
 /** This toString method returns the Links as a String with each Ethernet
  * interface in the vectorOfLinks returned as a separate line in the String.*/
    public String toString ( ) {    
          return (  " Interface number = " + this . interfaceNumber + " from Node "
              + this . nodeToWhichAttached  + " of type Ethernet" + CARRIAGE_RETURN ) ;
    }  // end of method toString

   public void sendEthernetFrame (
                   JNWBitSet frameToTransmit,
                   int sourceInterfaceNumber,
                   int destinationInterfaceNumber,
                   ModelEvents.SendFrameFromInterface sendFrameEvent) {

      // get reference to the bus
      ethernetBus = ( ModelPhysical.EthernetBus ) vectorOfLinks.get (0);

      // Increment the attempt number
      // If the transmission number exceeds the limit, give up
      if ( ( transmissionAttemptNumber = transmissionAttemptNumber + 1 )
             > ModelDataLink.EthernetState . maximumNumberOfTransmissionAttempts ) {
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
         Utility.err.println ("\nERROR : on interface "
            + sourceInterfaceNumber + " Ethernet exceeded "
            + ModelDataLink.EthernetState.maximumNumberOfTransmissionAttempts
            + " transmission attempts-Frame discarded");
         // Give up transmission attempts
         return;
      }  // end of if ( transmissionAttemptNumber > ModelDataLink.EthernetState...
      // Compute the time delay until the next transmission attempt
      timeDelay
         = ModelDataLink.EthernetUtilities.ComputeBackoff (transmissionAttemptNumber);

      ModelDataLink.Ethernet interfaceCard
         = ( ModelDataLink.Ethernet ) vectorOfInterfaces . get (sourceInterfaceNumber) ;
      // Does the transmission time occur while receiving another message?
      ////THE STUDENT SHOLD LOOK AT THIS TEST WHICH IS MENTIONED IN EXERCISE JLAN1////
      long eventStartTime = this.getTransmitTime();
      if (eventStartTime >= interfaceCard . startReceiveEthernetFrame &
          eventStartTime <= interfaceCard . stopReceiveEthernetFrame ) {
         // The interface is currently receiving a transmission so try again later
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
            Utility.war.println ("The transmission of the Ethernet message IS BLOCKED"
            + " at time " + eventStartTime + " Will try again at time "
            + ( eventStartTime + timeDelay ));
         // This transmission attempt should not be counted so reverse
         // the first statement in this method
         transmissionAttemptNumber = transmissionAttemptNumber - 1 ;
         // The interface is currently receiving a transmission.
         // The transmission is blocked.  Set the next transmission attempt
         sendFrameEvent.RestoreThisEventAtNewEventTime (eventStartTime + timeDelay);
      }  // end of if ( eventStartTime >= interfaceCard .startReceiveEthernetFrame &

      // START THE TRANSMISSION
      else { // Start the transmission sequence
         // Permission exists to transmit, but if the bus is already in use
         // a collision will occur.
         if (  ethernetBus.busInUse == true  ) {
            // A collision will occur so skip the transmission
            // This is a short-cut.  This message is never sent since you know
            // a priori that there will be a collision
            ethernetBus.collisionHasOccurred = true;
            // Try again at a randomly chosen time.  Randomly draw the time to try again.
            // Increment the event time to simulate the noise timeout.
            long eventStopTime = eventStartTime ////+ ModelDataLink.EthernetState.ethernetSlotSizeInTicks
            // Plus the time delay from the MAC layer-computed above
               + timeDelay;
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
           Utility.err.println ("ERROR : Ethernet Bus COLLISION AT TIME " + eventStartTime
             + " will retransmit at time " + eventStopTime );
           eventStartTime = eventStopTime ;
           sendFrameEvent.UpdateGlobalSimultationTime (
               eventStopTime );
           //  Set the next transmission attempt
           sendFrameEvent.RestoreThisEventAtNewEventTime (eventStartTime);
         }  // end of if (  ( ethernetBus.numberOfSimultaneousTransmissions ...
         else {  // The bus is clear so just transmit
            UpdateStartAndStopTimesAtOtherInterfaces (
                frameToTransmit,
                sendFrameEvent,
                sourceInterfaceNumber,
                destinationInterfaceNumber);
            ethernetBus.busInUse = true ;
            ModelEvents.TransmitViaEthernetBus tveb =
                new ModelEvents.TransmitViaEthernetBus (
                    eventStartTime ,
                    frameToTransmit ,
                    sourceInterfaceNumber ,
                    destinationInterfaceNumber ,
                    sendFrameEvent.eventSequenceNumber);
             //  Set the next transmission attempt
            // CAREFUL: RestoreThisEventAtNewEventTime MUST FOLLOW
            // new TransmitViaEthernetBus ( eventStartTime... BECAUSE RestoreThisEventAtNewEventTime
            // AS ITS FIRST OPERATION REDEFINES eventStartTime = newEventTime  ;
            sendFrameEvent.RestoreThisEventAtNewEventTime (
                eventStartTime + timeDelay );
         }  // end of else The bus is clear so just transmit
      }  // end of else Start the transmission sequence
   }  // end of public void ExecuteEvent ( )

   /** Updates the start and stop times that each interface will hear the tranmsitted
    *Ethernet frame.  The start time will be the eventTimeIn plus the delay time between
   the interfaces.  The stop time will be the start time plus the length of time required
   for the frame to pass the interface.*/
   private void UpdateStartAndStopTimesAtOtherInterfaces (
                JNWBitSet frameToTransmit,
                ModelEvents.SendFrameFromInterface sendFrameEvent,
                int sourceInterfaceNumber,
                int destinationInterfaceNumber) {
      //The transmit time for this frame
      long timeToTransmitFrameinTicks =
              ( long ) ( frameToTransmit.getJNWLength () *
                  sendFrameEvent.globalVariables.timeTicksPerSecond /
                  ( 1000.0 * ethernetBus.bitRateEthernetBus ) );
      for ( int i = 0 ; i < vectorOfInterfaces.size () ; i ++ ) {
         ModelDataLink.Ethernet EthernetInterface
            = ( ModelDataLink.Ethernet ) vectorOfInterfaces.get (i);
         // The time at which interface i will start hearing the transmission
         EthernetInterface.startReceiveEthernetFrame
            = sendFrameEvent.eventStartTime + ModelPhysical.EthernetBus.
                  transmissionDelayinTicks[sourceInterfaceNumber][i];
         // The time at which interface i will stop hearing the transmission
         EthernetInterface.stopReceiveEthernetFrame
            = sendFrameEvent.eventStartTime + ModelPhysical.EthernetBus.
                  transmissionDelayinTicks[sourceInterfaceNumber][i] + timeToTransmitFrameinTicks;
      }  // end of for ( int i = 0 ; i < vectorOfInterfaces.size() ; i++ )
   }  // end of UpdateStartAndStopTimesAtOtherInterfaces ( )
    
}
