/*
 * HDLC.java
 *
 * Created on June 8, 2007, 7:59 AM
 * Updated by JMP 12Oct2011 for JDLC3/HDLC
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */


package ModelDataLink;


import Utility.* ;


/**
 *
 * @version 1.202
 * @author dixonw
 * @modified extensively JMP 21Oct2011 for JDLC3/HDLC
 */
public class HDLC  extends HDLCUtilities {
    
   /** Creates a new instance of HDLC */
   public HDLC( int nodeNumber ) {
      super (  ) ;
      // Save to number of the node to which this interface is attached.
      nodeToWhichAttached = nodeNumber ;
   }

   /**instance of ModelEventsBaseClass we can refer to*/
   private ModelEvents.ModelEventsBaseClass hdlcEvent =
       new ModelEvents.ModelEventsBaseClass();

//   Everything that follows this line is
//   Copyright 2011 J.M.Pullen/George Mason University
//   Data Link Control Layer module for Network Workbench,
//   modified for Java Network Workbench. Where C++ NW used
//   pointers to structures (in particular, link_interface_state),
//   in Java we have HDLCState associated with each interface.
//   
//   This version designed by J.M. Pullen. Provides
//   interfaces for an HDLC-like DLC that uses go-back-n.
//
//   NOTES:
//   The ARQ used her is go-back-n.
//   1. There are two state variables that are very important
//   to the Workbench DLC, and occur at each interface:
//     dl_send_state indicates which of possible states the send
//       software for the interface occupies
//     link_active indicates whether the link is actively sending
//       data and/or supervisory frames (ACK frames in response to the
//       other end of the link do not count here); the link will
//       remain active if frames have been sent for which no ACK has 
//       been received, or if there are more frames on the queue.
//   2. The physical layer will automatically schedule a call to 
//   dlc_send each time a frame transmission is completed.
//   It is important to have dl_send_state correct so dlc_send picks
//   up in the right place when this happens.
//   3. The NW DLC send algorithm is conservative in that, if it has  
//   no new data to send and frames in the buffer have not been ACKed,
//   it simply retransmits them during idle line time.  This can result
//   in better performance in the presence of errors, in that a frame
//   that suffered an error may be retransmitted before a NACK is
//   received and thus arrive sooner than it otherwise would have (this 
//   is a reasonable practice for a point-to-point link, but not for a 
//   shared link layer such as Ethernet where other senders might use  
//   the capacity instead).  A consequence of this behavior is that when
//   there are no frames to send, a node will be sending the last frame 
//   sent, or a supervisory frame.
//
//   dl_init
//
//   The current Workbench DLC provides only an unreliable
//   link initiation. It just sets link_active to TRUE,
//   dl_send_state to dl_sending, and initializes the sequence
//   number state for this end of the link.
//
//   This is how a reliable DL initialization would work:
//
//   The before dl_send is called for a link, that link must be
//   initialized, connection is established using the normal
//   three-way handshake.  To do this:
//   a.  the sending end sends an init_link frame including its
//   initial sequence number SN=0 to the receiving end (continue
//   to send at intervals of MAX_LINK_RTT until response is received);
//   b.  the receiving end replies with a receiver_ready frame
//   including its initial sequence number SN=0, saves
//   the received SN as RN, and sets link_active to TRUE for its
//   end of this link;
//   c.  when the sending end receives this reply, it saves the
//   received SN as RN, sets SNmin=SNmax=0, and sets
//   link_active to TRUE for its end of  this link.  The link is
//   now ready to start sending (dl_send case sending).
//
//
// NW dl_send algorithm
// --------------------
//
//   This is how the DL send works after the link is initialized:
//
//   Each end can send up to DL_WINDOW_FRAMES frames while waiting
//   for an ACK.  The algorithm for the sending side is below (initial 
//   dl_send_state=dl_sending).  
//
//
// state waiting (in this state the DLC has been sending but has been blocked 
// by the fact the window is full; it wakes up here after a timeout and takes 
// some action):
// {
//   if the range between SNmin and SNmax is not smaller than DL_WINDOW_FRAMES
//   (the window is full so continue with the proper actions for waiting),
//   {
//     send the frame in buffer position SNmin;
//     after transmission continue with case waiting;
//   }
//   else (the range between SNmin and SNmax is smaller than the window),
//    an ACK has been received so continue with case dl_sending;
// }
//
// state dl_sending (in this state the DLC expects to dequeue and send 
// the next frame, the first step is to dequeue and buffer the frame):
// {
//   if the range between SNmin and SNmax is smaller than the window,
//   {
//     attempt to take a frame from the input queue;
//     if there is a frame available,
//     {
//       set SN=SNmax;
//       buffer the frame in position SNmax;
//       increment SNmax|mod DL_WINDOW_MAX;
//     }
//     else if frames remain unacked you need to continue re-sending the 
//     SNmin frame,
//       set SN=SNmin;
//     otherwise nothing remains to be sent,
//     { 
//       set link_active to FALSE; 
//       set send_state to waiting; 
//       escape from this case;
//     }
//     this leaves the next frame to be sent in buffer position SN,
//     finish that frame,
//     {
//       insert RN;
//       insert CRC-FCS;
//       stuff frame;
//     }
//     if the window becomes full with this frame,
//       set dl_send_state to waiting so no more frames are sent until ACKs     
//       are received;
//     finish by sending the frame that was just made to the physical 
//     layer;
//   }
// }
//
//
// state sending_supv (in this state send a receiver_ready supervisory frame 
// because a send_supv packet has been received indicating dl_receive found 
// a need for an ACK and queued that packet):
// {
//   make the supervisory packet;
//   send it;
// }.
//
//
// NW DLC receive algorithm
// ------------------------
//
// When a frame passing the CRC check is received,
// {
//    if RN of that frame is within the current window and above
//    SNmin in the window,
//       set SNmin = RN of that frame;
//    if SN of that frame == RN sent to other end,
//    {
//       release the contents to the network layer;
//       increment RN using modular arithmetic;
//    }
// }.
//

   //*********************************************************************
   //
   /**
    * queue/dequeue incoming packets for reliable DLC send
    * (due to the way JNW is built the packet is already in a frame)
    * isUrgent puts the packet at head if queue when true
    */
    public void enqueueHDLCFrame(
        JNWBitSet frameToTransmit,
        boolean isUrgent) {
        if(queueHead == null)
            queueHead = new frameQueueElement(frameToTransmit.clone(),null);
        else {
            if(isUrgent) {
                frameQueueElement newQueueHead =
                    new frameQueueElement(frameToTransmit.clone(),null);
                newQueueHead.nextInQueue = queueHead;
                queueHead = newQueueHead;
          } else {
              frameQueueElement findQueueTail = queueHead;
              while(findQueueTail.nextInQueue != null)
                  findQueueTail = findQueueTail.nextInQueue;
              findQueueTail.nextInQueue =
                  new frameQueueElement(frameToTransmit.clone(),null);
          }
      }
      // wake up the sender to read the queue
      sendHDLCFrame(makeSupervisoryFrame(this,true));
   }
   public JNWBitSet dequeue() {
       JNWBitSet returnBitSet;
       if(queueHead != null) {
          returnBitSet = queueHead.queuedFrame;
          queueHead = queueHead.nextInQueue;
       }
       else returnBitSet = null;
       return returnBitSet;
   }


   //*********************************************************************
   //
   //  DLC layer send (asynchronous process, which draws from queue)
   //
   //
   /**
    * sends data frame with ARQ process to provide reliability
    * @param frameToTransmit is HDLC frame with null FCS, unstuffed
    */
   public void sendHDLCFrame(JNWBitSet frameToTransmit) {

      JNWBitSet send_packet = null;

      // first time called, init link (make it active: unreliable init)
      if(!link_initialized)
      {
          link_initialized = true;
          
          // we've not sent anything
          // so send a supervisory frame as ACK
          dl_send_state = DLSendState.SENDING_SUPV;
          link_active = true;
      }

      // select appropriate DLC send state
      if(isSupervisoryFrame(frameToTransmit)) {
          // fake "wakeup" frame from this interface receiveHDLC
          if(GetDestinationHDLCAddress(frameToTransmit) == interfaceNumber) {
              // put the interface back into sending state
              dl_send_state = DLSendState.DL_SENDING;
              link_active = true;
          } else {
              // it's a real supervisory frame; send as such
              dl_send_state = DLSendState.SENDING_SUPV;
          }
      }
      
      // process this frame based on dl_send_state
      switch(dl_send_state)
      {
         case WAITING:
         //
         // (in this state the DLC has been sending but has been blocked
         // by the fact the window is full; it wakes up here after a timeout
         // and take some action):
         // {
         //   if the range between SNmin and SNmax is not smaller than
         //   DL_WINDOW_FRAMES (the window is full so continue with the proper
         //   actions for waiting),
         //   {
         //     send the frame in buffer position SNmin;
         //     after transmission continue with case waiting;
         //   }
         //   else (the range between SNmin and SNmax is smaller than the
         //   window), an ACK has been received so continue with case dl_sending;
         // }
         //
         if(window_full(SNmin,SNmax))
         {
             finishFrameAndSendToPhysicalLayer(
                 insertSNandRN(pending_frames[SNmin],SNmin));
             break;
         }

         // fall through to DL_SENDING
         dl_send_state = DLSendState.DL_SENDING;

         case DL_SENDING:
         //
         // in this state the DLC expects to dequeue and send the next frame
         //
         // the first step is to dequeue and buffer the frame:
         // {
         //   if the range between SNmin and SNmax is smaller than the window,
         //   {
         //     attempt to take a frame from the input queue;
         //     if there is a frame available,
         //     {
         //       set SN=SNmax;
         //       buffer the frame in position SNmax;
         //       increment SNmax|mod DL_WINDOW_MAX;
         //     }
         //
         if(!window_full(SNmin,SNmax))
         {
            send_packet = dequeue();
            if(send_packet != null) {
               SN = SNmax;
               SNmax = increment_SNmax(SNmax);
            }
         //
         //     else if frames remain unacked we need to continue
         //     re-sending the SNmin frame if there is one, so
         //       set SN=SNmin;
         //     otherwise nothing remains to be sent,
         //     {
         //       set link_active to FALSE;
         //       set send_state to waiting;
         //       return to escape from this case;
         //     }
         //   }
         //
            else // no packet was available to send
            {
               if(frames_remain_unacked(SNmin,SNmax))
                  SN = SNmin;
               else
               {
                  link_active = false;
                  dl_send_state = DLSendState.WAITING;
                  return;
               }
            }
         }// end if(!window_full(SNmin,SNmax))...
         //
         //   this leaves the next frame to be sent in buffer
         //   position SN; to finish that frame,
         //   {
         //     insert SN and RN;
         //     insert CRC-FCS;
         //     stuff frame;
         //   }
         //   and send it the the physical layer
         // }
         if(send_packet != null) {
             pending_frames[SN] = send_packet;
             finishFrameAndSendToPhysicalLayer(
                 insertSNandRN(pending_frames[SN],SN));
         } else {
         // else
         // {
         //   if frames remain unACKed (SNmin != SNmax)
         //    repeatedly resend the one in buffer position SNmin
         // }
            if(frames_remain_unacked(SNmin,SNmax)) {
               if(pending_frames[SN] != null)
                   finishFrameAndSendToPhysicalLayer(
                       insertSNandRN(pending_frames[SNmin],SNmin));
            }
         }// end else/if(send_packet != null)...
         break;

         case SENDING_SUPV:
         //
         // (in this state, send a receiver_ready supervisory frame
         // because a send_supv packet has been received indicating
         // dl_receive found a need for an ACK and queued that packet):
         // {
         //   make the supervisory packet;
         //   send it;
         //   return to WAITING state
         // }.
         //
         JNWBitSet hdlcSupervisoryFrame = // build zero payload frame
             AddHDLCHeaderAndTrailer(this,false,new JNWBitSet());
         setSupervisoryFrame(hdlcSupervisoryFrame);

         // supervisory frame, insert only RN
         JNWBitSet rnBitSet = new JNWBitSet(RN,4);
         hdlcSupervisoryFrame.setSubset(
             hdlcSupervisoryFrame.getJNWLength()-24,rnBitSet);
         finishFrameAndSendToPhysicalLayer(hdlcSupervisoryFrame);
         link_active = false;
         dl_send_state = DLSendState.WAITING;
         break;

         default: Utility.out.println(
           "\n*** Workbench internal error: illegal dl_send_state " +
           dl_send_state);
         return;

      }// end switch(dl_send_state)

   }// end sendHDLCFrame()


   /** sends frame to physical layer; return a copy */
   JNWBitSet insertSNandRN(JNWBitSet frameToTransmit,int insertSN) {
       // insert SN and RN (type 0 frame with P/F = 0)
       int frameToTransmitLength = frameToTransmit.getJNWLength();
       JNWBitSet snBitSet = new JNWBitSet(insertSN,4);
       frameToTransmit.setSubset(frameToTransmitLength-20,snBitSet);
       JNWBitSet rnBitSet = new JNWBitSet(RN,4);
       frameToTransmit.setSubset(frameToTransmitLength-24,rnBitSet);
       return frameToTransmit;
   }

   /** updates CRC for frame and sends it */
   JNWBitSet finishFrameAndSendToPhysicalLayer(JNWBitSet frameToTransmit) {

       // control field may have changed; strip off old
       // CRC-FCS, insert a new one, and stuff frame
       int sendLength = frameToTransmit.getJNWLength();
       JNWBitSet frameForCRC = 
            new JNWBitSet(frameToTransmit.getSubset(16,sendLength));
       frameForCRC.appendFCS(generatorPolynomial);
       JNWBitSet sendFrame = frameForCRC.stuff();

       // set simulation timing
       long eventStartTime = this.getTransmitTime();
       long eventStopTime = eventStartTime + timeToConvertPacketIntoFrame ;
       updateTransmitTime(eventStopTime);
       hdlcEvent.UpdateGlobalSimultationTime ( eventStopTime );
       
       // send using JNW mechanisms
       int interfaceAtOtherEndOfLink = ModelPhysical.Link.
            InterfaceNumberAtOtherEndOfLink ( interfaceNumber );
       ModelEvents.TransmitViaLink tvl =
         new ModelEvents.TransmitViaLink (
             eventStopTime ,
             interfaceNumber / 2, // link number
             this,
             interfaceAtOtherEndOfLink ,
             sendFrame );
       return sendFrame;
   }// end sendFrameToPhysicalLayer()

//************************************************************
//     DLC layer receive
//
   public void receiveHDLCFrame (
           JNWBitSet frameBeingReceived,
           int bitRateSourceToDestination) {

      // unstuff the bits
      JNWBitSet unstuffed_DLC_frame = frameBeingReceived.unstuff();
      int unstuffed_length = unstuffed_DLC_frame.getJNWLength();
      if(GetDestinationHDLCAddress(unstuffed_DLC_frame)!= interfaceNumber) {
          Utility.out.println("*******************************\n"+
              "HDLC misdelivered dest addr:"+
               GetDestinationHDLCAddress(unstuffed_DLC_frame)+
               " reached:"+interfaceNumber);
          return;
      }
      // calculate and check FCS -
      // if it is not zero then frame does not pass
      JNWBitSet checkFCS =
          unstuffed_DLC_frame.computeFCS(generatorPolynomial);
      if(!checkFCS.equals(zeroFCS)) {
          Utility.out.println("CRC failure receiving at interface " +
             interfaceNumber + " FCS:" + checkFCS.toString().trim());
          return;
       }

      // remove data from frame header
      int received_RN = GetRN(unstuffed_DLC_frame);

//
// When a frame passing the CRC check is received,
// {
//    if RN of that frame is within the current window and above
//    SNmin in the window,
//       set SNmin = RN of that frame;
//    if SN of that frame == RN sent to other end,
//    {
//       release the contents to the network layer;
//       increment RN using modular arithmetic;
//    }
// }.
// (if frame is not ACK'd, sender will retransmit)
//
      // received RN is an ACK if > SNmin
      int newSNmin = update_SNmin(received_RN);
      boolean SNminChanged = (newSNmin != SNmin);
      SNmin = newSNmin;

      // for supervisory ACK frames received, notify hdlcSend
      // that SNmin has advanced, opening window; then return
      if(SNminChanged) {
          if(isSupervisoryFrame(unstuffed_DLC_frame)) {
              sendHDLCFrame(makeSupervisoryFrame(this,true));
              return;
          }
      }

      // recover SN from frame header and use it to see if this is  
      // ezpected frame; if so:
      // {
      //   send payload to network layer
      //   update RN to send in next outgoing frame
      //   if this end of link is not actively sending, get
      //     hdlcSend to send a supervisory ACK frame
      // }
      int received_SN = GetSN(unstuffed_DLC_frame);
      if(accept_frame(received_SN,RN))
      {
         passFrameToNetworkLayer(
             unstuffed_DLC_frame,
             bitRateSourceToDestination);
         RN = update_RN(received_SN);
         if(!link_active) {
             sendHDLCFrame(makeSupervisoryFrame(this,false));
         }
      }
   }// end receiveHDLCFrame


   /** passes received frame to network layer */
   void passFrameToNetworkLayer(
        JNWBitSet frameToPass,
        int bitRateSourceToDestination) {
      long timeToReceiveFrameinTicks = ModelConstants.Node.
          timeToMovePacketFromInterfaceBufferToNodeBufferInTicks;
      long eventStartTime = getReceiveTime();
      long eventStopTime = eventStartTime +  timeToReceiveFrameinTicks ;
      receiveBuffer.add(frameToPass);
      ModelEvents.MovePacketFromInterfaceToNodeBuffer mpfitnb =
          new ModelEvents.MovePacketFromInterfaceToNodeBuffer (
              eventStopTime,
              interfaceNumber  ) ;
   }// end passToNetworkLayer()


   //************************************************************
   // 
   // methods implementing logic of go-back-n ARQ
   // these were in dllogic.cpp in C++ NW

   /** tests whether range between two numbers is smaller than
    *  the window (DLC_WINDOW_FRAMES), given DLC counter range (DL_WINDOW_MAX)
    *  Nmin is logically lower end of range, the lowest number not ACKed
    *  Nmax is logically higher end of range, the next number to be used */
   boolean LTwindow(int Nmin, int Nmax) {
      int testmax=Nmax;
      if(Nmin > Nmax) testmax = testmax+DL_WINDOW_MAX;
      return(testmax-Nmin < DL_WINDOW_FRAMES);
   }

   /**  tests whether range between two numbers is within the
    *   window (DLC_WINDOW_FRAMES), given DLC counter range (DL_WINDOW_MAX)
    *   Nmin is logically lower end of range, the lowest number not ACKed
    *   Nmax is logically higher end of range, the next number to be used */
   boolean INwindow(byte Nmin, byte Nmax) {
     //// student replaces FALSE with appropriate code
     return false;
   }

   /** tests whether the window has become full
    *  SNmin is SN of the lowest frame sent and not ACKed
    *  SNmax is one greater (mod DL_WINDOW_MAX) than SN of
    *  highest frame sent and not acked */
   boolean window_full(int SNmin, int SNmax) {
     //// student replaces FALSE with appropriate code
     return false;
   }

   /** tests whether frames have been sent for which ACKs
    *  have not been received
    *  SNmin is SN of the lowest frame sent and not ACKed
    *  SNmax is one greater (mod DL_WINDOW_MAX) than SN of
    *  highest frame sent and not acked
    */
   boolean frames_remain_unacked(int SNmin, int SNmax) {
     //// student replaces FALSE with appropriate code
     return false;
   }

   /** returns incremented SNmax for an interface within
    * the modular range, which is of size DL_WINDOW_MAX */
   int increment_SNmax(int SNmax) {
     //// student replaces 0 with appropriate code
     return 0;
   }

   /** returns updated value of SNmin in HDLCState */
   int update_SNmin(int newSNmin) {
       //// student replaces 0 with appropriate code
       return 0;
   }

   /** updates the RN of interface,
    *  based on received SN */
   int update_RN(int received_SN) {
      //// student replaces 0 with appropriate code
      return 0;
   }

   /** tests whether to accept frame, based on its SN
       and the RN associated with receiving interface*/
   boolean accept_frame(int received_SN,int interface_RN) {
      //// student replaces FALSE with appropriate code
      return false;
   }
   
}// end class HDLC
