/*
 * HDLCState.java
 *
 * Created on June 8, 2007, 8:04 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelDataLink;

import Utility.* ;

/**
 *
 * @version 1.202
 * @author dixonw
 * @modified JMP 21Oct2011 for JDLC3/HDLC
 *@see <a href="http://en.wikipedia.org/wiki/High-Level_Data_Link_Control">HDLC</a>
 *@see <a href="http://members.tripod.com/~vkalra/hdlc.html">HDLC - A Technical Overview
</a>
 */
public class HDLCState extends InterfaceBaseClass {

   /**Sequence number*/
    public int SN ;

    /**Request number*/
    public int RN ;

    /**Logically lowest sequence number buffered*/
    public int SNmin ;

    /**Logically one higher than highest sequence number buffered*/
    public int SNmax ;

    /**For DLC (no more frames than this pending ACK at any time)*/
    static int DL_WINDOW_FRAMES = 2 ;

    /**Max values of SN/RN (equivalent to 3-bit field important
     * for modulo arithmetic)*/
    static int DL_WINDOW_MAX = 8 ;

    /**Time in "ticks" to process an HDLC frame*/
    public static long DLProcessingTicks = 10 ;
    
    /**The possible protocols*/
    static public enum ARQProtocols  { Stop_and_Wait , GO_Back_N } ;
    
    static public ARQProtocols ARQProtocol   ;

    /**The transmit buffer - frames in window*/
    JNWBitSet[] pending_frames = new JNWBitSet[DL_WINDOW_MAX];

    public boolean link_initialized ;

    public boolean link_active  ;

    /** HDLC interface status */
    public static enum DLSendState { WAITING , DL_SENDING , SENDING_SUPV } ;

    public static DLSendState dl_send_state  ;

   /** generator polynomial, binary coded */
   JNWBitSet generatorPolynomial =
      new JNWBitSet("10001000000100010");

   /** zero FCS field indicating correct transmission */
   JNWBitSet zeroFCS =
      new JNWBitSet("0000000000000000");

    /** Creates a new instance of HDLCState */
    public HDLCState() {
       super ( ) ;
       SN = 0 ;
       RN = 0 ;
       SNmin = 0 ;
       SNmax = 0 ;
       dl_send_state = DLSendState.WAITING ;
       link_initialized = false;
       link_active = false ;
    }

    /** internal class for frame queue*/
    public class frameQueueElement {
        JNWBitSet queuedFrame = null;
        frameQueueElement nextInQueue = null;

        public frameQueueElement (
             JNWBitSet frameToQueue,
             frameQueueElement linkToNext) {
            queuedFrame = frameToQueue;
            nextInQueue = linkToNext;
        }
    }

    /** head of frameQueue*/
    frameQueueElement queueHead = null;

}
