/*
 * HDLCUtilities.java
 *
 * Created on June 21, 2007, 9:07 AM
 * Updated JMP Oct 20, 2011
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelDataLink;

import Utility.JNWBitSet;

/**
 *
 * @version 1.202
 * @author dixonw
 * @modified JMP 21Oct2011 for JDLC3/HDLC
 */
public class HDLCUtilities extends HDLCState implements ModelConstants.General {
    
    /** Creates a new instance of HDLCUtilities */
    public HDLCUtilities() {
        super ( ) ;
    }

   private static JNWBitSet CRC_16 = new JNWBitSet (
      "1 00010000 00010001" ) ;

   public static JNWBitSet AddHDLCHeaderAndTrailer (
           ModelDataLink.InterfaceBaseClass interfaceCard,
           boolean sendBackToSourceInterface,
           JNWBitSet packetIn )
   { // First clone the packet
      JNWBitSet HDLCFrame = packetIn.clone ( ) ;

      // Compute the frame length in bytes including the header and FCS
      int HDLCFrameLengthInBytes = HDLCFrame.getJNWLength () / BITS_PER_BYTE
         + ( 1 /*destination address length*/  + 1 /*source address length*/
         +   1 /*control length */             + 2 /* FCS length */)  ;
      
      // add null control field
      long controlField = 0;
      JNWBitSet controlBitSet = new JNWBitSet (controlField, 8);
      HDLCFrame . prepend (controlBitSet);

      // the source HDLC address
      JNWBitSet sourceAddressBitSet =
          new JNWBitSet ( interfaceCard.interfaceNumber , 8 ) ;
      HDLCFrame . prepend ( sourceAddressBitSet ) ;

      // the destination HDLC address
      JNWBitSet destinationAddressBitSet;
      if(sendBackToSourceInterface)
          destinationAddressBitSet =
              new JNWBitSet ( interfaceCard.interfaceNumber , 8  ) ;
      else
          destinationAddressBitSet =
              new JNWBitSet ( interfaceCard.oppositeInterfaceNumber , 8  ) ;
      HDLCFrame . prepend ( destinationAddressBitSet ) ;

      // Compute the CRC over the frame
      HDLCFrame . appendFCS ( CRC_16 ) ;
      // Add the preamble
      return  HDLCFrame ;
   }  // end of public static JNWBitSet AddHDLCHeaderAndTrailer

   public static JNWBitSet GetIPPacketFromHDLCFrame ( JNWBitSet HDLCFrameIn )
   {
      int HDLCHeaderLength = (
             8 /*destination address length*/  + 8 /*source address length*/
         +   8 /*control length */ )  ;
      return HDLCFrameIn .getSubset ( 16 /* FCS */
         , HDLCFrameIn . getJNWLength ( ) - HDLCHeaderLength /*2 byte FCS*/ ) ;
   }

   /**Returns the destination HDLC address from the Ethernet frame or
    * just the Ethernet frame header.*/
   public static int GetDestinationHDLCAddress ( JNWBitSet HDLCFrameIn )
   {
      int fromIndex = HDLCFrameIn . getJNWLength ( ) -  ( 
         + 1 /*destination address length*/ ) * BITS_PER_BYTE ;
      return (int) HDLCFrameIn .getSubset ( fromIndex , fromIndex + 8 )
         . convertJNWBitSetToLong ( ) ;
   }  // end of GetDestinationHDLCAddress


   /**Returns the source HDLC address from the HDLC frame or
    * just the HDLC frame header.*/
   public static int GetSourceHDLCAddress ( JNWBitSet HDLCFrameIn )
   {
      int fromIndex = HDLCFrameIn . getJNWLength ( ) -  ( 
         + 1 /*destination address length*/
         + 1 /*source address length*/ ) * BITS_PER_BYTE ;
      return (int) HDLCFrameIn . getSubset ( fromIndex , fromIndex + 8 )
         . convertJNWBitSetToLong ( ) ;
   }  // end of GetSourceHDLCAddress

   /**Returns Sequence Number (SN) from the input HDLC frame*/
   public static int GetSN ( JNWBitSet HDLCFrame) {
       int frameLength = HDLCFrame.getJNWLength();

       return (byte)HDLCFrame.getSubset(frameLength-20,frameLength-17).
           convertJNWBitSetToLong();
   }

   /**Returns Request Number (RN) from the input HDLC frame*/
   public static int GetRN ( JNWBitSet HDLCFrame) {
       int frameLength = HDLCFrame.getJNWLength();
       return (byte)HDLCFrame.getSubset(frameLength-24,frameLength-21).
           convertJNWBitSetToLong();
   }

   /**true if argument is a supervisory frame*/
   boolean isSupervisoryFrame(JNWBitSet testFrame) {
      int frameLength = testFrame.getJNWLength();
      if(frameLength < 20)return false;
      return (testFrame.getBitValue(frameLength-17) == 1) &&
             (testFrame.getBitValue(frameLength-18) == 0);
   }

   /** makes argument frame into supervisory frame*/
   public static JNWBitSet setSupervisoryFrame(JNWBitSet targetFrame) {
      int frameLength = targetFrame.getJNWLength();
      if(frameLength < 21)return null;
      targetFrame.setBitValue(frameLength-17,1);
      targetFrame.setBitValue(frameLength-18,0);
      targetFrame.setBitValue(frameLength-19,0);
      targetFrame.setBitValue(frameLength-20,0);
      targetFrame.setBitValue(frameLength-21,0);
      return targetFrame;
   }

   /** makes a new supervisory frame
    *  @parameter interfaceCard is the interface from which sent
    *  @parameter wakeupThisInterface true makes a
    *  frame that loops back to same interface*/
   public JNWBitSet makeSupervisoryFrame(
           ModelDataLink.HDLC interfaceCard,
           boolean wakeupThisInterface) {
       JNWBitSet supervisoryFrame =
           AddHDLCHeaderAndTrailer(
               interfaceCard,
               wakeupThisInterface,
               new JNWBitSet());
       setSupervisoryFrame(supervisoryFrame);
       return supervisoryFrame;
   }

}
