/*
 * InterfaceBaseClass.java
 *
 * Created on June 8, 2007, 8:02 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package ModelDataLink;

/**This is simply a "marker" class.  It will permit the Ethernet, TokenRing or
 * HDLC interfaces to be instantiated as a InterfaceBaseClass class.  The nodes
 *will instantiate InterfaceBaseClass class and the particular class can be used
 *depending on the input topology file.
 * @version 1.202
 * @author dixonw
 * @modified JMP 21Oct2011 for HDLC
 */
public abstract class InterfaceBaseClass implements
   ModelConstants.InterfaceCard , GUIConstants.General {

   /**The interface number to be shared by all the interfaces.  The interface
    *number is set to the staticInterfaceNumber in the constructor. This is
    *also reset to 0 in GUIActions.ReadTopologyFileAction, so that the
    *interface numbers always begin with 0.  This will only be necessary
    *if the user opens a second topology file.*/
   public static int staticInterfaceNumber = 0;

   /**InterfaceCard number.  This is assigned in the constuctor as the
    *interface is made.*/
   public int interfaceNumber;

   /**instance of HDLC class*/
   public ModelDataLink.HDLC hdlc = null;

   /**The time on the transmitter of the interface.  Interfaces need
    * two times, one for the transmitter and one for the receiver.*/
   public long transmitTime ;

   /**The time on the receiver fo the interface.  Interfaces need
    * two times, one for the transmitter and one for the receiver.*/
   public long receiveTime ;

   /**Number of interface card at other end of link.  This is useful
    *only for point-to-point connections.  The constructor will
    *compute an oppositeInterfaceNumber for a bus, but the number
    *is useless since all the interfaces will be on the one bus.*/
   public int oppositeInterfaceNumber;

   /**The node to which this interface is attached.*/
   public int nodeToWhichAttached;

   /**The transmit buffer.*/
   public Utility.JNWBuffer transmitBuffer = new Utility.JNWBuffer (
      interfaceTransmitBufferSizeInKilobytes );

   /**The receive buffer.*/
   public Utility.JNWBuffer receiveBuffer = new Utility.JNWBuffer (
      interfaceReceiveBufferSizeInKilobytes );

   /**Access to the print utilities.  These print output to the center panel
    *of the MainGUI.*/
   static Utility.err err = new Utility.err ();

   /**Creates a new instance of InterfaceBaseClass.*/
   public InterfaceBaseClass () {
      // Assign an interface number
      interfaceNumber = staticInterfaceNumber;
      staticInterfaceNumber = staticInterfaceNumber + 1;
      // Compute the index number of the interface card
      // at the opposite end of the link.
      oppositeInterfaceNumber = 2 * ( interfaceNumber / 2 ) + 1 -
         interfaceNumber % 2;
      // Time for the transmit buffer
      transmitTime = 0;
      // Time for the receive buffer
      receiveTime = 0;
   }

   /**Verifies the received message.  This is an abstract method that
    * will be overridden by the Ethernet, HDLC, and Token Ring
    * utility classes.*/
   public static boolean verifyAtReceiver ( Utility.JNWBitSet frameIn ) {
      return true;
   }

   /**Returns a Boolean indicating that the frame is NOT to be checked for
    * errors OR the frame IS to be checked for errors and it passes
    * the error check.*/
   public static boolean ProcessFrame ( Utility.JNWBitSet frameReceived ) {
      if (ModelData.GlobalVariables.verifyChecksumsAtReceiver == false ||
         ( ModelData.GlobalVariables.verifyChecksumsAtReceiver == true &&
         ModelDataLink.InterfaceBaseClass.verifyAtReceiver ( frameReceived ) ==
         true )) {
         return true;
      }
      else {
         return false;
      }
   }

   /**Updates the transmitter time.*/
   public long updateTransmitTime ( long timeIn ) {
      return transmitTime = java.lang.Math.max ( transmitTime
         , timeIn  ) ;
   }

   /**Returns current transmitter time.*/
   public long getTransmitTime () {
       return transmitTime;
   }


   /**Updates the receiver time.*/
   public long updateReceiveTime ( long timeIn ) {
      return receiveTime = java.lang.Math.max ( receiveTime
         , timeIn  ) ;
   }

   /**Returna current receiver time.*/
   public long getReceiveTime () {
       return receiveTime;
   }


   /**This resets the to 0.  This should be used when creating a new set of
    * Links.*/
   public static void RESET ( ) {
      staticInterfaceNumber = 0 ;
   }

}  // end of class
