/*
 * SendEthernetFrameFirstTime.java
 *
 * Created on August 13, 2007, 11:12 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelEvents;

/**Import some static variables from ModelDatas . LoadGeneratorVariables */
import static ModelData . LoadGeneratorVariables . messageGenerationRate ;
import static ModelData . LoadGeneratorVariables . messageMeanLength ;
import static ModelData . LoadGeneratorVariables . messageLengthDistribution ;
import Utility.JNWRandom;

/**A ModelEvents event to create a series of fixed or exponentially-distributed
 * length UDP messages at an exponential rate.  The data for the messages
 * comes from class ModelDatas . LoadGeneratorVariables.  The data in
 * ModelDatas . LoadGeneratorVariables
 *is input in the class GUIDatas.Load_Generator_Variables.
 *
 *<p><b>Versions</b>
 *<li><b>Version 1.001</b> of 20071118 substitutes a call to discreteEventSimulator
 *. RestoreEventAtNewEventTime( eventStopTime , this ) rether than creating a
 * new GenerateMessageLoad each time a new message is created.  This should
 * be much more efficient.</li>
 *<li><b>Version 1.002</b> of 20071204 adds import statement to make addressing
 * the static variables in ModelDatas . LoadGeneratorVariables much easier.</li>
 *
 * @version 1.202
 * @author dixonw
 * @modified JMP 16Nov2011 for JTRN3
 */
public class GenerateMessageLoad extends ModelEventsBaseClass
{
   /**The random number generator.*/
   private static JNWRandom randomNumberGenerator
      = JNWRandom.getInstance () ;
   
   /**The UDP packet.*/
   Utility . JNWBitSet UDPPacket = null ;
   
   /**The IP packet.*/
   Utility . JNWBitSet IPPacket = null ;
   
   /**Raw data is a bit set filled with 0s.  It is used to make a
    * message of the correct length.  The length of raw data plus the
    * UDP header plus the IP header must equal the length desired,
    * either fixed or exponential.*/
   Utility . JNWBitSet rawData = null  ;
   
   /**The source node buffer.*/
   Utility.JNWBuffer sourceNodeBuffer = null ;
   
   //**The time that the next load message will be generated.*/
   long nextEventTime = 0 ;
   
   /**The length of the message in bytes.  The length includes the
    *IP header ( 20 bytes ) and the UDP header ( 8 bytes ).  Therefore, the
    *message lengths cannot be less than 28 bytes.*/
   int messageLengthInBytes = 0 ;
   
   /**Constructor that creates a new instance of GenerateMessageLoad.*/
   public GenerateMessageLoad ( long eventTime )
   {
      super ( eventTime ) ;
      // Get a reference to the source node buffer
      sourceNodeBuffer = vectorOfNodes.get (
         ModelData.LoadGeneratorVariables.sourceNodeNumber ) .nodeBuffer ;
   }  // end of constructor
   
   /**Create the load message, put the load message into the source
    * node buffer, and schedules the generation of the next message.*/
   public void ExecuteEvent ( )
   {  // Compute the time of the next message creation event
      nextEventTime = eventStartTime
         + ( long ) ( ModelData.GlobalVariables.timeTicksPerSecond
         * randomNumberGenerator . nextExponential (
            messageGenerationRate ) ) ;
      messageLengthInBytes 
         = messageMeanLength ;
      // Is the length of the message exponentially distributed or fixed?
      if (  messageLengthDistribution
         == ModelData.LoadGeneratorVariables.possibleMessageLengthDistributions.Exponential )
      {  // It's exponential
         messageLengthInBytes = ( int ) randomNumberGenerator
            . nextExponential ( 1.0  / messageMeanLength ) ;
      }  // end of if
      // Create a blank message length.  The blank messages cannot be less than
      // 1 bytes long
      messageLengthInBytes = java.lang.Math.max ( ( messageLengthInBytes
         - ( 20 /*IP header length*/ + 8 /*UDP header length*/ ) ) , 1 ) ;
      // Create a blank message of the correct length
      rawData = new Utility.JNWBitSet (  8 * messageLengthInBytes  ) ;
      UDPPacket
         = ModelTransport.UDPUtilities.PrependUDPHeaderToData (rawData
         ,  ModelData.LoadGeneratorVariables.sourceNodeNumber
         ,  ModelData.LoadGeneratorVariables.destinationNodeNumber ) ;
      //Add the packet to the buffer of the node
      IPPacket
         =  ModelTransport.IPUtilities.PrependIPHeaderToMessage (  UDPPacket
         ,  ModelData.LoadGeneratorVariables.sourceNodeNumber
         ,  ModelData.LoadGeneratorVariables.destinationNodeNumber
         ,  ModelData.GlobalVariables.transmissionProtocols.UDP ) ;
      // Add the packet to the node buffer
      sourceNodeBuffer.add ( IPPacket ) ;
      new RoutePacketAtNodeOrProcessAtDestination ( eventStartTime  
         , ModelData.LoadGeneratorVariables.sourceNodeNumber , IPPacket ) ;
      // Put the GenerateMessageLoad event back into the discrete
      // event simulator at the next event time
      RestoreThisEventAtNewEventTime ( nextEventTime ) ;
   }  // end of public void ExecuteEvent ( )
}  // end of class GenerateMessageLoad