/*
 * ModelEventsBaseClass.java
 *
 * Created on February 6, 2007, 1:29 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelEvents;

import Utility.JNWRandom;

/**This is the base class for all the ModelEvents in the simulation.  Each
 *Event <b>MUST</b> inherit this class.  The class provides the Events a
 *reference to the nodes, links, and interfaces.
 *
 * <p><b>The first statement in every
 *constructor of a subclass <b><u>MUST</b></u> be <li>  super ( eventStartTime ) ;</li> This causes the
 * Event to be added to the discrete event simulator with
 *<li>  discreteEventSimulator . addEvent( this ) ;</li></b>
  *<p><b>Versions:</b>
 *<li><b>Version 1.200</b> of 20081001 adds the method  ComputeExecutionTime ( )
 * whihc is used for debugging.  Call the method twice in the Event and the method
 * will print the time difference.</li>
 * @version 1.200
 * @author dixonw*/
public class ModelEventsBaseClass  implements GUIConstants. General
   , ModelConstants. General 
{  /**Reference to the global variables.*/
   public static ModelData.GlobalVariables globalVariables
      = new ModelData.GlobalVariables ( ) ;
   
   /**Get a reference to the discrete event simulator.*/
   protected static ModelDES.DiscreteEventSimulator discreteEventSimulator
      = ModelDES.DiscreteEventSimulator.getInstance () ;
   
   /**A singleton vector containing all the Nodes*/
   static ModelData.JNWSingletonVectorOfNodes<ModelPhysical.Node> vectorOfNodes
      = ModelData.JNWSingletonVectorOfNodes.getInstance (  ) ;
   
   /**A singleton vector containing all the links*/
   static ModelData.JNWSingletonVectorOfLinks<ModelPhysical.Link> vectorOfLinks
      = ModelData.JNWSingletonVectorOfLinks.getInstance (  ) ;
   
   /**A singleton vector containing all the interfaces*/
   static ModelData.JNWSingletonVectorOfInterfaces 
      < ModelDataLink.InterfaceBaseClass >
      vectorOfInterfaces
      = ModelData.JNWSingletonVectorOfInterfaces.getInstance (  ) ;
   
   /** The random number generator.  Used only in the Event TransmitViaLink.*/
   static JNWRandom randomNumberGenerator
      = JNWRandom . getInstance ( ) ;
   
   /**Time that the event is scheduled to start.*/
   public long eventStartTime ;
   
   /**Time that the next Event is scheduled to start.  This will be
    * calculated in the Execute method of the particular event.*/
   public long eventStopTime ;
   
   /**Event sequence number unique to this Event.  Event sequence numbers 
    * are set to staticEventSequenceNumber in the base class constructor
    * and staticEventSequenceNumber is incremented.*/
   public int eventSequenceNumber ;
   
   /**Static event sequence number used to set <b>ALL</b>
    * the event sequence numbers.  It is incremented each time a
    new Event is created.*/
   static int staticEventSequenceNumber = 1 ;
   
   /**The number of the node on which the event executes.*/
   protected int thisNodeNumber = Integer . MIN_VALUE ;
   
   /**This node.*/
   private ModelPhysical.Node thisNode = null ;
   
   /**The number of the interface on which the event executes.*/
   protected  int interfaceNumber = Integer . MIN_VALUE ;
   
   /**This interface.*/
   private ModelDataLink . InterfaceBaseClass thisInterface = null ;
      
   /**The number of the link on which the event executes.*/
   protected  int linkNumber = Integer . MIN_VALUE ;
   
   /**Start time of the event.  Used for debugging only in the method 
    * ComputeExecutionTime.*/
   protected long eventDebuggingStartTime ; 
   
   /**The source node for the data*/
   int sourceNodeNumber = 0 ;
   
   /**The destination node for the data*/
   int destinationNodeNumber = 0 ;
   
   /**Access to the print utilities.  These print output to the center panel
    *of the MainGUI.*/
   static Utility.out out  = new Utility . out ( )  ;
   
   /**Access to the print utilities.  These print output to the center panel
    *of the MainGUI.*/
   static Utility.err err  = new Utility . err ( )  ;

   /**The Link being accessed by this Event.  If the Event is not accessing
    * the link in a TransmitViaLink Event, this value is MINUS_ONE*/
    static int linkInUse ;
   
   /**Creates a new instance of an ModelEventsBaseClass.  Sets and updates
    *the event sequence number, sets the event time to the parameter passed in,
    *and adds the event to the events list.  The constructor them puts this onto
    *the discrete event simulator.*/
   public ModelEventsBaseClass ( long eventTime )
   {
      eventSequenceNumber = staticEventSequenceNumber ;
      staticEventSequenceNumber = staticEventSequenceNumber + 1 ;
      eventDebuggingStartTime = MINUS_ONE ;
      // Set the event time
      eventStartTime = eventTime ;
      // Set the stop time to something-anything
      eventStopTime = eventTime ;
      // And put this event on the discrete time simulator
      discreteEventSimulator . addEvent ( this ) ;
   }
   
   /**The default constructor.  Used ONLY by
    *ModelDES.DiscreteEventSimulator.dummyEvent*/
   public ModelEventsBaseClass (   )
   {  }
   
   /**The static initializer*/
   static {
      linkInUse = MINUS_ONE ;
   }
   /**Gets the event Sequence Number of this event.  Used in
    *class TimeEventsList.*/
   public long getEventStartTime ( )
   {
      return eventStartTime ;
   }
   
   
   /**
    * Sets the event Sequence Number of this event.  Used in
    * class ModelDatas.ModelEventsBaseClass.  Used ONLY for testing
    */
   public void setEventStartTime ( long eventTime )
   {
      this . eventStartTime = eventTime ;
   }
   
   
   /**
    * Gets the event Sequence Number of this event.  Used in
    * class ModelDatas.ModelEventsBaseClass.
    */
   public long getEventSequenceNumber ( )
   {
      return eventSequenceNumber ;
   }
   
   
   /**Executes the event.  This methods should be overridden in every subclass.*/
   public void ExecuteEvent ( )
   {
      err.println ("IN ModelDatas.Event.Execute: You should not be here.  " +
         "Execute should have been overidden in the subclasses."  )   ;
   }
   
   
   /**Updates the global simulationTime.  The simulationTime is computed
    * as the maximum of the simulation time and the event time.  These times may
    * be different because the execution of an event may be delayed by the
    * execution of another event.  If the event is delayed, its execution time is
    * set to the simulation time.  If the event executes at its scheduled time, the
    * simulation time "hops" to the event time of the next event.*/
   public void UpdateGlobalSimultationTime ( long timeIn )
   {  // Read the simulation time and set the ModelDatas.GlobalVariables.simulationTime to
      // the maximum of the eventStartTime and the simulation time
      ModelData.GlobalVariables.simulationTime = java.lang.Math.max ( timeIn ////20080620eventStartTime
         , ModelData.GlobalVariables.simulationTime  ) ;
   }

   /**This computes the time in nanoseconds needed to execute this event.  Call the method at
    * the beginning of the event to set eventDebuggingStartTime <b><i>AND</b></i> call it
    * at the end of the event to set eventDebuggingStopTime and print the execution time.  It can also
    * be used inside the method to compute the execution time of individual statements. 
    */
   protected void ComputeExecutionTime ( ) {
      if ( eventDebuggingStartTime == MINUS_ONE ) {
         eventDebuggingStartTime = System.nanoTime() ;
      }
      else {
         System.out.println ( "Event " + this.toString ( )  
            + " required " + ( System.nanoTime() - eventDebuggingStartTime )
            + " nanoseconds for execution" ) ;
         eventDebuggingStartTime = MINUS_ONE ;
      }  // end of else
   }

   /**Resets the event time of this event to the time newEventTimeIn and adds 
    *this Event into the DES in order of the new event time.  This method
    *is intended for use with reoccurring Events like GenerateMessageLoad
    *that are polled from the discrete event simulator and then added back
    *into the DES at a new time.  This method can be called from within 
    *the ExecuteEvent() method of the event as
    * <li> RestoreThisEventAtNewEventTime ( newEventTimeIn )
    * ;</li><p>This will cause the event to be reentered in the DES at the new 
    * time.  The eventSequenceNumber is unchanged by this method.
    <p><b>Be VERY CAREFUL in using this method.  If this method is called 
    * from within an ExecuteEvent, remember that calling this 
    * method changes eventStartTime to newEventTimeIn.  So, if you are 
    * using eventStartTime and call this method, eventStartTime will be redefined 
    * by the first line of this method.  So this method shuld be called 
    * as the last stop in the program.</b><p>
    <p>ERROR:  This method is duplicative of 
    * ModelDES.DiscreteEventSimulator.RestoreEventAtNewEventTime
    * , and one or the other should be eliminated.</p>*/
   public void RestoreThisEventAtNewEventTime ( long newEventTimeIn  )
   {
      this . eventStartTime = newEventTimeIn  ;
      discreteEventSimulator. addEvent ( this ) ;
   }  // end of reorderEvent
   
   
   /**A toString method for the events*/
   public String toString ( )
   {
      return new String ( " Event " + eventSequenceNumber
         + " of type " + this . getClass ( ) . getName ( )
         + " to occur at time " + eventStartTime  ) ;
   }  // end of toString
   
   
   /**A toString method for the events*/
   public String toStringSequenceNumber ( )
   {
      return new String ( " Event " + eventSequenceNumber ) ;
   }  // end of toString
   
}
