 /*
 * StartEmail.java
 *
 * Created on July 6, 2007, 2:23 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package ModelEvents;

import Utility.* ;

/**Moves a packet from the receive buffer of the interface
 *card to the node buffer.
 * 
 * *<p><b>Versions:</b>
 * <li><b>Version 1.200</b> of 20080820 added section to ExecuteEvent,
 * starting "// Should the IP header checksum be verified?  If so is it true?"
 * that tests the IP header for validity before adding the packet to the 
 * Node buffer.</li>
 * @version 1.202
 * @author dixonw
 * @modified JMP 21Oct2011 for HDLC
 */
public class MovePacketFromInterfaceToNodeBuffer extends ModelEventsBaseClass
   implements ModelConstants.Node {

   /**The packet being routed.*/
   Utility.JNWBitSet frameToBeTransferred;

   /**This interface.*/
   ModelDataLink.InterfaceBaseClass interfaceCard = null;

   /**Kind of interface*/
   boolean thisInterfaceIsHDLC = false;

   /**The node to which the interface is attached.*/
   ModelPhysical.Node node = null;

   /** Creates a new instance of RouteFrameFromInterfaceToNodeBuffer */
   public MovePacketFromInterfaceToNodeBuffer (
      long eventTime ,
      int interfaceNumberIn ) {
      super ( eventTime );
      interfaceNumber = interfaceNumberIn;
   }

   /**Routes the BitSet from the receive buffer of the interface
    * to the node buffer.*/
   public void ExecuteEvent () { 
      ModelDataLink.InterfaceBaseClass interfaceCard 
         = vectorOfInterfaces.get ( interfaceNumber );
      thisInterfaceIsHDLC = interfaceCard instanceof ModelDataLink.HDLC;

      // Get the node number
      thisNodeNumber = interfaceCard.nodeToWhichAttached;
      // Get the node
      node = vectorOfNodes.get ( thisNodeNumber );
      // Update the node time
      eventStartTime = node.UpdateNodeTime ( eventStartTime );
      // Get the interface receive buffer
      Utility.JNWBuffer interfaceReceiveBuffer = interfaceCard.receiveBuffer;
      // Get the node buffer
      Utility.JNWBuffer nodeBuffer = node.nodeBuffer;
      frameToBeTransferred = /*( Utility.JNWBitSet )*/ interfaceReceiveBuffer.
         poll ();
      if (frameToBeTransferred == null) {
         Utility.err.println ( "ERROR IN ModelEvents" +
            ".MovePacketsFromInterfaceToNodeBuffer.Execute(): The " +
            "interfaceReceiveBuffer for interface " + interfaceNumber +
            " is unexpectedly empty." );
         return;
      }  // end of if ( frameToTransmit == null )
      if(interfaceCard instanceof ModelDataLink.Ethernet) {  
         long destinationAddress =
            ModelDataLink.EthernetUtilities.GetDestinationMACAddress ( 
            frameToBeTransferred );
         long nodeAddress = thisNodeNumber;
         //If this is not addressed to me, just toss it.
         //out.print ( "Node " + thisNodeNumber + " received Ethernet " +
         //   "message for node " + destinationAddress + CARRIAGE_RETURN );
         if (destinationAddress != nodeAddress) {  
            // If this is not destined for this node, forget it.
            return;
         }
      } 
      
      // if this is HDLC
      else if(thisInterfaceIsHDLC) {
          // HDLC supervisory frame is not routable as packet
          // send it direct to destination interface
          if(frameToBeTransferred.getJNWLength() < 200) {
              eventStopTime = eventStartTime;
              UpdateGlobalSimultationTime ( eventStopTime );
              new ReceivePhysicalFrameIntoInterface (
                  eventStartTime,
                  frameToBeTransferred,
                  interfaceCard.interfaceNumber );
              return;
          }
          JNWBitSet IPPacket =
              ModelDataLink.HDLCUtilities.GetIPPacketFromHDLCFrame(
                  frameToBeTransferred );
          long destinationAddress = ( int ) ModelTransport.IPUtilities .
              GetDestinationAddressFromIPHeader ( IPPacket );
      }
      
      // If this is a Token Ring
      else if(interfaceCard instanceof ModelDataLink.TokenRing) {
         Utility.err.println ( "ERROR IN ModelEvents"
            + ".RoutePacketsFromInterfaceToNodeBuffer.Execute(): Token Ring is not currently"
            + " implemented. Interface = " + interfaceNumber );
      }  // end of if ( frameToTransmit == null )

      // For a LAN node or an Ethernet bus node that is the destination
      Utility.JNWBitSet IPPacketToBeTransferred = null;
      if(thisInterfaceIsHDLC) {
          IPPacketToBeTransferred =
              ModelDataLink.HDLCUtilities.GetIPPacketFromHDLCFrame (
              frameToBeTransferred );
      } else if(interfaceCard instanceof ModelDataLink.Ethernet) {
          IPPacketToBeTransferred =
              ModelDataLink.EthernetUtilities.GetIPPacketFromEthernetFrame (
              frameToBeTransferred );
      }

      // Should the IP header checksum be verified?  If so is it true?
      boolean checksumIsValid =
         ModelTransport.IPUtilities.VerifyIPHeaderChecksum (
             IPPacketToBeTransferred );
      if (!ModelData.GlobalVariables.verifyChecksumsAtReceiver ||
          (ModelData.GlobalVariables.verifyChecksumsAtReceiver
              && checksumIsValid ) )
      {
         eventStopTime = eventStartTime +
         timeToMovePacketFromInterfaceBufferToNodeBufferInTicks;
         UpdateGlobalSimultationTime ( eventStopTime );
         // Create the next event
         new ModelEvents.RoutePacketAtNodeOrProcessAtDestination (
             eventStopTime ,
             thisNodeNumber,
             IPPacketToBeTransferred );
      }  // end of if (ModelData.GlobalVariables.verifyChecksumsAtReceiver == false ...
      else Utility.out.println("ERROR - INVALID HEADER CHECKSUM AT NODE:"+
          thisNodeNumber+" INTERFACE:"+interfaceCard.interfaceNumber);
   }   // end of Execute

}
