/*
 * ReceivePhysicalFrameIntoInterface.java
 *
 * Created on July 11, 2007, 3:59 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelEvents;


/**Receives the frame into the interface card.
 *
 * @version 1.202
 * @author dixonw
 * @modified JMP 21Oct2011 for JDLC3/HDLC
 */
public class ReceivePhysicalFrameIntoInterface extends ModelEventsBaseClass
   implements ModelConstants . Node , GUIConstants . General
{

   /**The received frame*/
   Utility.JNWBitSet frameBeingReceived  ;

   /**The interface.*/
   ModelDataLink.InterfaceBaseClass interfaceCard = null ;

   /** Creates a new instance of ReceivePhysicalFrameIntoInterface */
   public ReceivePhysicalFrameIntoInterface ( long eventStartTime
      , Utility.JNWBitSet incomingFrame , int interfaceReceivingFrame )
   {
      super ( eventStartTime ) ;
      frameBeingReceived = incomingFrame ;
      interfaceNumber = interfaceReceivingFrame ;
      interfaceCard = vectorOfInterfaces . get ( interfaceNumber ) ;
   }


   /**Executes the receive.  This accepts the frame from the link between the nodes.*/
   public void ExecuteEvent ( )
   {  // Update the receive time
      eventStartTime = interfaceCard.updateReceiveTime ( eventStartTime );
      // Update the stop time because the eventStopTime is needed 
      // for either branch of the if statement below
      int linkNumber = interfaceNumber / 2 ;
      ModelPhysical.Link link = vectorOfLinks . get ( linkNumber  ) ;
      long timeToReceiveFrameinTicks
         = ( long ) ( frameBeingReceived.getJNWLength ()
         * globalVariables.timeTicksPerSecond
         / ( 1000.0 * link . bitRateSourceToDestination ) );
      eventStopTime = eventStartTime +  timeToReceiveFrameinTicks ;
      // Should the checksum be verified?  If so is it true?
      if ( ModelData.GlobalVariables.verifyChecksumsAtReceiver == false
         || ( ModelData.GlobalVariables.verifyChecksumsAtReceiver == true
         && ModelDataLink.InterfaceBaseClass.verifyAtReceiver ( frameBeingReceived  ) == true ) )
      {
         // JNW treatment of Ethernet has a layer violation here- should fix
         // it moves frame as "packet" direct to network layer
         if(interfaceCard instanceof ModelDataLink . Ethernet ) {
         //Put the frame from the physical link into the interface receive buffer
         interfaceCard . receiveBuffer . add ( frameBeingReceived  ) ;
           // Create the next event
           new MovePacketFromInterfaceToNodeBuffer ( eventStopTime
             , interfaceNumber  ) ;
         }
         if(interfaceCard instanceof ModelDataLink . HDLC ) {
           // pass the frame to HDLC receive
           // (NOTE:this does not comply with JNW procedure which
           // would add the frame to buffer and poll it back out):needsfixed
           ((ModelDataLink.HDLC)interfaceCard).receiveHDLCFrame(
              frameBeingReceived,
              link . bitRateSourceToDestination);
         }
      }  // end of if ( ModelData.GlobalVariables.verifyChecksumsAtReceiver == false...

      // Important: simply drop the frame, no NACK will be sent (BZ)
      else
      {  // Else simply drop the frame and do NOT create a new Event
         if ( ModelData.GlobalVariables.verifyChecksumsAtReceiver == true &&
            ModelData.GlobalVariables.printReceiverErrors == true )
         {
            Utility.err.println ("ERROR: In receiving frame into interface " + interfaceNumber ) ;
         }
      }
      UpdateGlobalSimultationTime ( eventStopTime ) ;
   }  // end of Execute


   /**Removes the packet with the Ethernet, HDLC, or token ring header.
    THIS IS NEVER USED.*/
   public Utility.JNWBitSet RemovePacketFromFrame ( Utility.JNWBitSet frameReceived )
   {
      Utility.JNWBitSet packetReceived = new Utility.JNWBitSet ( ) ;
      if ( interfaceCard instanceof ModelDataLink . Ethernet )
      {
         Utility . out . print ( "Interface " + interfaceNumber + " converting packet to frame simulationTime = "
            + ModelData.GlobalVariables.simulationTime + DOUBLE_CARRIAGE_RETURN ) ;
         //Remove the Ethernet headers
         packetReceived
            = ModelDataLink . EthernetUtilities . GetIPPacketFromEthernetFrame ( frameReceived ) ;
      }
      else
      {  
         Utility.err.println  ("ERROR in action SendPhysicalFrameFromInterface for interface"
            + interfaceNumber + " unrecognized interfacce type: "
            + interfaceCard.getClass ( ) .getName ( ) ) ;
      }
      return frameReceived ;
   }
}
