/*
 * SendEthernetFrameFirstTime.java
 *
 * Created on August 13, 2007, 11:12 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package ModelEvents;

import ModelData.GlobalVariables.networkTypes;
import Utility.*;

/**A ModelEvents event to try to send an Ethernet
 * frame over a bus.
 * @version 1.200
 * @author dixonw*/
public class SendEthernetFrameFromInterface
       extends ModelEventsBaseClass {

   /**The frame to send.*/
   Utility.JNWBitSet frameToTransmit = null;

   /**The source interface number.*/
   int sourceInterfaceNumber;

   /**The destination interface number.*/
   int destinationInterfaceNumber;

   /** Transmission attempt*/
   int transmissionAttemptNumber;

   /** Time delay.*/
   long timeDelay;

   /**The Ethernet Bus*/
   ModelPhysical.EthernetBus ethernetBus
      = ( ModelPhysical.EthernetBus ) vectorOfLinks.get (0);

   /**Creates a new instance of SendEthernetFrameFromInterface.*/
   public SendEthernetFrameFromInterface (
           long eventTime ,
           Utility.JNWBitSet frameToBeTransmitted ,
           int sourceInterfaceNumberIn ,
           int destinationInterfaceNumberIn) {
      super (eventTime);
      eventStartTime = eventTime;
      frameToTransmit = frameToBeTransmitted;
      sourceInterfaceNumber = sourceInterfaceNumberIn;
      destinationInterfaceNumber = destinationInterfaceNumberIn;
      transmissionAttemptNumber = 0;
      timeDelay = 0;
   }

   /**Execute the transmission attempt.*/
   public void ExecuteEvent () {
      // Increment the attempt number
      // If the transmission number exceeds the limit, give up
      if ( ( transmissionAttemptNumber = transmissionAttemptNumber + 1 )
      > ModelDataLink.EthernetState . maximumNumberOfTransmissionAttempts ) {
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1      
         Utility.err.println ("\nERROR : on interface "
            + sourceInterfaceNumber + " Ethernet exceeded "
            + ModelDataLink.EthernetState.maximumNumberOfTransmissionAttempts
            + " transmission attempts-Frame discarded");
         // Give up transmission attempts
         return;
      }  // end of if ( transmissionAttemptNumber > ModelDataLink.EthernetState...
      // Compute the time delay until the next transmission attempt
      timeDelay 
         = ModelDataLink.EthernetUtilities.ComputeBackoff (transmissionAttemptNumber);

      ModelDataLink.Ethernet interfaceCard
         = ( ModelDataLink.Ethernet ) vectorOfInterfaces . get (sourceInterfaceNumber) ;
      // Does the transmission time occur while receiving another message?
      ////THE STUDENT SHOLD LOOK AT THIS TEST WHICH IS MENTIONED IN EXERCISE JLAN1////
      if (eventStartTime >= interfaceCard . startReceiveEthernetFrame &
          eventStartTime <= interfaceCard . stopReceiveEthernetFrame ) {
         // The interface is currently receiving a transmission so try again later
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1       
            Utility.war.println ("The transmission of the Ethernet message IS BLOCKED"
            + " at time " + eventStartTime + " Will try again at time " 
            + ( eventStartTime + timeDelay ));
         // This transmission attempt should not be counted so reverse 
         // the first statement in this method
         transmissionAttemptNumber = transmissionAttemptNumber - 1 ;
         // The interface is currently receiving a transmission.
         // The transmission is blocked.  Set the next transmission attempt
         RestoreThisEventAtNewEventTime (eventStartTime + timeDelay);       
      }  // end of if ( eventStartTime >= interfaceCard .startReceiveEthernetFrame &
      
      // START THE TRANSMISSION
      else { // Start the transmission sequence
         // Permission exists to transmit, but if the bus is already in use
         // a collision will occur.      
         if (  ethernetBus.busInUse == true  ) {
            // A collision will occur so skip the transmission
            // This is a short-cut.  This message is never sent since you know
            // a priori that there will be a collision
            ethernetBus.collisionHasOccurred = true;
            // Try again at a randomly chosen time.  Randomly draw the time to try again.
            // Increment the event time to simulate the noise timeout.
            eventStopTime = eventStartTime ////+ ModelDataLink.EthernetState.ethernetSlotSizeInTicks
            // Plus the time delay from the MAC layer-computed above
               + timeDelay;
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
////REMOVE THIS STATEMENT BELOW FOR EXERCISE JLAN1
           Utility.err.println ("ERROR : Ethernet Bus COLLISION AT TIME " + eventStartTime 
             + " will retransmit at time " + eventStopTime ); 
           eventStartTime = eventStopTime ;
           UpdateGlobalSimultationTime ( eventStopTime );
           //  Set the next transmission attempt
           RestoreThisEventAtNewEventTime (eventStartTime);           
         }  // end of if (  ( ethernetBus.numberOfSimultaneousTransmissions ...
         else {  // The bus is clear so just transmit  
            UpdateStartAndStopTimesAtOtherInterfaces (eventStartTime);
            ethernetBus.busInUse = true ;
            new TransmitViaEthernetBus (
                    eventStartTime ,
                    frameToTransmit ,
                    sourceInterfaceNumber ,
                    destinationInterfaceNumber ,
                    eventSequenceNumber);
             //  Set the next transmission attempt
            // CAREFUL: RestoreThisEventAtNewEventTime MUST FOLLOW 
            // new TransmitViaEthernetBus ( eventStartTime... BECAUSE RestoreThisEventAtNewEventTime
            // AS ITS FIRST OPERATION REDEFINES eventStartTime = newEventTime  ;
            RestoreThisEventAtNewEventTime ( eventStartTime + timeDelay );
         }  // end of else The bus is clear so just transmit
      }  // end of else Start the transmission sequence
   }  // end of public void ExecuteEvent ( )


   /** Updates the start and stop times that each interface will hear the tranmsitted
    *Ethernet frame.  The start time will be the eventTimeIn plus the delay time between
   the interfaces.  The stop time will be the start time plus the length of time required
   for the frame to pass the interface.*/
   private void UpdateStartAndStopTimesAtOtherInterfaces (long eventTimeIn) {
      //The transmit time for this frame
      long timeToTransmitFrameinTicks = ( long ) ( frameToTransmit.getJNWLength () * globalVariables.timeTicksPerSecond / ( 1000.0 * ethernetBus.bitRateEthernetBus ) );
      for ( int i = 0 ; i < vectorOfInterfaces.size () ; i ++ ) {
         ModelDataLink.Ethernet EthernetInterface 
            = ( ModelDataLink.Ethernet ) vectorOfInterfaces.get (i);
         // The time at which interface i will start hearing the transmission
         EthernetInterface.startReceiveEthernetFrame 
            = eventTimeIn + ModelPhysical.EthernetBus.transmissionDelayinTicks[sourceInterfaceNumber][i];
         // The time at which interface i will stop hearing the transmission
         EthernetInterface.stopReceiveEthernetFrame 
            = eventTimeIn + ModelPhysical.EthernetBus.transmissionDelayinTicks[sourceInterfaceNumber][i] + timeToTransmitFrameinTicks;
      }  // end of for ( int i = 0 ; i < vectorOfInterfaces.size() ; i++ )
   }  // end of UpdateStartAndStopTimesAtOtherInterfaces ( )

}  // end of class

