/*
 * SendFrameFromInterface.java
 *
 * Created on July 11, 2007, 2:57 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package ModelEvents;

import static ModelDataLink . HDLC . * ;


/**This class sends the frame from the interface card transmit buffer
 *to the link or Ethernet bus.
 *
 * @version 1.202
 * @author dixonw
 * @modified JMP 21Oct2011 for HDLC
 * and BZ April 2010
 */
public class SendFrameFromInterface extends ModelEventsBaseClass {

   /**The interface.*/
   ModelDataLink.InterfaceBaseClass interfaceCard = null;

   /**The source interface number. */
   int sourceInterfaceNumber;

   /**The packet to send*/
   Utility.JNWBitSet IPPacketOut = null;

   // The reliable transport entity
   //ModelTransport.Rtrans_Layer rtransport; // (BZ)

   /**The urgent state (for TCP urgent data)*/
   boolean isUrgent;

   /**Creates a new instance of SendFrameFromInterface.  The event time is
    * the time at which the frame starts to transmit.*/
   public SendFrameFromInterface (
              long eventStartTime ,
              int sourceInterfaceNumberInput ,
              Utility.JNWBitSet PacketOut /* IP Packet */,
              boolean isUrgentIn) { // (BZ)
      super ( eventStartTime );
      sourceInterfaceNumber = sourceInterfaceNumberInput;
      isUrgent = isUrgentIn;
      IPPacketOut = new Utility.JNWBitSet(PacketOut); // (BZ)
   }

   /**Executes the send.  This puts the frame on the link between the nodes.*/
   public void ExecuteEvent () {  // Get the interface
      interfaceCard = vectorOfInterfaces.get ( sourceInterfaceNumber );
      if(interfaceCard == null) {
          Utility.out.println("Null interface attached node: " + 
              interfaceCard.nodeToWhichAttached + "\n");
          return;
      }
      // Update the transmit time
      eventStartTime = interfaceCard.updateTransmitTime ( eventStartTime );
      //Get the frame from the transmit buffer (BZ)
      //Note that we bypass normal JNW buffering here
      Utility.JNWBitSet packetToTransmit = new Utility.JNWBitSet(IPPacketOut); // BZ
      //Utility.JNWBitSet packetToTransmit =
      //    interfaceCard . transmitBuffer . poll ();
      // packetToTransmit will never be null (BZ)
      if (packetToTransmit == null) {
         Utility.err.println ( "ERROR IN ModelEvents" +
            ".SendFrameFromInterface.Execute(): The sendBufferForPackets " +
            "for interface " + sourceInterfaceNumber + " is unexpectedly missing." +
            CARRIAGE_RETURN );
         return;
      }  // end of if ( frameToTransmit == null )

      // Make a DLC-layer frame from the packet retrieved (BZ)
      Utility.JNWBitSet frameToTransmit = WrapPacketIntoFrame ( packetToTransmit );
      eventStopTime = eventStartTime + timeToConvertPacketIntoFrame ;

      if(interfaceCard instanceof ModelDataLink.Ethernet) {
         // Send on an Ethernet Bus
         // Move first ethernet transmit time to next slot
         long slotSize = ModelDataLink.EthernetState.ethernetSlotSizeInTicks;
         long delay = ( ( eventStopTime + slotSize - 1 ) / slotSize ) * slotSize;
         eventStopTime = eventStopTime + delay;
         // Update the transmit to the end of this Event
         interfaceCard.updateTransmitTime ( eventStopTime );
         // Update the global simulation time to the maximum of the 
         // existing simulation time and the eventStopTime
         UpdateGlobalSimultationTime ( eventStopTime );
         // Get the destination interface needed only for Ethernet
         // Remember that the interface number on an Ethernet bus will be the
         // same as the node number
         int destinationInterfaceNumber = ( int ) ModelTransport.IPUtilities.
            GetDestinationAddressFromIPHeader ( packetToTransmit );
         // Create the next event
         SendEthernetFrameFromInterface se =
             new SendEthernetFrameFromInterface(
                 eventStopTime,
                 frameToTransmit,
                 sourceInterfaceNumber,
                 destinationInterfaceNumber);
      }  // end of Ethernet case

      if(interfaceCard instanceof ModelDataLink.HDLC) {
         // this is an HDLC serial link
         ModelPhysical.Link link = 
           vectorOfLinks.get ( sourceInterfaceNumber / 2 );
         // Get the number of the interface at the other end of the link
         int interfaceAtOtherEndOfLink = ModelPhysical.Link.
            InterfaceNumberAtOtherEndOfLink ( sourceInterfaceNumber );

         // Get the host node number (BZ)
         thisNodeNumber = interfaceCard.nodeToWhichAttached;
         sourceNodeNumber = ( int ) ModelTransport.IPUtilities.
             GetSourceAddressFromIPHeader ( packetToTransmit );

         // enqueue the packet for DLC transmit
         ModelDataLink.HDLC interfaceAtThisEndOfLink =
             (ModelDataLink.HDLC)vectorOfInterfaces.get(sourceInterfaceNumber);
         interfaceAtThisEndOfLink.enqueueHDLCFrame(frameToTransmit,isUrgent);
      } // end of HDLC case

         // Update the transmit to the end of this Event
         interfaceCard.updateTransmitTime ( eventStopTime );
         // Update the global simulation time to the maximum of the 
         // existing simulation time and the eventStopTime
         UpdateGlobalSimultationTime ( eventStopTime );
   }  // end of Execute


   /**Wraps the packet with the Ethernet, HDLC, or token ring header.  MOVE THIS TO THE
    * INTERFACE CLASS AND CALL BY POLYMORPHISM*/
   public Utility.JNWBitSet WrapPacketIntoFrame ( 
      Utility.JNWBitSet packetToTransmit ) {
      Utility.JNWBitSet frameToTransmit = new Utility.JNWBitSet (packetToTransmit);

      // Wrap the packet into an Ethernet frame
      if (interfaceCard instanceof ModelDataLink.Ethernet) {
         frameToTransmit = ModelDataLink.EthernetUtilities.
            AddEthernetHeaderAndTrailer ( packetToTransmit );
      }
      else if(interfaceCard instanceof ModelDataLink.HDLC) {
          frameToTransmit = ModelDataLink.HDLCUtilities.
            AddHDLCHeaderAndTrailer ( interfaceCard,false,packetToTransmit ) ;
      }
      else {
          
         Utility.err.println ( "ERROR in action SendPhysical" +
            "FrameFromInterface for interface" + sourceInterfaceNumber +
            " unrecognized interfacce type: " + interfaceCard.getClass ().
            getName () );
      }
      return frameToTransmit;
   }

}
