/*
 * DisplayEMail.java
 *
 * Created on July 14, 2007, 1:22 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelEvents;
import Utility.JNWBitSet ;

/**A class to start the sending of a TCP frame.  The class constructor
 *receives the frame and other necessary information for transmission.
 *Using information in the frame, the class instance also creates
 *a reference to the correct TCP connection.
 *<p><b>Versions:</b>
 * <li><b>Version 1.200</b> of 20081019 Changed      
 *    <li>new TransmitViaLink ( eventStartTime
        , linkNumber
        , interfaceAtOtherEndOfLink
        , frame.clone ( )  ) ;</li>
 * TO 
 *      <li>new TransmitViaLink ( eventStartTime
        , linkNumber
        , interfaceAtOtherEndOfLink
        , frame.clone ( )  ) ;</li> so that errors in TCP packets are not
 * automatically duplicated in the retransmissions.</li>
 * <li><b>Version 1.201</b> of 20081028 changed ExecuteEvent () to decrease
 * tcpConnection . unacknowledgedBytesInTransit when the number of transmission
 * attempts exceeds the limit.  Otherwise, the deleted IP packets will not
 * be replaced by other packets.
 * @version 1.201
 * @author dixonw
 * @modified BZ May 2010 and JMP 14Nov2011 for JTRN3
 */
public class SendTCPFrame extends ModelEventsBaseClass {

/**The frame to send.  This is a frame, not a packet.*/
Utility.JNWBitSet frame ; 
        
/**The number of the link over which this TCP frame is being sent.*/
int linkNumber ;

/**The interface at this end of the link*/
ModelDataLink.HDLC interfaceAtThisEndOfLink;

/**The number of the destination interface.*/
int interfaceNumberAtOtherEndOfLink ;

/**The number of transmission attempts.*/
int transmissionAttemptNumber ;

/**The length of data in the frame in bytes.  This is
*initialized in the constructor.  This may not be necessary.*/
int dataLength ;

/*The TCP connection for this frame.*/
 ModelTransport  . TCPConnection tcpConnection = null ;

 //JNWBitSet SyncPacket;
 
/**Get a reference to the discrete event simulator*/
 ModelDES.DiscreteEventSimulator discretEventSimulator
    = ModelDES.DiscreteEventSimulator.getInstance() ;

/**Creates a new instance of SendTCPFrame.*/
    public SendTCPFrame( 
               long eventTime ,
               int  linkNumberIn ,
               ModelDataLink.HDLC interfaceAtThisEndOfLinkIn ,
               int interfaceNumberAtOtherEndOfLinkIn ,
               Utility . JNWBitSet IPFrameIn ) {
        super ( eventTime ) ;
        linkNumber = linkNumber ;
        interfaceNumberAtOtherEndOfLink =
            interfaceNumberAtOtherEndOfLinkIn ;
        frame = IPFrameIn  ;
        // Initialize the number of transmission attempts      
        transmissionAttemptNumber = 1 ;
        dataLength = GetDataLength ( frame ) ;
        // Get a reference to the TCP connection which contains data
        // necessary to decide whether to transmit or not.
        long originalSourceNodeAddress 
           = ModelDataLink . EthernetUtilities .
               GetSourceAddressFromIPHeader( frame ) ;
        ModelPhysical . Node originalNode = vectorOfNodes .
            get ( ( int ) originalSourceNodeAddress );
        long originalDestinationNodeAddress 
           = ModelDataLink . EthernetUtilities .
               GetDestinationAddressFromIPHeader( frame ) ;

        // Get information about the TCP connection, in order to perform 
        // transport layer related operations (BZ)
        tcpConnection = originalNode . tcpConnection;         
        tcpConnection . lastTransmissionAttemptExceedWindowSize = false ;
    }  // end of constructor


/** Sends a frame containing a TCP segment packet.*/
 public void ExecuteEvent ( ) {
     // Has the maximum number Of TCP transmission Attempts been exceeded?
/*if ( ModelDataLink . EthernetUtilities . GetSourceAddressFromIPHeader( frame ) == 2 
   & ModelDataLink . EthernetUtilities . GetDestinationAddressFromIPHeader( frame ) == 4 ) {       
System.out.println ( "IN SendTCPFrame.ExecuteEvent source address  = " 
   + ModelDataLink . EthernetUtilities . GetSourceAddressFromIPHeader( frame )
   + " source address  = "
   + ModelDataLink . EthernetUtilities . GetDestinationAddressFromIPHeader( frame ) ) ;
System.out.println ( "IN SendTCPFrame.ExecuteEvent: DES = \n "
   + ModelDES.DiscreteEventSimulator.getInstance ().toString () ) ;
 }
 */

     // Maximum number of transmission attempts exceeded (bo)
     if ( ( transmissionAttemptNumber )        
        > tcpConnection . maximumNumberOfTCPTransmissionAttempts ) {
        // Decrease number of bytes unacknowledged
         tcpConnection . unacknowledgedBytesInTransit 
            = tcpConnection . unacknowledgedBytesInTransit - dataLength ; 
        // Print error message.
         Utility.err.println  ( "ERROR in ModelEvents.SendTCPFrame.ExecuteEvent:"
           + " Exceeded maximum number of transmission atttempts: "
           + tcpConnection . maximumNumberOfTCPTransmissionAttempts 
            + "\n  Terminating transmission, event number = " + this . eventSequenceNumber ) ;
         return ; // Return without creating another Event.  This stops transmission
     }  // end of if ( ( transmissionAttemptNumber  )

     // This is not the first attempt (BZ)
     if ( transmissionAttemptNumber > 1 ) {        
         // If this a later transmission attempt it means that the 
         // ACK message has not been received in time, otherwise 
         // the ACK message would delete this event.
         // If the retransmission was NOT due to exceeding the window size
         // then it was due to the failure to receive an ACK message in time
         // Therefore decrease the window sizes
         if ( tcpConnection . lastTransmissionAttemptExceedWindowSize == false ) {           
           tcpConnection . WindowSizeUpdateIncreaseIfTrueDecreaseIfFalse ( false ) ;    
         }  // end of if
     }  // end of if ( transmissionAttemptNumber > 1 )

     // this is the first attemp of the packet (bo)
     // Is the sliding window full?
     // dataLength: the length of the outgoing frame (bo)
     if ( ( dataLength + tcpConnection . unacknowledgedBytesInTransit 
         > tcpConnection . slidingWindowSize ) & transmissionAttemptNumber == 1 )   {      
     tcpConnection . lastTransmissionAttemptExceedWindowSize = true ;
     // Schedule next transmission re-attempt
     // Put this Event back into the DES delayed by time RTT
     // Wait one RTT before attempt to retransmit
     
     // This packet is out of the sliding window, do not transmit now, schedule a later transmission (bo)
     // The re-attempt of this frame is scheduled to be one RTT time later (bo)
     RestoreThisEventAtNewEventTime( eventStartTime + tcpConnection . RTT  ) ;    
     return ; // Return without starting transmission  
     }  // end of if ( dataLength + tcpConnection . unacknowledgedBytesInTransit

     // Otherwise, the frame to be transmitted does not exceed window size (bo)
     tcpConnection . lastTransmissionAttemptExceedWindowSize = false ;
     // Important (bo)
     // transmit now
     // Pass a clone of the frame.  If en error is introduced during transmission,
     // the samme error will be in all the retransmissions.
     new TransmitViaLink ( 
             eventStartTime ,
             linkNumber ,
             interfaceAtThisEndOfLink,
             interfaceNumberAtOtherEndOfLink ,
             frame.clone ( )  ) ;

     // Increment transmission attempt number
     transmissionAttemptNumber = transmissionAttemptNumber + 1 ;

     // Update the stored TCP packet information
     long dataSequenceNumber
        = ModelTransport.TCPUtilities.GetTCPSequenceNumberFromFrame( frame ) ;
     // Schedule a retransmission event first, if current transmission succeeds, then remove retransmission later (bo)
     // Schedule next transmission re-attempt
     // Put this Event back into the DES delayed by time
     // TRR plus 4 D before attempt to retransmit     
     RestoreThisEventAtNewEventTime( eventStartTime + tcpConnection . RTT 
        + 4 * tcpConnection . D   ) ;    
     return ; 
    }  // end of Execute

 
 /**Returns the data length in bytes from a frame.  The data length
  * is the frame length minus all the frame overhead minus the 
  * IP headers length minus the TCP header length.
  * <p>Currently this works <b><u>ONLY</b></u> for Ethernet frames.  <b><u>IS THIS RIGHT?</b></u>
  @see http://alevi.usc.edu/Html/datacom/sld024.htm*/
 private int GetDataLength ( Utility . JNWBitSet frameIn ) {
    // This is Ethernet
    dataLength = frameIn .getJNWLength ( ) / 8
       - ( 8 /*Ethernet preamble length*/
          + 6 /*Ethernet destination address length*/  
          + 6 /*Ethernet source address length*/ 
          + 2 /*Ethernet length field length*/ 
          + 4 /*Ethernet CRC length*/ )
       - 20 /*IP header length*/
       - 20 /*TCP header length*/ ;
    return dataLength ;
 }  // end of GetDataLength 
 
 
/**A method for testing only.*/    
    public static void main_Testing_Only ( String [ ] args ) {
System.out.println   ("Displaying test " ) ;
        Utility.JNWBitSet one = new Utility.JNWBitSet ( "00000000 00000000 00000000 00000000 " );
        Utility.JNWBitSet data = new Utility.JNWBitSet (  );
        data.append ( one) ;
        data.append ( one) ;
        data.append ( one) ;
        data.append ( one) ;
        data.append ( one) ;
        data.append ( one) ;
        data.append ( one) ;
        Utility.JNWBitSet theString = new Utility.JNWBitSet (  );
        theString.setJNWBitSetToString ( "NOW IS THE TIME!");
        data.append ( theString ) ; 
System.out.println   ("Displaying input " + data.toString ( )  ) ;        
        ProcessMessageAtDestination test = new  ProcessMessageAtDestination (  0
           , 5 
           , data ) ;
        test.ExecuteEvent ( ) ;
    } // end of main 
    
}  // end of class 


/*
     Utility.JNWBitSet TCP_IPPacket 
   = ModelDataLink.EthernetUtilities.GetIPPacketFromEthernetFrame ( frame ) ;     
System.out.println ("\n\nIN SendTCPFrame"
   + "\n source node = " + ModelTransport.IPUtilities.GetSourceAddressFromIPHeader ( TCP_IPPacket )
   + "\n destination node = " + ModelTransport.IPUtilities.GetDestinationAddressFromIPHeader ( TCP_IPPacket )
   + "\n sequence number = " + ModelTransport.TCPUtilities.GetTCPSequenceNumberFromIPPacket ( TCP_IPPacket  )
   + "\n message = \n " + ModelTransport.TCPUtilities.GetDataFromTCP_IPPacket ( TCP_IPPacket  ).convertJNWBitSetToString () ) ;
*/          
     

////Utility.JNWBitSet IPPacket 
////= ModelDataLink.EthernetUtilities.GetIPPacketFromEthernetFrame ( frame ) ;
////dataSequenceNumber = ModelTransport.TCPUtilities.GetSequenceNumberFromTCPPacket ( IPPacket ) ;  
////System.out.println ("IN SendTCPFrame: sending TCP packet sequence number = " + dataSequenceNumber ) ;  


////System.out.println ("\nIN SendTCPFrame: starting event number = " + eventSequenceNumber ) ;

////System.out.println ("IN SendTCPFrame: resending event = " + eventSequenceNumber ) ;  

////System.out.println ("IN SendTCPFrame: decreasing window size for event = " + eventSequenceNumber ) ;  

////System.out.println ( "IN SentTCPFrame, sliding window is full : dataLength = " + dataLength + " tcpConnection . unacknowledgedBytesInTransit = "
////   + tcpConnection . unacknowledgedBytesInTransit + " tcpConnection . slidingWindowSize = " + tcpConnection . slidingWindowSize ) ;  

////System.out.println ("IN SendTCPFrame: adding event number = " + eventSequenceNumber 
////   + " back into DES ") ;   