/*
 * SendPhysicalFrame.java
 *
 * Created on March 10, 2007, 11:19 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package ModelEvents;

import Utility.JNWRandom;

/**This sends a physical frame of data from one host to all the other hosts on
 * an Ethernet bus.  It first
 * adds random errors to the frame based on the link's bit error rate,
 * link . bitErrorRate.  It then computes the time needed to transmit the
 * frame plus the delay time for the link.  It then creates a ReceivePhysicalFrame
 * object for the time that the frame will finish reception at the receiver.
 * *<p><b>Versions:</b>
 * <li><b>Version 1.200</b> of 20081008 for some strange reason I had left
 * double randomNumber = ethernetBus . linkRandomNumberGenerator . nextDouble ();
 * in the code.  But ethernetBus has no linkRandomNumberGenerator, and execution of the
 * code did not throw a null pointer exception!!!  I changed to code to
 * double randomNumber = randomNumberGenerator . nextDouble ();</li>
 * @version 1.200
 * @author dixonw*/
public class TransmitViaEthernetBus extends ModelEvents.ModelEventsBaseClass {

   /**The frame to send */
   Utility.JNWBitSet frameBeingTransmitted = null;

   /**The random number generator used to generate bit errors.*/
   static JNWRandom randomNumberGenerator = JNWRandom.getInstance ();

   /**The source interface number. */
   int sourceInterfaceNumber;

   /**The destination interface number. */
   int destinationInterfaceNumber;

   /** Transmission attempt number*/
   int transmissionAttemptNumber;

   /**The number of the event which sent the Ethernet Frame.*/
   int sendEthernetFrameEventSequenceNumber;

   /**The Ethernet Bus*/
   ModelPhysical.EthernetBus ethernetBus = ( ModelPhysical.EthernetBus ) vectorOfLinks.get ( 0 );

   /**The Ethernet frame length in bits.*/
   int framelength;

   /**The ibt error rate on the Ethernet bus in error bits per bit.*/
   double bitErrorRate;

   /** Creates a new instance of SendPhysicalFrame and add itself to the discrete
   event simulator.*/
   public TransmitViaEthernetBus ( long eventTime , Utility.JNWBitSet frameBeingTransmitted , int sourceInterfaceNumber , int destinationInterfaceNumber , int sendEthernetFrameEventNumber ) {
      super ( eventTime );
      this.eventStartTime = eventTime;
      this.frameBeingTransmitted = frameBeingTransmitted;
      this.sourceInterfaceNumber = sourceInterfaceNumber;
      this.destinationInterfaceNumber = destinationInterfaceNumber;
      this.sendEthernetFrameEventSequenceNumber = sendEthernetFrameEventNumber;
   }

   /**Executes the send.  This introduces bit errors and a transmission delay.*/
   public void ExecuteEvent () {
      // Increment the number of transmissions   
      framelength = frameBeingTransmitted.getJNWLength ();        
      bitErrorRate = ethernetBus.bitErrorRate;     
      // Introduce errors
      // The algorithm below assumes that the bit error rate is small, 
      // that is: 1 - N *ber = ( 1 - ber) **N
      // This means that I draw the random number only once 
      double randomNumber = randomNumberGenerator . nextDouble ();        
      // Compute the number of bit errors.  The distribution is uniform with mean
      // value framelength * bitErrorRate
      int numberofErrors = ( int ) ( 2.0 * framelength * bitErrorRate * randomNumber + 0.5 );      
      // If there are errors, add them in     
      if (numberofErrors > 0) {
         for ( int i = 0 ; i < numberofErrors ; i ++ ) {
            int position = ethernetBus . linkRandomNumberGenerator . nextInt ( framelength );
            // Careful, I could reflip the same bit
            frameBeingTransmitted.flip ( position );
            // Output the error if wished
            if (ModelData.PrintPermissions.bitErrorGenerate == true) {
               Utility.war.println ( "ERROR in bit " + position + " of Ethernet frame " );
            }  // end of if ( ModelData.PrintPermissions.bitErrorGenerate == true )
         }  // end of for ( int i = 0 ; i < numberofErrors ; i++ )
      }  // end of if ( numberofErrors > 0  )        
      //Get the time delay table
      eventStopTime = eventStartTime 
         + ModelPhysical.EthernetBus.transmissionDelayinTicks[sourceInterfaceNumber]
         [destinationInterfaceNumber];
      new ReceivePhysicalFrameIntoEthernetBusInterface ( eventStopTime 
         , frameBeingTransmitted , sourceInterfaceNumber 
         , destinationInterfaceNumber , sendEthernetFrameEventSequenceNumber  );
   }  // end of Execute

} // end of class 