/*
 * SendPhysicalFrame.java
 *
 * Created on March 10, 2007, 11:19 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package ModelEvents;

import Utility.*;

/**This sends a physical frame of data from one host to another host.  It first
 * adds random errors to the frame based on the link's bit error rate,
 * link . bitErrorRate.  It then adds the delay time for the link.  It then
 * creates a ReceivePhysicalFrame object for the time that the frame will
 * start reception at the receiver.
  *<p><b>Versions:</b>
 *<li><b>Version 1.200</b> of 20080822 changes the distribution of
 * errors in a frame from the bastardized version of an Exponential
 * distribution to a Poisson distribution .</li>
 * @version 1.202
 * @author dixonw
 * @modified JMP 28Oct2011 for HDLC
 */
public class TransmitViaLink extends ModelEvents.ModelEventsBaseClass {

   /**The frame to send */
   Utility.JNWBitSet frameBeingTransmitted = null;

   /**The interface at this end of the link*/
   ModelDataLink.HDLC interfaceAtThisEndOfLink;

   /**The number of the interface at the other end of the link*/
   int interfaceNumberAtOtherEndOfLink = Integer.MIN_VALUE;

   /** Creates a new instance of SendPhysicalFrame and add itself to the discrete
    * event simulator.*/
   public TransmitViaLink ( 
              long eventStartTime ,
              int linkNumberIn ,
              ModelDataLink.HDLC interfaceAtThisEndOfLinkIn,
              int interfaceNumberAtOtherEndOfLinkIn ,
              JNWBitSet frameBeingTransmittedIn ) {
      super ( eventStartTime );
      frameBeingTransmitted = frameBeingTransmittedIn;
      linkNumber = linkNumberIn;
      interfaceAtThisEndOfLink = interfaceAtThisEndOfLinkIn;
      interfaceNumberAtOtherEndOfLink = interfaceNumberAtOtherEndOfLinkIn;
   }

   /**Executes the send.  This introduces bit errors and a transmission delay.*/
   public void ExecuteEvent () {
      ModelPhysical.Link link = vectorOfLinks.get ( linkNumber );
      int frameLength = frameBeingTransmitted.getJNWLength ();
      double bitErrorRate = link.bitErrorRate ;

      // Compute the number of bit errors.  This assumes that the bit error rate is small
      // so that 1 - N*ber =(approximately) ( 1 - ber)**N.  This means that
      // I draw the random number only once.  The number is a Poisson distribution.
      int numberofErrors
         = link.linkRandomNumberGenerator .nextPoisson ( frameLength * bitErrorRate ) ;
      // If there are errors, apply them (UDP only; TCP errors are in node)
      if(GUIActions.ReadEMailTextFilesAndPutIntoSourceNodeBufferAction.
          transmissionProtocol !=
            ModelData.GlobalVariables.transmissionProtocols.TCP)
      if (numberofErrors > 0) {
         for ( int i = 0 ; i < numberofErrors ; i ++ ) {
            // pick the position of the next bit error
            int position = link.linkRandomNumberGenerator.nextInt ( frameLength + 1 );
            // Careful, I could reflip the same bit
            frameBeingTransmitted.flip ( position );
            // Print the error if wished
            if (ModelData.PrintPermissions.printBitErrors == true) {
               Utility.err.println ( "ERROR in TransmitViaLink for Link "
                  + linkNumber + ": Error in bit "
                  + position + " of frame " );
            }  // end of if ( ModelDatas.PrintPermissions.printBitErrors == true )
         }  // end of for ( int i = 0 ; i < numberofErrors ; i++ )
      }  // end of if ( numberofErrors > 0  )
      // NOTE that here I assume that the rates are the same in both directions
      eventStopTime = eventStartTime + link . transmissionDelayinTicks ;
      ReceivePhysicalFrameIntoInterface rpfii =
          new ReceivePhysicalFrameIntoInterface (
              eventStopTime ,
              frameBeingTransmitted ,
              interfaceNumberAtOtherEndOfLink );
      // after send is done wake up sending interface hdlcSend
      // to send next frame (no timeout delay; just a wakeup)
      WakeupHDLC th =
          new WakeupHDLC(eventStopTime,interfaceAtThisEndOfLink);
      UpdateGlobalSimultationTime ( eventStopTime );
   }  // end of Execute

public static void main ( String [] args ) {
   
      int numberofErrors = 0 ;
      // If there are errors, add them in
////      if (numberofErrors > 0) {
         for ( int i = 0 ; i < numberofErrors ; i ++ ) {
            System.out.println (" Error number = " + i ) ;

         }  // end of for ( int i = 0 ; i < numberofErrors ; i++ )
////      }  // end of if ( numberofErrors > 0  )
System.exit ( 1 ) ; 
   for ( float x = ( float ) 0.0 ; x < 11000.0  ; x = x + ( float )1.0  ) {
      int j =  ( int ) ( x / 10000.0 ) ;
   System.out.println (" x = " + x + " j = " + j ) ;
   }
}  // end of main_For_Test_Only


}  // end of class TransmitViaLink



         
/*System.out.println ( "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nIn TransmitViaLink:                 XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX" 
   + "\nXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"  
   + " framelength = " + framelength 
   + " bitErrorRate = " + bitErrorRate 
   + " numberofErrors = " + numberofErrors ) ;
*/


////System.out.println ( "In TransmitViaLink: error in bit " + position  ) ; ////link = " + link   

/*System.out.println ( "In TransmitViaLink: passing frame to interface " ) ; 
Utility.JNWBitSet IPPacket 
   = ModelDataLink.EthernetUtilities. GetIPPacketFromEthernetFrame ( frameBeingTransmitted ) ;
Utility.JNWBitSet tcpPacket 
   = ModelTransport.TCPUtilities.CloneTCPPacketFromIPPacket ( IPPacket ) ;
Utility.JNWBitSet data 
   = ModelTransport.TCPUtilities.GetDataFromTCPPacket ( tcpPacket ) ;
String dataAsString = data.convertJNWBitSetToString ( ) ;
System.out.println ("IN TransmitViaLink:"
   + "TCP packet =\n" + IPPacket.toString () 
   + "\nas bit set, data = ***\n" + data.toString ( )+"***"
   + "\nas String, data = \n" + dataAsString +"***"); 
*/