/*

 * Routing.java

 *

 * Created on June 22, 2007, 7:28 AM

 *

 * To change this template, choose Tools | Template Manager

 * and open the template in the editor.

 */



package ModelNetwork;



import ModelPhysical.Link;



/**
 * Implements Dijkstra's algorithm.  This routing class Routing is intended 
 * to be used by the students to satisfy exercise JWAN2 of the textbook.
 * <p>There are two
 * differences between the Routing and Routing.
 * <li>1  Routing implements Dijkstra's algorithm
 * as shown on pages 109 and 110 of <i>Internet protocols Through Hands-On Programming</i>.
 * JNW implements Dijkstra's algorithm using a heap as described in <i>Introduction to 
 * Algorithms</i> bu Udi Manber.  The implementstaion with a help is faster but much more
 * complex.
 * 
 * <li>2  Routing returns a routing table, routingMatrix, containing the next interface in
 * the route to the destination.  To make the route more understandable by students, Routing
 * routes the frames by node number.  Since JNW reallly routes by interface number, an additional
 * method, routingMatrix = ConvertNodeRoutingToInterfaceRouting ( ) must be called as:
 * <li>routingMatrix = ConvertNodeRoutingToInterfaceRouting ( ) ;</li>
 * to convert the node-based routing to interface-based routing.
 * @author pullenm
 * @version 1.100
 * @see <a href="http://en.wikipedia.org/wiki/Dijkstra's_algorithm">Dijkstra's Algorithm</a>
 * @see <a href="http://www.personal.kent.edu/~rmuhamma/Algorithms/MyAlgorithms/GraphAlgor/dijkstraAlgor.htm">Dijkstra's Algorithm</a>
 * @see <a href="http://www.dgp.toronto.edu/people/JamesStewart/270/9798s/Laffra/DijkstraApplet.html">Dijkstra's Algorithm Interactive (REALLY GOOD)</a>
 * @see <a href="http://www.cs.sunysb.edu/~skiena/combinatorica/animations/dijkstra.html">Dijkstra's Algorithm Interactive</a>
 */

public class Routing implements GUIConstants . General {



/**A singleton vector containing all the Nodes*/

static ModelData.JNWSingletonVectorOfNodes<ModelPhysical.Node> vectorOfNodes

            = ModelData.JNWSingletonVectorOfNodes.getInstance (  ) ;



/**A singleton vector containing all the links*/

static ModelData.JNWSingletonVectorOfLinks<ModelPhysical.Link> vectorOfLinks

            = ModelData.JNWSingletonVectorOfLinks.getInstance (  ) ;



/**A singleton vector containing all the link interfaces*/

static ModelData.JNWSingletonVectorOfInterfaces < ModelDataLink.InterfaceBaseClass >

    vectorOfInterfaces

       = ModelData.JNWSingletonVectorOfInterfaces.getInstance (  ) ;



/**Number of nodes in the topology*/

static private int numberOfNodes  = vectorOfNodes.size ( ) ;

////static private int numberOfNodes /* = vectorOfNodes.size ( ) */;



/**Number of links in the topology*/

static private int numberOfLinks = vectorOfLinks.size ( ) ;

////static private int numberOfLinks /*= vectorOfLinks.size ( ) */;



/**A matrix of cost for transmitting from node j to node k as seen from a transmission

 originiating from node i.  The three indices are, in order, i, j, and k.  This is the

 cost input to the model.*/

static float [ ] [ ] [ ] cost  ;

/**The routing matrix for all the nodes.  This is generated by Dijkstra's algorithm.  The i index

 is the source Node and the j index is the destinaion Node.  Note that the node numbers in this

 matrix are the sequential node numbers*/

static int [ ] [ ] routingMatrix

          = new int [ numberOfNodes ] [ numberOfNodes ] ;



/** A very large floating point number, larger than any route cost */

static final float HUGEFLOAT = (float)1E30;



/** Creates a new instance of Routing */

    public Routing() {

        ComputeCostMatrix ( ) ;

    }



/**Loads the routing information into the routingTables of all node

   and prints the cost and routing values*/

    public static void ComputeRoutingMatrices ( ) {

       for(int i = 0; i < numberOfNodes; ++i) //{

          routingMatrix[i] = ComputeNodeRoutingTable(i);

          //for(int j=0; j<numberOfNodes; j++)
          //    System.out.println(routingMatrix[j] + "\n");
       //}

       PrintTestOutput();

       routingMatrix = ConvertNodeRoutingToInterfaceRouting ( );

 

       // Now load the routing table into the nodes

       for(int i = 0; i < numberOfNodes; ++i) {

         ModelPhysical.Node node = vectorOfNodes . get ( i ) ;

         node . routingTable =  routingMatrix[ i ] ;
       }

    }



/** Use Dijkstra's algorithm without a heap to compute the optimal cost of 
 * sending a packet from source node to each other node. */

    public static int[] ComputeNodeRoutingTable(int source) {
       
  int[] forward_routers = new int[numberOfNodes];

  //// NEXT TWO STATEMENTS ALLOW PROJECT JDLC TO WORK
  //// STUDENT MUST REMOVE THEM WHEN CODING DIJKSTRA'S
  forward_routers[0] = 0;
  forward_routers[1] = 1;


//// STUDENT FILLS IN
//// STUDENT FILLS IN
//// NOTE THAT SUBSCRIPTS FOR THE NODES RUN FROM 0 TO NNETS -1
//// NOT 1 TO NNETS AS IN THE TEXTBOOK
//// THE ALGORITHM DESCRIPTION BELOW IS FROM THE TEXTBOOK
/*
      Dijkstra's route optimization algorithm for JNW
      required storage:
      array R that will represent Dijkstra's set R
      array C of costs associated with route from n to source;
 
      initialization ;
      initialize R[0] to source node  ;
      do for all n from 1 to nnets
      {
         initialize C[n] to cost, as seen from the source of the link,
            from source to n;
         initialize forward_routers.router [n] to n ;
      }
 
      do for all h from 1 to nnets
      {
         add to R the router with the least-cost path to source;
         set cost_of_h to a very large number ;
         do for all n from 1 to nnets,
         {
            do for all j from 0 to nnets
            {
               if any R[j] contains n skip this n;
               if C[n] is less than cost_of_h
               {
                  set cost_of_h to C[n] ;
                  set R[h] to n ;
               }
            } // end of do for all j from 0 to nnets
         }  // end of do for all n from 1 to nnets,
 
      update the least-cost paths:
      do for all n from 1 to nnets
      {
         find cost_of_h_to_n, as seen from source,
            of link from R[h] to n;
         if C[R[h]] plus cost_of_h_to_n is less than C[n]
         {
            set C[n] to C[R[h]] plus cost_of_h_to_n ;
            set forward_routers.router[n] to
               forward_routers.router[R[h]] ;
          } // end of if
 
          stopping rule-quit when all routers are in R ;
          do for all n from 1 to nnets,
            do for all j from 0 to h ,
               look for an n that is not in some R[j];
               if no such n is found, break out of this (do for h) loop
      }  // end of do for all n from 1 to nnets
 
      }  // end of do for all h from 1 to nnets
 */

  return forward_routers;

 }  // end of ComputeNodeRoutingTable ( )



/** Print function*/

 public static void PrintRouting ( ) {

     Utility.out.println ( "\nROUTING MATRIX" ) ;

     for ( int i = 0 ; i  < numberOfNodes ; i++ ) {

        for ( int j = 0 ; j  < numberOfNodes ; j++ ) {

           Utility.out.print ( "\t" + routingMatrix  [ i ] [ j ] ) ;

     }

     Utility.out.println ("\n") ;

     }

 }

 /** Print function*/

public static void PrintTestOutput ( ) {

     Utility.out.println ( "COST MATRIX (INPUT)" ) ;

     for ( int i = 0 ; i  < numberOfNodes ; i++ ) {

        for ( int j = 0 ; j  < numberOfNodes ; j++ ) {

           Utility.out.print ( "\t" + cost  [0] [ i ] [ j ] ) ;

     } Utility.out.println ("") ;

     }

    PrintRouting ( )  ;

 }



/**Computes the cost matrix for transmitting from node j to node k as seem from a transmission

 originiating from node i.  The three indices are, in order, i, j, and k.  Presently, this is

 set to the transmission delay from node j to node k.*/

 public static void ComputeCostMatrix ( ) {

    numberOfNodes = vectorOfNodes . size ( ) ;

    numberOfLinks = vectorOfLinks . size ( ) ;

    cost

       = new float [ numberOfNodes ] [ numberOfNodes ] [ numberOfNodes ]  ;

    for ( int i = 0 ; i < numberOfNodes ; i ++ ) {

       for ( int j = 0 ; j < numberOfNodes ; j ++ ) {

          for ( int k = 0 ; k < numberOfNodes ; k ++ ) {

             if(j == k)cost [ i ] [ j ] [ k ] = 0;

             else cost [ i ] [ j ] [ k ] = HUGEFLOAT ;

          }

       }

    }

    // Set the cost to the delay time between nodes

    for ( int l = 0 ; l < numberOfLinks ; l ++ ) {

        ModelPhysical.Link link = vectorOfLinks . get ( l ) ;

        int source = link . sourceNodeNumber ;

        int destination = link .destinationNodeNumber   ;

        float costPerHop = link . transmissionDelay ;

        for ( int i = 0 ; i < numberOfNodes ; i ++ ) {

           // the cost matrix is the same when seen from all nodes

           // later the ith node might have a different value

           cost [ i ] [ source ] [ destination ] = costPerHop;

           cost [ i ] [ destination ] [ source ] = costPerHop ;

        }

    }  // end of for ( int i = 0 ; i < vectorOfLinks . size ( ) ; i ++ )

 }  // end of ComputeCostMatrix ( )

 

    /**Converts the routing by node number to the routing by interface number.

    * JNW routes by interface number.  The student version, JNWS, produces

    * routing by node number as part of exercise JWAN2.  But JNW-the 
    * superset of JNWS-
    
    * routes by interface number. Therefore, in JNWS this method must be called

    *to perform the translation.*/

   

   private static int [ ] [ ] ConvertNodeRoutingToInterfaceRouting ( )

   {

      int [ ] [ ] routingMatrixByInterface

         = new int [ numberOfNodes ] [ numberOfNodes ] ;

      // clone the routingMatrix

      routingMatrixByInterface = routingMatrix . clone ( ) ;

      int arraySize = routingMatrix .length ;

      ModelPhysical.Node node = null ;

      for ( int i = 0 ; i < arraySize ; i++ )

      {

         node = vectorOfNodes . get ( i );

         for ( int j  = 0 ; j < arraySize ; j++ )

         {

            int destinationNodeNumber = routingMatrix [ i ] [ j ] ;

            if ( i != j )

            {

               for ( int k = 0 ; k < node . vectorOfInterfacesForThisNode . size ( ) ; k++ )

               {

                  int interfaceNumber =  node . vectorOfInterfacesForThisNode . get ( k ) . interfaceNumber ;

                  int linkNumber = interfaceNumber / 2 ;

                  ModelPhysical.Link link = vectorOfLinks . get ( linkNumber  ) ;

                  int otherNodeNumber = link . nodeAtOtherEndOfLink ( i ) ;

                  if ( routingMatrix [ i ] [ j ] == otherNodeNumber )

                  {

                     routingMatrixByInterface [ i ] [ j ] = interfaceNumber ;

                     // Skip the rest of the search

                     break ;

                  }  // end of if

               }  // end of for ( int k = 0 ; k < node . vectorOfInterfacesForThisNode.size ( ) ; k++ )

            }  // end of if ( i != j )

         else

         {

            routingMatrixByInterface [ i ] [ j ] = -1 ;

         }

         }  // end of for ( int j  = 0 ; j < arraySize ; j++ )            

      }  // end of for ( int i = 0 ; i < arraySize ; i++ )

      return routingMatrixByInterface ;

      

   }// end of  ConvertNodeRoutingToInterfaceRouting ( 

   

}  // end of class Routing

