/*
 * Link.java
 *
 * Created on December 20, 2006, 2:26 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package ModelPhysical;

/**Creates an Ethernet bus.  Only one Ethernet bus can be created
 * in a topology.  All the data is stored in vectorOfLinks.get ( 0 ) 
  *<p><b>Versions:</b>
 *<li><b>Version 1.200</b> of 20081105 changes the 
 * toString () method.</li>
 * @version 1.200
 * @author dixonw*/
public class EthernetBus extends ModelPhysical.Link implements GUIConstants.General {

   /**The singleton vector containing all the nodes*/
   ModelData.JNWSingletonVectorOfNodes<ModelPhysical.Node> vectorOfNodes
      = ModelData.JNWSingletonVectorOfNodes.getInstance();
   /**The singleton vector containing all the links*/
   ModelData.JNWSingletonVectorOfLinks<ModelPhysical.Link> vectorOfLinks
      = ModelData.JNWSingletonVectorOfLinks.getInstance();
   /**The singleton vector containing all the links*/
   ModelData.JNWSingletonVectorOfInterfaces<ModelDataLink.InterfaceBaseClass> vectorOfInterfaces
      = ModelData.JNWSingletonVectorOfInterfaces.getInstance();
   
////  /**Data rate in kbps on the Ethernet bus.  Equals the
////    *bit rate for the first link.*/
////   public int bitRateEthernetBus;

   /** Transmission delay on the Ethernet bus in microseconds between
    * one i and node j */
   public float[][] transmissionDelay;
   /** Transmission delay on the Ethernet bus in ticks between node i and node j */
   public static long[][] transmissionDelayinTicks;

   /**Has a collision occurred?*/
   public boolean collisionHasOccurred;
   
   /**The bus is already in use.*/
   public boolean busInUse ;

   /** Creates a new instance of an Ethernet Bus.
    *
    *<p>First the constructor accesses the links to compute the time delays 
    * between the different transceivers.  The port
    * numbers are not initialized here.  The port numbers are computed in
    * MainGUI_Actions.ReadTopologyFileAction.CreateArray ( ).  The link also
    * add the interface cards to the singletonVectorOfInterfaces.</p>*/
   public EthernetBus() {  //First compute the delay times between the different nodes

      ComputeTimeDelaysBetweenNodesOnEthernetBus();
      // Remove any nodes, links, or interfaces from the Vectors.  This could happen if the user
      // opens a second topology file.  Also reset the static numbers so that the links
      // and interfaces are numbered from 0                                                                                                                                                                                                                                                                                                                                                                                                                                                   
      // First set the data rate to the link ( 0 ) rate
      ModelPhysical . Link link = vectorOfLinks.get(0) ;
      bitRateEthernetBus = link.bitRateSourceToDestination;
      bitErrorRate = link. bitErrorRate ;
      interfaceType =  ModelData.GlobalVariables.interfaceTypes.Ethernet ;

      // Remove all the Links
      vectorOfLinks.removeAllElements();
      // Remove all the interfaces
      vectorOfInterfaces.removeAllElements();
      // Add this Ethernet bus to the vector of Links
////      vectorOfLinks.add(this);
      // Update the roouting tables and the vector of links
      // and the vector of interfaces for each node.
      for (int i = 0 ; i < vectorOfNodes.size() ; i++) {
         ModelPhysical.Node node = vectorOfNodes.get(i);
         // First remove all the links for this Node
         node.vectorOfLinksForThisNode.removeAllElements();
         // Add this Ethernet bus to the vector of Links for this Node
         node.vectorOfLinksForThisNode.add(this);
         // Now create the routing table so that all the packets are
         // routed throught the one interface attached to the node
         node.routingTable = new int[vectorOfNodes.size()];
         for (int j = 0 ; j < vectorOfNodes.size() ; j++) {
            // All the packets are routed to the single interface
            node.routingTable[j] = i;
         }  // end of for (int j = 0 ; j < vectorOfNodes.size() ; j++ )
         // First remove all the interfaces for this Node

         node.vectorOfInterfacesForThisNode.removeAllElements();
         // Add the Ethernet interface card back
         ModelDataLink.Ethernet ethernetInterface = new ModelDataLink.Ethernet(i);
         vectorOfInterfaces.add(ethernetInterface);
         // First remove all the links for this Node
         node.vectorOfInterfacesForThisNode.add(ethernetInterface);
      }  // end of for (int i = 0 ; i < vectorOfNodes.size() ; i++ )
      collisionHasOccurred = false ;
      busInUse = false ;
////      this . interfaceType = ModelData.GlobalVariables.interfaceTypes.Ethernet ;
      // Add this Ethernet bus to the vector of Links
System.out.println ( "IN2 EthernetBus constructor: " + this.toString ( ) )  ; 
////System.out.println ( "IN EthernetBus constructor: vectorofLinks.size ( ) = " + vectorOfLinks.size ()) ;  
      vectorOfLinks.add( this );      
////System.out.println ( "IN EthernetBus constructor: vectorofLinks.size ( ) = " + vectorOfLinks.size ()) ;      
     
ModelPhysical.EthernetBus ethernetBus
         = ( ModelPhysical.EthernetBus ) vectorOfLinks.get ( 0 );
         java.lang.StringBuilder stringBuilder
         = new java.lang.StringBuilder ( 2000 ) ;
      stringBuilder . append (
      " Ethernet bus has rate = " 
         + ethernetBus . bitRateEthernetBus + " with type  = "
         + ethernetBus . interfaceType + " with bit error rate " 
         + ethernetBus . bitErrorRate + "\n and transmission delay in microseconds\n"  ) ;
////         + this . transmissionDelay + " microseconds with interfaces\n" );
////            stringBuilder . append ( "\nCOST MATRIX TO GO TO GO FROM NODE I TO NODE J USING THE ROUTING TABLE" + CARRIAGE_RETURN ) ;
for ( int i = 0 ; i  < ethernetBus . transmissionDelay.length ; i++ )
      {
         for ( int j = 0 ; j  < ethernetBus . transmissionDelay [0].length ; j++ )
         {
            stringBuilder . append ( TAB + ethernetBus . transmissionDelay  [ i ] [ j ] ) ;
         }  // end of for ( int j = 0 ; j  < numberOfNodes ; j++ )
         stringBuilder . append ( CARRIAGE_RETURN ) ;
      }  // end of  for ( int i = 0 ; i  < numberOfNodes ; i++ )
      String string =   new String ( stringBuilder ) ;
System.out.println ( string ) ;  



   }  // end of  EthernetBus constructor


   /**Computes the delay times between the nodes on an Ethernet bus.*/
   public void ComputeTimeDelaysBetweenNodesOnEthernetBus() {
      int arraySize = vectorOfNodes.size();
      transmissionDelay = new float[arraySize][arraySize];
      transmissionDelayinTicks = new long[arraySize][arraySize];
      // Zero the arrays
      for (int i = 0 ; i < arraySize ; i++) {
         for (int j = 0 ; j < arraySize ; j++) {
            transmissionDelay[i][j] = ZERO_FLOAT;
            transmissionDelayinTicks[i][j] = ZERO;
         }  // end of for ( int j = 0 ; j < arraySize ; j++ )
      }  // end of for ( int i = 0 ; i < arraySize ; i++ )

      for (int i = 0 ; i < arraySize ; i++) {
         ModelPhysical.Node node = vectorOfNodes.get(i);
         /**A Vector containing all the links to this node*/
         java.util.Vector<Link> vectorOfLinks = node.vectorOfLinksForThisNode;
         // Iterate down the vector of links
         for (int j = 0 ; j < vectorOfLinks.size() ; j++) {
            ModelPhysical.Link link = vectorOfLinks.get(j);
            int source = link.sourceNodeNumber;
            int destination = link.destinationNodeNumber;
            float transmissionDelayTime = link.transmissionDelay;
            transmissionDelay[source][destination] = transmissionDelayTime;
         }  // end of for ( int j = 0 ; j < vectorOfLinks . size ( )  ; j++ )
      }  // end of for ( int i = 0 ; i < arraySize ; i++ )

      // Print matrix if displayCostMatrixAndRoutingMatrices == true
      if (ModelData.PrintPermissions.displayCostMatrixAndRoutingMatrices == true) {
         Utility.out.println ("Ethernet Bus Routing Input Delay Matrix");
         for (int i = 0 ; i < arraySize ; i++) {
            for (int j = 0 ; j < arraySize ; j++) {
               Utility.out.print("\t " + transmissionDelay[i][j]);
            }  // end of for ( int j = 0 ; j < i ; j++ )
            System.out.println();
         }  // end of for ( int i = 0 ; i < arraySize ; i++ )
      } // end of if ( ModelDatas . PrintPermissions . displayCostMatrixAndRoutingMatrices

      // Compute upper half of matrix
      for (int i = 0 ; i < arraySize ; i++) {
         for (int j = i + 2 ; j < arraySize ; j++) {
            transmissionDelay[i][j] =
               transmissionDelay[i][j - 1] + transmissionDelay[j - 1][j];
         }  // end of for ( int i = 0 ; i < arraySize ; i++ )
      }  // end of for ( int i = 0 ; i < arraySize ; i++ )

      // Compute lower half of matrix
      for (int i = 0 ; i < arraySize ; i++) {
         for (int j = 0 ; j < i ; j++) {
            transmissionDelay[i][j] =
               transmissionDelay[j][i];
         }  // end of for ( int j = 0 ; j < i ; j++ )
      }  // end of for ( int i = 0 ; i < arraySize ; i++ )

      // Compute matrix in ticks
      for (int i = 0 ; i < arraySize ; i++) {
         for (int j = 0 ; j < arraySize ; j++) {
            transmissionDelayinTicks[i][j] = (long) (transmissionDelay[j][i] // Input is in microseconds so conversion is
               // ModelDatas . GlobalVariables . timeTicksPerSecond ) / 1000000
               * ModelData.GlobalVariables.timeTicksPerSecond) / 1000000;
         }  // end of for ( int j = 0 ; j < i ; j++ )
      }  // end of for ( int i = 0 ; i < arraySize ; i++ )

      // Print matrix
      if (ModelData.PrintPermissions.displayCostMatrixAndRoutingMatrices == true) {
         // Print upper half of matrix
         Utility.out.println ("Ethernet Bus Routing Cost Matrix");
         for (int i = 0 ; i < arraySize ; i++) {
            for (int j = 0 ; j < arraySize ; j++) {
               Utility.out.print("\t " + transmissionDelayinTicks[i][j]);
            }  // end of for ( int j = 0 ; j < i ; j++ )
            Utility.out.println ( "" ) ;
         }  // end of for ( int i = 0 ; i < arraySize ; i++
      }  // end of if ( ModelDatas . PrintPermissions . displayCostMatrixAndRoutingMatrices

   }  // end of ComputeTimeDelaysBetweenNodesOnEthernetBus


   /** This toString method returns the Ethernet bus as a String.*/
   public String toString() {
      java.lang.StringBuilder stringBuilder
         = new java.lang.StringBuilder ( 2000 ) ;      
      stringBuilder . append (  
      " Ethernet bus has rate = " 
         + this . bitRateEthernetBus + " with type  = "
         + this . interfaceType + " with bit error rate " 
         + this . bitErrorRate + "\n and transmission delay in microseconds\n" ) ;
      for ( int i = 0 ; i  < transmissionDelay.length ; i++ )
      {
         for ( int j = 0 ; j  < transmissionDelay [0].length ; j++ )
         {
            stringBuilder . append ( TAB + transmissionDelay  [ i ] [ j ] ) ;
         }  // end of for ( int j = 0 ; j  < numberOfNodes ; j++ )
         stringBuilder . append ( CARRIAGE_RETURN ) ;
      }  // end of  for ( int i = 0 ; i  < numberOfNodes ; i++ )
      return  new String ( stringBuilder ) ;
   }  // end of method toString

}  // end of class


   /**Sets the arrival times for the frame at all the interface cards.  When
    * the frame of length frameLength is transmitted at time eventTime
    * from interface sourceInterface the frame starts to arrive at
    * startReceiveEthernetFrame and stop arriving at stopReceiveEthernetFrame.
    * StartReceiveEthernetFrame and stopReceiveEthernetFrame are computed
    * for every Ethernet cord on the bus.
   <p>THIS APPEARS TO NOT BE USED ANYWHERE AND IS SUSCIPTIBLE TO REMOVAL.*/
   /*public void SetArrivalTimesAtAllInterfacesOnEthernetBus ( long eventTime
   , int frameLength
   , int sourceInterface )
   {
   long tranmissionLength =  frameLength / bitRateSourceToDestination;
   for ( int i = 0 ; i < vectorOfNodes.size ( ) ; i++ )
   {
   ModelDataLink.Ethernet ethernetCard
   = ( ModelDataLink.Ethernet ) singletonVectorOfInterfaces . get ( i ) ;
   //Get the time delay time
   long delay = transmissionDelayinTicks [ sourceInterface ] [ i ] ;
   ethernetCard . startReceiveEthernetFrame = eventTime + delay ;
   ethernetCard . stopReceiveEthernetFrame = eventTime + delay + tranmissionLength ;
   }  // end of for ( int i = 0 ; i < vectorOfNodes.size ( ) ; i++ )
   }  // end of SetArrivalTimesAtAllInterfaces ( )
    */

/*
   /**Determines if a collission has occurred on the Ethernet bus.  
    * Used for Ethernet CSMA/CD.*/
/*   public boolean HasEthernetCollissionOccurred() {
      if (numberOfSimultaneousTransmissions > 1) {
         return true ;
      }
      else {
         return false ;
      }
   }
*/
   /**FOR TESTING ONLY*/
////   public static void main(String[] args) {
////      for (int i = 0 ; i < 100 ; i++) {
 ////     }
 ////  }  // end of main