/*
 * Link.java
 *
 * Created on December 20, 2006, 2:26 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelPhysical;


/**A link.  This class is mostly a container for information about the link.  
 * The Link number is set by staticLinkNumber becasue the Links are numbered
 * as they are created: This is NOT ture for Nodes.
 * *<p><b>Versions:</b>
 * <li><b>Version 1.001</b> of 20080616 add a random number generator
 * for each link.</li>
 * @version 1.101
 * @author dixonw*/
public class Link implements GUIConstants . General
{
   
   /**The singleton vector containing all the links*/
   ModelData.JNWSingletonVectorOfLinks<ModelPhysical.Link> singletonVectorOfLinks
      = ModelData.JNWSingletonVectorOfLinks.getInstance ( ) ;
   
   /**The singleton vector containing all the interfaces*/
   ModelData.JNWSingletonVectorOfInterfaces<ModelDataLink.InterfaceBaseClass> 
      singletonVectorOfInterfaces
      = ModelData.JNWSingletonVectorOfInterfaces.getInstance ( ) ;
   
   /**This link number is used to assign a unique llink number to each Link.  This number 
    * increments as new Links are added to the topology.  This number is reset to 0 in
    * GUIActions.ReadTopologyFileAction, so that the links numbers always begin
    * with 0.  This will only be necessary if the user opens a second topology file.*/
   public static int staticLinkNumber ;
   
   /**Link number.  This number is assigned by the constructor, and it starts
    *at 0.*/
   public int sequentialLinkNumber ;
   
   /**Source node.  Assigned by the user.  The Link does not care which Node is
    *the source and which Node is the destination.  The first node mentioned
    *in the input topology table is the source.  This number is the Node number
    *entered by the user.*/
   public int sourceNodeNumber ;
   
   /**Destination node.  Assigned by the user.  The Link does not care which Node is
    *the source and which Node is the destination.  The first node mentioned
    *in the input topology table is the source.  This number is the Node number
    * entered by the user.*/
   public int destinationNodeNumber ;
   
   /**Source port number*/
   public int sourcePort ;
   
   /**Destination port*/
   public int destinationPort ;
   
   /**Data rate in kbps from the destination to the source*/
   public int bitRateSourceToDestination ;
   
   /**Data rate in kbps from the source to the destination*/
   public int bitRateDestinationToSource ;
   
   /**Data rate in kbps on the Ethernet bus.  Equals the
    *bit rate for the first link.  This is only unsed in the
    subclass EthernetBus.  All data for the Ethernet bus
    is stored in the first Link of the topology.*/
   public int bitRateEthernetBus ;
   
   /** Bit error rate for transmissions on this link*/
   public float bitErrorRate ;
   
   /** Transmission delay on this link in microseconds*/
   public float transmissionDelay ;
   
   /** Transmission delay on this link in microseconds expressed in
    ticks.*/
   public long transmissionDelayinTicks ;
   
   /** Link type: HDLC, Ethernet or token ring*/
   public ModelData . GlobalVariables . interfaceTypes interfaceType ;
   
   /**The interface for the source.*/
   public ModelDataLink.InterfaceBaseClass sourceInterface = null ;
   
   /**The interface for the destination.*/
   public ModelDataLink.InterfaceBaseClass destinationInterface = null ;
   
   /**The random number generator for this link.  This is different from
    * the singleton random number generator, JNWRandom.*/
   public Utility.JNWRandom_NonSingleton linkRandomNumberGenerator = null ;
    
  
////  /**The time on this Link*/
////   public long linkTime ;
   
   /** Creates a new instance of Link and save the link number, the source and
    * destination node numbers, and the data rates in both directions.  The port
    * numbers are not initialized here.  The port numbers are computed in
    * MainGUI_Actions.ReadTopologyFileAction.CreateArray ( ).  The link also
    * add the interface cards to the singletonVectorOfInterfaces. 
    <p>It sets ths seed for the random number generator for this Link.*/
   public Link (int sourceNodeNumber , int destinationNodeNumber
     /* , int sourceNodeSequentialNumber, int destinationNodeSequentialNumber*/
      , int dataRate , float bitErrorRate , float transmissionDelay
      ,  ModelData . GlobalVariables . interfaceTypes  interfaceType )
   {  // Assign a link number
      sequentialLinkNumber = staticLinkNumber ;
      staticLinkNumber = staticLinkNumber + 1 ;
 ////     linkTime = 0 ;
      this . sourceNodeNumber = sourceNodeNumber ;
      this . destinationNodeNumber = destinationNodeNumber ;
      this . bitRateSourceToDestination = dataRate ;
      this . bitRateDestinationToSource = dataRate ;
      this . bitRateEthernetBus = dataRate ; 
      this . interfaceType = interfaceType ;
      this . bitErrorRate = bitErrorRate  ;
      this . transmissionDelay = transmissionDelay ;
      //  Add the interfaces
      if (  interfaceType . equals ( ModelData.GlobalVariables.interfaceTypes.Ethernet )   )
      {  // Create the Ethernet interfaces
         sourceInterface = new ModelDataLink.Ethernet ( sourceNodeNumber ) ;
         destinationInterface = new ModelDataLink.Ethernet ( destinationNodeNumber ) ;
      }
      else if (  interfaceType . equals ( ModelData.GlobalVariables.interfaceTypes.HDLC )   )
      {  // Create the HDLC interfaces
         sourceInterface = new ModelDataLink.HDLC ( sourceNodeNumber ) ;
         destinationInterface = new ModelDataLink.HDLC ( destinationNodeNumber ) ;
      }
      else if (  interfaceType . equals ( ModelData.GlobalVariables.interfaceTypes.TokenRing )   )
      {  // Create the Token Ring interfaces
         sourceInterface = new ModelDataLink.TokenRing ( sourceNodeNumber ) ;
         destinationInterface = new ModelDataLink.TokenRing ( destinationNodeNumber ) ;
      }  
      else
      {  // The interface is unrecognized
         Utility.err.println ( "ERROR in ModelPhysical.Link: link type, "
            + interfaceType . toString ( ) + " unrecognized " ) ;
      }
      // Add them to the vector of interfaces.  The interfaces 
      // must be added to the vector in the order they were created.
      singletonVectorOfInterfaces .add ( sourceInterface ) ;
      singletonVectorOfInterfaces .add ( destinationInterface ) ;
      // add this Link to the vector of all Links
      singletonVectorOfLinks . add ( this ) ;
      //Compute the transmission delay in ticks
      transmissionDelayinTicks
         =  ( long ) ( 0.0000001 * transmissionDelay 
         * ModelData.GlobalVariables.timeTicksPerSecond );
      // Create a random number generator for this link.
      linkRandomNumberGenerator 
         = new Utility . JNWRandom_NonSingleton ( ) ;
      // Sets the seed for this random number generator
      // JNWRandom is accessed in GUIActions.ReadTopologyFileAction
      // to make sure seedForJNWRandomNumberGenerator is defined
      // The random number sequence for each Link will be different
      linkRandomNumberGenerator .setSeed ( 
         Utility . JNWRandom . seedForJNWRandomNumberGenerator
         + 1000 * ( sequentialLinkNumber + 1 ) );
/*////System.out.println ("IN constructor for Link " + sequentialLinkNumber
////   + " first 20 random numbers = " ) ;      
////      for ( int i = 0 ; i < 20 ; i++ ) {
////System.out.println ("Random number " + i
////   + " = " + linkRandomNumberGenerator.nextInt ()) ;          
////      }
      */
   }  // end of Link constructor
   
   
   /**A constructor that will only be called by the Ethernet bus subclass*/
   public Link ( )
   {   }
   
   
   /**Returns the interface number for the node.*/
   public int getOppositeInterfaceNumber ( int nodeNumber )
   {
      if ( sourceInterface . nodeToWhichAttached == nodeNumber )
      {
         return sourceInterface . interfaceNumber ;
      }  // end of if ( sourceInterface . nodeToWhichAttached == nodeNumber )
      else if ( destinationInterface . nodeToWhichAttached == nodeNumber )
      {
         return destinationInterface . interfaceNumber ;
      }  // end of if ( destinationInterface . nodeToWhichAttached == nodeNumber )
      else
      { Utility.err.println ("ERROR IN ModelPhysical.Link.getInterfaceNumber:"
           + " the input node number " + nodeNumber + " is not attached to link "
           + sequentialLinkNumber  )  ;
        return Integer.MIN_VALUE ;
      }
   }  // end of getOppositeInterfaceNumber
   
   
   /**Return the index of the node at the other end of the link.  If 
    *input node number does not match either of the nodes for 
    *this link, it returns Integer.MIN_VALUE.*/
   public int nodeAtOtherEndOfLink (  int nodeNumber )
   {
      if ( nodeNumber == sourceNodeNumber )
      { return this . destinationNodeNumber ; }
      else if ( nodeNumber == destinationNodeNumber )
      { return this . sourceNodeNumber ; }
      else
      { Utility.err.println("ERROR IN ModelPhysical.Link.NodeAtOtherEndOfLink: inputNodeNumber does"
           + " not match either the source Node or the destination Node!\n"  )  ;
        return Integer.MIN_VALUE ; }
   }  // end of NodeIndexAtOtherEndOfLink
   
   
   /**Return the interface number of the interface at the other end of the link.  This
    method works only for a point-to-point link, not a bus.*/
   public static int InterfaceNumberAtOtherEndOfLink ( int interfaceNumber )
   {  //The interface numbers are always even odd pairs like 0 and 1, 2 and 3
      if ( interfaceNumber % 2 == 1 )
      {
         return interfaceNumber - 1 ;
      }
      else return interfaceNumber + 1 ;
   }

   
   /**This toString method returns the Link as a String with each object in
    * the vector returned as separate lines.*/
   public String toString ( )
   {
      return " Link number = " + this . sequentialLinkNumber + " from Node "
         + this . sourceNodeNumber + " to Node " + this . destinationNodeNumber 
         + " at rate "
         + this . bitRateSourceToDestination + " with return rate "
         + this . bitRateDestinationToSource + " is link type "
         + this . interfaceType . toString ( ) + " with bit error rate "
         + this . bitErrorRate + " and transmission delay "
         + this . transmissionDelay  +  " microseconds" + CARRIAGE_RETURN  ;
   }  // end of method toString
   
   
   /**This resets the to 0.  This should be used when creating a new set of 
    * Links.*/
   public static void RESET ( ) {
      staticLinkNumber = 0 ;
   }
   
}
