/*
 * Node.java
 *
 * Created on December 20, 2006, 2:26 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelPhysical;

import Utility.* ;
import ModelTransport.*;

/**A node.  The node number cannot be set using a static incrementer since
 * the node number is read from the topology file.
 *
 * @version 1.202
 * @author dixonw
 * @modified BZ May2010 and JMP 12Nov2011 for JTRN3
 */
public class Node implements
   ModelConstants.Node , GUIConstants.General
{
   /**A singleton vector containing all the Nodes*/
   ModelData.JNWSingletonVectorOfNodes<Node> singletonVectorOfNodes
      = ModelData.JNWSingletonVectorOfNodes.getInstance ( ) ;
   
   /**Node number passed into constructor.  This is the node number that
    *the user assigns to the node.*/
   public  int thisNodeNumber ;
   
   /**The time on this node.*/
   public  long nodeTime ;
   
   /**The X and Y positions of the Node on the graphical 
    * display.  XPos, YPos >= 0 */
   public int XPos , YPos ;
   
   /**A Vector containing all the links connecting to this node*/
   public  java.util.Vector<Link> vectorOfLinksForThisNode
      = new java.util.Vector<Link> ( ) ;
   
   /**A Vector containing all the interfaces belonging to this node*/
   public  java.util.Vector<ModelDataLink.InterfaceBaseClass> 
      vectorOfInterfacesForThisNode
      = new java.util.Vector<ModelDataLink.InterfaceBaseClass> ( ) ;

   /**A singleton vector containing all the Nodes*/
   static ModelData.JNWSingletonVectorOfNodes<ModelPhysical.Node> vectorOfNodes
      = ModelData.JNWSingletonVectorOfNodes.getInstance (  ) ;
   
   /**A Buffer for storing messages for transmission-level processing.  
    * sent message are put into this buffer for processing by the transmission level.*/
   public  Utility.JNWBuffer nodeBuffer
      = new Utility.JNWBuffer ( nodeBufferSizeInKilobytes ) ;
   
   /**A matrix containing the interface number for the minimum cost route to the
    *specified node.  The size of this table is set in ModelRouting.Routing.*/
   public int [ ] routingTable ;

   /** the current email TCP connection from this node
    *  (only one is allowed at any given time) */
   public ModelTransport.TCPConnection tcpConnection = null;
   
   /** Stores the outgoing email */
   public JNWBitSet[] MessageBuffer; // BZ

   /** The length of current email in bytes */
   public int emailLength = 0;

   /** The destination node of current email*/
   public Node emailDestinationNode;
   public int emailDestinationNodeNumber = -1;

   /** Semaphore that current email has been processed*/
   public boolean thisTCPEmailIsDone = false;

   /** Email converted from characters to a bitset*/
   Utility.JNWBitSet emailBitSet ;

   /** Linked list of all pending emails from this node */ // JMP                     // JMP
   public class EmailQueueElement {

       /** email in this queue position */
       JNWBitSet thisEmailBitSet;

       /** Length of the email in the buffer */
       int thisEmailLength;

       /** Source node of the email in the buffer */
       int thisEmailSourceNode;

       /** Destination of the email in the buffer */
       int thisEmailDestinationNodeNumber;

       /** Next email in the queue */
       EmailQueueElement nextQueueElement = null;

       public EmailQueueElement(
              JNWBitSet newEmailBitSet,
              int newEmailLength,
              int newEmailSourceNode,
              int newEmailDestinationNode) {
           
           // this EmailQueueElement will hold the new email
           thisEmailBitSet = newEmailBitSet;
           thisEmailLength = newEmailLength;
           thisEmailSourceNode = newEmailSourceNode;
           thisEmailDestinationNodeNumber = newEmailDestinationNode;
       }

       public void extractHeadElement() {
           if(emailQueueHead == null) {
               emailBitSet = null;
               emailLength = 0;
               emailDestinationNodeNumber = -1;
           } else {
               emailBitSet =
                   emailQueueHead.thisEmailBitSet;
               emailLength = 
                   emailQueueHead.thisEmailLength;
               emailDestinationNodeNumber =
                   emailQueueHead.thisEmailDestinationNodeNumber;
               emailQueueHead =
                   emailQueueHead.nextQueueElement;
           }
       }// end extractHeadElement

   }// end class BufferQueueElement

   /** Head of the BufferQueue */ // JMP
   public EmailQueueElement emailQueueHead = null;

   /** Loads a new email from the node's queue*/ // JMP
   public void loadNextEmail() {

       // pull data from queue
       if(emailQueueHead == null)return;
       emailQueueHead.extractHeadElement();
       Utility.out.println(
           "\n*** new email entered at node:"+thisNodeNumber+
           " dest:"+emailDestinationNodeNumber+
           " length:"+emailLength+"\n");
       // create both ends of the TCP connection
       int emailArraySize = emailLength/20;
       if(emailLength%20 > 0)emailArraySize++;
       tcpConnection =
           new ModelTransport.TCPConnection (
               thisNodeNumber ,
               emailDestinationNodeNumber ,
               emailLength);
       emailDestinationNode =
           vectorOfNodes.get(emailDestinationNodeNumber);
       emailDestinationNode.tcpConnection =
           new ModelTransport.TCPConnection (
               emailDestinationNodeNumber,
               thisNodeNumber,
               emailLength);

       // active open of connection
       tcpConnection.rtrans_send();

       // make TCP segments from the bitset
       MessageBuffer  = tcpConnection.PrependTCPHeadersToData (
           emailBitSet,
           thisNodeNumber,
           emailDestinationNodeNumber ) ;

       // Add headers to make these IP packets
       for ( int i = 0 ; i < MessageBuffer.length ; i++  ){
           // Prepend IP header
           MessageBuffer [ i ] =
               ModelTransport.IPUtilities.PrependIPHeaderToMessage (
                   MessageBuffer [ i ] ,
                   thisNodeNumber ,
                   emailDestinationNodeNumber ,
                   ModelData.GlobalVariables.transmissionProtocols.TCP );
       }

       // set the new MessageBuffer in TCPConnection
       tcpConnection.setMessageBuffer(MessageBuffer);
       
   }// end loadNextEmail()

   // stores received packets in ring buffer
   int IPBufferSize = 10;
   public JNWBitSet[] IPPacketReceivedFromInterface = new JNWBitSet[IPBufferSize];
   int IPBufferAddIndex = 0;
   int IPBufferRemoveIndex = 0;
   
   /** Constructor creates a new instance of Node and adds it to the vectorOfAllNodes */
   public Node ( int nodeNumber )
   {
      this.thisNodeNumber = nodeNumber ;
      singletonVectorOfNodes . add ( this ) ;
      nodeTime = 0 ;
      XPos = YPos = MINUS_ONE ;
   }

   public void addToIPLayerBuffer(JNWBitSet FrameIn) {

       // buffer drops packet when full
       IPPacketReceivedFromInterface[IPBufferAddIndex] = FrameIn;
       IPBufferAddIndex = (IPBufferAddIndex+1)%IPBufferSize;
       //System.out.println("added to IP layer buffer\n");
   }

   public JNWBitSet getPacketFromIPBuffer() {
       
       JNWBitSet returnValue = IPPacketReceivedFromInterface[IPBufferRemoveIndex];
       IPBufferRemoveIndex = (IPBufferRemoveIndex+1)%IPBufferSize;
       return returnValue;
   }

   public void addMessageBuffer(
               JNWBitSet emailBitSetIn,
               int emailLengthInBytes,
               int emailDestinationNodeNumberIn) { // BZ
       emailBitSet = emailBitSetIn;
       emailLength = emailLengthInBytes;
       emailDestinationNodeNumber = emailDestinationNodeNumberIn;
       EmailQueueElement eqe =
           new EmailQueueElement(
               emailBitSet,
               emailLength,
               thisNodeNumber,
               emailDestinationNodeNumber);

       // append this instance to end of queue
       if(emailQueueHead == null)
           emailQueueHead = eqe;
       else {
           EmailQueueElement emailQueueTail = emailQueueHead;
           while(emailQueueTail.nextQueueElement != null)
               emailQueueTail = emailQueueTail.nextQueueElement;
           emailQueueTail.nextQueueElement = eqe;
       }
   }

   public long getMessageSize() {

       return (long)emailLength;
   }

   public Utility.JNWBitSet getPacketFromMessageBufferBySequenceNumber(long sequence_number) { // BZ

       Utility.JNWBitSet IPPacket = null;
       for(int i=0; i<MessageBuffer.length; i++) {
           Utility.JNWBitSet TCPSegment = ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(MessageBuffer[i]);
           if(ModelTransport.TCPUtilities.GetSequenceNumberFromTCPSegment(TCPSegment) == sequence_number)
           {
               IPPacket = MessageBuffer[i];
               break;
           }
       }
       return IPPacket;
   }

   /**Adds a link to the vector of links associated with this node.  Add the interface
    * to the Vector of interfaces for this node.*/
   public void addLink ( Link linkToBeAddedToNode )
   {  // Add the link to the vector of llinks for this node
      vectorOfLinksForThisNode . add ( linkToBeAddedToNode ) ;
      // Add the interface to the vector of interfaces for this node
      if ( linkToBeAddedToNode.sourceNodeNumber == thisNodeNumber )
      {
         vectorOfInterfacesForThisNode
            . add ( linkToBeAddedToNode .sourceInterface ) ;
      }
      else
      {
         vectorOfInterfacesForThisNode
            . add ( linkToBeAddedToNode .destinationInterface ) ;
      }
   }  // end of addLink
   
   
   /**Updates the time on the node.*/
   public long UpdateNodeTime ( long timeIn )
   {
      return nodeTime = java.lang.Math.max ( nodeTime , timeIn ) ;
   }  // end of UpdateNodeTime
   
   
   /**Initializes the routingMatrix.*/
   public void InitializeRoutingMatrix ( )
   {  // One entry for every node in the LAN
      routingTable = new int [ singletonVectorOfNodes .size () ] ;
      // Initialized to route the message back to itself
      for ( int i = 0 ; i < routingTable.length ; i++ )
      {
         routingTable [ i ] = MINUS_ONE ;
      }
   }  // end of InitializeRoutingMatrix ( )
   
   
   /**Sets the X and Y position of the Node in the graphical display.*/
   public void setPosition ( int xPosIn , int yPosIn ) {
      XPos = xPosIn ;
      YPos = yPosIn ;
   }   
   
   /**Gets the X position of the Node in the graphical display.*/
   public int getXPos (  ) { return XPos  ; }   
   
      /**Gets the Y position of the Node in the graphical display.*/
   public int getYPos (  ) {  return YPos  ; }   
   
   /** This toString method returns the Node as a String with each object in
    the vector returned as separate lines.*/
   public String toString ( )
   {  // a generic object
      Object object = null ;
      // Create an iterator
      java.util.Iterator iterator = vectorOfLinksForThisNode . iterator ( ) ;
      // Create a stringBuffer
      java.lang.StringBuffer stringBuffer = new java.lang.StringBuffer ( 200 ) ;
      stringBuffer . append ( " Node number " + thisNodeNumber + " has "
         + vectorOfLinksForThisNode . size ()
         + " link(s) attached\n  The link(s) are numbered ") ;
      // If the Vector has more elements
      while ( iterator . hasNext ( ) )
      {  // Add a separator if there are already link numbers to be printed
         object = iterator . next ( ) ;
         int linkNumber = ( (Link) object ) . sequentialLinkNumber ;
         stringBuffer . append ( linkNumber  ) ;
         // Is there another number?
         if ( iterator . hasNext ( ) )
         {
            stringBuffer . append ( ", " ) ;
         }
      } // end of  while ( iterator.hasNext ( ))
      stringBuffer . append ( CARRIAGE_RETURN  ) ;
      return new String ( stringBuffer )  ;
   }  // end of method toString
}
