/*
 * Link.java
 *
 * Created on December 20, 2006, 2:26 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelPhysical;


/**A link contining a Token Ring
 *
 * @version 1.100
 * @author dixonw*/
public class TokenRing implements GUIConstants . General {
    
    /**The singleton vector containing all the links*/
    ModelData.JNWSingletonVectorOfLinks<ModelPhysical.Link> singletonVectorOfLinks
            = ModelData.JNWSingletonVectorOfLinks.getInstance ( ) ;
    
       /**The singleton vector containing all the links*/
    ModelData.JNWSingletonVectorOfInterfaces<ModelDataLink.InterfaceBaseClass> singletonVectorOfInterfaces
            = ModelData.JNWSingletonVectorOfInterfaces.getInstance( ) ;
    
    /**The first numjber to be used by the Links.  This is reset to 0 in 
    *GUIActions.ReadTopologyFileAction, so that the links numbers always begin
    *with 0.  This will only be necessary if the user opens a second topology file.*/
    public static int startingLinkNumber ; 
    
    /**Link number.  This number is assigned by the constructor, and it starts
    *at 0.*/
    public int sequentialLinkNumber ;
    
   /**Source node.  Assigned by the user.  The Link does not care which Node is 
   *the source and which Node is the destination.  The first node mentioned 
   *in the input topology table is the source.  This number is the Node number
   *entered by the user.*/
    public int sourceNode ; 
    
    /**Destination node.  Assigned by the user.  The Link does not care which Node is 
    *the source and which Node is the destination.  The first node mentioned 
    *in the input topology table is the source.  This number is the Node number
    *entered by the user.*/
    public int destinationNode ;
    
    /**The number of messages on this link.  Used for Ethernet CDMA.*/
    int numberOfMessagesOnThisLink = 0 ;
    
    /**Source port number*/
    public int sourcePort ;
    
    /**Destination port*/
    public int destinationPort ;
    
    /**Data rate in kbps from the destination to the source*/
    public int bitRateSourceToDestination ;
    
    /**Data rate in kbps from the source to the destination*/
    public int bitRateDestinationToSource ;
    
    /** Number of bits passed through this link*/
    public long numberOfBitspassed = 0 ;

    /** Number of it errors on this link.*/
    public long numberOfBitErrors = 0 ;
    
    /** Bit error rate for transmissions on this link*/
    public float bitErrorRate ;

    /** Transmission delay on this link in microseconds*/
    public float transmissionDelay ;

    /** Transmission delay on this link in microseconds*/
    public long transmissionDelayinTicks ;
    
   /** Link type: HDLC, Ethernet or token ring*/
  public ModelData . GlobalVariables . interfaceTypes interfaceType ;
    
    /**The interface for the source.*/
    public ModelDataLink.InterfaceBaseClass sourceInterface = null ;   
    
    /**The interface for the destination.*/
    public ModelDataLink.InterfaceBaseClass destinationInterface = null ;
    
    /**Cost of going from the source to the destination on this link.  Used
     *by Dijkstra's algorithm.*/
    public float sourceToDestinationCost = ( float ) 0.0 ;
    
    /**Cost of going from the destination to the source on this link.  Used
     *by Dijkstra's algorithm.*/
    public float destinationToSourceCost = ( float ) 0.0 ; 
    
    
    /** Creates a new instance of Link and save the link number, the source and
     destination node numbers, and the data rates in both directions.  The port
     numbers are not initialized here.  The port numbers are computed in 
     MainGUI_Actions.ReadTopologyFileAction.CreateArray ( ).  The link also 
     add the interface cards to the singletonVectorOfInterfaces. */
    public TokenRing(int sourceNode, int destinationNode
       , int sourceNodeSequentialNumber, int destinationNodeSequentialNumber
       , int dataRate 
       , float bitErrorRate , float transmissionDelay 
       , ModelData.GlobalVariables.interfaceTypes interfaceType ) {
        // Assign a link number
        sequentialLinkNumber = startingLinkNumber ;
        startingLinkNumber = startingLinkNumber + 1 ;
        this . sourceNode = sourceNode ;
        this . destinationNode = destinationNode ;   
        this . bitRateSourceToDestination = dataRate ;
        this . bitRateDestinationToSource = dataRate ;
        this . interfaceType = interfaceType ; 
        this . bitErrorRate = bitErrorRate  ;
        this . transmissionDelay = transmissionDelay ;
        //  Add the interfaces    
        if (  interfaceType . equals ( ModelData.GlobalVariables.interfaceTypes.Ethernet )   ) {
           // Create the Ethernet interfaces
           sourceInterface = new ModelDataLink.Ethernet ( sourceNode ) ;
           destinationInterface = new ModelDataLink.Ethernet ( destinationNode ) ;     
        }
        else if (  interfaceType . equals ( ModelData.GlobalVariables.interfaceTypes.HDLC )   ) {
           // Create the HDLC interfaces
           sourceInterface = new ModelDataLink.HDLC ( sourceNode ) ;
           destinationInterface = new ModelDataLink.HDLC ( destinationNode ) ;            
        } 
        else if (  interfaceType . equals ( ModelData.GlobalVariables.interfaceTypes.TokenRing )   ) {
            // Create the HDLC interfaces
           sourceInterface = new ModelDataLink.TokenRing ( sourceNode ) ;
           destinationInterface = new ModelDataLink.TokenRing ( destinationNode ) ;
        } 
        else {
            Utility.err.println ( "ERROR in ModelPhysical.Link "  ) ;
        }
        // Add them to the vector of interfaces  
        // The interfaces must be added to the vector in the order they 
        // were created.
        singletonVectorOfInterfaces .add( sourceInterface ) ;
        singletonVectorOfInterfaces .add( destinationInterface ) ;           
        // add this Link to the vector of all Links
        //Compute the transmission delay in ticks
        transmissionDelayinTicks 
           = ( long ) ( ( transmissionDelay 
           * ModelData.GlobalVariables.timeTicksPerSecond ) / 1000000.0 );
    }  // end of Link constructor
    
    
  /**Returns the interface number for the node.*/
    public int getInterfaceNumber ( int nodeNumber ) {
        if ( sourceInterface . nodeToWhichAttached == nodeNumber ) {
            return sourceInterface . interfaceNumber ;
        }  // end of if
        else if ( destinationInterface . nodeToWhichAttached == nodeNumber ) {
            return destinationInterface . interfaceNumber ; 
        }  // end of else if
        else { Utility.err.println ("ERROR IN ModelPhysical.Link.getInterfaceNumber:"
           + " the input node number " + nodeNumber + " is not attached to link "
           + sequentialLinkNumber  )  ;
           return Integer.MIN_VALUE ;
        }  // end of else     
    }
  
    
  /**Return the index of the node at the other end of the link.  If input node
  number does not match either of the nodes for this link, it returns 
  Integer.MIN_VALUE.*/
 public int nodeAtOtherEndOfLink (  int nodeNumber ) {
////     if ( inputNodeNumber == link . sourceNodeSequentialNumber ) 
////        { return link . destinationNodeSequentialNumber ; }
////     else if ( inputNodeNumber == link . destinationNodeSequentialNumber ) 
////        { return link . sourceNodeSequentialNumber ; }
     if ( nodeNumber == sourceNode ) 
        { return this . destinationNode ; }
     else if ( nodeNumber == destinationNode ) 
        { return this . sourceNode ; }
     else { Utility.err.println ("ERROR IN ModelPhysical.Link.NodeAtOtherEndOfLink: "
             + " inputNodeNumber does"
             + " not match either the source Node or the destination Node!\n" )  ;
        return Integer.MIN_VALUE ; } 
 }  // end of NodeIndexAtOtherEndOfLink   

 
 /**Return the interface number of the interface at the other end of the link*/
 public static int InterfaceAtOtherEndOfLink ( int interfaceNumber ) {
     if ( interfaceNumber % 2 == 1 ) {
          return interfaceNumber - 1 ;
     }
     else return interfaceNumber + 1 ;
 }
 
 
/** This toString method returns the Link as a String with each object in the vector returned as 
*separate lines.*/
    public String toString ( ) {
          return " Link number = " + this . sequentialLinkNumber + " from Node "
              + this . sourceNode + " to Node " + this . destinationNode + " at rate " 
              + this . bitRateSourceToDestination + " with return rate " 
              + this . bitRateDestinationToSource + " is link type "
              + this . interfaceType . toString ( ) + " with bit error rate "                  
              + this . bitErrorRate + " and transmission delay "
              + this . transmissionDelay  +  " microseconds with interfaces\n" ; 
    }  // end of method toString  

    
/**FOR TESTING ONLY*/    
    public static void main_Testing_Only ( String [ ] args ) {
    for ( int i = 0 ; i < 100 ; i++ ) {
 System.out.println ( "The other interface from " + i + " is " + InterfaceAtOtherEndOfLink (  i )) ; 
    } 
    }
          
}
