/*
 * IPUtilities.java
 *
 * Created on July 7, 2007, 1:29 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelTransport;


import Utility.JNWBitSet ;


/**
 * Utilities for the creation of IP headers for packets.
 *<p><b>Versions</b>
 * <li><b>Version 1.001</b> of 20071118 adds the method
 * SetIPPacketPrecedenceToNetworkControl ( JNWBitSet IPPacketIn ) which sets the
 * type of service precedence to "Network Control."</li>
 *</i><b>Version 1.002</b> os 20081121 adds the method
 * ComputeIPCheckSum( JNWBitSet IPPacketIn, JNWBitSet oldChecksum ) which
 * computes the hacked checksum of only the source and address fields.
 * THIS HACK DOES NOT FOLLOW RFC AND NEEDS FIXED.</li>
 * @author dixonw
 * @version 1.202
 * @modified JMP 29Oct11 for HDLC
 * @see <a href="http://www.freesoft.org/CIE/Course/Section3/7.htm">IP PacketStructure</a>
 * @see <a href="http://www.networksorcery.com/enp/protocol/ip.htm#Protocol"><u>Really GOOD</u> IP Reference plus all other protocols</a>
 * @see <a href="http://www.faqs.org/rfcs/rfc1700.html">RFC 1700 - Assigned Numbers</a>
 *
 */
public class IPUtilities implements GUIConstants.General
{

   /**The transmission protocol*/
   static ModelData.GlobalVariables.transmissionProtocols
      transmissionProtocol ;

   /** Creates a new instance of IPUtilities */
   public IPUtilities ()
   {
   }

   /**Constructs an IP header for a message.  It creates JNWBitSets in the
    *order that they appear in the IP header.  It then appends the fields together.
    *The checksum is first set to " 00000000 00000000 ".  After the IP header is
    *assembled the correct checksum is calculated and stored in the checksum
    *field of the IP header.
    *
    * <p>NOTE:  It should be possible to eliminiate most of the class line
    *parameters below.
    *Only JNWBitSet packetIn is really needed.  All the other parameters are set in
    *ModelTransport.IPUtilities.PrependIPHeaderToMessage whihc saves the parameters
    *and then calls ConstructIPHeader.  HOWEVER in
    *ModelTransport.TCPUtilities.ComposeACKPacket this method is called with all these
     parameters.  Until that call is revised the parameters must include all the
    *variables.
    *<p>This header has NO options available.  The IP header length is 40 bytes.
    *@see <a href="http://www.freesoft.org/CIE/Course/Section3/7.htm">IP Packet Structure</a>*/
   public static JNWBitSet ConstructIPHeader ( JNWBitSet TCPorUDPPacketIn
      , long sourceNodeAddressIn, long destinationNodeAddressIn
      , ModelData.GlobalVariables.transmissionProtocols
         transmissionProtocolIn  )
   {  // Construct the IP header
      // Version = 4; Internet Header Length = 5; Precedence = 000
      // Delay, Throughput, and Reliability = 000
      JNWBitSet versionIHLAndTypeOfService = new JNWBitSet ( "0100 0101 000 000 00" ) ;
      JNWBitSet totalLength = new JNWBitSet (
          ( TCPorUDPPacketIn . getJNWLength () / 8 ) + 20 /*IP header length*/ , 16  ) ;
      JNWBitSet identification = new JNWBitSet ( "00000000 00000000" ) ;
      // 0 , DF bit = Don't Fragment = 1 ; MF bit = Last Fragment = 0
      JNWBitSet IPflags = new JNWBitSet ( "010" ) ;
      // Fragment offset 0
      JNWBitSet fragmentOffset = new JNWBitSet ( " 00000 00000000 " ) ;
      // 17 for UDP and 6 for TCP.  See reference
      // 17 for UDP
      JNWBitSet protocol = new JNWBitSet ( 17 , 8  ) ;
      if ( transmissionProtocolIn
         == ModelData.GlobalVariables.transmissionProtocols.TCP)
      {  // 6 for TCP
         protocol = new JNWBitSet ( 6 , 8 ) ;
      }
      // Load a dummy checksum which will later be replaced
      JNWBitSet sourceBitSet  = new JNWBitSet ( sourceNodeAddressIn , 32 ) ;
      JNWBitSet destinationBitSet = new JNWBitSet ( destinationNodeAddressIn , 32 ) ;
      // Now construct the IPHeader
      JNWBitSet IPHeader = new JNWBitSet ( ) ;
      // Construct the IP Header
      IPHeader . append ( versionIHLAndTypeOfService ) ;
      IPHeader . append ( totalLength ) ;
      IPHeader . append ( identification ) ;
      IPHeader . append ( IPflags ) ;
      IPHeader . append ( fragmentOffset ) ;
      // Set the time to live to 255
      IPHeader . append ( new JNWBitSet ( 255 , 8 )  ) ;
      IPHeader . append ( protocol ) ;
      // Insert a dummy checksum of all zeros
      IPHeader . append ( new JNWBitSet ( " 00000000 00000000 " )) ;
      IPHeader . append ( sourceBitSet ) ;
      IPHeader . append ( destinationBitSet ) ;
      // Now add in the correct IP header checksum
      IPHeader = ComputeIPHeaderChecksumAndInsertIntoHeader (
         IPHeader ) ;
      return IPHeader ;
   }  // end of ConstructUDPHeader ( )
   
   /**Calculates the checksum for the IP header.
    *At present this is a hack that considers only the source
    *and destination fields. MUST BE FIXED.
    */
   public static JNWBitSet ComputeIPChecksum( JNWBitSet IPPacketIn,
                                              JNWBitSet oldChecksum )
   {
     // build the checksum in this
     JNWBitSet checksum = new JNWBitSet(oldChecksum);
      
     // locate start of address fields 
      int fromIndex = IPPacketIn . getJNWLength () - 4 * 32 ;
      
      // extract source address and include in checksum
      JNWBitSet sourceAddressBitSet
         = IPPacketIn  . getSubset ( fromIndex , fromIndex + 32) ;
      checksum = sourceAddressBitSet . xorAndShift ( checksum ) ;
      
      // extract destination address and include in checksum
      JNWBitSet destinationAddressBitSet
         = IPPacketIn . getSubset ( fromIndex -32 , fromIndex ) ;
      checksum = destinationAddressBitSet . xorAndShift ( checksum ) ;
      
      // return the result
      return checksum;
   }// end method ComputeIPChecksum


   /**Adds the UDP UDPHeader to the message.  Takes the source node number and
    *the destination node number as parameters.
    *<p>You may laters wish to combine this with ConstructIPHeader and eliminate
    *ConstructIPHeader as a separate method.*/
   public static JNWBitSet PrependIPHeaderToMessage ( JNWBitSet TCP_or_UDPPacketIn
      , long sourceNodeAddressIn, long destinationNodeAddressIn
      , ModelData.GlobalVariables.transmissionProtocols
         transmissionProtocolIn  )
   {
      JNWBitSet IPHeader = IPUtilities . ConstructIPHeader  ( TCP_or_UDPPacketIn
         , sourceNodeAddressIn , destinationNodeAddressIn
         , transmissionProtocolIn )  ;
      TCP_or_UDPPacketIn. prepend ( IPHeader ) ;
      return TCP_or_UDPPacketIn ;
   }  // end of PrependIPHeaderToMessage


   /**Computes the IP header checksum.  Computers the 1s compliment sum of the
    *header.  It then performs the 1s compliment of the checksum and then inserts
    *the checksum into the IP header.*/
   public static JNWBitSet ComputeIPHeaderChecksumAndInsertIntoHeader (
      JNWBitSet IPHeaderIn )
   {  // Make a 16 bit ehccksum initialized to zeros
      JNWBitSet IPChecksum = new JNWBitSet ( 16 ) ;
      // Form the checksum
      IPChecksum = IPHeaderIn.xorAndShift ( IPChecksum ) ;
      // Invert the checksum
      IPChecksum . invertBitSet ( ) ;
      IPHeaderIn . setSubset ( IPHeaderIn.getJNWLength ( ) - 3 * 32 , IPChecksum ) ;
      return IPHeaderIn ;
   }


   /**Gets the destination from the IP header.*/
   public static long GetSourceAddressFromIPHeader ( JNWBitSet IPPacketOrHeaderIn )
   {  // Location of source address relative ot IP header
      int fromIndex = IPPacketOrHeaderIn.getJNWLength () - 4 * 32  ;
      return IPPacketOrHeaderIn . getSubset ( fromIndex , fromIndex + 32)
         . convertJNWBitSetToLong ( ) ;
   }  // end of GetSourceAddressFromIPHeader


   /**Gets the destination from the IP header.*/
   public static long GetDestinationAddressFromIPHeader ( JNWBitSet IPPacketOrHeaderIn )
   {  // Location of dest address relative to IP header
      int fromIndex = IPPacketOrHeaderIn.getJNWLength () - 5 * 32  ;
      return IPPacketOrHeaderIn . getSubset ( fromIndex , fromIndex + 32)
         . convertJNWBitSetToLong ( ) ;
   }  // end of GetDestinationAddressFromIPHeader


   /**Gets the protocol from the IP header.  Currently UDP, 17, and TCP,
    * 6, are implemented.
    * @see <a href="http://www.faqs.org/rfcs/rfc1700.html">RFC 1700 - Assigned Numbers</a>*/
   public static ModelData . GlobalVariables . transmissionProtocols
      GetProtocolFromIPHeader ( JNWBitSet IPpacketIn )
   {  // Location of protocol in IP headers
      int fromIndex = IPpacketIn.getJNWLength ( ) - 2 * 32 - 16 ;
      int protocolNumber = ( int ) IPpacketIn . getSubset ( fromIndex
         , fromIndex + 8  ) . convertJNWBitSetToLong ( ) ;
      // TCP
      if ( protocolNumber == 6 ) {
         return ModelData.GlobalVariables.transmissionProtocols.TCP ;
      }  // UPD
      else if ( protocolNumber == 17 ) {
         return ModelData.GlobalVariables.transmissionProtocols.UDP ;
      }  // An unrecognized protocol
      Utility.err.println ("ERROR in ModelTransport.IPUtilities."
         + "GetProtocolFromIPHeader: unrecognized protocol = "
         + protocolNumber  + " returning UDP " ) ;
      return ModelData.GlobalVariables.transmissionProtocols.UDP ;
   }  // end of GetProtocolFromIPHeader



  /**Verifies the IP header checksum at the receiver.  The input can
  *be either the IP header or the whole IP packet.*/
   public static boolean VerifyIPHeaderChecksum ( JNWBitSet IPPacketOrHeaderIn )
   {
      // if this is an HDLC supervisory frame it has no IP header
      if(IPPacketOrHeaderIn.getJNWLength() < 200)return false;
      int fromIndex = IPPacketOrHeaderIn . getJNWLength ( ) - 20 * 8 /*IP header length*/ ;
      // Get the topmost 20 bytes, the IP header
      JNWBitSet IPPacketHeader
         =  IPPacketOrHeaderIn . getSubset ( fromIndex , fromIndex + 20 * 8  ) ;
      // perform a shifting XOR on the header  The result must be
      // a bit sets of all 1s
      JNWBitSet result = IPPacketHeader . xorAndShift ( new JNWBitSet ( 16 ) ) ;
      if ( IPPacketHeader . xorAndShift ( new JNWBitSet ( 16 ) )
         . cardinality () !=  16 )
      {
          Utility.out.println("IP Header checksum failed:\n"+
              IPPacketOrHeaderIn);
          return false ;  // If all the bits are not set then there is an error
      }
      else
      { 
          return true ;
      }
   }  // end of VerifyIPHeaderChecksum


   /**Converts a String of the form XXX.XXX.XXX.XXX to a Java long.  This method is
    no used anywhere as of 20081015.*/
   public static long ConvertIPV4AddressToLong ( String IPV4AddressStringIn )
   {
      long longAddress = 0 ;
      java.util.StringTokenizer stringTokenizer
         = new java.util.StringTokenizer ( IPV4AddressStringIn , "." ) ;
      long multiplier = 256 * 256 * 256 ;
      while ( stringTokenizer .hasMoreTokens () )
      {
         longAddress = longAddress
            + new Long ( stringTokenizer . nextToken ( ) . trim ( )  )
            * multiplier ;
         multiplier = multiplier / 256 ;
      }  // end of while ( stringTokenizer .hasMoreTokens() )
      return longAddress ;
   }  // end of ConvertIPV4AddressToLong


   /**Converts a Java long to a String of the form XXX.XXX.XXX.XXX.  This method is
    no used anywhere as of 20081015.*/
   public static String ConvertLongToIPV4Address ( long IPV4AddressIn )
   {
      String string = EMPTY_STRING ;
      long remainder = 0 ;
      long divisor = 256 ;
      for ( int i = 0 ; i < 4 ; i++ )
      {
         remainder =  ( IPV4AddressIn % divisor )  ;
         string = new Long ( remainder ) . toString ( ) + "." + string  ;
         IPV4AddressIn = IPV4AddressIn / divisor ;
      }  // end of for ( int i = 0 ; i < 4 ; i++ )
      // remove the last "."
      string = string .substring ( 0 , string . length () -1 ) ;
      return string ;
   }  // end of ConvertIPV4AddressToLong


   /**Sets the IP header priority to "Network Control."  The precedence bits
    *of the type of service are set to "111."  This method is
    no used anywhere as of 20081015.*/
   public static JNWBitSet SetIPPacketPrecedenceToNetworkControl
      ( JNWBitSet IPPacketOrHeaderIn )  {
      JNWBitSet networkControlPrecedence = new JNWBitSet ( "111" ) ;
      int fromIndex = IPPacketOrHeaderIn.getJNWLength ( ) - 8 - 3 ;
      IPPacketOrHeaderIn .setSubset ( fromIndex , networkControlPrecedence ) ;
      return IPPacketOrHeaderIn ;
   }  // end of SetIPPacketPrecedenceToNetworkControl


   /**A main method for testing the JNWBitSet.
    * @param args the command line arguments*/
   public static void main_Testing_Only ( String[] args )
   {
     /*
      Utility.JNWBitSet test = new Utility.JNWBitSet ( "00000000 00000000 00000000 ");
      test = SetIPPacketPrecedenceToNetworkControl ( test ) ;
      System.out.println ( " test = " + test.bitSetToString ( )) ;


         JNWBitSet data = new JNWBitSet ("00000000 00000001 00000000 00000011 "
+"00000000 00000111 00000000 00001111 00000000 00011111"
+"00000000 00111111 00000000 01111111 00000000 11111111 "
+"00000000 10101010 11111111 11111111 ") ;
         data = SetIPPacketPrecedenceToNetworkControl ( data ) ;
        System.out.println (" data = " + data.bitSetToString() ) ;
 System.exit ( 0 ) ;

        JNWBitSet dataz = new JNWBitSet ("00000000 00000001 00000000 00000011 "
+"00000000 00000111 00000000 00001111 00000000 00011111"
+"00000000 00111111 00000000 01111111 00000000 11111111 "
+"00000000 10101010 11111111 11111111 ") ;
        JNWBitSet XOR = new JNWBitSet ( 16 ) ;
        System.out.println (" XOR = " + XOR.bitSetToString() ) ;
        XOR = data . xorAndShift( XOR  )  ;
        System.out.println (" XOR = " + XOR.bitSetToString() ) ;
        System.out.println ("The data is valid " + VerifyIPChecksum( data )) ;

 System.exit ( 0 ) ;


      long t = 256 ;
      long IPV4Address = 144 * t * t * t
         + ( long ) 69 * t * t
         + ( long ) 115* t
         + ( long ) 43;

      String string =   ConvertLongToIPV4Address (IPV4Address ) ;
      System.out.println ( " IPV4Address = " + IPV4Address + " string representation = " +  string ) ;
      System.out.println ( " string representation = " + string  + " long = " + ConvertIPV4AddressToLong ( string ) ) ;
      System.out.println ( " long = " + IPV4Address  + " string = " + ConvertLongToIPV4Address ( ConvertIPV4AddressToLong ( string ) ) ) ;
      System.exit ( 0 ) ;


      String eMail = "abcde" ;
      JNWBitSet eMailBitSet = new JNWBitSet (  ) ;

      JNWBitSet zeros = new JNWBitSet ( "00000000 00000000 00000000 00000000" ) ;
      JNWBitSet IPPacket = PrependIPHeaderToMessage ( zeros
      , 1024 , 512
      , ModelData.GlobalVariables.transmissionProtocols.TCP ) ;
      // Source
      // Destination
      System.out.println ( " IPPacket = " + IPPacket.toString ( 16 ) ) ;
      System.out.println ( "Verify = " + VerifyIPHeaderChecksum( IPPacket )
      + "\n\n\n") ;
      for ( int i = 0   ; i < IPPacket.getJNWLength ( ) ; i++ ) {
         IPPacket . flip ( i ) ;
      System.out.println ( "Verify for i = " + i + " : " + VerifyIPHeaderChecksum( IPPacket ) ) ;
         IPPacket . flip ( i ) ;

      }



/*
      System.out.println ( " Message = " + message.bitSetToString () ) ;
      int protocol = GetProtocolFromIPHeader( message  ) ;
      System.out.println ( " protocol = " + protocol ) ;
      long address = GetDestinationAddressFromIPHeader ( message  ) ;
      System.out.println ( " destination address = " + address ) ;
      address = GetSourceAddressFromIPHeader ( message  ) ;
      System.out.println ( " source address = " + address ) ;
  */
        /*
        eMailBitSet . setJNWBitSetToString ( eMail  ) ;
        eMailBitSet .setJNWBitSetToZerosAndOnes( "11111111 00000000 ") ;
        ModelDatas . GlobalVariables . eMailMessage  =  eMailBitSet ;
        ModelDatas . GlobalVariables .sourceNodeNumber = 32 ;
        ModelDatas . GlobalVariables . destinationNodeNumber = 64 ;
        JNWBitSet result ;
        ConstructIPMessage ( eMailBitSet ) ;
         */
   } // end of main
} // end of class 


//*Computes the 1s complement UDP UDPHeader checksum*/
/*
    private static JNWBitSet ComputeUDPChecksum ( JNWBitSet bitSetToBeChecksummed ) {
    System.out.println ( "Message to be checksummed is : "
       + bitSetToBeChecksummed . bitSetToString() ) ;
        JNWBitSet UDPChecksum = new JNWBitSet ( " 00000000 00000000 " ) ;
        boolean booleanValue = false ;
        for ( int bitSetIndex = 0 ; bitSetIndex < bitSetToBeChecksummed .getJNWLength ( )
           ; bitSetIndex = bitSetIndex + 16  ) {
           for ( int i = 0 ; i < 16 ; i++ ) {
               // The value of the bit in the dataset
               booleanValue = bitSetToBeChecksummed . get ( i + bitSetIndex ) ;
               booleanValue = ( UDPChecksum . get ( i ) ) ^ booleanValue ;
               UDPChecksum . set ( i , booleanValue) ;
           }
        }  // end of for ( int bitSetIndex = 0 ;
        System.out.println ( "Checksum is : " + UDPChecksum . bitSetToString() ) ;
        return UDPChecksum ;
    }
 */



//*Computes the pseudoheader*/
/*
    private static JNWBitSet ComputePseudoHeaderAtSource ( JNWBitSet messageCloneIn ) {
        int source = ModelDatas . GlobalVariables .sourceNodeNumber ;
        JNWBitSet sourceBitSet  = new JNWBitSet (  ) ;
        sourceBitSet . setJNWBitSetToLongAndTruncate( source , 32 );
        int destination = ModelDatas . GlobalVariables . destinationNodeNumber ;
        JNWBitSet destinationBitSet  = new JNWBitSet (  ) ;
        destinationBitSet . setJNWBitSetToLongAndTruncate( destination , 32 );
        JNWBitSet eigthZeros = new JNWBitSet ( " 00000000 " ) ;
        JNWBitSet seventeen  = new JNWBitSet ( ) ;
        seventeen . setJNWBitSetToLongAndTruncate( 17 , 8 );
        JNWBitSet lengthBitSet  = new JNWBitSet (  ) ;
        int UDPLength = messageCloneIn . getJNWLength( ) / 8 ;
        lengthBitSet . setJNWBitSetToLongAndTruncate( UDPLength , 16 );
        JNWBitSet pseudoheader = new JNWBitSet (  ) ;
        pseudoheader . append ( sourceBitSet ) ;
        pseudoheader . append ( destinationBitSet ) ;
        pseudoheader . append ( eigthZeros ) ;
        pseudoheader . append ( seventeen ) ;
        pseudoheader . append ( lengthBitSet ) ;
System.out.println ( "source =  " + source  + " destination = " + destination + " length = " + UDPLength
   +"\n pseudoheader = " + pseudoheader . bitSetToString() )  ;
        return pseudoheader ;
    }
 */




 /////  /**Adds the UDP UDPHeader to the message*/
////   public static JNWBitSet PrependIPHeaderToMessage ( JNWBitSet messageIn  )
////   {
////      JNWBitSet IPHeader = IPUtilities . ConstructIPHeader  ( messageIn
////         , ModelDatas . GlobalVariables . sourceNodeNumber
 ////        , ModelDatas . GlobalVariables . destinationNodeNumber )  ;
////      messageIn . prepend ( IPHeader ) ;
//// System.out.println ( "\nThe message with IP header is : " + messageIn . bitSetToString() ) ;
 ////     return messageIn ;
////   }