/* VERSION 20111125
 * TCPConnection.java
 *
 * Created on July 6, 2007, 2:16 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package ModelTransport;

import Utility.*;
import ModelEvents.* ;

/**Creates a TCP connection.  This is instantiated in the Node array 
 * tcpConnections in class ModelPhysical.Node
 * NOTE: Because the email buffer is in Node, only one email TCP
 * connection from each node is possible at a time. NEEDS FIXED.
 * @see <a href="http://www.freesoft.org/CIE/Course/Section4/8.htm">TCP Header Format</a>
 * @see <a href="http://www.faqs.org/rfcs/rfc1700.html">RFC 1700 - Assigned Numbers</a>
 * @version 1.202
 * @author dixonw
 * @modified BZ May 2010 to match Understanding Internet Protocols
 */
public class TCPConnection extends TCPUtilities {

   /**The JNW random number generator.*/
   static Utility.JNWRandom randomNumberGenerator = Utility.JNWRandom.
      getInstance ();

   /**The times unACKed segments were last sent */
   long[] segment_send_times;

   /**Node of host for this TCPConnection*/
   ModelPhysical.Node hostNode;

   /**IP address of host for this TCPConnection*/
   int hostNodeNumber;

   /**HDLC address of interface we route to on that node*/
   int interfaceNumber;

   /**Buffer for received segments*/
   JNWBitSet[] emailBuffer = new JNWBitSet[100];

   /**Initialized state*/
   boolean nodeConnectionInitialized = false;

   /**Latches initialized state to prevent repeat*/
   public boolean initializeMessageBuffer(int bufferLength) {
       boolean returnValue = !nodeConnectionInitialized;
       nodeConnectionInitialized = true;
       return returnValue;
   }

   /**instance of ModelEventsBaseClass we can refer to*/
   private ModelEvents.ModelEventsBaseClass tcpEvent =
       new ModelEvents.ModelEventsBaseClass();

   /** SYN packet, ready to send */
   JNWBitSet synPacket = null;

   /** SYN-ACK packet, ready to send */
   JNWBitSet synAckPacket = null;

   /** FIN packet, ready to send */
   JNWBitSet finPacket = null;

   /** FIN-ACK packet, ready to send */
   JNWBitSet finAckPacket = null;

   /** time a SYN is sent */
   long synSentTime = 0;

   /** indicates SYN-ACK was received */
   boolean receivedSYNACK = false;

   /** time a FIN is sent */
   long finSentTime = 0;

   /** indicates FIN was received and FIN-ACK sent*/
   boolean receivedFIN = false;

   /** period for scheduling TCPTimeout wakeup */
   int timeoutWakeupPeriod = 5000;

   /** time when TCPTimeout will be called next */
   long nextScheduledTimeout = timeoutWakeupPeriod;

   /** Creates a new instance of TCPConnection.*/
   public TCPConnection ( 
       int hostNodeNumberIn ,
       int destinationNodeNumber ,
       int emailArrayLength) {
       super ( hostNodeNumberIn , destinationNodeNumber );
       hostNode = vectorOfNodes.get(hostNodeNumberIn);
       hostNodeNumber = hostNodeNumberIn;
       interfaceNumber = hostNode.routingTable[(int) destinationNodeNumber];

       // initialize incoming email buffer and saved_segments
       saved_segments = new JNWBitSet[emailArrayLength];
       for(int i=0; i<emailArrayLength; ++i)
           saved_segments[i] = null;
       for(int i=0; i < emailBuffer.length; ++i)
           emailBuffer[i] = null;

       // make SYN and SYN-ACK packets for use when needed
       synPacket = ComposeSYNPacket(
           hostNodeNumber,
           destinationNodeNumber,
           false // not SYN-ACK
       );
       synAckPacket = ComposeSYNPacket(
           hostNodeNumber,
           destinationNodeNumber,
           true // is SYN-ACK
       );

       // make FIN and FIN-ACK packets for use when needed
       finPacket = ComposeFINPacket(
           hostNodeNumber,
           destinationNodeNumber,
           false // not FIN-ACK
       );
       finAckPacket = ComposeFINPacket(
           hostNodeNumber,
           destinationNodeNumber,
           true // is FIN-ACK
       );
   }  // end of constructor

   /** accept reference to the buffer for TCP segments */
   public void setMessageBuffer(JNWBitSet[] MessageBuffer)
   {
       saved_segments  = new JNWBitSet[MessageBuffer.length];
       segment_send_times = new long[MessageBuffer.length];
       for(int i=0; i < saved_segments.length; ++i) {
           if(MessageBuffer[i] != null)
               saved_segments[i] =
                   CloneTCPSegmentFromIPPacket(MessageBuffer[i]);
           else saved_segments[i] = null;
           segment_send_times[i] = -1;
       }
   }// end initializeMessageBuffer

   /** get a packet out of the buffer */
   public JNWBitSet getPacketFromBuffer(int segmentIndex) {
      return hostNode.MessageBuffer[segmentIndex];
   }

   /** consolidates MessageBuffer back into an email and prints it*/
   public void printSegmentsAsEmail() {
       JNWBitSet emailBitSet = new JNWBitSet();
       for( int i = 0;
            i < emailBuffer.length;
            ++i)
           if(emailBuffer[i] != null)
             emailBitSet.append(emailBuffer[i]);
       String eMail = emailBitSet.convertJNWBitSetToString();
       out.print ( "\n\n E-Mail TCP Message RECEIVED at node "+
           hostNode.thisNodeNumber + " is:\n"  +
           eMail + DOUBLE_CARRIAGE_RETURN ) ;
   }

   /**Find a segment in saved_segments that matches a sequence number
      returns the segment number, or -1 if none matches */
   public int getPacketIndexFromMessageBufferBySequenceNumber(
                  long sequence_number) { // BZ
       Utility.JNWBitSet IPPacket = null;
       int bufferIndex = -1;
       for(int i=0;
           i < hostNode.MessageBuffer.length;
           i++) {
           if(saved_segments[i] != null)
           {
               if(GetTCPSequenceNumberFromIPPacket(hostNode.MessageBuffer[i])
                   == sequence_number)
               {
                   IPPacket = hostNode.MessageBuffer[i];
                   bufferIndex = i;
                   break;
               }
           }
       }
       return bufferIndex;
   }

   /** creates a new frame send event at present time */
   void sendNewFrame(
           int sourceInterfaceNumber,
           Utility.JNWBitSet PacketOut,
           boolean isUrgent)
   {
       // set simulation timing
       ModelDataLink.InterfaceBaseClass interfaceCard =
           (ModelDataLink.HDLC)vectorOfInterfaces.get (
               sourceInterfaceNumber );
       long eventStartTime = interfaceCard.getTransmitTime();
       long eventStopTime = eventStartTime ;
       interfaceCard.updateTransmitTime(eventStopTime);
       tcpEvent.UpdateGlobalSimultationTime ( eventStopTime );

       // send frame
       ModelEvents.SendFrameFromInterface sffi =
           new ModelEvents.SendFrameFromInterface (
               ModelData.GlobalVariables.simulationTime+100,
               sourceInterfaceNumber,
               PacketOut,
               isUrgent);
   }


   /**Accepts a received TCP packet; if data, adds to buffer and ACKs it.
      if ACK, updates send state*/
   public void ProcessReceivedTCPPacket ( long eventTime , JNWBitSet IPPacketIn, boolean failedCRC
           /*, int nodeNumber*/ ) {
      // retrieve the TCP segment from a IP packet (BZ)
      JNWBitSet TCPSegment = ModelTransport.TCPUtilities.
              CloneTCPSegmentFromIPPacket ( IPPacketIn ); 
      long destinationNodeNumber = 
          ModelTransport.IPUtilities.
              GetDestinationAddressFromIPHeader(IPPacketIn); // BZ

      // if there was a transmission error, impose SN of last_char_acked
      // on the packet so that it causes that to be acked => NACK for any higher SN
      JNWBitSet imposeSN;
      if(failedCRC)
      {
          long replaceSN = last_char_acked+1-(IPPacketIn.getJNWLength()-320)/BITS_PER_BYTE;//IP=TCP hdrs
          if(replaceSN < 0)
              imposeSN = new JNWBitSet((long)0,32);
          else
              imposeSN = new JNWBitSet(replaceSN,32);
          int fromIndex = IPPacketIn . getJNWLength ( )
              - 32 * ( 5 /*Length of IP header*/
              + 2 /*Location of sequence number in TCP header*/) ;
          IPPacketIn.setSubset(fromIndex,imposeSN);
      }

      // Let the method rtrans_receive deal with the processing of received packet
      rtrans_receive(IPPacketIn); // BZ

   }  // end method ProcessReceivedTCPPacket ( JNWBitSet IPPacketIn )

   // length of a TCP Segment payload in bits
   public long getTCPSegmentPayloadLength(JNWBitSet TCPSegment)
   {
      return TCPSegment.getJNWLength() -
             ModelData.GlobalVariables.SEGMENT_HEADER_OFFSET*8;
   }

   /**Creates JNW data structures to support a TCP connection between
    * the source node and the destination node.
    * If the entry in the table tcpConnections [ destinationNodeNumber ]
    * is NOT null then a connection has already been opened. */
   // reliable transport aend
   public void rtrans_send() {

     JNWBitSet send_packet = null;
     JNWBitSet reply_packet = null;
     JNWBitSet reply_segment = null;
     int segmtno;
     long ackno;
     boolean isSYNACK = false, isFINACK = false;

     // response to input depends on state of TCP send
     switch(send_state) {

// state closed (tl_send is inactive at this host; this is active_open
// {
//   copy the message to outgoing_message;
//   send a segment with ack=0, syn=1, fin=0;
//   set timer for net_rtt_timeout;
//   set tl_send_state to syn_sent;
//   initialize state for send side of connection
// }
//
    case closed:
//  tl_send is inactive at this host; this is active_open
//  (a) copy the message to outgoing_message[]
//  (b) send a segment with ack=0, syn=1, fin=0
//  (c) set timer for net_rtt_timeout
//  (d) set tl_send_state to 'syn_sent'
//  (e) initialize state for send side of connection to syn_sent
//  (f) initialize state for receive side to receiving

         isSYNACK = false;
         isFINACK = false;

         // Schedule SYN transmission
         sendNewFrame(interfaceNumber,synPacket,false);
         synSentTime = ModelDES.DiscreteEventSimulator.getSimulationTime();
         send_state = ModelData.GlobalVariables.tl_send_state.syn_sent;
         receive_state = ModelData.GlobalVariables.tl_receive_state.receiving;
         last_reply = null;
         last_char_sent = -1;
         send_window_start = 0;// NOTE: real TCP chooses at random
         number_of_fins_sent = 0;

         break;

// state syn_sent (a SYN has been sent to open connection; 
// either an ACK has been received or the timer has expired):
// {
//   if ACK has not been received
//   {
//     send the connection request SYN again;
//     set timer for net_rtt_timeout;
//     send the SYN again;
//   }
//   else
//   {
//     set tl_send_state to tl_sending;
//     start transmit;
//   }
//
    case syn_sent:
//  a SYN has been sent to open connection; either a SYN-ACK has been
//  received or the timer has expired;
//    if SYN-ACK has not been received
//    (a) send the connection request SYN again
//    (b) set timer for net_rtt_timeout
//    (c) send the SYN again
//  otherwise set tl_send_state to 'tl_sending' and start transmit

        reply_segment = this.last_reply;
        // If the received packet is not a SYN-ACK, resend a SYN
        isSYNACK = false;
        if(reply_segment != null) {
            isSYNACK = GetTCPSYNBitFromTCPSegment(reply_segment) &&
                       GetTCPACKBitFromTCPSegment(reply_segment) &&
                       !GetTCPFINBitFromTCPSegment(reply_segment);
        }
        if(!isSYNACK) {
            // Schedule SYN transmission
            sendNewFrame(interfaceNumber,synPacket,false);
            synSentTime =
                ModelDES.DiscreteEventSimulator.getSimulationTime();
            this.last_reply = null;
            break;
        } else {
            RTT = ModelDES.DiscreteEventSimulator.getSimulationTime() -
                synSentTime;
        }
        this.send_state =
            ModelData.GlobalVariables.tl_send_state.tl_sending;

// state tl_sending (connection has been established and tl is sending the
// message in outgoing_message; when any segment is sent, save a copy in 
// tl_buffer; send_rtl_segments saves the segments
// {
//   update send_window_start;
//   update send_window_size;
//   send to NL any TL segments that need sent;
//   schedule timeout;
//   escape from this case;
// }
//
    case tl_sending:
//  connection has been established and tl is sending the message in
//  outgoing_message[]; (when any segment is sent, 
//  set a timer to call trans_send back after RTT
//  time units; send_rtl_segments() does this)
//
//  (a) begin processing here by using any segment that has been
//  received from other end to update
//  send_window_start and send_window_size.

        // if we have a reply from other end, extract and use 
        // its window_size and ackno
        reply_segment = last_reply;
        last_reply = null;
        if(reply_segment != null) {

            // "send_window_size" is set by the destination, hence must be
            // updated in any case:SYN-ACK (= initialWindowSize), ACK
            // (the calculated window size by destination)
            // NOTE: this is for completeness; JNW currently does not use window
            // for congestion control
            send_window_size = GetTCPWindowSizeFromTCPSegment(reply_segment);

            // ACK received - remove any saved_segments below ackno
            // and move send_window_start up to ackno
            // also update RTT
            ackno = GetAcknowledgementNumberFromTCPSegment(reply_segment);
            if( ackno > send_window_start) {
                for(segmtno=0; segmtno < saved_segments.length; segmtno++) {
                    if((saved_segments[segmtno] != null) &&
                        (segment_send_times[segmtno] >= 0)) {
                        // Delete all segments with seqno smaller than the ackno
                        // of the received ACK, since they have been delivered
                        // successfully and ACKed
                        if((GetSequenceNumberFromTCPSegment(saved_segments[segmtno]) +
                            getTCPSegmentPayloadLength(
                                saved_segments[segmtno])/BITS_PER_BYTE)
                            <= ackno) {
                            saved_segments[segmtno] = null;
                            // send_window_start is the next expected segment
                            // at the destination
                            send_window_start =
                                GetAcknowledgementNumberFromTCPSegment(reply_segment);
                            RTT = ModelDES.DiscreteEventSimulator.getSimulationTime() -
                                segment_send_times[segmtno];
                        }//end if((GetSequence...
                    }// end if(saved_segments[segmtno] != null)
                }// end for(segmtno=0;...
            }// end if(GetAcknowledgementNumber...
        } // end if(reply_segment != null)

        // now send the segments implied by the last reply
        send_rtl_segments(reply_segment);
        break;

// state close_wait (tl has sent all the data, received ACKs for all of it, and
// sent off a FIN; either another segment has been received, or the timer has expired):
// {
//   if another segment has been received with ack=1, syn=0, fin=1 (FIN-ACK)
//   {
//      set tl_send_state to closed;
//      send_rtl_segments(reply_segment);
//      obtain another email if there is one;
//   }
//   else send the FIN packet again
//
    case close_wait:
         if(last_reply != null)
           isFINACK = GetTCPFINBitFromTCPSegment(last_reply) &&
                      GetTCPACKBitFromTCPSegment(last_reply);
         if(isFINACK)
             close_connection();
         else
             sendNewFrame(interfaceNumber,finPacket,true);

//  A more complete solution:
//  tl has sent all the data, received ACKs for all of it, and
//  sent off a FIN; either another segment has been received,
//  or the timer has expired
//  (a) if another segment and it has ack=1, syn=0, fin=1 (FIN-ACK)
//  set tl_send_state to 'closed' and schedule read of another email;
//  (b) otherwise send the FIN twice, setting a timer for net_rtt_timeout
//  (c) after sending and waiting on two FINs, set tl_send_state to 'closed'

         break;
         
    }// end switch(send_state)

    // use JNW event to set a timeout that will wake rtrans_send()
    // if no response to our sending is received (the timeouts can't
    // be triggered without this (could do this at shorter time
    // interval but that slows down the simulation since it causes
    // a fair amount of processing in send_rtl_segments() )
    if(send_state != ModelData.GlobalVariables.tl_send_state.closed) {
        if(ModelDES.DiscreteEventSimulator.getSimulationTime() >=
            nextScheduledTimeout) {
            nextScheduledTimeout =
                ModelDES.DiscreteEventSimulator.getSimulationTime() +
                timeoutWakeupPeriod;
            ModelEvents.TimeoutTCP ttcp =
                new ModelEvents.TimeoutTCP(nextScheduledTimeout,this);
            tcpEvent.UpdateGlobalSimultationTime(nextScheduledTimeout);
        }
    }


}// end rtrans_send()

    // this is the portion of reliable transport send that is left for
    // student completion; see comments below for functions
    public void send_rtl_segments(JNWBitSet reply_seg) { // BZ

       JNWBitSet send_segment;
       int segmtno;
       long ackno;
       JNWBitSet IPPacket;
       int interfaceNumber = hostNode.routingTable[(int) destinationNodeNumber];
       ModelDataLink.InterfaceBaseClass transmitInterface = 
           vectorOfInterfaces.get(interfaceNumber);
       
// the TCP segments of the email to be transmitted are available in 
// two forms when send_rtl_segments is invoked:
// 1. as IP Packets in JNWBitSet[] hostNode.MessageBuffer
// 2. as TCP segments in JNWBitSet[] saved_segments
// CAUTION: when saved_segments elements are ACKed, they are set to null
//
// algorithm for sending TL segments (method send_rtl_segments)
//
// Step 1:
//
// if reply_seg is not null and contains an ackno (syn==0, fin==0)
//   indicating that a previously sent segment has not been received
//   (ackno<last_char_sent), where ackno can be obtained using
//   GetAcknowledgementNumberFromTCPSegment() :
//   do for all non-null segments in the buffer (number is saved_segments.length),
//     if that segment has had time to reach its destination
//     (segment send time is greater than zero and
//      ModelDES.DiscreteEventSimulator.getSimulationTime() minus last send time
//       is greater than 2*RTT),
//       send that segment again from the tl_buffer and also update RTT
//       and its segment_sent_time;
//     during this process update the RTT for this TCPConnection as the
//       difference between the aend time of the segment with matching seqno
//       and the global "clock" 
//       ModelDES.DiscreteEventSimulator.getSimulationTime()
//
       // Step 1: Receive a NACK, resend the failed segment, and update RTT
       if(reply_seg != null)
       if(reply_seg.getJNWLength() > 0) {
           // Is this a NACK (ACK of previous send when later segs had time to arrive)?
           long seqnoAcked =
               GetAcknowledgementNumberFromTCPSegment(reply_seg);
           if(!GetTCPSYNBitFromTCPSegment(reply_seg) &&
              !GetTCPSYNBitFromTCPSegment(reply_seg) &&
              ( seqnoAcked < last_char_sent)) {
               // if so resend all non-null segments old enough to have caused it
               long simTime = ModelDES.DiscreteEventSimulator.getSimulationTime();
               for(segmtno=0; segmtno < saved_segments.length; segmtno++) {
                   if(saved_segments[segmtno] != null)
                   if(segment_send_times[segmtno] >= 0)
                   if(simTime - segment_send_times[segmtno] > 2*RTT) {
                       sendNewFrame(
                           interfaceNumber,
                           getPacketFromBuffer(segmtno),
                           false);
                       RTT = simTime - segment_send_times[segmtno];
                       segment_send_times[segmtno]= simTime;
                   }// end if(simTime...
               }// end for(segmtno...
           }// end if(GetTCPACK...
       }// end if(reply_seg...

// Step 2:
// calculate window_end as the smaller of (send_window_start+send_window_size-1)
// and (sending_message.size-1);
// while last_char_sent is less than window_end,
// {
//   send another segment from the message, available as
//     getPacketFromBuffer(segment_number)
// }

       // Step 2: Send a segment within window size
       // Update window end
       send_window_end = send_window_start + send_window_size-1;
       if(send_window_end > hostNode.getMessageSize())
           send_window_end = hostNode.getMessageSize()-1;

       // send if window has room
       while(last_char_sent < send_window_end) {
           int send_segment_index =
               getPacketIndexFromMessageBufferBySequenceNumber(
                   last_char_sent+1);
           if(send_segment_index >= 0)
           {
               JNWBitSet packetToSend =
                   getPacketFromBuffer(send_segment_index);
               sendNewFrame(
                   interfaceNumber,
                   packetToSend,
                   false);
               last_char_sent += 
                   GetDataLengthInBytesFromTCPIPPacket(
                       packetToSend);
               segment_send_times[send_segment_index] =
                   ModelDES.DiscreteEventSimulator.getSimulationTime();
           }
           else break;
       }//end while(last_char_sent < send_window_end)
       
// Step 3:
// resend timed-out segments: for each non-null segment in saved_segments,
//   if segment has timed out waiting for ACK, retransmit its packet
// packet is available from getPacketFromBuffer(); send only
//   for non-negative segment_send_times[]
// for timeout use segment_send_times[] + 2*RTT + 5000, not TCP's RTT+4*D)
//   compared to ModelDES.DiscreteEventSimulator.getSimulationTime()
//
// for every segment tranmitted, save segment_send_time from
//   ModelDES.DiscreteEventSimulator.getSimulationTime())
//   and update last_char_sent by adding GetSequenceNumberFromTCPSegment()
//   its getTCPSegmentPayloadLength()/8 and subtracting one
//
// students should modify the code immediately following this to
// implement the above algorithm
//
// to send a segment:
//   sendNewFrame(interfaceNumber,bitsetOfPacket,false);
//
       // Step 3: send timeout segments
       // check past sent segments for timeouts

       //// STUDENT SOLUTION FOR CS555 FALL 2011 GOES HERE
       //for all the segments in saved segments


// Step 4:
// if send_window_start is greater than or equal to size of message being sent
// (all data has been sent and ACKed),
// {
//   send FIN;
//   set tl_send_state to close_wait;
// }.
//
       // Step 4: All segments sent, send FIN
       if(send_window_start >= hostNode.getMessageSize()) {
           sendNewFrame(interfaceNumber, finPacket, true);
           finSentTime =
               ModelDES.DiscreteEventSimulator.getSimulationTime();
           send_state = ModelData.GlobalVariables.tl_send_state.close_wait;
       }
   }// end send_rtl_segments

   // reliable transport receive
   public void rtrans_receive(JNWBitSet IPPacketIn) { // BZ

       JNWBitSet ack_segment;
       long ackno;
       JNWBitSet received_segment =
           ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(IPPacketIn);

       // destroy a fraction of the packets arriving,
       // based on randomNumberGenerator, to simulate
       // losses due to congestion; use error rate of first
       // link attached to node
       float randomNumber = randomNumberGenerator.nextFloat();
       float errorRate =
           hostNode.vectorOfLinksForThisNode.get(0).bitErrorRate;
       if(randomNumber < errorRate) {
           Utility.out.println("*** DROPPED A PACKET AT NODE:"+
               hostNode.thisNodeNumber+" interface:"+interfaceNumber+
               " sim time:"+
               ModelDES.DiscreteEventSimulator.getSimulationTime());
           return;
       }

       // extract flag bit values
       boolean ACKbit = GetTCPACKBitFromTCPSegment(received_segment);
       boolean SYNbit = GetTCPSYNBitFromTCPSegment(received_segment);
       boolean FINbit = GetTCPFINBitFromTCPSegment(received_segment);
       boolean isSYN = SYNbit && !ACKbit && !FINbit;
       boolean isACK = ACKbit && !SYNbit && !FINbit;
       boolean isFIN = FINbit && !ACKbit && !SYNbit;
       boolean isSYNACK = SYNbit && ACKbit && !FINbit;
       boolean isFINACK = FINbit && ACKbit && ! SYNbit;

       // If received any sort of ACK for sent segment
       // let rtrans_send deal with it - then return immediately
       if(ACKbit ) {
           last_reply = received_segment;
           rtrans_send();
           return;
       }

       switch(receive_state) {

// state listening (tl_receive is listening at this host; only a SYN 
// or SYN-ACK should be received:
// if a SYN without ACK is received,
// {
//   reply with a SYN-ACK (ack=1, syn=1, fin=0,ackno=seqno+1);
//   set tl_receiving_from_net and tl_receiving_from_host;
//   set tl_receive_state to receiving;
//   escape from this state;
// }
// if a FIN is received (because FIN-ACK from this host gets lost),
//   reply with FIN-ACK;
//   break;
       case listening:

           // Received a SYN, reply with a SYN-ACK
           if(isSYN) {
               sendNewFrame(interfaceNumber, synAckPacket, false);
               synSentTime =
                   ModelDES.DiscreteEventSimulator.getSimulationTime();
               last_char_acked = -1;
               receive_state =
                   ModelData.GlobalVariables.tl_receive_state.receiving;
               break;
           }
               
           if(isFIN)
               sendNewFrame(interfaceNumber, finAckPacket, true);
           break;

       case receiving:

// state receiving
// (tl_receive has received a SYN and replied SYN-ACK at some
// point, now it is accepting data segments; or is receiving ACKs):
// {
//   if a SYN arrives from the connected net-host (this means the
//   SYN-ACK previously sent had not reached the sending end before
//   timeout, so it is necessary to repeat the reponse to SYN above),
//   {
//     reply with an ACK-SYN (ack=1, syn=1, fin=0,ackno=seqno+1);
//     set tl_receiving_from_net and tl_receiving_from_host;
//     set tl_receive_state to receiving;
//     escape from this case;
//   }
//
           if(isSYN) {
               JNWBitSet respond_packet = synAckPacket;
               // Schedule SYN-ACK transmission
               //transmitInterface.transmitBuffer.add(reply_segment);
               sendNewFrame(interfaceNumber, respond_packet, true);
               last_char_acked = -1;
               receive_state =
                   ModelData.GlobalVariables.tl_receive_state.receiving;
               break;
           }
//   if it is an ACK and its ackno is greater than
//   send_window_end, update send_window_start to its ackno and
//   escape from this case
          if(isACK || isSYNACK) {
              ackno = GetAcknowledgementNumberFromTCPSegment(received_segment);
              int data_size = received_segment.getJNWLength()-320;//TCP+IP headers
              if(ackno > send_window_end+1)
              {  
                 send_window_end = ackno + data_size - 1;
                 send_window_start = ackno;
 
                 // call rtrans_send to respond to NACKs
                 // and send any remaining segments in window
                 rtrans_send();
              }
              break;
          }

//   if this is a data segment (fin == syn == ack == 0) and seqment length in
//       bytes is greater than SEGMENT_HEADER_OFFSET,
//   {
//     if this segment starts with last_char_acked + 1
//     {
//         create an ACK segment (ack=1,syn=0, fin=0, ackno=seqno of this
//             segment+maximumTCPPayloadInBits);
//         update last_char_acked with length of data in this segment
//         send the ACK segment;
//     }
//     escape from this case;
//   }
           // Received a data segment
           // Copy received segment to message buffer
           if(getTCPSegmentPayloadLength(received_segment) > 0) {
               // send an ACK for highest continuous byte received;
               // if this is the immediate next segment after the one just
               // received (its seqno is last_char_acked+1) also append to buffer
               if((last_char_acked + 1) ==
                  GetSequenceNumberFromTCPSegment(received_segment)) {
                   last_char_acked += getTCPSegmentPayloadLength(received_segment)/8;
                   
                   // add the received segment to the incoming message payload
                   int segmentIndex =
                       (int)GetSequenceNumberFromTCPSegment(received_segment)/20;
                   emailBuffer[segmentIndex] = GetDataFromTCPSegment(received_segment);
               }// end if(last_char_acked ==...
               
               // send the ACK
               ackno = last_char_acked + 1;
               JNWBitSet DataACKPacket = ComposeDataACKPacket(ackno);
               sendNewFrame(interfaceNumber, DataACKPacket, false);

               // check for need to allocate new buffer
               long offset = GetSequenceNumberFromTCPSegment(received_segment);
               if(offset == 0) {

                   // this is the first segment; allocate the buffer
                   incoming_message = new IncomingMessage(
                       IPUtilities.GetSourceAddressFromIPHeader(IPPacketIn),
                       IPUtilities.GetDestinationAddressFromIPHeader(IPPacketIn));
               }

           }//end if(received_segment.getJNWLength() >...

//   if it is a FIN (all data has been sent and ACKed),
//   {
//     send a FIN-ACK segment (ack=1,syn=0,fin=1);
//     pass the message to the application layer;
//     set tl_receive_state back to listening;
//     escape from this case;
//   }
//   if it is a FIN-ACK (ack=1, syn=0, fin=1) (done with connection)
//     set tl_send_state to closed.
           
           if(isFIN) {
               printSegmentsAsEmail();
               sendNewFrame(interfaceNumber, finAckPacket, true);
               this.receive_state =
                   ModelData.GlobalVariables.tl_receive_state.listening;
               break;
           } // end receive FINACK

       }// end switch(this.receive_state)

   }

   // close connection and read next email, if there is one
   void close_connection() {
       Utility.out.println("\nCLOSE CONNECTION AT NODE:"+
           hostNode.thisNodeNumber+" INTERFACE:"+interfaceNumber);
       this.send_state =
           ModelData.GlobalVariables.tl_send_state.closed;
       hostNode.thisTCPEmailIsDone = true;
   }
   
}  // end of class TCPConnection
