/* VERSION 20071203
 * UDPState.java
 *
 * Created on July 6, 2007, 2:16 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelTransport;

import Utility.JNWBitSet ;

/**The state of a TCP connection.  <p>The TDCPState also contains some methods that are
 * directly related to updating the TCP variables, RRT and D.
 * @see <a href="http://www.freesoft.org/CIE/Course/Section4/8.htm">TCP Header Format</a>
 * @see <a href="http://www.faqs.org/rfcs/rfc1700.html">RFC 1700 - Assigned Numbers</a>
 * @see <a href="http://cities.lk.net/tcp.html">Lots of TCP Algorithms</a>
 * * 
 * *<p><b>Versions:</b>
 * <li><b>Version 1.101</b> of 20080418 removes the error message at the end of
 * DeleteTransmittedPacketInformation.  The message is presnetly commended out.</li>
 * * <li><b>Version 1.200</b> of 20080822 changes the dimension of array
 * arrayOfTransmittedPacketInformation from [20] [5] to [40] [5].</li>
 * <li><b>Version 1.201</b> of 20081021 changes  
 *       for ( int i = 0 ; i < arrayOfTransmittedPacketInformation .length ; i++ )
      { 
         if ( arrayOfTransmittedPacketInformation [ i ] [ 0 ] == 0 )
         {
            openRow = i ;
            break ;  // Now put the data into the row
         }  // Is this row valid and is the data sequence numnber is already in the array?
 *       else if ...
 * TO
 *       LOOP: for ( int i = 0 ; i < arrayOfTransmittedPacketInformation .length ; i++ )
      {   // Is this row empty?
         if ( arrayOfTransmittedPacketInformation [ i ] [ 0 ] == 0 )
         {  // The first open row
            openRow = i ;
            break LOOP ;  // Now put the data into the row
         }  // end of if ( arrayOfTransmittedPacketInformation [ i ] [ 0 ] == 0 ) 
      for ( int i = 0 ; i < arrayOfTransmittedPacketInformation .length ; i++ )
         if ( arrayOfTransmittedPacketInformation [ i ] [ 1 ]
            == dataSequenceNumber )
 * The break was in the wrong place with the else if obscuring the conditions for branching
 * when a duplicate packet arrived.</li> 
 * @version 1.201
 * @author dixonw
 * @modified JMP 24Nov11 for JTRN3
 */
public class TCPState implements GUIConstants . General , ModelConstants . General
{
   /**Get a reference to the discrete event simulator.*/
   protected static ModelDES.DiscreteEventSimulator discreteEventSimulator
      = ModelDES.DiscreteEventSimulator.getInstance () ;
   
   /**A singleton vector containing all the Nodes*/
   static ModelData.JNWSingletonVectorOfNodes<ModelPhysical.Node> vectorOfNodes
      = ModelData.JNWSingletonVectorOfNodes.getInstance (  ) ;
   
   /**A singleton vector containing all the interfaces*/
   static ModelData.JNWSingletonVectorOfInterfaces < ModelDataLink.InterfaceBaseClass >
      vectorOfInterfaces
      = ModelData.JNWSingletonVectorOfInterfaces.getInstance (  ) ;
   
   /**Destination port.  This assume that the destination is the
    *HTTP port, 80.  I also assume that the source and
    *destination ports are identical.*/
   protected static int TCPDestinationPort = 80 ;
   
   /**Source port.  This assume that the source is the HTTP port,
    *80.  I also assume that the source and destination ports are identical.*/
   protected static int TCPSourcePort = TCPDestinationPort ;
   
   /**Time to process the data into a TCP packet.  Here it is assumed
    *to be 10 microseconds.*/
   protected static long TCPProcessingTime = ( long ) ( .000010
      * ModelData.GlobalVariables.timeTicksPerSecond ) ;
   
   /**The maximum payload size for a TCP packet in JNW.  A real TCP
    *packet can have up to 65,495 bytes.*/
   public static int maximumTCPPayloadInBytes ;
   
   /**The maximum TCP payload in bits.  Initialized in the TCPUtilities
    *constructor.*/
   static int maximumTCPPayloadInBits ;
   
   /**Sequence number of the data in the TCP payload.  This is the number
    *of the next byte being transmitted by this node of a TCP connection.*/
   public long sequenceNumber ;
   
   /**Acknowledgement number.  The acknowledgement number acknowledges
    *that all bytes before this sequence number have been successfully
    *received.  This is the next byte that this node is expecting.
    *<p>The acknowledgement number is updated by the TCPDestinationBuffer.
    * This means that the initial value of the acknowledgementNumber,
    * set in the TCPState constructor must be passed to the 
    * TCPDestinationBuffer in the constructor of TCPState */
   protected long acknowledgementNumber ;
   
   /**Source node address for this TCP connection*/
   int sourceNodeNumber = 0 ;
   
   /**Destination node number for this TCP connection*/
   int destinationNodeNumber = 0 ;
   
   /**Source node address.  This is the 64-bit version of the source node
    *number.*/
   long sourceNodeAddress  ;
   
   /**Destination node address.  This is the 64-bit version of the destination
    *node number.*/
   long destinationNodeAddress  ;
   
   /**The transmit window size. Only this number of bytes of data can be
    *unacknowledged.  The transmitter cannot send more bytes until receiving
    *acknowledgement of data within window size of the byte to be transmitted*/
   public long slidingWindowSize ;
   
   /**The number of bytes that have been transmitted but are unacknowledged.
    *   This number must remain less than the slidingWindowSize or the new packet
    *cannot be transmitted.*/
   public int unacknowledgedBytesInTransit ;
   
   /**The maximum number of transmission attempts.*/
   public static int maximumNumberOfTCPTransmissionAttempts = 16 ;
   
   /**Initial window size*/
   protected static int initialSlidingWindowSize ;
   
   /**
    * Window threshold above which the window size increases linearly.  Below the
    * window threshold each successfully acknowledged packet doubles the window
    * size.  Above the window threshold each successfully acknowledged packet
    * increaes the window size by slidingWindowSizeIncrementAboveThreshold
    */
   protected static int slidingWindowThreshold ;
   
   /**Increment in the window size when the window size exceeds
    *slidingWindowThreshold and another packet is successfully acknowledged. */
   protected static int slidingWindowSizeIncrementAboveThreshold ;
   
   /**The value of alpha used in the exponential average of round trip time.
    *0.0 <= alpha <= 1.0
    * @see <i>Computer Networks </i>, Fourth Edition, Andrew S. Tannenbaum, pages 551
    * and 552*/
   protected static double alpha ;
   
   /**Value of the exponential average of the round trip time in ticks
    *for messages to be acknowledged.*/
   public long RTT ;
   
   /**Deviation of the round trip time in ticks.  Used to compute the maximum
    *wait time until the sender assumes that the TCP packet has been lost, and
    *retransmission starts.*/
   public long D ;// not used with NW (BZ)
   
   /**The last transmission attempt exceeded the sliding window szie.  If true
    *the window size should NOT be decreased.  This is used in SendTCPFrame.*/
   public boolean lastTransmissionAttemptExceedWindowSize ;
   
   /**Creates a TCPDestinationBuffer to receive the ordered the TCP packets.*/
   ModelTransport.TCPDestinationBuffer TCPDestinationBuffer
      = null  ;

   public ModelData.GlobalVariables.tl_send_state send_state;

   public ModelData.GlobalVariables.tl_receive_state receive_state;

   /** last segment received from other end */
   public JNWBitSet last_reply = null;

   public long send_window_start; // Start of Send window (BZ)

   public long last_char_sent;    // Last Character sent (BZ)

   public long send_window_size;  // The size of send window (BZ)

   public long send_window_end;   // The last character in current send winodw (BX)

   public long last_char_acked;   // BZ

   public long number_of_fins_sent; //BZ

   public JNWBitSet[] saved_segments; // A buffer storing unACKed TCP segments BZ

   public ModelTransport.IncomingMessage incoming_message; // A transport-layer buffer for storing incoming message BZ
   
   /** Initializes a new instance of TCPState */
   public TCPState ( )
   {
      // This is instantiated first to ensure the
      // bufferAcknowledgementNumber can be set in the
      // TCPDestinationBuffer
      TCPDestinationBuffer = new ModelTransport.TCPDestinationBuffer ( ) ;
      IncomingMessage incomingMessage; // BZ
      send_state = ModelData.GlobalVariables.tl_send_state.closed; // The initial send state is closed (BZ)
      receive_state = ModelData.GlobalVariables.tl_receive_state.listening;
         
         //The sequence number and the acknowldegement number
      // should be set by first setting the sequence number on each end of
      // each TCP connection to a random
      // number between 0 and 2**32-1 incluse and then sending the 
      // sequence number to the other end of the connection via a three-way
      // handshake

      // Important: Initial value of TCP state variables are set here (bo)
      sequenceNumber = 0 ;
      acknowledgementNumber = 0 ;

      // Now initialize the "real" acknowledgementNumber in the destination buffer
      TCPDestinationBuffer . bufferAcknowledgementNumber 
         = acknowledgementNumber ;
      
      // Data used in Student exercises
      unacknowledgedBytesInTransit = 0 ;
      slidingWindowThreshold = 2000 ;
      slidingWindowSizeIncrementAboveThreshold = 200 ;
      // START OF STUDENT CHANGES NEEDED FOR EXERCISES 
      // START OF STUDENT CHANGES NEEDED FOR EXERCISES 
      initialSlidingWindowSize = 100  ;
      slidingWindowSize = initialSlidingWindowSize ;
      unacknowledgedBytesInTransit = 0 ;
      slidingWindowThreshold = 800 ;
      slidingWindowSizeIncrementAboveThreshold = 100 ;
      maximumTCPPayloadInBytes = 20 ;
      maximumTCPPayloadInBits = BITS_PER_BYTE * maximumTCPPayloadInBytes  ;
      RTT = 10000 ; // 1000 microseconds
      // END OF STUDENT CHANGES NEEDED FOR EXERCISES 
      // END OF STUDENT CHANGES NEEDED FOR EXERCISES 
      alpha =  0.5 ;
      D = 0 ;
      lastTransmissionAttemptExceedWindowSize = false ;
   }  // end of constructor TCPState()
   
   static
   {
   }
   
   public void InitSendWindowAfterTCPConnectionEstablishment() { // BZ

       send_window_start = 0;
       last_char_sent = -1;
   }

   public void InitReceiverWindowAfterTCPConnectionEstablishment() { // BZ

       last_char_acked = -1;
       send_window_end = 0;
   }

   /**Updates the round trip time as an exponential average,
    * and updated the deviation of the round trip time.
    * @see " Computer Networks, Fourth Edition, Andrew S. Tannenbaum, pages 551 and 552"
    */
   protected void UpdateRRTAndD ( long measuredDelay )
   {
      D =  ( long ) ( alpha * D + ( ( 1.0 - alpha )
      * java.lang.Math.abs ( RTT - measuredDelay ) ) );
      RTT = ( long ) ( alpha * RTT + ( ( 1.0 - alpha ) * measuredDelay ) ) ;    
   }  // end of UpdateRRTAndD

   
   /**Updates the TCP window size.  It increases the window size if
    * increaseWindowSize == true.  It decrease the slidingWindowSize back to
    * its initial value if increaseWindowSize == false.
    * @see " Computer Networks, Fourth Edition, Andrew S. Tannenbaum, pages 551 and 552"
    */
  public void WindowSizeUpdateIncreaseIfTrueDecreaseIfFalse (
      boolean increaseWindowSize)
   {
      // not used in JNWSv1.202
      
      Utility.out.print("increaseWindowSize="+increaseWindowSize+
                        " entry windowSize="+slidingWindowSize);

   }  // end of WindowSizeUpdateIncreaseIfTrueDecreaseIfFalse
   
   

   /**For testing only.*/
   public static void main_Testing_Only ( String [ ] args )
   {
      
   }  // end of main
   
}  // end of class TCPState
