/*
 *
 * Created on July 6, 2007, 2:15 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelTransport;

import Utility.* ;
import static ModelTransport.IPUtilities.* ;

/**Utilities for the creation of TCP headers for IP packets.
 * @see <a href="http://www.freesoft.org/CIE/Course/Section4/8.htm">TCP Header Format</a>
 * @see <a href="http://www.faqs.org/rfcs/rfc1700.html">RFC 1700 - Assigned Numbers</a>
 * 
 *  * *<p><b>Versions:</b>
 * <li><b>Version 1.101</b> of 20080609 adds method
 * GetTCPAcknowledgementNumberFromIPPacket.</li>
 * <li><b>Version 1.103</b> of 20081121 cleans
 * up errors in names.</li>
 * @version 1.202
 * @author dixonw
 * @ modified BZ May 2010 for JTRN3
 */
public class TCPUtilities extends TCPState implements ModelConstants.General
{

   /**The JNW random number generator.*/
   static Utility.JNWRandom randomNumberGenerator = Utility.JNWRandom.getInstance (); // BZ
   /**The source port bit set for the TCPHeader.  This is initialized once
    *in the static initializer.*/
   static JNWBitSet sourcePortBitSet = null ;
   
   /**The source port bit set for the TCPHeader.  This is initialized once
    *in the static initializer.*/
   static JNWBitSet destinationPortBitSet = null ;
   
   /**The bit set representing the source and destination ports.  This is
    *the concatination of sourceNodeBitSet and destinationNodeBitSet.  It is
    *useful for speeding the creation of the TCP header.  It is defined
    *in the static initializer.*/
   static JNWBitSet portBitSets = null ;
   
   /**
    * The TCP header length in 32-bit blocks.  The TCP header has no options,
    * so TCPHeaderLengthBitSet has value 5.
    */
   static JNWBitSet TCPHeaderLengthBitSet =  null ;
   
   /**Six zeros.  This is the six bit field of 0s between the
    * TCP header length and the TCP flags.  This variable is made
    * a static class variable.*/
   static JNWBitSet sixZeros = null ;
   
  /*The TCP flags, URG, ACK, PSH, RST, SYN, and FIN.  This variable
   *is made a static class variable. Set to "000000".  The last fragment
   *in a message has the FIN bit set to "1".*/
   static JNWBitSet TCPFlagsDefaultValue;

   // The TCP flag bits of various type segment
   static JNWBitSet TCPFlagBitsofSYN;    // BZ
   static JNWBitSet TCPFlagBitsofSYNACK; // BZ
   static JNWBitSet TCPFlagBitsofFIN;    // BZ
   static JNWBitSet TCPFlagBitsofFINACK; // BZ
   static JNWBitSet TCPFlagBitsofACK;    // BZ
   
   /*The TCP flags, URG, ACK, PSH, RST, SYN, and FIN.  This variable
    *is made a static class variable. Set to "000001".*/
   static JNWBitSet ACKFlags = null ;
   
   /**The sliding window default value, "00000000".  This default value will
    *probably not ever be changed.*/
   static JNWBitSet windowSizeDefaultValue = null ;
   
   /**The TCP checksum default value, "00000000"  The default value is later
    *updated.*/
   static JNWBitSet checksumDefaultValue = null ;
   
   /**The  urgent pointer, "00000000".  This default value will probably not
    *ever be changed.*/
   static JNWBitSet urgentPointerDefaultValue = null ;
   
   /**Source node bit set.  The source node and destination node for all the
    *fragments in a message are the same so they are computed only once in
    *PrependTCPHeadersToData.  These bit set are used only to compute the
    *pseudoheader XOR.*/
   JNWBitSet sourceNodeBitSet  = null;
   
   /**Destination node bit set.  The source node and destination node for all
    *the fragments in a message are the same so they are computed only once in
    *PrependTCPHeadersToData.  These bit set are used only to compute the
    *pseudoheader XOR.*/
   JNWBitSet destinationNodeBitSet  = null ;
   
   /**The XOR of the sourceNodeBitSet and the destinationNodeBitSet and the
    *TCP ptorocol, 6.  Used to
    *compute the pseudoheader checksum.  It is faster to compute this partial
    *sum of the TCPPsuedoheaderXOR once in the TCPUtillities constructor
    *and use the XOR in all computations.*/
   JNWBitSet TCPPseudoheaderPartialXOR  = null ;
   
   /**The pseudoheader checksum.  The checksum is initially set to all 0s.
    *The pseudoheader checksum is the XOR of the source and destination
    *addresses, the TCP protocol, 6, and the total length.  It is
    *then XORed with the XOR of the data and TCP
    *header to form the real checksum.  The real checksum is then inverted
    *and set into the TCP header.*/
   JNWBitSet TCPPsuedoheaderXOR = null ;
   
   /**The TCP protocol number, 6, as a bit set.  Used in computing the
    *pseudoheader checksum.  The field include 8 preceeding "0"s.*/
   static JNWBitSet TCPProtocolBitSet = null ;
   
   /**
    * The fields of the TCP header after the acknowledgement number,
    * TCPHeaderLengthBitSet, the reserved sixZeros, the TCPFlagsDefaultValue
    * , the default TCP flags, the advertised window size, the default
    * checksum, the default urgent pointer.  Since these default fields
    * never change, they are concatinated in the static initializer only
    * once for repeated use, speedin up the model.
    */
   static JNWBitSet allTCPHeaderAfterAcknowledgementNumber = null ;
   
   /** Creates a new instance of TCPUtilities */
   public TCPUtilities ( int sourceNodeNumber , int destinationNodeNumber  )
   {
      super ( ) ;
      // Save the source node number and the destination node number
      this . sourceNodeNumber = ( int ) sourceNodeNumber  ;
      this . destinationNodeNumber = ( int ) destinationNodeNumber  ;
      // Save the source node and destination node bit sets
      // These are declared in TCPUtilities
      sourceNodeBitSet  = new JNWBitSet ( sourceNodeNumber , 32 ) ;
      destinationNodeBitSet  = new JNWBitSet ( destinationNodeNumber , 32 ) ;
      // Compute the partial XOR of the pseudoheader
      // This in computed only once for each TCP connection
      TCPPseudoheaderPartialXOR = sourceNodeBitSet
         . xorAndShift ( TCPProtocolBitSet ) ;
      TCPPseudoheaderPartialXOR = destinationNodeBitSet
         . xorAndShift ( TCPPseudoheaderPartialXOR ) ;  
   }  // end of constructor
   
   
   //Static initializer which is called only once
   static
   {  // The source and destination ports never change
      sourcePortBitSet = new JNWBitSet ( TCPSourcePort , 16 ) ;
      // Assumes that we are sending between ports 80
      destinationPortBitSet = new JNWBitSet ( TCPDestinationPort , 16 ) ;
      portBitSets = sourcePortBitSet . clone ( ) ;
      portBitSets . append ( destinationPortBitSet ) ;
      // The TCP header length  which is ALWAYS 5
      TCPHeaderLengthBitSet = new JNWBitSet ( 5 , 4 ) ;
      // Six zeros
      sixZeros = new JNWBitSet ( "000000" ) ;
      // The default value of the TCP flags
      TCPFlagsDefaultValue = new JNWBitSet ( "000000" ) ;
      TCPFlagBitsofSYN = new JNWBitSet("000010");    // BZ
      TCPFlagBitsofSYNACK = new JNWBitSet("010010"); // BZ
      TCPFlagBitsofFIN = new JNWBitSet("000001");    // BZ
      TCPFlagBitsofFINACK = new JNWBitSet("010001"); // BZ
      TCPFlagBitsofACK = new JNWBitSet("010000");    // BZ

      // The window size, checksum, and urgent pointer are
      // set initially to all zeros
      windowSizeDefaultValue = new JNWBitSet (  60, 16 ) ;
      checksumDefaultValue = new JNWBitSet ( 16 ) ;
      urgentPointerDefaultValue = new JNWBitSet ( 16 ) ;
      allTCPHeaderAfterAcknowledgementNumber = TCPHeaderLengthBitSet . clone ( ) ;
      allTCPHeaderAfterAcknowledgementNumber . append ( sixZeros ) ;
      allTCPHeaderAfterAcknowledgementNumber . append ( TCPFlagsDefaultValue ) ;
      allTCPHeaderAfterAcknowledgementNumber . append ( windowSizeDefaultValue  ) ;
      allTCPHeaderAfterAcknowledgementNumber . append ( checksumDefaultValue ) ;
      allTCPHeaderAfterAcknowledgementNumber . append ( urgentPointerDefaultValue ) ;
      // TCP protocol has value 6  and it will contain the eight leading 0s
      TCPProtocolBitSet = new JNWBitSet ( 6 , 16 ) ;
   }
   
   /**Constructs a TCP Header for a message.  The input is a JNWBitSet.  The
    *output is the TCP header.*/
   private JNWBitSet ConstructTCPHeader ( JNWBitSet messageIn )
   {  // Now construct the TCPHeader
      int messageLengthInBytes = messageIn . getJNWLength ( ) / BITS_PER_BYTE ;
      // Get the sequence number
      JNWBitSet sequenceNumberBitSet = new JNWBitSet ( sequenceNumber , 32 ) ;
      
      // Update the sequence number
      sequenceNumber = sequenceNumber + messageLengthInBytes ;
      // Update the acknowldegement number
      acknowledgementNumber = 0;
      JNWBitSet acknowledgementNumberBitSet
         = new JNWBitSet ( acknowledgementNumber , 32 ) ;
      JNWBitSet TCPHeader = new JNWBitSet ( ) ;
      TCPHeader . append ( sourcePortBitSet ) ;
      TCPHeader . append ( destinationPortBitSet ) ;
      TCPHeader . append ( sequenceNumberBitSet ) ;
      TCPHeader . append ( acknowledgementNumberBitSet ) ;
      TCPHeader . append ( allTCPHeaderAfterAcknowledgementNumber  ) ;
      return TCPHeader  ;
   }  // end of ConstructUDPHeader ( )
   
   
   /**Appends the TCP Header to the head of the message.  The method first
    *breaks the message into umberOfTCPPackets data messages with maximum length
    *maximumTCPPayloadInBytes.  IT then calls ConstructTCPHeader to construct
    *the TCP header.  It appends the TCP headers to the data.  It then
    *computer the correct XOR and puts the correct XOR into the TCP packet.*/
   public JNWBitSet [ ] PrependTCPHeadersToData ( JNWBitSet messageIn
      , int sourceNodeNumberIn , int destinationNodeNumberIn )
   {  // This sets the class variables to the input
      sourceNodeAddress = sourceNodeNumberIn ;
      destinationNodeAddress  = destinationNodeNumberIn ;
      // Source node bit set
      sourceNodeBitSet  = new JNWBitSet ( sourceNodeNumberIn, 32 );
      // Destination node bit set
      destinationNodeBitSet  = new JNWBitSet ( destinationNodeNumberIn , 32 );
      // Compute the number of TCP packets in the message including
      // the last packet which may not be full  
      int numberOfTCPPackets    
      = ( messageIn .getJNWLength ( ) 
         + maximumTCPPayloadInBits - 1 )
         / maximumTCPPayloadInBits ;
 
      // Create the correct number of TCP packets
      JNWBitSet TCPMessagesOut [ ] = new JNWBitSet [ numberOfTCPPackets ] ;
      for ( int i = 0 ; i < numberOfTCPPackets ; i++ )
      {  //Get the portion of the message for this packet
         TCPMessagesOut [ i ] = messageIn . getSubset (
            java.lang.Math.max ( 0 , messageIn.getJNWLength ( )
         - ( i + 1 )  * maximumTCPPayloadInBits   )
            , messageIn.getJNWLength ( ) - i * maximumTCPPayloadInBits  )  ;
         JNWBitSet TCPHeader = ConstructTCPHeader  ( TCPMessagesOut [ i ] )  ;
         TCPMessagesOut [ i ] . prepend ( TCPHeader ) ;
         // Compute the XOR over the TCP header and the data
         JNWBitSet messageXOR = ComputeTCPXOR ( TCPMessagesOut [ i ] ) ;
         // Get the pseudoheader XOR
         ComputeTCPPseudoHeaderXOR ( TCPMessagesOut [ i ] . getJNWLength ( )/ BITS_PER_BYTE );        
         messageXOR . xor ( TCPPsuedoheaderXOR ) ;
         // Invert the checksum
         messageXOR . invertBitSet ( ) ;
         // Put the checksum into the TCP message header
         TCPMessagesOut[i].setSubset ( TCPMessagesOut[i] . getJNWLength ()  - 18 * BITS_PER_BYTE
            , messageXOR ) ;
      }  // end of for ( int i = 0 ; i < numberOfTCPPackets ; i++ )
      return TCPMessagesOut;
   }  // end of PrependTCPHeadersToData
   
   
   /**Return the TCP packet from the IP packet.  The returned value is a clone of the
    *original so that the returned packet can be changed without chainging the
    *original.*/
   public static JNWBitSet CloneTCPSegmentFromIPPacket ( JNWBitSet IPPacketIn )
   {
      JNWBitSet clone =  IPPacketIn . clone ()  ;
      // Remove the 20 bytes of the IP header
      clone . setJNWLength ( IPPacketIn .getJNWLength ( )
      - IP_HEADER_LENGTH_IN_BITS  ) ;
      return clone ;
   }  // end of CloneTCPSegmentFromIPPacket
   
   
   /**Computes the modulo 2 sum of the TCP pseudoheader which is used to
    * compute the TCP checksum.  
    * See http://www.freesoft.org/CIE/Course/Section4/8.htm
    * Leon-Garcia and Widjaja, Second Edition, pages 601 and 602 for the 
    * algorithm.
    * <p>The source number and destination numbers are from the user input.  At 
    * the destination, the model reads the source number and destination numbers 
    * from the IP header.*/
   private void ComputeTCPPseudoHeaderXOR ( int TCPSegmentLengthInBytes )
   {  // Include TCP header in length '     
      TCPPsuedoheaderXOR
         = TCPPseudoheaderPartialXOR 
         . xorAndShift ( new JNWBitSet ( TCPSegmentLengthInBytes , 16 ) ) ;
   }  // end of ComputeTCPPseudoHeaderXOR
   
   
   /**Computes the checksum of the TCP packet.  The TCPPacket should include
    *the TCP header since the XOR is computed over the pseudo header
    *including the source and destination addresses. 
    *THIS HACK DOES NOT COMPLY WITH RFC AND MUST BE REPLACED.
    *The checksum in the
    *TCP headers should be initialized to "00000000 00000000" 
    *with new JNWBitSet ( 16 ) as the input value to xorAndShift.*/
   private static JNWBitSet ComputeTCPXOR ( JNWBitSet TCPPacketIn )
   {
      return TCPPacketIn . xorAndShift ( new JNWBitSet ( 16 ) )  ;
   }  // end of ComputeTCPXOR
   
   
   /**Sixteen ones; used to invert a JNWBitSet of size 16 */
   static JNWBitSet sixteenOnes = new JNWBitSet(65535,16);
   
   /**Verifies the TCP checksum at the receiver.  Notice that I include the
    *source and destination nodes from the IP header.  Therefore, if an error
    *occurs in the source or destination addresses of the IP header, an error
    *will be reported. THIS HACK DOES NOT COMPLY WITH RFCs AND MUST BE REPLACED */
   public static boolean VerifyTCPXORFromIPPacket ( JNWBitSet IPPacketIn )
   {  // Start the checksum at all zeros
      JNWBitSet checksum = new JNWBitSet ( 16 ) ;
      // Get the TCP segment    
      JNWBitSet TCPSegment = CloneTCPSegmentFromIPPacket ( IPPacketIn )  ;
      checksum = ComputeTCPXOR ( TCPSegment )  ;

      // include source and destination IP addresses in the checksum
      // THIS HACK DOES NOT FOLLOW RFC AND MUST BE FIXED
      checksum = ComputeIPChecksum( IPPacketIn, checksum );
      // the following does not apply to ACK segments which
      // consist only of TCP header and IP header (total 40 bytes)
      if(IPPacketIn.size() > 40 * BITS_PER_BYTE )
      {
        // TCP protocol has value 6  and it will contain the eight leading 0s
        checksum = checksum . xorAndShift ( TCPProtocolBitSet ) ;
        JNWBitSet lengthBitSet  
         = new JNWBitSet ( TCPSegment.getJNWLength ( ) / BITS_PER_BYTE , 16 )  ;
        checksum = checksum . xorAndShift ( lengthBitSet ) ;
      }
      // for ACK segments, just invert the checksum
      else checksum . xor(sixteenOnes);

      // Is the checksum all ones?
      return ( checksum.cardinality () ==  16 );
   }
   
   
   /**Returns a clone of the data from a TCP segment.*/
   public static JNWBitSet GetDataFromTCPSegment (JNWBitSet TCPSegmentIn)
   {
      JNWBitSet data = TCPSegmentIn . clone ( ) ;
      data . setJNWLength ( data . getJNWLength ( )
      - 20 * 8 /*TCP header length */) ;
      return data ;
   }  // end of getDataFromUDPPacket
   
   
   /**Returns the TCP sequence number from an Ethernet frame but NOT and HDLC frame.*/
   public static long GetTCPSequenceNumberFromFrame (JNWBitSet FrameIn)
   {
      int fromIndex = FrameIn . getJNWLength ( )
      - ( 8 /*Ethernet preamble length*/
         + 6 /*Ethernet destination address length*/
         + 6 /*Ethernet source address length*/
         + 2 /*Ethernet length field length*/
         + 20 /*IP header length*/
         + 8 /*TCP header location of sequence number*/) * BITS_PER_BYTE  ;
      JNWBitSet data = FrameIn.getSubset ( fromIndex, fromIndex + 32 ) ;
      long sequenceNumber = data . convertJNWBitSetToLong ( ) ;
      return sequenceNumber ;
   }  // end of getDataFromUDPPacket
   
   
   /**Returns the TCP sequence number from a TCP packet.*/
   public static long GetSequenceNumberFromTCPSegment (JNWBitSet TCPSegmentIn )
   {
      int fromIndex = TCPSegmentIn . getJNWLength ( )
      - 32 * ( 2 /*Location of sequence number in TCP header*/) ;
      return TCPSegmentIn . getSubset ( fromIndex , fromIndex + 32 )
      . convertJNWBitSetToLong ( ) ;
   }  // end of getDataFromUDPPacketket
   
   
   /**Returns the TCP sequence number from a IP packet.  The TCP packet in is
    * unchanged.*/
   public static long GetAcknowledgementNumberFromTCPSegment (JNWBitSet TCPSegmentIn )
   {
      int fromIndex = TCPSegmentIn . getJNWLength ( )
      - 32 * ( 3 /*Location of sequence number in TCP header*/) ;
      return TCPSegmentIn . getSubset ( fromIndex , fromIndex + 32 )
      . convertJNWBitSetToLong ( ) ;
   }  // end of getDataFromUDPPacket
   
   //public static long GetWindowSizeFromTCPSegment( JNWBitSet TCPPacketIn ) { // BZ

   //    return 10; //
   //}
    /**Returns the TCP sequence number from an IP packet.*/
   public static long GetTCPSequenceNumberFromIPPacket (JNWBitSet IPPacketIn)
   {
      int fromIndex = IPPacketIn . getJNWLength ( )
      - 32 * ( 5 /*Length of IP header*/
         + 2 /*Location of sequence number in TCP header*/) ;
      return IPPacketIn . getSubset ( fromIndex , fromIndex + 32 )
      . convertJNWBitSetToLong ( ) ;
   }  // end of getDataFromUDPPacket
   
   /**Returns the TCP sequence number from an IP packet.*/
   public static long GetTCPAcknowledgementNumberFromIPPacket (JNWBitSet IPPacketIn)
   {
      int fromIndex = IPPacketIn . getJNWLength ( )
      - 32 * ( 5 /*Length of IP header*/
         + 3 /*Location of sequence number in TCP header*/) ;
      return IPPacketIn . getSubset ( fromIndex , fromIndex + 32 )
      . convertJNWBitSetToLong ( ) ;
   }  // end of getDataFromUDPPacket


   /**Returns the TCP sequence number from an IP packet.*/
   public static long GetTCPAcknowledgementNumberFromTCPSegment (JNWBitSet TCPSegmentIn)
   {
      int fromIndex = TCPSegmentIn . getJNWLength ( )
      - 32 * ( 3 /*Location of sequence number in TCP header*/) ;
      return TCPSegmentIn . getSubset ( fromIndex , fromIndex + 32 )
      . convertJNWBitSetToLong ( ) ;
   }  // end of getDataFromUDPPacket
   
   
   /**Returns the number of bytes of data in the TCP/IP packet.  The data excludes
    * the IP header and the TCP header.*/
   public static int GetDataLengthInBytesFromTCPIPPacket (JNWBitSet IPPacketIn)
   {
      int length = IPPacketIn . getJNWLength ( )  ;
      return ( IPPacketIn . getJNWLength ( ) / BITS_PER_BYTE
         - ( 20 /*IP header length in bytes*/ + 20 /*TCP header length in bytes */)  ) ;
   }  // end of GetDataLengthInBytesFromTCPIPPacket
   
   
   /**Returns a clone of the data from a IP packet.  The data excludes the TCP header.*/
   public static JNWBitSet GetDataCloneFromTCP_IPPacket (JNWBitSet IPPacketIn)
   {
      JNWBitSet data = IPPacketIn . clone ( ) ;
      data . setJNWLength ( data . getJNWLength ( )
      - BITS_PER_BYTE * ( 20 /*IP header length */ + 20 /*TCP header length */) ) ;
      return data ;
   }  // end of getDataFromUDPPacket

    // Create SYN packet (BZ)
    public JNWBitSet ComposeSYNPacket(int sourceNodeNumber, 
                                      int destinationNodeNumber,
                                      boolean isSYNACK) {
        JNWBitSet SynHeader = ComposeSYNSegmentHeader(isSYNACK);
        SynHeader = ModelTransport.IPUtilities.PrependIPHeaderToMessage (
                        SynHeader ,
                        sourceNodeNumber ,
                        destinationNodeNumber ,
                        ModelData.GlobalVariables.transmissionProtocols.TCP);
        return SynHeader;
    }

    // Compose TCP header for a SYN packet (BZ)
    public JNWBitSet ComposeSYNSegmentHeader(boolean isSYNACK) {

        JNWBitSet synSegmentHeader = new JNWBitSet(0);

        synSegmentHeader.append(sourcePortBitSet);
        synSegmentHeader.append(destinationPortBitSet);

        // Don't generate a random initial sequence number = use 0 for this version
        //long initialSequenceNumber = randomNumberGenerator.nextInt(Integer.MAX_VALUE);
        //JNWBitSet sequenceNumberBitSet = new JNWBitSet(initialSequenceNumber, 32);
        JNWBitSet sequenceNumberBitSet = new JNWBitSet(0, 32);
        
        // The acknowledgement number of a SYN segment equals 0, correct?
        JNWBitSet acknowledgementNumberBitSet = new JNWBitSet(0, 32);

        synSegmentHeader.append(sequenceNumberBitSet);
        synSegmentHeader.append(acknowledgementNumberBitSet);

        JNWBitSet bitsAfterACKNUM = TCPHeaderLengthBitSet.clone( );
        bitsAfterACKNUM.append(sixZeros);
        if(isSYNACK) {
            bitsAfterACKNUM.append(TCPFlagBitsofSYNACK);
            // Set initial window size in a SYN ACK message
            JNWBitSet window_size = new JNWBitSet(initialSlidingWindowSize, 16);
            bitsAfterACKNUM.append(window_size); 
        }

        else {
            // TCPFlagBitsofSYN is the TCP flag bits for SYN segment
            bitsAfterACKNUM.append(TCPFlagBitsofSYN);

            // default window size = 20
            bitsAfterACKNUM.append(windowSizeDefaultValue); 
        }
        bitsAfterACKNUM.append(checksumDefaultValue);
        bitsAfterACKNUM.append(urgentPointerDefaultValue);
        synSegmentHeader.append(bitsAfterACKNUM);
        return synSegmentHeader;
    }

    // we send only header for FIN packet
    public JNWBitSet ComposeFINPacket(int sourceNodeNumber,
                                      int destinationNodeNumber,
                                      boolean isFINACK) { // BZ

        JNWBitSet FinHeader = ComposeFINSegmentHeader(isFINACK);

        FinHeader =
            ModelTransport.IPUtilities.PrependIPHeaderToMessage(
                FinHeader,
                sourceNodeNumber,
                destinationNodeNumber,
                ModelData.GlobalVariables.transmissionProtocols.TCP);

        return FinHeader;
    }

    public JNWBitSet ComposeFINSegmentHeader(boolean isFINACK) { // BZ

        JNWBitSet finSegmentHeader = new JNWBitSet(0);

        finSegmentHeader.append(sourcePortBitSet);
        finSegmentHeader.append(destinationPortBitSet);

        // Sequence number = 0 
        JNWBitSet sequenceNum = new JNWBitSet(0, 32);
        finSegmentHeader.append(sequenceNum);
        // Acknowledgement number = 0
        JNWBitSet ackNum = new JNWBitSet(0, 32);
        finSegmentHeader.append(ackNum);

        JNWBitSet bitsAfterACKNUM = TCPHeaderLengthBitSet.clone();
        bitsAfterACKNUM.append(sixZeros);

        if(isFINACK)
            bitsAfterACKNUM.append(TCPFlagBitsofFINACK);
        else
            bitsAfterACKNUM.append(TCPFlagBitsofFIN);
        
        bitsAfterACKNUM.append(windowSizeDefaultValue);
        bitsAfterACKNUM.append(checksumDefaultValue);
        bitsAfterACKNUM.append(urgentPointerDefaultValue);

        finSegmentHeader.append(bitsAfterACKNUM);
        finSegmentHeader.set(0 + 16 + 16); // ?

        return finSegmentHeader;
    }

    // Re-implementation of the function for composing ACKKNOWLEDGEMENT segment
    public JNWBitSet ComposeDataACKPacket(long ackno) { // BZ

        JNWBitSet dataACKSegmentHeader = ComposeDataACKSegmentHeader(ackno);

        dataACKSegmentHeader =
            ModelTransport.IPUtilities.PrependIPHeaderToMessage(
                dataACKSegmentHeader,
                sourceNodeNumber,
                destinationNodeNumber,
                ModelData.GlobalVariables.transmissionProtocols.TCP);

        return dataACKSegmentHeader;
    }

    // Re-implementation of the function for composing ACKKNOWLEDGEMENT Header
    public JNWBitSet ComposeDataACKSegmentHeader(long ackno) { // BZ

        JNWBitSet dataACKHeader = new JNWBitSet(0);

        dataACKHeader.append(sourcePortBitSet);
        dataACKHeader.append(destinationPortBitSet);

        // Sequence number = 0
        JNWBitSet sequenceNum = new JNWBitSet(0, 32);
        dataACKHeader.append(sequenceNum);
        // Acknowledgement number = 0
        JNWBitSet ackNum = new JNWBitSet(ackno, 32);
        dataACKHeader.append(ackNum);
        // HLEN = 5 (default value with no data)
        JNWBitSet hlen = new JNWBitSet(5, 4);
        dataACKHeader.append(hlen);
        // Reserved = 0
        JNWBitSet reserved = new JNWBitSet(0, 6);
        dataACKHeader.append(reserved);
        // Code bits = 010000
        dataACKHeader.append(TCPFlagBitsofACK);
        // Set Window
        JNWBitSet window = new JNWBitSet(TCPState.initialSlidingWindowSize, 16);
        dataACKHeader.append(window);
        // Checksum = 0 (bypass checking)
        JNWBitSet checksum = new JNWBitSet(0, 16);
        dataACKHeader.append(checksum);
        // Urgent pointer = 0
        JNWBitSet urgPoint = new JNWBitSet(0, 16);
        dataACKHeader.append(urgPoint);

        return dataACKHeader;
    }

    /**Composes the IP packet containing the ACK reply message indicating
     * that the bytes up to and including byte acknowledgementNumber - 1
     * have been successfully received; no data in this segment*/
    public JNWBitSet ComposeACKPacket( JNWBitSet IPPacketIn ) {
        // obtain the ACK Segment Header from student-written method
        JNWBitSet TCPSegmentIn = CloneTCPSegmentFromIPPacket(IPPacketIn);
        JNWBitSet ackSegmentHeader = ComposeACKSegmentHeader(TCPSegmentIn);
        
        // extract the source address from destination of the incoming packet
        int fromIndex = IPPacketIn . getJNWLength( ) - 4 * 32 ;
        JNWBitSet sourceAddressNodeNumberBitSet
                =  IPPacketIn .getSubset( fromIndex , fromIndex + 32 ) ;
        long replyDestinationNodeNumber
                = sourceAddressNodeNumberBitSet . convertJNWBitSetToLong( ) ;
        
        // extract the destination address from source of the incoming packet
        int toIndex = IPPacketIn . getJNWLength( ) - 5 * 32 ;
        JNWBitSet destinationAddressNodeNumberBitSet
                =  IPPacketIn .getSubset( toIndex , toIndex + 32 ) ;
        long replySourceNodeNumber
                = destinationAddressNodeNumberBitSet . convertJNWBitSetToLong( ) ;
     
        // use this to create an IP packet header with source and destination
        JNWBitSet IPPacketOut
                = ModelTransport . IPUtilities . ConstructIPHeader(
                ackSegmentHeader,
                replySourceNodeNumber ,
                replyDestinationNodeNumber,
                ModelData.GlobalVariables.transmissionProtocols . TCP ) ;
     
        // append the segment header to make ACK packet
        IPPacketOut.append(ackSegmentHeader);
            
        // calculate the checksum and replace it in the TCP header
        // INCLUDING IP ADDRESSES IS A HACK THAT DOES NOT FOLLOW RFC AND MUST BE FIXED
        JNWBitSet messageXOR = ComputeTCPXOR ( ackSegmentHeader );
        messageXOR = ComputeIPChecksum( IPPacketIn, messageXOR );
        IPPacketOut.setSubset ( IPPacketOut . getJNWLength ()  - 38 * BITS_PER_BYTE
            , messageXOR ) ;
        
        return IPPacketOut  ;
    }  // end of ComposeACKPacket

    /** Composes the TCP header for an ACK packet */
    public JNWBitSet ComposeACKSegmentHeader(JNWBitSet TCPSegmentIn) {
        
        JNWBitSet ackSegmentHeader = new JNWBitSet(0);
        
        acknowledgementNumber = TCPDestinationBuffer.getAcknowledgementNumber ();
        
        // student writes this to compose the
        // 20-byte TCP ACK segment header by appending bits
        
        // Source port and destination port = 023
        // Sequence number = 0 (the sequence number of the ACK segment is ignored)
        // Acknowledgement number (see assignment)
        // HLEN = 5 (default value with no data)
        // Reserved = 0
        // Code bits = 0
        // Window = 0
        // Checksum = 0 (bypass checking)
        // Urgent pointer = 0
        // Source port and destination port = 023
        // Source port and destination port = 023
        JNWBitSet sourcePort = new JNWBitSet(023, 16);
        ackSegmentHeader.append(sourcePort);
        JNWBitSet destinationPort = new JNWBitSet(023, 16);
        ackSegmentHeader.append(destinationPort);
        
        // Sequence number = 0 (the sequence number of the ACK segment is ignored)
        JNWBitSet sequenceNum = new JNWBitSet(0, 32);
        ackSegmentHeader.append(sequenceNum);
        
        // Acknowledgement number (see assignment)
        JNWBitSet ackNum = new JNWBitSet(acknowledgementNumber, 32);
        ackSegmentHeader.append(ackNum);
        
        // HLEN = 5 (default value with no data)
        JNWBitSet hlen = new JNWBitSet(5, 4);
        ackSegmentHeader.append(hlen);
        
        // Reserved = 0
        JNWBitSet reserved = new JNWBitSet(0, 6);
        ackSegmentHeader.append(reserved);
        
        // Code bits = 010000
        JNWBitSet codeBits = new JNWBitSet(16, 6);
        ackSegmentHeader.append(codeBits);
        
        // Window = 0
        JNWBitSet window = new JNWBitSet(0, 16);
        ackSegmentHeader.append(window);
        
        // Checksum = 0 (bypass checking)
        JNWBitSet checksum = new JNWBitSet(0, 16);
        ackSegmentHeader.append(checksum);
        
        // Urgent pointer = 0
        JNWBitSet urgPoint = new JNWBitSet(0, 16);
        ackSegmentHeader.append(urgPoint);

        return ackSegmentHeader;        
    } // end of ComposeACKSegmentHeader
    
   /**Returns the TCP window size.*/
   public static int GetTCPWindowSizeFromTCPSegment ( JNWBitSet TCPSegmentIn )
   {
      int fromIndex = TCPSegmentIn .getJNWLength ( ) - 4 * 32 ;
      return ( int ) TCPSegmentIn . getSubset ( fromIndex , fromIndex + 16 )
      . convertJNWBitSetToLong ( ) ;
   }  // end of GetTCPWindowSizeFromTCPSegment
   
   
   /**Returns the FIN bit from the TCP header.  The input is an IP packet, not
    *a TCP packet.*/
   public static boolean GetTCPFINBitFromTCPSegment ( JNWBitSet TCPSegmentIn )
   {
      return TCPSegmentIn.get ( TCPSegmentIn . getJNWLength ()
      - 7 * 16 /*position of FIN bit in TCP header*/ ) ; // see "http://www.freesoft.org/CIE/Course/Section4/8.htm" (BZ)
                                                         // Also, see page 36 of the tutorial for bits ordering in JNWBitSet (BZ)
   }  // end of GetTCPFINBitFromTCPSegment
   
   public static boolean GetTCPACKBitFromTCPSegment(JNWBitSet TCPPacketIn) { // BZ

       return TCPPacketIn.get(TCPPacketIn.getJNWLength() - 7 * 16 + 4); // The position of ACK bit
   } // end GetTCPACKBitFromTCPSSegment

   public static boolean GetTCPSYNBitFromTCPSegment(JNWBitSet TCPPacketIn) { // BZ

       return TCPPacketIn.get(TCPPacketIn.getJNWLength() - 7 * 16 + 1); // The position of SYN bit
   }// end GetTCPSYNBitFromTCPSegment

   /**Returns the FIN bit from the IP packet.  The input is an IP packet, not
    *a TCP segment.*/
   public static boolean GetTCPFINBitFromIPPacket ( JNWBitSet IPPacketIn )
   {
      return IPPacketIn . get ( IPPacketIn . getJNWLength ()
      - 5 * 32 /*IP header length*/
         - 7 * 16 /*position of FIN bit in TCP header*/ ) ;
   }  // end of GetTCPFINBitFromIPPacket
   
   // BZ
   public static JNWBitSet GetTCPControlBitsFromIPPacket(JNWBitSet IPPacketIn) {

       int fromIndex = IPPacketIn.getJNWLength() - 5 * 32 - 7 * 16;
       JNWBitSet controlbits = IPPacketIn.getSubset(fromIndex, fromIndex + 6);

       return controlbits;
   }

   public static boolean IsPacketAnSYNMessage(JNWBitSet IPPacketIn) {

       return GetTCPSYNBitFromTCPSegment(ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(IPPacketIn)) &&
              !GetTCPACKBitFromTCPSegment(ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(IPPacketIn));
   }// end IsPacketAnSYNMessage

   public static boolean IsPacketAnSYNACKMessage(JNWBitSet IPPacketIn) {

       return GetTCPSYNBitFromTCPSegment(ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(IPPacketIn)) &&
              GetTCPACKBitFromTCPSegment(ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(IPPacketIn));
   }// end IsPacketAnSYNACKMessage

   public static boolean IsPacketAnFINACKMessage(JNWBitSet IPPacketIn) {

       return GetTCPFINBitFromTCPSegment(ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(IPPacketIn)) &&
              GetTCPACKBitFromTCPSegment(ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(IPPacketIn)) ;
   }

   public static boolean IsPacketAnFINMessage(JNWBitSet IPPacketIn) {

       return GetTCPFINBitFromTCPSegment(ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(IPPacketIn));
   }

   public static boolean IsPacketAnACKMessage(JNWBitSet IPPacketIn) {
           
       return GetTCPACKBitFromTCPSegment(ModelTransport.TCPUtilities.CloneTCPSegmentFromIPPacket(IPPacketIn)) ;
   }// end IsPacketAnACKMessage
   
   /**A main method for testing the JNWBitSet.
    * @param args the command line arguments*/
   public static void main ( String[] args )
   {

   } // end of main
   
}  // end of class TCPUtilities

