/*
 * UDPUtilities.java
 *
 * Created on July 6, 2007, 2:15 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package ModelTransport;

import Utility.JNWBitSet ;


/**Utilities for the creation of UDP packets.
 *  *<p>Versions
 *<li><b>Version 1.001</b> of 20071116 changes JNWBitSet
 * AddUPDHeaderToMessage ( JNWBitSet messageIn  ) to
 * AddUPDHeaderToMessage ( JNWBitSet messageIn , int
 * sourceNodeNumberIn , int destinationNodeNumberIn )
 * @see <a href="http://www.freesoft.org/CIE/RFC/768/index.htm">User Datagram Protocol</a>
 * @see <a href="http://www.faqs.org/rfcs/rfc768.html">RFC 768</a>
 * @version 1.100
 * @author dixonw*/
public class UDPUtilities extends UDPState
{
   
   /**Source node address*/
   static long sourceNodeAddress  ;
   
   /**Destination node address*/
   static long destinationNodeAddress  ;
   
   /** Creates a new instance of UDPUtilities */
   public UDPUtilities ()
   {
   }
   
   /**Constructs a UDP UDPHeader for a message.  The input is a JNWBitSet.  The
    * output is the UDP packet with the UDP header appended.
    * @see <a href="http://www.freesoft.org/CIE/RFC/768/index.htm">User Datagram Protocol</a>
    * @see <a href="http://www.faqs.org/rfcs/rfc768.html">RFC 768</a>
    */
   public static JNWBitSet PrependUDPHeaderToData ( JNWBitSet dataIn
      , int sourceNodeNumberIn , int destinationNodeNumberIn  )
   { // set the source node number and the destinationn node number
      sourceNodeAddress = sourceNodeNumberIn ;
      destinationNodeAddress = destinationNodeNumberIn ;
      // Clone the bit set
      JNWBitSet dataClone =  dataIn . clone (  )  ;
      // Now construct the UDPHeader
      // Create the source port bit set
      JNWBitSet sourcePortBitSet = new JNWBitSet ( UDPSourcePort , 16  ) ;
      // Create the destination port bit set
      JNWBitSet destinationPortBitSet = new JNWBitSet ( UDPDestinationPort , 16 ) ;
      // Now include the eight bytes for the UDPHeader in the length
      JNWBitSet UDPLengthBitSet = new JNWBitSet ( dataClone . getJNWLength ( ) / 8
         + /* UDP header length in bytes*/ 8 , 16 ) ;
      // Create a checksum of all zeros
      JNWBitSet UDPChecksum = new JNWBitSet ( 16 ) ;
      // Create a clone message with the UPD header, except the
      // checksum is all zeros
      // Load all zeros for a dummy checksum
      dataClone . prepend ( UDPChecksum ) ;
      dataClone . prepend ( UDPLengthBitSet) ;
      dataClone . prepend ( destinationPortBitSet ) ;
      dataClone . prepend ( sourcePortBitSet ) ;
      // Start to compute the real checksum
      // Start to compute the real checksum
      // First compute the XOR over the data and UDP header
      // with the dummy checksum
      UDPChecksum = dataClone .xorAndShift ( UDPChecksum ) ;
      JNWBitSet sourceNodeBitSet = new JNWBitSet ( sourceNodeAddress , 32 ) ;
      UDPChecksum = sourceNodeBitSet . xorAndShift ( UDPChecksum ) ;
      // Compute the checksum over the source node address
      JNWBitSet destinationNodeBitSet
         = new JNWBitSet ( destinationNodeAddress , 32 ) ;
      UDPChecksum = destinationNodeBitSet . xorAndShift ( UDPChecksum ) ;
      // Add the UDP protocol number 17 to the pseudoheader
      // The first 8 bits of this expansion are zeros as required
      // by the UDP pseudoheader checksum
      JNWBitSet seventeen  = new JNWBitSet ( 17 , 16 ) ;
      UDPChecksum = seventeen . xorAndShift ( UDPChecksum ) ;
      UDPChecksum = UDPLengthBitSet . xorAndShift ( UDPChecksum ) ;
      // Invert the checksum
      UDPChecksum . invertBitSet (  );
      // Put the real checksum inot the UDP header
      dataClone . setSubset ( dataClone . getJNWLength ( )
      - 8 * 8 /* position of the checksum in the UDP header*/ , UDPChecksum ) ;
 ////     JNWBitSet UDPHeader  = dataClone.getSubset ( dataClone . getJNWLength ( ) - 8 * 8  , dataClone . getJNWLength ( ) ) ;
////System.out.println ( "UDPHeader =  " + UDPHeader . toString() )  ;
      return dataClone ;
   }  // end of ConstructUDPHeader ( )
   
   
   /**Verifies the TCP checksum at the receiver.*/
   public static boolean VerifyUDPXORFromIPPacket ( JNWBitSet IPPacket )
   {  // Get the UDP packet
      JNWBitSet UDPPacket = GetUDPPacketFromUDP_IPPacket ( IPPacket )  ;
      // Start the checksum at all zeros
      JNWBitSet checksum = new JNWBitSet ( 16 ) ;
      // Compute the checksum over the UDP packet
      checksum = UDPPacket.xorAndShift ( checksum ) ;
      // Get the source node bit set from the IP packet
      int fromIndex = IPPacket.getJNWLength ()
      - 4 * 32 /*Position of the source address in the IP header */ ;
      JNWBitSet sourceNodeAddressBitSet
         = IPPacket  . getSubset ( fromIndex , fromIndex + 32 ) ;
      checksum = sourceNodeAddressBitSet.xorAndShift ( checksum ) ;
      // Get the destination node bit set from the IP packet
      JNWBitSet destinationNodeAddressBitSet
         = IPPacket . getSubset ( fromIndex - 32 , fromIndex ) ;
      checksum = destinationNodeAddressBitSet.xorAndShift ( checksum ) ;
      // UDP protocol has value 17  and it will contain the eight leading 0s
      JNWBitSet seventeen = new JNWBitSet ( 17 , 16 ) ;
      checksum.xor ( seventeen ) ;
      fromIndex = UDPPacket . getJNWLength () - 3 * 16 ;
      JNWBitSet lengthBitSet  = UDPPacket . getSubset ( fromIndex , fromIndex + 16 ) ;
      checksum.xor (lengthBitSet ) ;
      if ( checksum.cardinality () !=  16 )
      {  // If all the bits are not set then there is an error
         return false ;
      }
      else
      { return true ; }
   }
   
   
   /**Returns a clone of the UDP packet from the IP packet.  The returned
    *value is a clone of the original so that the returned packet
    *can be changed without changing the original.*/
   public static JNWBitSet GetUDPPacketFromUDP_IPPacket ( JNWBitSet IPPacketIn )
   {  // Make clone
      JNWBitSet clone = IPPacketIn  . clone ( ) ;
      // Remove the 20 bytes of the IP header
      clone . setJNWLength ( IPPacketIn .getJNWLength ( ) - 20 * 8 ) ;
      return clone ;
   }
   
   
   /**Returns a clone of the data from a UDP/IP packet.  Clones the packet
    *before returning the data.  The data does not include either the IP
    header of the UDP header.*/
   public static JNWBitSet GetDataFromUDP_IPPacket (JNWBitSet UDP_IPPacketIn )
   {  // Make clone
      JNWBitSet data = UDP_IPPacketIn . clone ( ) ;
      data . setJNWLength ( data . getJNWLength ( )
      - 32 * ( 5 /*IP header length */ + 2 /*UDP header length */) ) ;
      return data ;
   }  // end of getDataFromUDPPacket
 
 
   /**A main method for testing the JNWBitSet.
    * @param args the command line arguments*/
   public static void main_Testing_Only ( String[] args )
   {
      System.out.println ( " Starting UPD " ) ;
      // obtain input data
      
      String inputString = "" ;
      JNWBitSet result = new JNWBitSet ( );
////        while ( ! inputString .equals ( "END" ) ) {
      JNWBitSet dataBits = new Utility.JNWBitSet ( ) ;
/*
      try
      {
         java.io.BufferedReader bufferedReader = new
            java.io.BufferedReader ( new java.io.InputStreamReader (
            System .in ) ) ;
         inputString = bufferedReader . readLine ( ) ;
         dataBits.setJNWBitSetToString ( inputString ) ;
      }
      catch ( Exception e )
      {
         System.err.println ( " ERROR IN MainGUI_Actions.ReadTopologyFileAction.actionPerformed:"
            + "\nException "+ e. getClass ( ) . getName ( ) + " thrown while trying to "
            + " read line \"" + inputString + "\" in input topology file"
            + "\n    Data should be in format\n\"SOURCE NODE , DESTINATION NODE , "
            + "DATA RATE, BIT ERROR RATE , TRANSMISSION DELAY\"" ) ;
      }
 */
      JNWBitSet inputStringBitSet = new JNWBitSet (  ) ;
      inputStringBitSet . setJNWBitSetToString ( inputString  ) ;
      System.out.println ( " inputString = " + inputStringBitSet . toString ( )  ) ;
      if ( ( ( inputStringBitSet.getJNWLength ( )  % 16 ) != 0 ) )
      {
         // Append additional bits
////System.out.println ("IN ReadEMailTextFilesAndPutIntoSourceNodeBufferAction() before padding message.  Message length = "
////        + eMailBitSet.getJNWLength ( ) )  ;
         inputStringBitSet = inputStringBitSet.shift ( 16 - ( inputStringBitSet.getJNWLength ( )  % 16 ) ) ;
         System.out.println ( " Adjusted inputString = " + inputStringBitSet . toString ( )  ) ; ////System.out.println ("IN ReadEMailTextFilesAndPutIntoSourceNodeBufferAction() after padding message.  Message length = "
////        + eMailBitSet.getJNWLength ( ) )  ;
      }
      
      
////        ModelDatas . GlobalVariables . eMailMessage  =  inputStringBitSet ;
////        ModelDatas . GlobalVariables .sourceNodeAddress = 32 ;
      ////       ModelDatas . GlobalVariables . destinationNodeAddress = 64 ;
      System.out.println ( " calling AddUPDHeaderToMessage " ) ;
      result = PrependUDPHeaderToData ( inputStringBitSet  , 1024  , 316 ) ;
      
      System.out.println ( " result message with UDP header = " + result.toString ( )  ) ;
      // Pretend this is a UDP/IP packet
      result . setJNWLength ( result.getJNWLength () + 8 * 20 ) ;
      ////     System.out.println ( " shifter message message with UDP header = " + result.toString ( )  ) ;
      JNWBitSet bitSet = GetDataFromUDP_IPPacket ( result ) ;
      String string = bitSet . convertJNWBitSetToString ( ) ;
      System.out.println ( " Resulting message  = " + string  ) ;
      result . setJNWLength ( result.getJNWLength () - 8 * 20 ) ;
      System.out.println ( "\n\n" ) ;
      result = ModelTransport.IPUtilities.PrependIPHeaderToMessage ( result , 1024  , 316
         , ModelData.GlobalVariables.transmissionProtocols.UDP) ;
      System.out.println ( " Resulting message with IP header = " + result.toString ( )  ) ;
      bitSet = GetDataFromUDP_IPPacket ( result ) ;
      string = bitSet .convertJNWBitSetToString ( ) ;
      System.out.println ( " Resulting message  = " + string  ) ;
      ////     result . flip ( 10 ) ;
////       }
////System.out.println ( " The " + i + "th message = " + result.toString()  ) ;
      boolean passedTest  = VerifyUDPXORFromIPPacket ( result  ) ;
      System.out.println ( " With no bits flipped, the IP packet is correct = " + passedTest   ) ;
      //result.getJNWLength ( )
      for ( int j = 0 ; j <  result.getJNWLength ( )  ; j++ )
      {
         result  . flip ( j ) ;
///////////////System.out.println ( "\n\n\n\n" ) ;
////       result [ i ] . flip ( result [ i ] .getJNWLength ( ) - 12 * 8 - 1 ) ;
         ////       result  . flip ( 0 ) ;
         passedTest  = VerifyUDPXORFromIPPacket ( result  ) ;
         System.out.println ( " With the " + j + "th bit flipped, the IP packet is correct = " + passedTest   ) ;
         result . flip ( j ) ;
      }
      
      ////    for ( int j = 0 ; j < result.getJNWLength () ; j++ )
      ////    {
      ////       result  . flip ( j ) ;
///////////////System.out.println ( "\n\n\n\n" ) ;
////       result [ i ] . flip ( result [ i ] .getJNWLength ( ) - 12 * 8 - 1 ) ;
////        result [ i ] . flip ( 0 ) ;
////         boolean passedTest  = ModelTransport.IPUtilities.VerifyIHeaderChecksum ( result  ) ;
////         System.out.println ( " With the " + j + "th bit flipped, the UDP header is correct = " + passedTest   ) ;
      ////        result . flip ( j ) ;
      ////     }
   } // end of main
   
   
}  // end of class UDPUtilities
