/*
 * JNWBitSet.java
 *
 * Created on January 6, 2007, 10:01 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package Utility;

/**
 * JNWBitSet is a subclass of java.util.BitSet designed to handle
 * Vectors of bits.  Unfortunately, the length method in
 * java.util.BitSet returns the location of the first 1 in the
 * BitSet.  For example, the length of "00000000 00001000" is 4,
 * but the JNWLength is 16.  We need to know the "real" length
 * including preceding zeros.  Therefore, this package and these
 * methods were necessary.</p>
 * <p>The methods include the ability to initialize BitSets to
 * ASCII strings and to strings of 0s and 1s.</p>
 * <p>JNWBitSet implements interfaces Utility.Frame and Utility.Packet.
 *   Therefore, if one wishes, the programmer can instantiate
 * Utility.Frame and Utility.Packet to clarify the difference
 * between a frame and a packet.  The performance is identical.</p>
 *
 * *<p><b>Versions:</b>
 * <li><b>Version 1.005</b> of 20071116 adds the constructor
 * JNWBitSet( long inputIn , int numberOfBits ) and the methods
 * public int compare ( JNWBitSet bitSetForComparison ),
 * public JNWBitSet xorBitSet  ( JNWBitSet bitSetIn ), and
 * public String convertJNWBitSetToHEXString ( ).
 * Method public long convertJNWBitSetToLong ( ) was modified
 * for faster performance.</li>
 * <li><b>Version 1.005</b> of 20071119 modifies the compare ( )
 * method to greatly increase the speed by using length versus
 * JNWLength.</li>
 * <li><b>Version 1.006</b> of 20071127 replaces the CRC computations,
 * which previously used a shift register, with much faster methods
 * that perform polynomial divisions, without the shifts.  These new
 * versions are 500 to 700 times faster.
 * </li>
 * <li><b>Version 1.101</b> of 20071215 added method
 * removeNullCharactersAtEnd. </li>
 * *<li><b>Version 1.102</b> of 20080304 found error in getBitValue
 * Changed
 *       if ( bitSetIndex < 0 | bitSetIndex >= 63 )
 *      {  Utility.err.println (
 *            "ERROR in Utility.JNWBitSet.getBitValue: "
 *            + " bitSetIndex = " + bitSetIndex  ) ; }
 * to
 *       if ( bitSetIndex < 0 | bitSetIndex >= JNWLength )
 *      {  Utility.err.println (
 *            "ERROR in Utility.JNWBitSet.getBitValue: "
 *            + " bitSetIndex = " + bitSetIndex  ) ;
 *         return -1 ; } </li>
 * <li><b>Version 1.103</b> of 20080609 implements Utility.Frame
 * and Utility.Packet</li>
 * <li><b>Version 1.200</b> of 20080919 changed toString method to ensure
 * all the columns align correctly.</li>
 * @author dixonw
 * @version 1.200
 */
public class JNWBitSet extends java.util.BitSet
   implements GUIConstants.General
////, Utility.Frame, Utility.Packet
{
   /**The length of the bit set.  The bits in JNWBitSet are numbered 01 from the
    * least significant bit, at the right, to JNWLength - 1, the most significant bit
    * at the left.
    * <p>The default length method of java.util.BitSet
    * returns the location of the highest 1.  The bits are numbered 0 to
    * length - 1.  If a bit set to a String of 16 0s, the default length method
    * will return a length of 0.  This length method will return a length of
    * 16 by updating the "length" correctly.  JNWLength must be set in methods
    * setToZerosAndOnes and setToString.  JNWLength is changed in method shift ( ) ,
    * set ( ), setToZerosAndOnes ( ) , and  setToString ( String stringOfCharacters )*/
   private int JNWLength  ;


   /**Creates a new instance of JNWBitSet with no initialization.*/
   public JNWBitSet ( )
   {
      super ( ) ;
      JNWLength = 0 ;
   }

   /**A copy constructor.  This creates a clone of a JNWBitSet.  It copies
    * the length and then is copies the bits.  This constructor is needed
    * because the clone method does not work for java.util.BitSet.
    * <p>Later the clone ( ) method was implemented for Utility.JNWBitSet.*/
   public JNWBitSet ( JNWBitSet JNWBitSetIn  )
   {
      this . JNWLength = JNWBitSetIn . getJNWLength () ;
      // Make sure there is enough room for the JNWBitSet
      if ( JNWLength > 0 )
      {  // Is the JNWBitSet NOT of length 0
         this . set ( JNWLength - 1 , true ) ;
         this . set ( JNWLength - 1 , false ) ;
         for ( int i = 0 ; i < this . JNWLength ; i++ )
         {
            this . set ( i , JNWBitSetIn .get ( i )  ) ;
         }  // end of for ( int i = 0 ; i < this . JNWLength ; i++ )
      }  // end of if ( JNWLength > 0 )
   }  // end of copy constructor


   /**Creates a bit set whose initial size is large enough to explicitly
    * represent nbits with subscripts in the range 0 through nbits - 1.  The
    * JNWBitSet is initialized to all zeros.*/
   public JNWBitSet (  int nbits )
   {
      super ( nbits  ) ;
      JNWLength = nbits ;
      // Set and unset the bit to ensure that the superclass actually
      // allocates the memory.  The superclass will NOT allocate
      // additional memory if the bit value is NOT set to true.
      if ( nbits == 0 )
      { return ; }
      this. set ( nbits - 1 , true );
      this. set ( nbits - 1 , false );
   } // end of constructor JNWBitSet(  int nbits )


   /**Creates a new instance of JNWBitSet initialized to a number, inputIn,
    * which is then truncated the requested number of bits, numberOfBits.  This
    * does the same thing as
    * <li><t> JNWBitSet data = new JNWBitSet ( ) ;</li>
    * <li><t> data . setJNWBitSetToLongAndTruncate ( inputIn , numberOfBits ) ;</li>
    */
   public JNWBitSet ( long inputIn , int numberOfBits )
   {
      super ( ) ;
      long number = inputIn  ;
      int bitLocation = 0 ;
      // Allocate a minimum of 64 bits for the result
      this . clear ( ) ;
      if ( inputIn < 0 )
      {
         Utility.err.println ( "ERROR in Utility.JNWBitSet: the number "
            + inputIn + " is negative which is prohibited." ) ;
         return ;
      }
      do
      {
         if ( number % 2 == 1 )
         {
            this . set ( bitLocation , true ) ;
         }
         number = number / 2 ;
         bitLocation = bitLocation + 1 ;
      } while ( number > 0 ) ;
      if ( bitLocation > numberOfBits )
      {
         Utility.err.println ( "ERROR in Utility.JNWBitSet constructor: the number "
            + inputIn + " cannot be represented in " + numberOfBits
            + " bits." ) ;
      }
      // Truncate the result to numberOfBits bits
      this . setJNWLength (  numberOfBits ) ;
   }


   /**Creates a new instance of JNWBitSet initialized to a string of
    * zeros and ones.  Spaces are permitted in the string of zeros and ones,
    *  but no other characters are permitted.*/
   public JNWBitSet ( String stringOfZerosAndOnes )
   {
      super ( ) ;
      this . setJNWBitSetToZerosAndOnes  ( stringOfZerosAndOnes ) ;
   }


   /**Returns the length of the BitSet including any leading zeros.  This is
    * necessary since the default length method  of BitSet returns the
    * location of the first 1 in the bit set.  If a bit set to a String
    * of 16 0s, the default length method would return a length of 0.  This
    * length method will return a length of 16.*/
   public int getJNWLength (  )
   {
      return this . JNWLength  ;
   }

    public String getSubset(int i) {
        throw new UnsupportedOperationException("Not yet implemented");
    }


   /**Sets the JNWBitSet to a length i.  None of the values of the bit
    * set are changed for index < i-1.  If the lenght is increased the i - 1
    * position of the bit set is set to ture and then to false.  This forces the
    superclass java.util.BitSet to allocate the necessary memory.  java.util.BitSet
    will not allocate memory unless at least one of the bits is 1.*/
   public void setJNWLength ( int i  )
   {
      if ( i <= JNWLength )
      {
         JNWLength = i ;
      }
      else
      {
         //Make sure there is SPACE by setting the ith bit to true
         this . set ( i - 1 ) ;
         //Reset the ith bit to false
         this . set ( i - 1 , false ) ;
      }  // end of else
   }  // end of public void setJNWLength ( int i  )


   /**Sets the bit at the specified index to true.  Overrides the default
    * set method to ensure the JNWLength is correct.  bitSetIndex >= 0.
    * <p><b>set(int bitSetIndex) can change the JNWLength</b>*/
   public void set (int bitSetIndex)
   {
      if ( bitSetIndex < 0 )
      {
         Utility.err.println ( "ERROR in Utility.JNWBitSet.set; bitSetIndex = "
            + bitSetIndex + " < 0. ") ;
      }
      super . set ( bitSetIndex  ) ;
      // Update the JNWLength.  JNWLength >= bitSetIndex + 1  since
      // the BitSet begins at index 0
      JNWLength = java.lang.Math.max ( JNWLength , bitSetIndex + 1 ) ;
   }


   /**Sets the value of the bitSetIndexth bit as a 0 or 1, not
    * a true or false.  JNWLength - 1 >= bitSetIndex >= 0.
    * <p><b>setBitValue ( int bitSetIndex , int value )
    * can change the JNWLength</b>*/
   public void setBitValue (int bitSetIndex , int value )
   {
      if ( bitSetIndex < 0 )
      {
         Utility.err.println ( "ERROR in Utility.JNWBitSet.setBitValue; bitSetIndex = "
            + bitSetIndex + " < 0. ") ;
      }
      if ( value == 0 | value == 1 )
      {
         // Set the bitSetIndex to make sure it is available
         super . set ( bitSetIndex  ) ;
         if ( value == 0 )
         {
            super . set ( bitSetIndex , false ) ;
         }
         // Update the JNWLength.  JNWLength >= bitSetIndex + 1  since
         // the BitSet begins at index 0
         JNWLength = java.lang.Math.max ( JNWLength , bitSetIndex + 1 ) ;
      }  // end of if ( value == 0 | value == 1 )
      else
      {
         Utility.err.println ("ERROR: in Utility.JNWBitSet.setBitValue, value = " + value ) ;
      }
   }


   /**Sets the bit at the specified index to the specified value.  Overrides
    * the default set method to ensure the length is correct.  If the
    * value is false and the bitIndex is NOT in the range of the existing
    * bit set, java.util.BitSet will <b>NOT</b> automatically expand its
    * size to include the value at bitIndex.  Therefore, I first force the
    * bit set to expand in size by setting the bitIndex value to true, and
    * then it is set to the correct boolean value.  bitSetIndex >= 0.
    * <p><b>set(int bitSetIndex, boolean value) changes the JNWLength</b>*/
   public void set ( int bitSetIndex, boolean value )
   {  // First set the bit to force the BitSet to expand in size if needed.
      if ( bitSetIndex < 0 )
      {
         Utility.err.println ( "ERROR in Utility.JNWBitSet.set; bitSetIndex = "
            + bitSetIndex + " < 0. ") ;
      }
      super . set ( bitSetIndex  , true ) ;
      super . set ( bitSetIndex  , value) ;
      // Update the JNWLength.  JNWLength >= bitSetIndex + 1  since
      // the BitSet begins at index 0
      JNWLength = java.lang.Math.max ( JNWLength , bitSetIndex + 1 ) ;
   }  // end of set


   /**Sets all of the bits in this BitSet to false.  Overrides the default
    * set method to ensure the length is correct.  This method is necessary
    * since the default clear method clears all the bits for
    * 0<= index < BitSet.length.  But BitSet.length is not the
    * same as JNWBitSet.JNWLength.
    * <b>JNWLength is unchanged.</b>*/
   public void clear ( )
   {  // Note that  0 <=  index <= JNWLength - 1
      // which is the numbering for the superclass java.util.BitSet.
      for ( int i = 0 ; i < JNWLength ; i++ )
      {
         this . set ( i , false ) ;
      }
   }


   /**Shifts the bits within a JNWBitSet.  Returns a JNWBitSet equal to
    * the  original with all the bits logically shifted j positions to the
    * higher bit (to the left) if j is positive, and shifted j positions
    * to the lower bit (to the right) if j is negative.  Zeros are
    * padded into the lowest bit positions for left shifts.  Bits
    * just drop off the right end for right shifts.
    * <p><b>The length of the JNWBitSet is changed by a shift.</b>*/
   public JNWBitSet shift ( int shift )
   {
      // The original length of this JNWBitSet-Be careful, this may be
      // changed  during this operation
      int lengthAtStartOfShift = JNWLength ;
      if ( shift == 0 )
      { return this ; }  // No shift
      // If this is a right shift larger than the length of the bit set,
      //clear the bit set
      if ( - shift  >= lengthAtStartOfShift )
      {
         this . clear ( ) ;  // Return the empty bitSet
         JNWLength = 0 ;
         return this ;
      }
      if ( shift > 0 )
      {  // Shift to the higher bit
         for ( int i = lengthAtStartOfShift - 1 ; i >= 0 ; i-- )
         {
            // If needed, JNWLength is automatically changed in set
            this . set ( i + shift  , this . get ( i  ) ) ;
         } // end of for ( int i = this. size ( ) - 1 ; i >= j ; i-- )
         // Load zeros into the lowest bits
         for ( int i = shift - 1 ; i >= 0 ; i-- )
         {
            this . set ( i , false ) ;
         } // end of for ( int i = j - 1 ; i >= 0 ; i-- )
      }  // end of if ( j > 0 )
      else
      {  // Shift to the lower bit
         for ( int i = - shift ; i <= lengthAtStartOfShift - shift - 1  ; i++ )
         {
            this . set ( i + shift  ,  this . get ( i ) ) ;
         } // end of for
         // the JNWLength must be updated here for left shifts
         this.JNWLength = this.JNWLength + shift  ;
      }  // end of else
      return this ;
   }  // end of shift ( int shift )


   /**Sets a BitSet to a String of 0s and 1s.  Blank spaces are acceptable,
    * and ignored, in the String.
    * <p><b>The length of the JNWBitSet is changed by setToZerosAndOnes.</b>*/
   public void setJNWBitSetToZerosAndOnes ( String stringOfZerosAndOnes )
   {
      this . clear ( ) ;
      int bitLocation = 0 ;
      for ( int i = stringOfZerosAndOnes . length ( ) - 1 ; i >= 0  ; i-- )
      {
         char zeroOrOne = stringOfZerosAndOnes.charAt ( i );
         if ( zeroOrOne == '0')
         {
            this . set ( bitLocation , false ) ;
            bitLocation = bitLocation + 1 ;
         }
         else if ( zeroOrOne == '1')
         {
            this . set ( bitLocation , true ) ;
            bitLocation = bitLocation + 1 ;
         }
         else if ( zeroOrOne == ' ')
         {  // Ignore spaces
         }
         else Utility.err.println ( "ERROR in Utilities.JMWBitSet.set "
            + "( String ): input string " + stringOfZerosAndOnes + " contains "
            + zeroOrOne + "\" neither a 0 nor a 1 nor a space" ) ;
      }  // end of for ( int i = length ; i >= 0 ; i-- )
      JNWLength = bitLocation  ;
   }  // end of set


   /**Sets a BitSet to a long, an integer a short, or a byte and truncates the bit
    * set to numberOfBits length.  The input is a
    * long.  The JNWBitSet result it truncated to the length specified, usually 4, 8
    * or 16 bits.  <b>This method assumes that the number input is a non-negative
    * number.</b>  Negative inputs will result in an error message.
    * <p><b>The length of <i>this</i> JNWBitSet is changed by
    * setJNWBitSetToLongAndTruncate.</b>*/
   public void setJNWBitSetToLongAndTruncate ( long inputIn , int numberOfBits )
   {
      long number = inputIn  ;
      int bitLocation = 0 ;
      // Allocate a minimum of 64 bits for the result
      this . clear ( ) ;
      if ( inputIn < 0 )
      {
         Utility.err.println( "ERROR in Utility.JNWBitSet: the number "
            + inputIn + " is nagative" ) ;
         return ;
      }
      do
      {
         if ( number % 2 == 1 )
         {
            this . set ( bitLocation , true ) ;
         }
         number = number / 2 ;
         bitLocation = bitLocation + 1 ;
      } while ( number > 0 ) ;
      if ( bitLocation > numberOfBits )
      {
         Utility.err.println ( "ERROR in Utility.JNWBitSet: the number "
            + inputIn + " cannot be represented in " + numberOfBits
            + " bits."  ) ;
      }
      // Truncate the result to numberOfBits bits
      this . setJNWLength (  numberOfBits ) ;
   }  // end of set


   /**Converts a JNWBitSet to a byte array.  If the bitset does not contain
    * an integer number of bytes, the top-most byte is prepended with zeros
    * to fill the topmost byte.*/
   public byte [ ]  getBytes ( )
   {
      int byteValueAsInteger = 0 ;
      // How long is the BitSet?
      int bitSetLength = this . getJNWLength ( ) ;
      int byteArrayLength = ( bitSetLength + 7 ) / 8 ;
      int extraBits = bitSetLength % 8 ;
      // Allocate SPACE for the byte array
      // HOW ABOUT USIING byteArrayLength = ( bitSetLength + 7 ) / 8
      byte [ ] byteArray = new byte [ byteArrayLength ] ;
      int multiplier = 1 ;
      int accumulator = 0 ;
      // Convert the topmost bits if there are bits not completing a byte.
      if ( extraBits != 0 )
      {
         for ( int i = bitSetLength - extraBits  ; i < bitSetLength  ; i++ )
         {
            if ( this . get ( i  ) == true /*bit == true */ )
            { accumulator = accumulator +  multiplier ; }
            multiplier =  2 * multiplier  ;
         }
         byteArray [ 0 ] = ( byte ) accumulator ;
      }   // end of if ( extraBits != 0 )
      accumulator = 0 ;
      multiplier =  1 ;
      int bitsProcessed = 0 ;
      // Convert the remianing bytes starting at the rightmost bit.
      for ( int i = 0 ; i < bitSetLength - extraBits  ; i++  )
      {
         boolean bit = this . get ( i  ) ;
         bitsProcessed = bitsProcessed + 1 ;
         if ( this . get ( i  )  == true )
         { accumulator = accumulator + multiplier ;
         }
         multiplier = multiplier * 2 ;
         if ( bitsProcessed == 8 )
         {
            int j = byteArrayLength - ( ( i + 1 ) / 8 ) ;
            byteArray [ j ] = ( byte ) accumulator ;
            multiplier = 1 ;
            accumulator = 0 ;
            bitsProcessed = 0 ;
         }  /// end of if ( bitsProcessed == 8 )
      }  // end of for ( int i = byteArray . length -1 ; i >= 0  ; i-- ) {
      return byteArray ;
   }  // end of convertToByteArray (  )


   /**Converts a JNWBitSet to the hexidecimal String that it represents.*/
   public String convertJNWBitSetToHEXString ( )
   {
      char [ ] characterArray
         = { '0' , '1' , '2' , '3'
            , '4' , '5' , '6' , '7'
            , '8' , '9' , 'A' , 'B'
            , 'C' , 'D' , 'E' , 'F' } ;
      JNWBitSet clone = this . clone ( ) ;
      java.lang.StringBuilder stringBuilder
         = new java.lang.StringBuilder  (  ) ;
      // Fill in "0" if the JNWBitSet is not a multiple of octets
      while ( ( clone . getJNWLength ( ) % 4 ) != 0 )
      {
         clone. prepend ( "0" ) ;
      }
      int index = 0 ;
      for ( int i = clone . getJNWLength ( ) - 1 ; i >= 0 ; i = i - 4   )
      {
         index = 8 * clone .getBitValue ( i )
         + 4 * clone .getBitValue ( i - 1  )
         + 2 * clone .getBitValue ( i - 2  )
         + 1* clone .getBitValue ( i - 3 )  ;
         stringBuilder . append ( characterArray  [ index ] ) ;
      }  // end of for ( int i = clone . getJNWLength ( ) - 1 ; i >= 0 ; i = i - 4   )
      return new String ( stringBuilder ) ;
   }  // end of convertJNWBitSetToString


   /**Converts a JNWBitSet to the String that it represents.*/
   public String convertJNWBitSetToString ( )
   {
      return new String ( this . getBytes ( ) ) ;
   }  // end of convertJNWBitSetToString


   /**Converts a JNWBitSet to the long that it represents.  The
    * user can then convert the resulting long to a byte, short,
    * int or long.  A bit set of length 64 or more will overflow
    * multiplier, so an error is reported, but the conversion
    * continues.
    * <p>This method creates powers of 2 ( 2, 4, 8, 16, etc.)
    * and adds them to the result if the corresponding bit
    * of <i><b>this</i></b> is set.  But
    * for an unknown reason an overflow, multiplier > 2**64 sets
    * multiplier to 0.  This is inherent to Java, and this lack of an
    * error message on overflow is not written into this method: It is part
    * of Java.  Therefore, any bits higher than the 64th bit
    * are ignored.  For example
    * "111111000 00000000 00000000 00000000 00000000 00000000 00000000 00000111 00000000"
    * is converted to 1792.  If bit 63 is set, the result is negative.
    * For example
    * "10000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000"
    * is -9223372036854775808.  No exception is thrown by the
    * bit set exceeding 63 bits, but an error message is printed.
    * <p>On 20071111 this method was changed from
    * <li>for( int i = 0 ; i < this . JNWLength ; i++ )</li>
    * to <li>for( int i = 0 ; i < this.length() ; i++ )</li>
    * and it was tested.  The "improved" algorithm searches from bit 0 to
    * bit length rather than bit JNWLength.  This improves speed
    * since most of the 32-bit numbers, for example sequence numbers,
    * have low numerical value (less than 2** 16).  This "improved"
    * algorithm will only search until the highest 1 is found.  It will
    * NOT search the complete bit set.*/
   public long convertJNWBitSetToLong ( )
   {
      long result = 0 ;
      long multiplier = 1 ;
      if ( this . length () > 63 )
      {
         System.err.println ( " ERROR IN Utility.JNWBitSet.convertJNWBitSet"
            + "ToLong ( ): length " + this . length () + " exceeds "
            + "63 bits." ) ;  }
      // Note that the index i runs from 0 to this . length() not
      // this . JNWLength since none of the bits above this . length() are 1
      for( int i = 0 ; i < this . length () ; i++)
      {
         if (  this .get ( i ) == true )
         {
            result = result + multiplier ; }
         multiplier = 2 * multiplier ;
      }  // end of for( int i = 0 ; i < this.length() ; i++ )
      return result ;
   }  // end of convertJNWBitSetToString


   /**Sets a String to a JNWBitSet representing the String.
    * The method first clears the JNWBitSet and then set the JNWBitSet
    * to the input String.  <p> <b>The length of the JNWBitSet is changed
    * by setToString.</b>*/
   public void setJNWBitSetToString ( String characterString )
   {
      byte [] byteArray  = characterString . getBytes ( ) ;
      byte singleByte ;
      this . clear ( ) ;
      int bitSetIndex = 0 ;
      int byteValueAsInteger = 0 ;
      // Start at the rightmost character
      for ( int i = byteArray . length - 1 ; i >= 0  ; i-- )
      {
         singleByte = byteArray [ i ] ;
         byteValueAsInteger = singleByte ;
         for ( int j = 0 ; j < 8 ; j++ )
         {
            // Get the least significant bit
            int bit = byteValueAsInteger - 2 * ( byteValueAsInteger / 2 ) ;
            if ( bit == 1 )
            {
               this . set ( bitSetIndex , true )  ;
            }
            else
            {
               this . set ( bitSetIndex , false )  ;
            }
            byteValueAsInteger = byteValueAsInteger / 2 ;
            bitSetIndex = bitSetIndex + 1 ;
         } // end of for
      }  // end of for ( int i = byteArray . length -1 ; i >= 0  ; i-- ) {
      JNWLength = 8 * byteArray . length ;
   }  // end of setJNWBitSetToString


   /**Writes a complete BitSet as octets, four octets per line as a
    * String.  The highest order bit is written first.*/
   public String toString ( )
   {  // Number of bits to be printed in each row
      int columnWidth = 32 ;
      return this . toString ( columnWidth ) ;
   } // end of WriteBitSet


   /**Writes a complete BitSet as octets with columnWidth
    * bits per row.  A space is inserted after eight
    * character.  The highest order bit is written first.*/
   public String toString ( int columnWidth )
   {
      java.lang.StringBuilder stringBuilder
         = new java.lang.StringBuilder ( 2 * JNWLength  + columnWidth ) ;
      // How many bits will be displayed in the first line?
      int firstLineBits =  JNWLength % columnWidth ;
      // How many blank spaces should be inserted in the first line?
      int firstLineSpaces = columnWidth + ( columnWidth / 8 ) - 1
         - ( firstLineBits + firstLineBits / 8 ) ;
      ////20080919 Addition////
      if ( firstLineBits == 0 ) {
         firstLineSpaces = 0 ;
      } // end of if
      ////20080919////
      // Pad blanks to start the message with firstLineSpaces spaces
      for ( int i = 0 ; i < firstLineSpaces ; i++ )
      {
         stringBuilder . append ( SPACE ) ;
      }
      for ( int i = JNWLength - 1 ; i >= 0  ; i-- )
      {
         // Add a newline except for the first bit
         if ( ( i + 1 ) % columnWidth == 0 && i != JNWLength - 1 )
         { stringBuilder . append ( CARRIAGE_RETURN ) ;
         }
         // Add a SPACE between the octets but NOT at the beginning of a line
         else
         ////20080919 Change////
         { if ( ( i + 1 ) % 8 == 0 && ( i + 1 ) % columnWidth != 0 )
           {
              stringBuilder . append ( SPACE ) ;
           }  // end of if ( ( i + 1 ) % 8 == 0 && ( i + 1 ) % columnWidth != 0 )
         }  // end of else
         ////20080919////
         // Add a one
         if ( this . get ( i ) == true )
         { stringBuilder . append ( "1" ) ;  }
         else
         { stringBuilder . append ( "0" ) ; } // Add a zero
      }  // end of for ( int i = 0 ; i < bitSetIn.length( ) ; i++ )
      return new String ( stringBuilder ) ;
   } // end of WriteBitSet


   /**Writes the bits representing a subset of a JNWBitSet.  Returns a String
    * of the JNWBitSet of the bits fromIndex(inclusive) to the specified
    * toIndex(exclusive).  An error message is printed if the requested
    * range is outside 0 to JNWLength -1*/
   public String writeSubSetOfBitSet (int fromIndex , int toIndex  )
   {
      if ( toIndex < 0 | fromIndex >= JNWLength | fromIndex > toIndex )
      {
         Utility.err.println ( "ERROR in JNWBitSet.getSubset of bit set: "
            + this . toString () + " JNWLength = " + JNWLength
            + " fromIndex = " + fromIndex + " toIndex = "
            + toIndex  ) ;
      }  // end of if
      JNWBitSet subset = this . getSubset (  fromIndex , toIndex  ) ;
      return subset . toString ( ) ;
   } // end of WriteBitSet


   /**Writes the lowest N bits of a BitSet as eight-bit bytes, five bytes per
    * line.  The highest order bit is written first.  This may be useful
    * for checking the CRC of the arrived BitSet.*/
   public String writeLowestNBitsOfBitSet (int numberOfBitsToWrite )
   {
      JNWBitSet lowestNBits = this . clone ( )  ;
      lowestNBits . truncateToLowestNBits ( numberOfBitsToWrite ) ;
      return lowestNBits . toString ( ) ;
   } // end of WriteBitSet



   /**Computes the CRC Cyclical redundancy Code for a data BitSet with any CRC
    * polynomial.  CRCPolynomialIn is the generator polynomial for the CRC.  The
    * original data is unchanged.*/
   public JNWBitSet computeFCS ( JNWBitSet generatorPolynomial  )
   {
      // This is the data set.  This is cloned because the original data
      // set is shifted below to compute the CRC
      JNWBitSet clone =   this . clone ( )    ;
      // Find the length of the generator polynomial.  Use length, not
      // JNWLength since the generator polynomial could have leading
      // zeros, for example " 00000001 00010000 00100001 "
      // Shift the result
      clone = clone . shift ( generatorPolynomial .length ( ) - 1 ) ;
      // Find the remainder of the original bit set with its pad of zeros are the end
      JNWBitSet remainder = clone . computeRemainder ( generatorPolynomial ) ;
      return  remainder  ;
   } // end of computeCRC


   /**Computes the CRC Cyclical redundancy Code for a data BitSet with any CRC
    * polynomial.  CRCPolynomialIn is the generator polynomial for the CRC.*/
   public JNWBitSet computeFCS_Student ( JNWBitSet generatorPolynomial  )
   {
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      //// THIS JUST RETURNS AN ALL-ZERO POLYNOMIAL OF THE CORRECT LENGTH
      return new JNWBitSet ( generatorPolynomial . getJNWLength ( ) - 1 ) ;
   } // end of computeCRC

   /**Computes the CRC Cyclic redundancy Code for a data BitSet with any CRC
    * polynomial.  CRCPolynomialIn is the generator polynomial for the CRC.*/
   public JNWBitSet computeFCSUsingShiftRegister ( JNWBitSet generatorPolynomial  )
   {
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      return  ( JNWBitSet ) null ;
   } // end of computeCRC


   /**Computes the CRC of a frame and returns a boolean indicating a
    *whether the CRC indicates a transmission error.  True means that the
    * CRC is correct, and false indicates a bit error somewhere.*/
   public boolean verifyFCSUsingShiftRegister( JNWBitSet generatorPolynomial )
  {
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      return true ;
   }  // end of verifyCRC


   /**Computes the remainder polynomial of dividing <i><b>this</b></i> by the
    * generatorPolynomial.  GeneralPolynomial is the generator polynomial for
    * the CRC.*/
   public JNWBitSet computeRemainder ( JNWBitSet generatorPolynomial  )
   {  // This is the data set.  This is cloned because the original data
      // set is destroyed during the processing
      JNWBitSet register =   this . clone ( )  ;
      // Find the length of the generator polynomial.  Use length, not
      // JNWLength since the generator polynomial could have leading
      // zeros, for example " 00000001 00010000 00100001 "
      int first1Position = generatorPolynomial . length ( ) ;
      //Now set the leftmost bit of the generator polynomial to 0
      generatorPolynomial . set ( first1Position - 1 , false ) ;
      // Now find the position of the second 1 in the divisor
      int second1Position = generatorPolynomial . length ( ) - 1 ;
      //Now reset the leftmost bit of the generator polynomial to 1
      generatorPolynomial . set ( first1Position - 1  ) ;
      // Skip all the leading zeros in this
      for ( int i = register.length ( ) - 1 ;  i >= first1Position - 1 ; i-- )
      {
         if ( register . get ( i ) == true )
         {  // is the feedback bit true?
            for ( int j = second1Position ; j >= 0 ; j-- )
            {
               if ( generatorPolynomial . get ( j  ) == true )
               {
                  register . flip ( i - ( first1Position - 1 ) + j  ) ;
               }  // end of if ( CRCPolynomialIn .get ( j  ) == true )
            }  // end of for ( int j = first1Position   ; j >= 0 ; j-- )
         }  // end of if ( feedbackBit == true )
      }  // end of for ( int i = resultSet.length() - 1 ;  i >= newLength - 1 ; i++ )
      register . setJNWLength ( first1Position - 1  ) ;
      return  register . clone ( )  ;
   } // end of computeCRC

   /**Computes the remainder polynomial of dividing <i><b>this</b></i> by the
    * generatorPolynomial.  GeneralPolynomial is the generator polynomial for
    * the CRC.*/
   public JNWBitSet computeRemainder_Student ( JNWBitSet generatorPolynomial  )
   {  // This is the data set.  This is cloned because the original data
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      ////STUDENT INSERT CODE HERE
      //// THIS JUST RETURNS AN ALL-ZERO POLYNOMIAL OF THE CORRECT LENGTH
      return new JNWBitSet ( generatorPolynomial . getJNWLength ( ) - 1 ) ;
   } // end of computeCRC


   /**Appends a cyclical redundancy code to the end of <i><b>this</b></i> JNWBitSet
    * using a JNWBitSet to represent the CRC.*/
   public void appendFCS ( JNWBitSet generatorPolynomialIn )
   {  // Remove any preceeding zeros from the generator polynomial
      JNWBitSet generatorPolynomial = generatorPolynomialIn . clone ( ) ;
      generatorPolynomial . setJNWLength ( generatorPolynomial . length ( ) ) ;
      JNWBitSet CRCResult
         = this . computeFCS ( generatorPolynomial ) ;
      this . append ( CRCResult  ) ;
   }  // end of appendCRC( JNWBitSet crcCode )


   /**Appends a JNWBitSet to the end of the existing BitSet.*/
   public void append ( JNWBitSet bitSet )
   {  // Shift the JNWBitSet
      this. shift ( bitSet . getJNWLength ( ) ) ;
      for ( int i = 0 ; i < bitSet . getJNWLength ( ) ; i++ )
      {
         this . set ( i , bitSet.get ( i ) ) ;
      }  // end of for
   }  // end of append


   /**Appends a string of 0s and 1s to the end of the existing BitSet.*/
   public void append ( String stringOfZerosAndsOnes  )
   {  // Shift the JNWBitSet
      this . append ( new JNWBitSet ( stringOfZerosAndsOnes ) ) ;
   }  // end of append


   /**Appends a JNWBitSet to the beginning of the existing BitSet.*/
   public void prepend ( JNWBitSet bitSet )
   {  // Save JNWLength since it will change
      int oldNWBLength = JNWLength ;
      // Not that the index runs from 0 to getNWBlength( ) - 1
      for ( int i = bitSet . getJNWLength ( ) - 1 ; i >=0   ; i-- )
      {
         // set ( ) will automatically update the JNWLength
         this . set ( oldNWBLength + i , bitSet.get ( i ) ) ;
      }  // end of for
   }  // end of prepend


   /**Appends a String of 0s and 1s to the beginning of the existing BitSet.*/
   public void prepend ( String stringOfZerosAndsOnes  )
   {  // Save JNWLength since it will change
      this . prepend ( new JNWBitSet ( stringOfZerosAndsOnes ) ) ;
   }  // end of prepend


   /**An method to find the start of pattern in a JNWBitSet.  It returns the address in
    * this bitSet of the first bit of the first pattern occurrence.  For example, for
    *      JNWBitSet dataBitSet = new JNWBitSet (
    * "11111111 11101111 10101010 11110111 11100011") ;
    * JNWBitSet patternBitSet = new JNWBitSet (
    * "11110111111") ;
    *  dataBitSet.findpattern (patternBitSet) returns 15.  If there
    * is no matching pattern in this bit set, the method returns -1.*/
   public int findPattern ( JNWBitSet patternBitSet )
   {
      if ( patternBitSet . getJNWLength ( ) > this . JNWLength )
      {
         Utility.err.println (
            "ERROR in Utilities.JNWBitSet.findPattern: "
            + "Matching pattern length, " + patternBitSet . getJNWLength ( )
            + ", exceeds length of JNWBitSet, " + this . JNWLength  )  ;
         return -1 ;
      }  // end of if
      else
      {
         for ( int i = this . JNWLength - 1
            ; i >= patternBitSet . getJNWLength ( ) - 1  ; i-- )
         {
            boolean foundPattern = true ;
            // This is NOT an efficient way of doing this
            // but it works
            for ( int j = 0 ; j < patternBitSet . getJNWLength ( )  ; j++ )
            {
               if ( ! ( this . get ( i + j - patternBitSet . getJNWLength ( )  + 1 )
               == patternBitSet . get ( j ) ) )
               {
                  foundPattern = false ;
                  break ;
               }  // end of if ( this.get( i  + j - bitSet . getNWBlength( ))
            }  // end for ( int j = bitSet . getNWBlength() ; j >= 0 ; j-- )
            if ( foundPattern == true )
            { return i ; }
         }  // end of for ( int i = this . JNWLength - 1 ; i >= bitSet . getJNWLength( ) - 1  ; i-- )
         // pattern not found, so return an illegal number
         return -1 ;
      }  // end of else
   }  // end of public int findPattern


   /**Returns the value of the bitSetIndexth bit as a 0 or 1, not
    * a true or false.  JNWLength - 1 >= bitSetIndex >= 0 */
   public int getBitValue ( int bitSetIndex )
   {
      if ( bitSetIndex < 0 | bitSetIndex >= JNWLength )
      {  Utility.err.println (
            "ERROR in Utility.JNWBitSet.getBitValue: "
            + " bitSetIndex = " + bitSetIndex  ) ;
         return -1 ; }
      boolean booleanValue = this . get ( bitSetIndex ) ;
      if ( booleanValue == true )
      { return   1 ; }
      return  0 ;
   }


   /**Computes the CRC of a frame and returns a boolean indicating a
    *whether the CRC indicates a transmission error.  True means that the
    * CRC is correct, and false indicates a bit error somewhere.*/
   public boolean verifyFCS ( JNWBitSet generatorPolynomial )
   {  // Compute remainder  which should be all 0s if no error has occurred
      JNWBitSet remainder = this . computeRemainder ( generatorPolynomial ) ;
      for ( int i = 0 ; i < remainder.JNWLength ; i++ )
      {  // Are the computed remainder and the remainder in this the same?
         if ( remainder . get ( i ) == true )
         {
            return false ;  // If any bit of the remainder is 1 then an error has occurred.
         }  // end of if ( remainder.get ( i ) == true )
      }  // end of for ( int i = 0 ; i < remainder.JNWLength ; i++ )
      return true ;
   }  // end of verifyCRC


   /**Computes the CRC of a frame and returns a boolean indicating a
    *whether the CRC indicates a transmission error.  True means that the
    * CRC is correct, and false indicates a bit error somewhere.*/
   public boolean verifyFCS_Student ( JNWBitSet generatorPolynomial )
   {
      return true ;
   }  // end of verifyCRC

   /**Truncates the JNWBitSet to the lowest n bits.  If JNWLength < n,
    * then JNWLength is unchanged. */
   public void truncateToLowestNBits ( int n )
   {  // Simply setting the JNWBitSet length to n will
      // cause it to ignore all the bits above the nth
      // bit-in location n-1
      JNWLength = java.lang.Math.min ( JNWLength , n ) ;
   }


   /**Stuffs a bit set.  The method looks for five consecutive ones,
    *11111, in a row.  When it finds the five ones, it stuffs a 0 after the
    *fifth one.  This is used in HDLC transmission since the header and
    *footer for HDLC contains six 1 in a row 01111110, and it is necessary
    *to avoid the occurrence of the same pattern in the data.*/
   public JNWBitSet stuff (  )
   {  // Compute the maximum possible length of the result
      int length = this . getJNWLength () ;
      length = ( int ) ( ( 1.2 * length ) + 1 ) + 16 ;
      // Make a result of that length
      JNWBitSet bitstuffedResult = new JNWBitSet ( length ) ;
      int counterForSuccessiveOnes = 0 ;
      int indexForStuffedResult = length - 1 ;
      boolean bitValue = false ;
      for ( int i = this . getJNWLength () - 1 ; i >= 0 ; i-- )
      {
         bitValue = this . get ( i ) ;
         bitstuffedResult . set (  indexForStuffedResult , bitValue )  ;
         // Decrement the index for the stuffed result
         indexForStuffedResult =  indexForStuffedResult - 1 ;
         if ( bitValue == true )
         {
            counterForSuccessiveOnes = counterForSuccessiveOnes + 1 ;
         }  // end of if ( bitvalue == true )
         else
         {
            counterForSuccessiveOnes = 0 ;
         }  // end of else
         if ( counterForSuccessiveOnes == 5 )
         {  // put in a 0 after the fifth 1
            bitstuffedResult . set (  indexForStuffedResult , false )  ;
            counterForSuccessiveOnes = 0 ;
            indexForStuffedResult =  indexForStuffedResult - 1 ;
         } // end of if ( counterForSuccessiveOnes == 5 )
      }  // end of for ( int i = dataBitsIn.getJNWLength() - 1 ; i >= 0 ; i-- )
      //The result may have some data padded at the bottom
      bitstuffedResult . shift ( - (  indexForStuffedResult + 1 ) )  ;
      return bitstuffedResult ;
   }


   /**Stuffs a bit set.  The method looks for five consecutive ones,
    *11111, in a row.  When it finds the five ones, it stuffs a 0 after the
    *fifth one.  This is used in HDLC transmission since the header and
    *footer for HDLC contains six 1 in a row 01111110, and it is necessary
    *to avoid the occurrence of the same pattern in the data.*/
   public JNWBitSet stuff_Student (  )
   {
      JNWBitSet unstuffedBitSet = this  ;
      JNWBitSet stuffedBitSet = null ; 
////STUDENT INSERTS CODE HERE
////STUDENT INSERTS CODE HERE
////STUDENT INSERTS CODE HERE
////STUDENT INSERTS CODE HERE
      return stuffedBitSet ;
   }


   /**Unstuffs a bit set.  The method looks for five consecutive ones,
    *11111, in a row.  It then removes the 0 that was stuffed after the
    * five ones.  This is used in HDLC transmission since the header and
    *footer for HDLC contains six 1 in a row 01111110, and it is necessary
    *to avoid the occurrence of the same pattern in the data.*/
   public JNWBitSet unstuff (  )
   {
      int indexForUnstuffedResult = this . getJNWLength ( ) - 1 ;
      JNWBitSet unstuffedResult
         = new JNWBitSet ( indexForUnstuffedResult + 1 ) ;
      int counterForSuccessiveOnes = 0 ;
      boolean bitValue = false ;
      for ( int i = this . getJNWLength () - 1 ; i >= 0 ; i-- )
      {
         bitValue = this . get ( i ) ;
         unstuffedResult . set ( indexForUnstuffedResult , bitValue )  ;
         indexForUnstuffedResult = indexForUnstuffedResult - 1 ;
         if ( bitValue == true )
         {
            counterForSuccessiveOnes = counterForSuccessiveOnes + 1 ;
         }
         else
         {
            counterForSuccessiveOnes = 0 ;
         }
         if ( counterForSuccessiveOnes == 5 )
         {  // Decrement the index of the stuffed bit set
            // effectively put in a 0 after the fifth 1
            i = i - 1 ;
            counterForSuccessiveOnes = 0 ;
         }  // end of if ( counterForSuccessiveOnes == 5 )
      }  // end of for ( int i = dataBitsIn.getJNWLength() - 1 ; i >= 0 ; i-- )
      //The result will be shorter that the stuffed data
      unstuffedResult . shift ( - ( indexForUnstuffedResult + 1 ) )  ;
      return unstuffedResult ;
   }


   /**Stuffs a bit set.  The method looks for five consecutive ones,
    *11111, in a row.  When it finds the five ones, it stuffs a 0 after the
    *fifth one.  This is used in HDLC transmission since the header and
    *footer for HDLC contains six 1 in a row 01111110, and it is necessary
    *to avoid the occurrence of the same pattern in the data.*/
   public JNWBitSet unstuff_Student (  )
   {
      JNWBitSet stuffedBitSet = this  ;
      JNWBitSet unstuffedBitSet = null  ;
////STUDENT INSERTS CODE HERE
////STUDENT INSERTS CODE HERE
////STUDENT INSERTS CODE HERE
////STUDENT INSERTS CODE HERE
      return unstuffedBitSet ;
   }


   /**Overrides the cardinality method of java.util.BitSet
    * to return the number of bits set to true in this JNWBitSet.  This
    * is necessary since the length is JNWLength, not length, as in BitSet.*/
   public int cardinality ( )
   {
      int cardinality = 0 ;
      for ( int i = 0 ; i < JNWLength ; i++ )
      {
         cardinality = cardinality + this .getBitValue ( i ) ;
      }
      return cardinality ;
   }


   /**Returns a clone of a subset of a JNWBitSet.  Returns a JNWBitSet
    * of the bits fromIndex(inclusive) to the specified toIndex(exclusive).
    * An error message is printed if the requested range is outside
    * 0 to JNWLength -1.*/
   public JNWBitSet getSubset (int fromIndex, int toIndex)
   {
      JNWBitSet result = null  ;
      if ( toIndex < 0 | fromIndex >= JNWLength | fromIndex > toIndex )
      {
         Utility.err.println ( "ERROR in JNWBitSet.getSubset of bit set: "
            + this . toString () + " JNWLength = " + JNWLength
            + " fromIndex = " + fromIndex + " toIndex = "
            + toIndex  ) ;
      }  // end of if
      else
      {
         result = new JNWBitSet ( toIndex - fromIndex ) ;
         for ( int i = fromIndex ; i < toIndex ; i++ )
         {
            result . set ( i - fromIndex , this . get ( i )) ;
         }  // end of for ( int i = fromIndex ; i < toIndex ; i++ )
      }  // end of else
      return result ;
   }  // end of getSubset (int fromIndex, int toIndex)


   /**Returns a clone of the topmost n bits of a JNWBitSet.  Useful
    * for extracting headers.  An error message is printed if n
    * exceeds the length of the bit set.*/
   public JNWBitSet getNHighestBits ( int n )
   {
      JNWBitSet result = null  ;
      if ( n > this . getJNWLength ( )  )
      {
         Utility.err.println ( "ERROR in JNWBitSet.getNHighestBits: "
            + " JNWLength = " + JNWLength
            + " n = " + n ) ;
      }  // end of if
      else
      {
         int toIndex = this . getJNWLength ( ) ;
         result = this . getSubset ( toIndex - n , toIndex ) ;
      }  // end of else
      return result ;
   }  // end of getSubset (int fromIndex, int toIndex)



   /**Sets a subset of this to a JNWBitSet starting at bit position
    * fromIndex.  An error message is printed if the requested range
    * is outside 0 to JNWLength -1.*/
   public void setSubset ( int fromIndex, JNWBitSet data )
   {
      if ( fromIndex + data.getJNWLength ( ) > this. getJNWLength ( )
      | fromIndex < 0 )
      {
         Utility.err.println ( "ERROR in JNWBitSet.setSubset of bit set: "
            + data . toString () + " JNWLength = " + this. getJNWLength ( )
            + " fromIndex = " + fromIndex ) ;
      }  // end of if
      else
      {
         for ( int i = 0 ; i < data.JNWLength ; i++ )
         {
            this . set ( i + fromIndex , data . get ( i )) ;
         }  // end of for ( int i = fromIndex ; i < toIndex ; i++ )
      }  // end of else
   }  // end of setSubset


   /**Clones a JNWBitSet.  This performs the same function as the copy
    * constructor.*/
   public JNWBitSet clone ( )
   {
      return new JNWBitSet ( this ) ;
   }


   /**Applies the xor successively to the complete value of <i>this.</i>.
    *<p> This method is used to compute the UDP and TCP checksum for the source and
    * destination addresses in the pseudoheader.</p>*/
   public JNWBitSet xorAndShift  ( JNWBitSet xorIn )
   {
      JNWBitSet resultClone = xorIn. clone ( ) ;
      // since any bits above position
      // length are all zeros and will not affect the xor
      for ( int i = 0 ; i < this.getJNWLength ( )  ; i++ )
      {
         if ( this . get (i) == true )
         {
            resultClone . flip ( i % resultClone.JNWLength ) ;
         }  // end of if
      }
      return resultClone ;
   }  // end of xorAndShift2


   /**Inverts each bit of <i>this.</i>.*/
   public void invertBitSet ( )
   {
      for ( int i = 0 ; i < this. JNWLength ; i++ )
      {
         this. flip ( i ) ;
      }  // end of for
   }  // end of invertBitSet


   /**Performs a XOR on every bit of two bit sets.  The result has the
    * length of the
    * longest bit set either <i><b>this</i></b> or bitSetIn.  This is
    * different from the java.lang.BitSet.xor.  In java.lang.BitSet.xor, the
    * return type is void so <i><b>this</i></b> is not changed in length.
    * <p>This method was built to help with testing the CRC methods.*/
   public JNWBitSet xorBitSet  ( JNWBitSet bitSetIn )
   {
      int thisLength = this .getJNWLength ( ) ;
      int bitSetInLength = bitSetIn .getJNWLength ( ) ;
      JNWBitSet resultBitSet
         = new JNWBitSet ( java.lang.Math.max ( thisLength  , bitSetInLength) ) ;
      if ( this . getJNWLength ( ) > bitSetIn . getJNWLength ( ) )
      {
         // Check if this has any bits set
         for ( int i = this . getJNWLength ( ) - 1
            ; i >= bitSetIn . getJNWLength ( )  ; i-- )
         {
            resultBitSet . set ( i , this . get ( i ) ) ;
         }  // end of for ( int i = this . getJNWLength ( )
      } // end of if ( this . getJNWLength ( ) > bitSetForComparison.getJNWLength( ) )
      if ( this . getJNWLength ( ) < bitSetIn.getJNWLength ( ) )
      {
         // Check if bitSetForComparison has any bits set
         for ( int i = bitSetIn . getJNWLength ( ) - 1
            ; i >= this . getJNWLength ( )  ; i-- )
         {
            // If so, bitSetForComparison is greater than this
            resultBitSet . set ( i , bitSetIn . get ( i ) ) ;
         }  // end of for ( int i = this . getJNWLength ( )
      } // end of if ( this . getJNWLength ( ) > bitSetForComparison.getJNWLength( ) )
      int start = java.lang.Math.min ( this . getJNWLength ( )
      , bitSetIn . getJNWLength ( ) )  ;
      for ( int i = 0   ; i < start   ; i++ )
      {
         resultBitSet . set ( i , this . get ( i ) ^ bitSetIn . get ( i ) ) ;
      }  // end of for ( int i = this . getJNWLength ( )
      return resultBitSet ;
   }  // end of xorBitSet



   /**Compares two bit sets.  It <b>ASSUMES</b> that the bit sets represent positive
    * ints.  Returns +1 if <i>this</i> is greater than
    *the passed parameter.  It returns 0 if they are equal, and it returns -1 if
    *<i><b>this</b></i> is less than the passed parameter.  This is used in the
    * private class
    * Utility.TCPDestinationBuffer.TCPComparitor to order TCP packets
    *by sequence numbers in reconstructing messages.  With this method it is
    *not necessary to convert the sequence numbers to ints before making
    *the ordering.
    * ,<p>20071119Revised algorithm completely to test lengths versus JNWLengths
    * to more rapidly make the comparison.*/
   public int compare ( JNWBitSet bitSetForComparison )
   {  // Is this longer than bitSetForComparison
      if ( this . length ( ) > bitSetForComparison . length ( ) )
      {
         return 1 ;
      } // end of if ( this . length ( ) > bitSetForComparison . length ( ) )
      else if ( bitSetForComparison . length ( ) > this . length ( )  )
      {
         return -1 ;
      }  // end of if ( bitSetForComparison . length ( ) > this . length ( )  )
      // So the lengths are equal
      // If both the bit set are all 0s return 0
      if ( this . length ( ) == 0 )
      { return 0 ; }
      // Return a +1 if this has a bit set before bitSetForComparison
      // Return a -1 if bitSetForComparison has a bit set before this
      for ( int i = this . length ( ) ; i >= 0  ; i-- )
      {
         boolean me = this . get ( i ) ;
         // Is only one of the booleans true?
         if ( ( me  ^  bitSetForComparison . get ( i ) ) == true )
         {
            // Only one bit is set, if me is set then this is greater
            if ( me == true )
            { return 1 ; }
            // Otherwise bitSetForComparison is greater
            else return -1 ;
         }  // end of if ( ( me  ^  bitSetForComparison . get ( i ) ) == true ) {
      }  // end of for ( int i = this . length ( ) ; i >= 0  ; i-- )
      return 0 ;
   }  // end of compare


   /**Returns a boolean telling whether a JNWBitSet begins with the parameter passed.*/
   public boolean beginsWith ( JNWBitSet bitSetIn )
   {
      int indexThis = this . getJNWLength ( ) ;
      int indexbitSetIn = bitSetIn . getJNWLength ( ) ;
      if ( indexbitSetIn > indexThis )
      {
         Utility.err.println ( "ERROR in Utility.BitSet.beginsWith: length of"
            + " parameter bit set larger than this bit set." ) ;
         return false ;
      }  // end of if ( fromIndexbitSetIn > fromIndexThis )
      for ( int i = indexbitSetIn ; i >= 0 ; i-- )
      {
         if ( this . get ( indexThis ) != bitSetIn . get ( i ) )
         {
            return false ;
         }
         indexThis = indexThis - 1 ;
      } // end of for ( int i = bitSetIn . getJNWLength ( ) ; i >= 0 ; i-- )
      return true ;
   }  // end of beginsWith


   /**Returns a boolean telling whether a JNWBitSet ends with the parameter passed.*/
   public boolean endsWith ( JNWBitSet bitSetIn )
   {
      int indexbitSetIn = bitSetIn . getJNWLength ( ) ;
      if ( indexbitSetIn > this . getJNWLength ( ))
      {
         Utility.err.println ( "ERROR in Utility.BitSet.endsWith: length of"
            + " parameter bit set larger than this bit set." ) ;
         return false ;
      }  // end of if ( fromIndexbitSetIn > fromIndexThis )
      for ( int i = indexbitSetIn ; i >= 0 ; i-- )
      {
         if ( this . get ( i ) != bitSetIn . get ( i ) )
         {
            return false ;
         }
      } // end of for ( int i = bitSetIn . getJNWLength ( ) ; i >= 0 ; i-- )
      return true ;
   }  // end of beginsWith



   /**Removes all the NULL characters ( "00000000" ) from the end of a
    * JNWBitSet.  Null characters that may have been inserted
    * by Ethernet, which requires a message of at least 46 octets.  A TCP
    * ACK message will have only 40 octets, and Ethernet will pad
    * it with null characters.*/
   public void removeNullCharactersAtEnd (  )
   {
      // index is the location of the first 1 in the bit set
      int index = MINUS_ONE ;
      for ( int i = 0 ; i <= this .getJNWLength ( ) ; i++ )
      {
         index = i ;
         // Search for the first 1
         if ( this  . get ( i ) == true )
         {
            break ;
         }
      }  // end of for ( int i = 0 ; i < 8 * 6 ; i++ )
      // Does index fall on an octet boundary?
      // If not decrease it to the octet boundary
      index  = 8 * ( ( index ) / 8 ) ;
      this . shift ( - index ) ;
   }  // end of removeNullCharactersAtEnd



   /**Decrements a JNWBitSet by 1 where the bit set represents
    * a non-negative number.  If the bit set represents a base-2 integer,
    * decrementByOne will decrement the bit set representation by one
    * without the need of first converting the bit set to an integer,
    *  decrementing the integer, and then converting the decremented
    * integer back into a bit set.
    * <p>This is used to decrement the TTL field in IP packets quickly.*/
    public int decrementByOne (  ) {
        int bitNumber ;
        // Check for 0 length bit Set
        if ( this.JNWLength > 0 ) {
            for ( bitNumber = 0 ; bitNumber < this.JNWLength ; bitNumber++ ) {
                if ( this.get ( bitNumber ) == false ) {
                    this.set( bitNumber );
                    if ( bitNumber == this.JNWLength - 1 ) {
                        Utility.err.println/*Utility.err.*/( "ERROR in Utility."
                           + "JNWBitSet.decrementByOne-bits all 0 representing "
                           + "int value 0-cannot decrement") ;
                    }
                }  // end of if ( this.get ( bitNumber ) == false )
                else {
                    this.set ( bitNumber , false ) ;
                    return bitNumber + 1 ;
                  // end of else
            }  // end of for ( bitNumber = 0 ; bitNumber < this.JNWLength ; bitNumber++ )
        }  // end of if ( this.JNWLength > 0 )

    }  // end of if ( this.JNWLength > 0 )
        return 0 ;
    }


    /**This flips the N lowest bits of a JNWBitSet.  It is used with
     * decrementByOne to decrement the TTL field in a TCP header.*/
   public JNWBitSet flipNLowestBits ( int n ) {
      JNWBitSet result = this . clone ( ) ;
      for ( int i = 0 ; i < n ; i++ ) {
         result . flip ( i ) ;
      }
      return result ;
   }


   /**Removes the N highest (leftmost) bits from a JNWBitSet by
    * resetting the JNWLength.*/
   public void removeNHighestBits ( int n ) {
      if ( n > this . JNWLength ) {
         Utility.err.println ( "ERROR in Utility.JNWBitSet."
            + "removeNHighestBits in trying to remove " + n
            + " bits from\n" + this . toString ( )) ;
      }  // end of if
      else {
         this . JNWLength = this .JNWLength - n ;
      }  // end of else
   }  // end of public void removeNHighestBits ( int n )


   /**Removes the N lowest (rightmost) bits from a JNWBitSet by
    * shifting the bitSet.*/
   public void removeNLowestBits ( int n ) {
       if ( n > this . JNWLength ) {
         Utility.err.println ( "ERROR in Utility.JNWBitSet."
            + "removeNLowestBits in trying to remove " + n
            + " bits from\n" + this . toString ( )) ;
      }  // end of if
      else {
          this . shift ( -n ) ;
      }  // end of else
   }  // end of public void removeNHighestBits ( int n )



   /**A main method for testing the JNWBitSet.
    * @param args the command line arguments*/
   public static void main_For_Testing_Only ( String[] args )
   {
      

 
      
      
 JNWBitSet dataBitSet = new JNWBitSet ( "11111111 11111111 ") ;
 JNWBitSet polynomial = new JNWBitSet ( " 10101010 " ) ;
 dataBitSet.xor ( polynomial ) ;
 System.out.println ("resulting bitSet = " + dataBitSet )  ;
 System.exit ( 1 ) ;
 JNWBitSet bitSet = new JNWBitSet ( "11111111 11 " ) ;    
      for ( int i = 1 ; i < 180 ; i++ ) {
          bitSet = new JNWBitSet ( i );
          for ( int j = 0 ; j < bitSet.getJNWLength () ; j++ ) {
             bitSet .set (  j  );
         }

System.out.println ("bitSet =\n" + bitSet.toString ( 40 ) );
       }


System.exit ( 1 ) ;
int length = bitSet .getJNWLength ( ) ;
System.out.println ( " bitSet length = " + bitSet .getJNWLength ( ) ) ;
bitSet . shift ( - length ) ;
System.out.println (" bitSet =\n" + bitSet.toString ( ) );
System.out.println ( " bitSet length = " + bitSet .getJNWLength ( ) ) ;
 System.exit ( 1 ) ;


       for ( int i = 1 ; i < 5 ; i++ ) {
         bitSet. removeNHighestBits ( 15 ) ;
         System.out.println (" i = " + i + " bitSet = " + bitSet.toString ( ) );
       }
System.out.println (" bitSet =\n" + bitSet.toString ( ) );
       for ( int i = 1 ; i < 5 ; i++ ) {
         bitSet. removeNLowestBits ( 15 ) ;
         System.out.println (" i = " + i + " bitSet = " + bitSet.toString ( ) );
       }
 System.exit ( 1 ) ;



      JNWBitSet test13 = null ;
       JNWBitSet test14 = null ;
       JNWBitSet test15 = null ;
       for ( int i = 0 ; i < 100 ; i++ ) {
           test13 = new JNWBitSet ( i , 16 ) ;
           test14 = new JNWBitSet ( 255 * 256 + 255 , 16 ) ;
           System.out.println ( "\n i = " + i  + "\n"+ test13 . toString ( 32 ) ) ;
           System.out.println ( " Unflipped bit set = " + test15  ) ;
           int j = test13.decrementByOne ( ) ;
           test15 = test14 . flipNLowestBits ( j ) ;
           test15 . flipNLowestBits ( j ) ;
           System.out.println ( " j = " + j + " Flipped bit set = " + test15 + " \n" ) ;
           System.out.println ( " Changed " + j + " bits \n" +  test13 . toString (  ) ) ;
           System.out.println ( " Value now " + test13 . convertJNWBitSetToLong ( ) ) ;
   ////        System.out.println ( " Flipped bit set = " + test15 + " \n" ) ;
       }
      System.exit ( 1 ) ;


      JNWBitSet test3 = new JNWBitSet ( " 11110000 00000000 10101010 01010101 10000000") ;
      JNWBitSet zeros = new JNWBitSet ( "00000000" ) ;
      zeros . append ( zeros ) ;

      test3.append (zeros ) ;
      System.out.println ( "result = " + test3.toString () ) ;
      test3.removeNullCharactersAtEnd (  ) ;
      System.out.println ( "result = " + test3.toString () ) ;

      System.exit ( 1 ) ;



      JNWBitSet parameter = new JNWBitSet ("11111111") ;
      JNWRandom randomNumberGenerator2 = JNWRandom .getInstance () ;
      int flipBit  ;
      for (int i = 0 ; i < 10000 ; i++ )
      {
         flipBit = randomNumberGenerator2 .nextInt ( 8 ) ;
         parameter . flip ( flipBit  );
         if ( test3.endsWith ( parameter ) == true )
         {
            System.out.println ( " parameter = " + parameter .toString () );
         }

      }

      System.exit ( 1 ) ;


      ////       JNWBitSet tester3 = new JNWBitSet ( " 1 00100011 01000101 01100111 10001001 10101011 11001101 1110 1111") ;
      ////       System.out.println ( "Hex represnetation = " + tester3.convertJNWBitSetToHEXString ( ) ) ;

////  System.exit ( 0 ) ;
 /*
            JNWBitSet dataa = new JNWBitSet ( ) ;
            JNWBitSet bitSetForComparisona = new JNWBitSet (1,1 ) ;
            System.out.println ( "Ccomprison of data1 to bitSetForComparison = " + dataa.compare ( bitSetForComparisona)) ;
 System.exit ( 0 ) ;
////            data1.setJNWBitSetToLongAndTruncate( 000 , 1 ) ;
////            bitSetForComparison.setJNWBitSetToLongAndTruncate( 00 , 1 ) ;




            JNWBitSet data1 = new JNWBitSet ( ) ;
            JNWBitSet bitSetForComparison = new JNWBitSet ( ) ;
            data1.setJNWBitSetToLongAndTruncate( 000 , 1 ) ;
            bitSetForComparison.setJNWBitSetToLongAndTruncate( 00 , 1 ) ;
            System.out.println ( "Ccomprison of data1 to bitSetForComparison = " + data1.compare ( bitSetForComparison)) ;
  System.exit ( 0 ) ;
  */

 /*

        JNWBitSet length = new JNWBitSet (" 00000000  1111111111 00001111 ") ;
        JNWBitSet JNWLength = new JNWBitSet ("00000001 00000000  1111111111 00001111 ") ;
         length . setJNWLength ( length .length () ) ;
         JNWLength . setJNWLength ( JNWLength .length () ) ;
        System.out.println ( " length = " + length .toString() + "\nJNWLength = "
                + JNWLength.toString() ) ;
        System.exit ( 0 ) ;
  */


 /*

        JNWBitSet one = new JNWBitSet ( " 11111111 11111111 11111111 11111111 11111111 1111111" ) ;
        JNWBitSet two = new JNWBitSet ( " 00000000 00000000 11111111" ) ;
  ///      one.xor ( two  ) ;
        one = two. xorBitSet ( one ) ;
        System.out.println ( "\none = " + one .toString()
           + "\ntwo= " + two .toString()
           + "\n   Xor = " + two .toString());

 System.exit ( 0 ) ;
  */
      String string3 = "abc" ;
      JNWBitSet tester = new JNWBitSet ( ) ;
      tester . setJNWBitSetToString ( string3  ) ;
      JNWBitSet  generatorPolynomial
         = new JNWBitSet ( "1 00000100 11000001 00011101 10110111" ) ;


      generatorPolynomial
         = new JNWBitSet ( "1 01101010 00000100 11111111 11000001 00011101 10110111" ) ;
      ////     generatorPolynomial = new JNWBitSet ( "10011" ) ;
      ////       JNWBitSet polynomial = new JNWBitSet ( "1" ) ;
      JNWBitSet accumulator   ;
      JNWBitSet cloneDuplicate ;
      JNWBitSet generatorClone ;
      JNWBitSet remainderAdded ;
      JNWBitSet result  ;
      JNWBitSet remainder ;
      JNWBitSet remainderComputed ;
      ////       int maximumShift = 50 ;
      JNWRandom randomNumberGenerator = JNWRandom .getInstance () ;

      for ( int i = 0 ; i < 100 ; i++ )
      {
         accumulator = generatorPolynomial . clone ( ) ;
         /////           generatorClone = generatorPolynomial . clone ( ) ;
         for ( int k = 0 ; k < 100 ; k++ )
         {
            ////               cloneDuplicate = generatorPolynomial . clone ( ) ;
            int shift2 = randomNumberGenerator .nextInt ( 500 ) ;
            generatorClone = generatorPolynomial . clone ( ) ;
            generatorClone . shift ( shift2  ) ;
            accumulator.xorBitSet ( generatorClone )  ;
         }

////            int shift = randomNumberGenerator .nextInt ( generatorClone .getJNWLength ( ) - 1  ) + 1  ;
         remainderAdded = new JNWBitSet ( generatorPolynomial .getJNWLength ( ) - 1 ) ;
         ///           remainderAdded = new JNWBitSet ( shift ) ;
////            if ( shift > generatorPolynomial .getJNWLength ( ) ) {
////                generatorClone . shift ( shift - generatorPolynomial . getJNWLength ( ) -1  ) ;
////            }
////System.out.println ("remainderAdded = " + remainderAdded.toString() ) ;
////System.out.println ("remainderAdded length = " + remainderAdded.length() ) ;
////System.out.println ("remainderAdded JNWLength = " + remainderAdded.getJNWLength() ) ;
         for ( int j = 0 ; j < 50 ; j++ )
         {
            int toFlip = randomNumberGenerator .nextInt ( generatorPolynomial .getJNWLength ( ) - 1 ) ;
////System.out.println ( "flipping bit = " + toFlip ) ;
            remainderAdded . flip ( toFlip ) ;
         }
         accumulator . append ( remainderAdded ) ;
         accumulator . appendFCS ( generatorPolynomial ) ;
////         accumulator = new JNWBitSet ( " 1101011011 " );
////System.out.println ("remainderAdded = " + remainderAdded.toString() ) ;
         ////        JNWBitSet computedReaminer = accumulator . computeRemainder ( generatorPolynomial ) ;
////System.out.println ("computed remainder = " + computedReaminer.toString() ) ;
         System.out.println ("\nCheck CRC = " + accumulator .verifyFCS ( generatorPolynomial  )  ) ;
         for ( int j = 0 ; j < accumulator . getJNWLength ( ) ; j++ )
         {
            accumulator. flip ( j ) ;
            if ( accumulator .verifyFCS ( generatorPolynomial  ) == true )
            {
               System.out.println ("ERROR in CRC computation: bit set = " + accumulator.toString ()
               + "\n for bit " + j )  ;
               accumulator. flip ( j ) ;
            }
         }
////         remainderAdded . flip ( 8 ) ;
////         computedReaminer = accumulator . computeRemainder ( generatorPolynomial ) ;
////System.out.println ("remainder comparisons = " + remainderAdded .compare ( computedReaminer ) ) ;

////System.out.println ("remainderAdded length = " + remainderAdded.length() ) ;
////System.out.println ("remainderAdded JNWLength = " + remainderAdded.getJNWLength() ) ;
////         remainderComputed = accumulator . computeRemainder ( generatorPolynomial ) ;
////System.out.println ("\n\nremainder added to accumulator = " + remainderAdded.toString() ) ;
////       accumulator. appendCRC ( generatorPolynomial ) ;
////System.out.println ("Computed packet = " + accumulator .toString () ) ;
////System.exit (  1 ) ;
////System.out.println ("Compare remainder and computed remained = " + accumulator  . computeRemainder ( generatorPolynomial) . compare ( generatorPolynomial )) ;
////         System.out.println (" Verify = " + accumulator . verifyCRC ( generatorPolynomial )  ) ;
////System.out.println ("Computed remainder = " + accumulator  .computeRemainder ( generatorPolynomial)  .toString () ) ;
////         accumulator . flip ( 0 ) ;
//// System.out.println ("Computed remainder = " + accumulator  .computeRemainder ( generatorPolynomial) .toString ()) ;
////         System.out.println (" Verify = " + accumulator. verifyCRC ( generatorPolynomial ) ) ;
////         System.out.println ("compare= " + remainderComputed . compare ( remainderAdded ) ) ;
////          if ( remainderComputed . compare ( remainderAdded ) != 0 ) {
////        System.out.println ( "\nremainderAdded = " + remainderAdded .toString()
////           + "\nremainder= " + remainderremainderComputed .toString()
         ////           + "\n   comparison = " + remainderAdded . compare ( remainder   ) );
////            + "\n   comparison = " + remainder. compare ( remainderAdded    ) );  }
      }  // end of  for ( int i = 0 ; i < 1 ; i++ )
      System.out.println ( "FINISHED");
      System.exit (  0 ) ;





      long longNumber = 1 ;
      for ( int i = 0 ; i < 63 ; i++ )
      {
         JNWBitSet test1 = new JNWBitSet ( ) ;
         test1.setJNWBitSetToLongAndTruncate ( longNumber , 64 ) ;

         JNWBitSet test2 = new JNWBitSet ( longNumber , 64 ) ;
         longNumber = 2 * longNumber ;

         System.out.println (" Compare = " + test1.compare ( test2 ) ) ;
      }
      System.exit (1 );


      String string2 = "        JNWBitSet number = new JNWBitSet ( ) "
         + " JNWBitSet number2 = new JNWBitSet ( ) ; "
         + " long numberLong = 1 ; "
         + " long Number ;"
         + " long Number2 ;"
         + " for ( int i = 0 ; i < 63 ; i++ ) {"
         + " number.setJNWBitSetToLongAndTruncate( numberLong , 63 ) ;"
         + " number2.setJNWBitSetToLongAndTruncate( numberLong , 63 ) ; "
         + " Number = number.convertJNWBitSetToLong() ;"
         + " Number2 = number2.convertJNWBitSetToLong() ;"
         + " if ( Number != Number2 | numberLong != Number2 | Number != numberLong ) {"
////        + "     System.out.println ("ERROR in convertJNWBitSetToLong2 for number " + numberLong ) "
         + "  }"
         + "         numberLong = 2 * numberLong ;"
         + " }";

      long number5 = 1024 ;
////      JNWBitSet bitSet = new JNWBitSet ( ) ;
      bitSet.setJNWBitSetToString ( string2 ) ;
      string2 = bitSet.convertJNWBitSetToString ();
      System.out.println ( " string2 = \n" + string2 ) ;

      System.exit ( 0 ) ;
////        bitSet = new JNWBitSet (
////        " 10000000 00000000 00000000 00000000"
////              +  "00000000 00000000 00000000 00000000") ;
      byte [ ] bytes = bitSet.getBytes () ;

      for ( int i = 0 ; i < bytes.length ; i++ )
      {
         System.out.println ( "Byte " + i + " =  " + bytes [ i ]  );
      }
      ////   }

      System.exit ( 0 ) ;
      ////       for ( long i = 0 ; i <= number5 ; i++ ) {
////        bitSet.setJNWBitSetToLongAndTruncate( i - 10 , 16 ) ;
      System.out.println ( "Number  is bit set " + bitSet.convertJNWBitSetToLong () );
      ////   }
      System.exit ( 0 ) ;














      JNWBitSet number = new JNWBitSet ( ) ;
      JNWBitSet number2 = new JNWBitSet ( ) ;
      long numberLong = 1 ;
      long Number ;
      long Number2 ;
      for ( int i = 0 ; i < 63 ; i++ )
      {
         number.setJNWBitSetToLongAndTruncate ( numberLong , 63 ) ;
         number2.setJNWBitSetToLongAndTruncate ( numberLong , 63 ) ;
         Number = number.convertJNWBitSetToLong () ;
         Number2 = number2.convertJNWBitSetToLong () ;
         if ( Number != Number2 | numberLong != Number2 | Number != numberLong )
         {
            System.out.println ("ERROR in convertJNWBitSetToLong2 for number " + numberLong ) ;
         }
         numberLong = 2 * numberLong ;
      }
      System.out.println ( "NO ERRORS") ;
      System.exit ( 0 ) ;







////  Utility . TCPComparator  TCPComparator = new Utility.TCPComparator ( ) ;
////  Utility.TCPComparator  TCPComparator = new Utility.TCPComparator ( ) ;
      java.util.PriorityQueue priorityQueue = new java.util.PriorityQueue ( 10 /* , TCPComparator */)  ;
      java.util.LinkedList linkedList = new java.util.LinkedList ( ) ;
      JNWBitSet data2 = new JNWBitSet ( ) ;
      JNWBitSet data3 = new JNWBitSet ( ) ;
      for ( int i = 0 ; i < 100 ; i++ )
      {
         data2 = new JNWBitSet ( ) ;
         data2.setJNWBitSetToLongAndTruncate (  i , 55 ) ;
         priorityQueue.add ( data2 ) ;
         data3 = data2.clone ( ) ;
         linkedList . add ( data3 ) ;
      }

      for ( int i = 0 ; i < 100 ; i++ )
      {
         data2 = ( JNWBitSet ) linkedList.poll ( ) ;
         int j = ( int ) data2.convertJNWBitSetToLong ( ) ;
         System.out.println ( i + "th bit set polled from linked list has value " + j ) ;
      }
      System.out.println ( 5 ) ;


      for ( int i = 0 ; i < 100 ; i++ )
      {
         data2 = ( JNWBitSet ) priorityQueue.poll ( ) ;
         int j = ( int ) data2.convertJNWBitSetToLong ( ) ;
         System.out.println ( i + "th bit set polled from prioirty queue has value " + j ) ;
      }


      System.exit ( 1 ) ;
      JNWBitSet CRC_32 = new JNWBitSet (
         "1 00000100 11000001 00001101 10110111" ) ;

      String string = "00000000 00010111 00000000 00010111 0000000000000000 00000000 00000000 00000000 0000000000000000 00000000 01010000 00000000 0000000011001000 00000000 00000000 00000000 0000000001000001 01000010 01000011 01000100 0100010100110110 00110111 00111000 00111001 0011000001000001 01000010 01000011 01000100 0100010100110110 00110111 00111000 00111001 0011000001000001 01000010 01000011 01000100 0100010100110110 00110111 00111000 00111001 0011000001000001 01000010 01000011 01000100 0100010100110110 00110111 00111000 00111001 00110000" ;
      string = "00000000 00000000 11111111 11111111" ;
      JNWBitSet data = new JNWBitSet ( string ) ;
      data = CRC_32. clone ( ) ;
      System.out.println ( "The data length = " + data.JNWLength ) ;
      System.out.println ( "The data = " + data.toString () ) ;
      data.shift ( 3 ) ;
      data . flip ( 5 ) ;
      data.appendFCS ( CRC_32 ) ;
      System.out.println ( "The data length = " + data.JNWLength ) ;
      System.out.println ( "The data = " + data.toString () ) ;

      System.out.println ( "Verify = " + data.verifyFCS ( CRC_32 ) ) ;
      for ( int i = 0 ; i < data.getJNWLength ( ) ; i++ )
      {
         data . flip ( i ) ;
         System.out.println ( "Verify = " + data.verifyFCS ( CRC_32 ) ) ;
         data . flip ( i ) ;
      }


      System.exit ( 1 ) ;


      JNWBitSet subset = data.getSubset ( 15 , 37 ) ;
      System.out.println ( "The data length = " + data.JNWLength ) ;
      System.out.println ( "The data = " + data.toString () ) ;
      System.out.println ( "The subset JNWLength = " + subset.JNWLength ) ;
      System.out.println ( "The subset length = " + subset.length () ) ;
      System.out.println ( "The subset = " + subset . toString () ) ;

      System.exit ( 1 ) ;


      JNWBitSet CRC = new JNWBitSet ( 16 ) ;
      System.out.println ( "The data length = " + data.JNWLength ) ;
      System.out.println ( "The data = " + data.toString () ) ;
      System.out.println ( "The CRC length = " + CRC.JNWLength ) ;
      System.out.println ( "The CRC = " + CRC.toString () ) ;
////            CRC = data.xorAndShift ( data, CRC ) ;
      System.out.println ( "The CRC length = " + CRC.JNWLength ) ;
      System.out.println ( "The CRC = " + CRC.toString () ) ;
/*
////        JNWBitSet xor = new JNWBitSet ( "1111 00" ) ;
////        JNWBitSet xor2 = xor. clone ( ) ;System.out.println ( "dataTotal = " + dataTotal .toString() )  ;
        dataTotal.invertBitSet ( ) ;
System.out.println ( "dataTotal = " + dataTotal .toString() )  ;
System.exit ( 1 ) ;
////System.out.println ( "resultant XOR = " + dataTotal.xorAndShift( xor ) .toString() )  ;
////System.out.println ( "resultant XOR2 = " + dataTotal.xorAndShift2( xor2 ) .toString() )  ;
System.out.println ( "dataTotal = " + ( dataTotal ). toString() )  ;
////System.out.println ( "xor = " + ( xor ). toString() )  ;
System.exit ( 1 ) ;
        JNWBitSet clone = dataTotal ;
System.out.println ( "dataTotal = " + ( dataTotal ). toString() )  ;
System.out.println ( "dataTotal = " + ( clone ). toString() )  ;
        clone . flip ( 8 ) ;
System.out.println ( "dataTotal = " + ( dataTotal ). toString() )  ;
        clone = dataTotal.clone( );
////        clone = new JNWBitSet ( dataTotal ) ;
        clone . flip ( 8 ) ;
        clone . flip ( 9 ) ;
                clone . flip ( 10 ) ;
                        clone . flip ( 11 ) ;
                                clone . flip ( 12 ) ;
                                        clone . flip ( 13 ) ;
System.out.println ( "dataTotal = " + ( clone ). toString() )  ;
 System.exit (1);
System.out.println ( "Original data = " + ( ( Utility.JNWBitSet) dataTotal ). toString() )  ;
        java.util.BitSet data7 = new JNWBitSet ( "11111111 00000000 11111111 00000000 00111100" ) ;
        Utility.JNWBitSet data = new JNWBitSet ( "10101111 00000000  11111111" ) ;
        dataTotal.xor ( data7 ) ;
        int fromIndex = 28 ;
        int toIndex = 32 ;
////        dataTotal . setSubset ( fromIndex , data ) ;
System.out.println ( "altered data = " + ( ( Utility.JNWBitSet) dataTotal ).toString() )  ;
        System.exit ( 0 ) ;


        String string = "Now it the time for all good me to come to the aid of their country.\n" ;
        string = string + string ;
        string = string + string ;
        string = string + string ;
        string = string + string ;
System.out.println ( "\"" + string + "\"" ) ;
        JNWBitSet originalData = new JNWBitSet (  ) ;
         originalData .setJNWBitSetToString( string )  ;
        System.out.println ( "Original = " + originalData.toString(  )) ;
        System.out.println ( "Lowest N bits = " + originalData.writeLowestNBitsOfBitSet( 240 )) ;
        System.exit ( 0 ) ;



////        JNWBitSet data2 = null ;
        final JNWBitSet originalGeneratorPolynomial = new JNWBitSet (
                "1 00010000 00100001") ;
////System.out.println ( "data length = " + data ) ;
        long stop = 0 ;
        int limit = 1 ;
        int numberOfIterations = 2 ;
        long start = System.nanoTime() ;
        for ( int i = 0 ; i < numberOfIterations ; i++ ) {
 ////           data = new JNWBitSet ( originalData ) ;
////            length = data .getJNWLength() ;
////            if ( i == 0 ) {
////                System.out.println ("Data length = " + length ) ;
////                System.out.println ("Data without CRC = " + data.toString() ) ;
////            }
        }
        long loopTime  = System.nanoTime()  - start ;
        data = new JNWBitSet ( originalData ) ;
System.out.println ( "\n\nInefficient algorithm CRC = " + data.computeCRC( originalGeneratorPolynomial ).toString ( )  ) ;
        start = System.nanoTime() ;
        for ( int i = 0 ; i < numberOfIterations; i++ ) {
            data = new JNWBitSet ( originalData ) ;
////            length = data .getJNWLength() ;
            data.appendCRC( originalGeneratorPolynomial ) ;
////            if ( i == 0 ) {
////                System.out.println ("Data length = " + length ) ;
////                System.out.println ("Data and CRC = " + data.toString() ) ;
////            }
        }
        stop = System.nanoTime() ;
        long inefficientAlgorithmTime = stop - start - loopTime;


        data = new JNWBitSet ( originalData ) ;
System.out.println ( "Efficient algorithm CRC = " + data.computeCRCWithoutShifts( originalGeneratorPolynomial ) .toString ( ) ) ;
        start = System.nanoTime() ;
        for ( int i = 0 ; i < numberOfIterations; i++ ) {
            data = new JNWBitSet ( originalData ) ;
////        int    length = data .getJNWLength() ;
            data.appendCRCWithoutShifts( originalGeneratorPolynomial ) ;
             if ( i == 0 ) {
 ////               data.flip ( 842 ) ;
 ////               JNWBitSet remainder = data.computeRemainderWithoutShifts( originalGeneratorPolynomial ) ;
                if ( data.verifyCRCWithoutShifts( originalGeneratorPolynomial )) { System.out.println ("NO ERROR IN TRANSMISSION") ; }
                else { System.out.println ("ERROR IN TRANSMISSION") ; }
            }
////               System.out.println ("Data length = " + length ) ;
////                System.out.println ("Data and CRC = " + data.toString() ) ;
////            }
        }
        stop = System.nanoTime() ;
        long efficientAlgorithmTime = stop - start - loopTime ;
        double ratio = (double) efficientAlgorithmTime / (double ) inefficientAlgorithmTime ;
System.out.println (" loop time = " + loopTime + " nanoseconds \n"
     + " inefficientAlgorithmTime = " + inefficientAlgorithmTime + " nanoseconds \n"
     + " efficientAlgorithmTime = " + efficientAlgorithmTime+ " nanoseconds\n"
     + " ratio = " + ratio
     + "\n speed up = " + ( 1.0/ ratio ) );
System.exit ( 1 ) ;




     JNWBitSet generatorPolynomial = new JNWBitSet (  originalGeneratorPolynomial ) ;
 ////       JNWBitSet data = new JNWBitSet ( originalData ) ;
        data.setJNWLength( 1000 );
        data.appendCRCWithoutShifts( generatorPolynomial )  ;
 System.out.println ("data plus CRC = " + data.toString() )  ;
 ////       dataWithCRC.flip ( 20 ) ;
        JNWBitSet resultCRC = data.computeCRCWithoutShifts( originalGeneratorPolynomial ) ;
  System.out.println ("CRC = " + resultCRC . toString() )  ;
  boolean dataOK = data.verifyCRC ( generatorPolynomial ) ;
  System.out.println ("The data passed error free = " + dataOK )  ;
System.exit ( 1 ) ;
////        data = new JNWBitSet ( dataOriginal ) ;
        data . appendCRC ( generatorPolynomial )  ;
 System.out.println ("data with CRC = " + data . toString() )  ;
        generatorPolynomial = new JNWBitSet (  originalGeneratorPolynomial ) ;
 ////       data = new JNWBitSet ( dataOriginal ) ;
////         generatorPolynomial = new JNWBitSet ( "11 000000000000 101"  ) ;
        JNWBitSet result2 = data . computeCRCWithoutShifts ( generatorPolynomial)  ;
  System.out.println ("result = " + result2 . toString() )  ;

System.exit ( 1 ) ;



/*


System.out.println ( new java.lang.Long( 1 ) . MIN_VALUE) ;
System.out.println ( new java.lang.Long( 1 ) . MAX_VALUE) ;
        JNWBitSet dataBitSet = new JNWBitSet (
                "11111111 11101111 10101010 11110111 11100011") ;
        JNWBitSet patternBitSet = new JNWBitSet (
                "11110111111") ;
System.out.println ( "unstuffed bit set = " + dataBitSet  .toString ( ) ) ;
System.out.println ( "patternBitSet = " + patternBitSet  .toString ( ) ) ;
System.out.println ( "The pattern occurs at position = "
        + dataBitSet .findPattern ( patternBitSet ) ) ;
System.exit ( 1 ) ;
////        dataBitSet2 = dataBitSet2 .stuff( ) ;
////System.out.println ( "stuffed bit set = " + dataBitSet2  .toString ( ) ) ;
////        dataBitSet2 = dataBitSet2 .unstuff( ) ;
////System.out.println ( "unstuffed bit set = " + dataBitSet2  .toString ( ) ) ;

System.exit ( 1 ) ;
JNWBitSet dataBitSet3 = null ;
 ////       JNWBitSet dataBitSet = new JNWBitSet ( dataBitSet2 ) ; ;
        dataBitSet3 = dataBitSet.shift ( 3 ) ;
System.out.println ( "shifted bit set = " + dataBitSet3  .toString ( ) ) ;
////        dataBitSet = new JNWBitSet ( dataBitSet2 ) ;
        dataBitSet3 = dataBitSet.shift ( -3 ) ;
System.out.println ( "shifted bit set = " + dataBitSet3 .toString ( ) ) ;
        System.exit ( 1 ) ;
////        dataBitSet.flip( 0 , 8 ) ;
System.out.println ( "dataBitSet.JNWLength = " + dataBitSet.JNWLength ) ;
        dataBitSet.flip( dataBitSet.JNWLength - 8  , dataBitSet.JNWLength  ) ;
System.out.println ( "Original bit set = " + dataBitSet .toString ( ) ) ;
////System.out.println ( "Clone bit set = " + cloneBitSet  .toString ( ) ) ;
        System.exit ( 1 ) ;


dataBitSet.setJNWLength( 8 ) ;
System.out.println ( "Bitset cardinality = " + dataBitSet .cardinality ( ) ) ;
        System.exit ( 1 ) ;

        System.out.println ( " Starting CRC " ) ;
 ////       JNWBitSet dataBitSet = new JNWBitSet (  ) ;
////        int number = 1024 ;
        for ( int i = 0 ; i <= 1024 ; i++) {
        dataBitSet.setJNWBitSetToLongAndTruncate( i, 8) ;
        System.out.println ( "JNWBitSet representation for  dataBitSet "
           + i + " is " + dataBitSet . toString( ) + " and backward conversion is "
           + dataBitSet . convertJNWBitSetToLong ( ) ) ;
        }

 */
/*        JNWBitSet CRC = new JNWBitSet (  ) ;
        CRC.setJNWBitSetToZerosAndOnes( "1 10000000 00000101" ) ;
        JNWBitSet zeros = new JNWBitSet ( ) ;
        zeros.setJNWBitSetToZerosAndOnes( " 00000000 00000000 " ) ;
        dataBitSet.setJNWBitSetToString( string ) ;
        System.out.println ( "CRC = "
           + CRC.toString()) ;
        System.out.println ( "zeros = "
           + zeros.toString()) ;
        System.out.println ( "BitSet for String " + string + " is: "
           + dataBitSet.toString()) ;
        dataBitSet.append( zeros ) ;
        System.out.println ( "BitSet for String " + string + " is now: "
           + dataBitSet.toString()) ;
       JNWBitSet output = new JNWBitSet (  ) ;
       output = dataBitSet.computeCRC( CRC );
        System.out.println ( "CRC for String " + string + " is: "
           + output.toString( )) ;

/*
        int i ;
        for ( int j = 0 ; j < 100 ; j++ ) {
            stringBuilder = new java.lang.StringBuilder ( ) ;
        for (  i = 0 ; i < 40 ; i++ ) {
            if ( i %8 == 0 ) { stringBuilder.append ( " " ) ; }
            if ( random.nextBoolean() ) { stringBuilder.append ( "0" ) ; }
            else { stringBuilder.append ( "1" ) ; }
         }  // end of for ( int i = 0 ; i < 24 ; i++ )

        string = new String (stringBuilder  ) ;
////        string = "01010101 10101010" ;
////        System.out.println ( "\n\nString =         " + string ) ;
        dataBitSet.setToZerosAndOnes( string ) ;
        System.out.println ( "\n length = " + i + " JNWLength = " + dataBitSet.getJNWLength()) ;
        System.out.println (dataBitSet . toString ( ) ) ;
        byte [ ] bytes = dataBitSet .getBytes() ;
        System.out.println (dataBitSet . writeLowestNBitsOfBitSet ( 48 ) ) ;
        for ( int k = 0 ; k < bytes.length ; k++ ) {
        System.out.print ( bytes [ k ] + "     " ) ;
        }
        System.out.println ( ) ;
 */
/*
        String inputString = "" ;
        while ( ! inputString . equals ( "END" ) ) {
        try {
      java.io.BufferedReader bufferedReader = new
         java.io.BufferedReader ( new java.io.InputStreamReader (
            System .in ) ) ;
            inputString = bufferedReader . readLine ( ) ;
            dataBitSet.setToString ( inputString ) ;
            String outputString = dataBitSet .convertToString() ;
            System.out.println ( outputString ) ;
        } catch ( Exception e ) {
        System.err.println ( " ERROR IN MainGUI_Actions.ReadTopologyFileAction.actionPerformed:"
            + "\nException "+ e. getClass ( ) . getName ( ) + " thrown while trying to "
            + " read line \"" + inputString + "\" in input topology file"
            + "\n    Data should be in format\n\"SOURCE NODE , DESTINATION NODE , "
            + "DATA RATE, BIT ERROR RATE , TRANSMISSION DELAY\"" ) ;
        }
 */
////        int shift = random.nextInt ( 10 ) ;
////        if ( random.nextBoolean()) { shift = - shift ;  }
////        System.out.println ( " shift = " + shift ) ;
////        dataBitSet.shift ( shift ) ;
////        System.out.println (dataBitSet . toString ( ) ) ;
/////        }  // end of for ( int j = 0 ; j < 10 ; j++ )


/*        JNWBitSet dataBitSet = new JNWBitSet ( 55 ) ;
        dataBitSet = new JNWBitSet ("111 00000001 00000010 00000100 10000000" 00000010 00000100 00001000 00010000 00100000 01000000 10000000") ;
        byte byteArray [ ] = dataBitSet.getBytes() ;
        System.out.println ( "BYTE ARRAY = ") ;
        for ( int i = 0 ; i < byteArray. length ; i++ ) {
            System.out.println ( " byte = " + byteArray [ i ] ) ;
        }
        dataBitSet. setToString ( "ABCDEFG " ) ;
///;/        String outputString = dataBitSet.convertToString ( ) ;
        System.out.println ( "OUTPUT = " + dataBitSet.convertToString ( ) ) ;
 */
/*
        String string = "ABCdefgHIJklmnopqrstuv" ;
        string = "ABC" ;
        dataBitSet .setToString ( string ) ;
        System.out.println ( "Original bit set = " + dataBitSet . toString ( ) + " with length = " +  dataBitSet.getJNWLength( ) ) ;
        dataBitSet.append( new JNWBitSet ( ( "01"))) ;
        System.out.println ( "Appended bit set = " + dataBitSet . toString ( ) + " with length = " +  dataBitSet.getJNWLength( ) ) ;
        System.out.println ( "Original String = " + string + "\ndata bit set = " + dataBitSet.toString()
          + "\ndata bit set as String = "  + dataBitSet . toString ( )  ) ;
        System.exit ( 1 ) ;

System.out.println ( "\nWrite 55 0s");
System.out.println ( "Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet . toString ( )) ;
dataBitSet.setToString( "01234567890");
System.out.println ( "\nWrite 01234567890");
System.out.println ( "Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet . toString ( )) ;
        dataBitSet . clear ( ) ;
System.out.println ( "\nWrite all 0s");
System.out.println ( "Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet . toString ( )) ;
        dataBitSet . setToZerosAndOnes
                ("0011 0001 0011 0010 0011 0011 " ) ;
System.out.println ( "\nWrite 0011 0001 0011 0010 0011 0011");
System.out.println ( "Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet . toString ( )) ;
 dataBitSet .append( new JNWBitSet ( " 1111 1111 0000 0000 1111 1111 0000 0000 1111 1111 0000 0000")) ;
System.out.println ( "\nAppend 11111111 00000000 and write 55 bits");
System.out.println ( "Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet.writeLowestNBitsOfBitSet ( 55 ) ) ;
System.out.println ( "\nWrite all bits");
System.out.println ( "Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet . toString ( )) ;
 dataBitSet .setNWBLength( ( 55)) ;
System.out.println ( "Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet . toString ( )) ;
        dataBitSet . setToZerosAndOnes
                ("0000 0000 0011 0010 0000 0000 " ) ;
System.out.println ( "Original Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet . toString ( )) ;
dataBitSet .append( new JNWBitSet ( "01111110" ) ) ;
dataBitSet .prepend( new JNWBitSet ( "01111110" ) ) ;
JNWBitSet newBitSet = ( JNWBitSet ) dataBitSet . clone ( ) ;
System.out.println ( "Apended and prepended Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet . toString ( )) ;
    newBitSet .setNWBLength( 27 ) ;
 System.out.println ( "\n\n\nSet Data set length = " + newBitSet.getJNWLength( )
   + " Data set = " + newBitSet . toString ( )) ;
for ( int i = 0 ; i < 16 ; i++ ) {
    dataBitSet . shift ( i ) ;
System.out.println ( "Shifted Data set length = " + dataBitSet.getJNWLength( )
   + " Data set = " + dataBitSet . toString ( )) ;
}
for ( int i = 0 ; i < 16 ; i++ ) {
    newBitSet . shift ( -i ) ;
System.out.println ( "Shifted Data set length = " + newBitSet.getJNWLength( )
   + " Data set = " + newBitSet . toString ( )) ;
 */
   } // end of main
}


/**Overrides the default toString methods of BitSet to return the BitSet as an ASCII string.  If
 * the JNWBitSet is not a multiple of 8 bits long, it is 0 padded at the end.  This is similar
 * to convertToString ( ).  This method has been replaced by convertToAString ( ) */
/*     public String toString ( ) {
////System.out.println("In toString: ")  ;
        JNWBitSet bitSetToConvert = ( JNWBitSet ) this . clone ( ) ;
        int byteArrayLength = bitSetToConvert .JNWLength / 8 ;
////System.out.println("In toString: byteArrayLength = " + byteArrayLength )  ;
// In case the bit set is not a multiple of 8 bits long  assume zeros fill the bit set.
        int shiftSize = bitSetToConvert .JNWLength % 8 ;
        if ( ! ( shiftSize == 0 ) ) {
            byteArrayLength = byteArrayLength + 1 ;
////System.out.println("In toString: shiftSize = " + shiftSize )  ;
            bitSetToConvert . shift ( shiftSize ) ; }
////System.out.println ( "Shifted bit set = " + bitSetToConvert . toString ( ) + " with length = " +  bitSetToConvert . getJNWLength( ) ) ;
        // remainder bits are zero padded
        byte [ ] byteArray  = new byte [ byteArrayLength ] ;
        for ( int i = 0 ; i < byteArrayLength  ; i++ ) {
           int dummyint = bitSetToConvert .JNWLength - 8 * i ;
////System.out.println("In toString: dummyint value = " + dummyint )  ;
           byteArray  [ i ] =  ( byte ) (
                128 * BitValue ( bitSetToConvert . get ( dummyint - 1 ) )
              +  64 * BitValue ( bitSetToConvert . get ( dummyint - 2 ) )
              +  32 * BitValue ( bitSetToConvert . get ( dummyint - 3 ) )
              +  16 * BitValue ( bitSetToConvert . get ( dummyint - 4 ) )
              +   8 * BitValue ( bitSetToConvert . get ( dummyint - 5 ) )
              +   4 * BitValue ( bitSetToConvert . get ( dummyint - 6 ) )
              +   2 * BitValue ( bitSetToConvert . get ( dummyint - 7 ) )
              +       BitValue ( bitSetToConvert . get ( dummyint - 8 ) ) );
////System.out.println("In toString: integer value = " + integerValue )  ;
       }
        return new String ( byteArray ) ;
     }  // end of toString
 */


////     private static int BitValue ( boolean value ) {
////         if ( value == true ) {
////             return   1 ; }
////         return  0 ;
////     }



// FROM findPattern

//// System.out.println ( "\nI = " + i ) ;
//// System.out.println ( "Start search at = "  + " pattern index = " )
////for ( int j = patternBitSet . getJNWLength( ) - 1 ; j >= 0 ; j-- ){

////System.out.println ( "Index for this = " + ( i + j + ( patternBitSet . getJNWLength( ) - 1  ) ) ) ;
////            }
////             if ( ! ( this.get( i  + j - patternBitSet . getJNWLength( ) + 1 )
////System.out.println ( "this . get ( i - j + patternBitSet . getJNWLength( )  ) = " + this . get ( i + j - patternBitSet . getJNWLength( )  )  ) ;
////System.out.println ( "patternBitSet . get( j ) = " + patternBitSet . get( j )  ) ;















//// ALL THE SHIFT REGISTER METHODS



   /**Computes the CRC of a frame and returns a boolean indicating a
    *whether the CRC indicates a transmission error.  True means that the
    * CRC is correct, and false indicates a bit error somewhere.*/
////   public boolean verifyFCSUsingShiftRegister ( JNWBitSet generatorPolynomial )
////   {  // If the CRC is not all 0s, there is an error
////      if ( this . computeFCSUsingShiftRegister ( generatorPolynomial ) . cardinality ( ) > 0 )
////      {
////         return false ;
////      }
////      else
////      { return true ; }
////   }  // end of verifyCRC


   /**Computes the CRC of a frame and returns a boolean indicating a
    *whether the CRC indicates a transmission error.  True means that the
    * CRC is correct, and false indicates a bit error somewhere.*/
////   public boolean verifyFCSUsingShiftRegister_Student ( JNWBitSet generatorPolynomial )
////   {  // If the CRC is not all 0s, there is an error
 ////     return true ;
////   }  // end of verifyCRC



   /**Appends a cyclic redundancy check (CRC) frame check sequence (FCS)
    * to the end of this JNWBitSet using a JNWBitSet to represent the CRC.*/
/*   public void appendFCSUsingShiftRegister ( JNWBitSet generatorPolynomial )
   {
      int length = generatorPolynomial .getJNWLength () - 1 ;
      JNWBitSet clone = new JNWBitSet ( this ) ;
      JNWBitSet CRCResult  = clone . computeFCSUsingShiftRegister ( generatorPolynomial ) ;
      // shift 0s into the bit set
      this . shift ( length ) ;
      // Copy the crc into the lowest bits of this
      for ( int i = 0 ; i < CRCResult .getJNWLength ( ) ; i++ )
      {
         this . set ( i , CRCResult . get ( i )) ;
      }  /// end of for
   }  // end of appendCRC( JNWBitSet crcCode )

      /**Computes the CRC Cyclical redundancy Code for a data BitSet with any CRC
    * polynomial.  CRCPolynomialIn is the generator polynomial for the CRC.*/
/*   public JNWBitSet computeFCSUsingShiftRegister ( JNWBitSet CRCPolynomialIn  )
   {  // This is the data set
      JNWBitSet dataSet =  new JNWBitSet ( this ) ;
      // Find the length of the generator polynomial.  Use length, not
      // JNWLength since the gnnerator polynomial could have leading
      // zeros, for example " 00000001 00010000 00100001 "
      dataSet . shift ( CRCPolynomialIn.length ( ) - 1) ;
      // Create a shift register for computing the CRC
      JNWBitSet shiftRegister = new JNWBitSet (  ) ;
      boolean feedbackBit ;
      // Note that the feedback bit is contained in location
      // CRCpolynomial . length ( ) - 2 not CRCpolynomial . length ( ) - 1
      // because the bit is sampled BEFORE because the bit is sampled
      // BEFORE the shift register is shifted.
      int feedbackBitLocation = CRCPolynomialIn . getJNWLength ( ) - 1  ;
      // The length of the dataSet changes during the processing
      int lengthOfdataSet = dataSet . getJNWLength ( ) ;
      // i runs over all the data set
      for ( int i = 0 ; i < lengthOfdataSet  ; i++ )
      {
         shiftRegister . shift ( 1 ) ;  // Shift the shift register up one bit
         // Get the feedback bit
         feedbackBit = shiftRegister . get ( feedbackBitLocation ) ;
         // Set the lowest shift regiter bit to the topmost, unprocessed
         // data bit
         shiftRegister . set ( 0 , dataSet . get ( lengthOfdataSet - 1  - i ) ) ;
         // If the feedback bit is 1
         if ( feedbackBit == true )
         {
            // XOR the bits corresponsing to the CRC's polynomial coefficients
            // which are 1
            shiftRegister . xor ( CRCPolynomialIn )  ;
            // Flip the top-most bit of the CRC to its original value
            shiftRegister  .set ( CRCPolynomialIn . getJNWLength ( ) ) ;
         }  // end of if
         if ( lengthOfdataSet - 2  - i >= 0 )
         {  ////WHY IS THIS HERE 20071205
////System.out.println ( " next input = " + dataSet . getBitValue ( lengthOfdataSet - 2  - i ) ) ;
         }
      }  // end of for ( int i = 0 ; i < dataSet . length ( ) ; i++ )
      /** Create return GMUBitSet*/
 /*     JNWBitSet CRC = new JNWBitSet (  ) ;
      for ( int i = 0 ; i < CRCPolynomialIn . getJNWLength ( ) - 1 ; i++ )
      {
         CRC . set ( i , shiftRegister . get ( i ) ) ;
      }
      return  CRC ;
   } // end of computeCRC
  */