/*
 * JNWBuffer.java
 *
 * Created on June 1, 2007, 1:26 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package Utility;

////import Utility.JNWBitSet ;

/** Creates a FIFO buffer for use in storing frames, packets, etc. in the
 *interface or elsewhere.  The buffer has a finite size in bytes.  JNWBitSet
 * arriving will cause the buffer to compare the remaining memory to the
 * size of the arriving JNWBitSet.  If the new data cannot fit in the
 * buffer, it is discarded, and the data is lost.  The buffer will
 *accept ONLY String or JNEBitSets; any other class will cause an error message
 *
 *<p>Since the buffer size is saved as an integer, the maximum buffer size is
 *268 megabytes.
 *
 *Version 1.001 adds a test for a null value from the poll with
 * <li>if ( ( object =  super . poll ( )  )  == null ) return object ;</li>
 * *<p><b>Versions</b>
 *<li><b>Version 1.001</b> adds a test for a null value from the poll with
 * <li>if ( ( object =  super . poll ( )  )  == null ) return object ;</li></li>
 *<li><b>Version 1.002</b> of 20071118 adds the boolean addFirst, which if set
 * will add the Object to the head of the buffer rather than the tail.
 * <p>It also contains the method FrameOrNot ( JNWBitSet objectIn ) which
 * tests whether the bit set is an HDLC frame or an Ethernet frame.</li>
 *<li><b>Version 1.100</b> of 20071205 removes the storage capability for
 * Strings.  This capability  is never used.</li>
 * @author dixonw
 * @version 1.002
 */
public class JNWBuffer extends java.util.LinkedList 
   implements GUIConstants . General
{
   /** Current buffer size in bytes*/
   int currentBufferSizeInBytes = 0  ;
   
   /** Maximum buffer size in bytes*/
   int maximumBufferSizeInBytes  ;  
   
   /**Storage for TCP packets transmitted, but not acknowledged*/
   java.util.LinkedList TCPList = new java.util.LinkedList (  );
   
   /** Creates a new instance of JNWBuffer with a specified maximum
    * buffer size specified in bytes.*/
   public JNWBuffer ( int maximumBufferSizeInKilobytes )
   {
      super ( ) ;
      this . maximumBufferSizeInBytes = 1024 * maximumBufferSizeInKilobytes ;
   }  // end of public JNWBuffer( int maximumBufferSizeInBits )
   
   
   /**Adds either a JNWBitSet or a String to the buffer is possible,
    * otherwise, the JNWBitSet is discarded.  Bit lengths for a JNWBitSet
    * are rounded up to the next largest byte; that is, 7 bits will
    * occupy one byte.  The buffer will accept ONLY a String or a JNEBitSet;
    * any other class will cause an error message*/
   public boolean add ( JNWBitSet bitSetIn )
   {
      int length = ( bitSetIn . getJNWLength ( ) + 7 ) / 8 ;
      // is the BitSet too long?
      if ( ( currentBufferSizeInBytes + length  )
      > maximumBufferSizeInBytes )  // Is there an overflow?
      {  // Yes.  Should I print an error message?
         if ( ModelData.PrintPermissions.printBufferOverflows == true )
         {
            Utility.err.println("ERROR IN Utility.JNWBuffer.add: buffer overflow "
               + "for buffer instantiated in class "
               + this . getClass () . getName ( ) ) ;
         }   // end of if ( ModelDatas.PrintPermissions.printBufferOverflows == true )
         return false ; }  // end of if ( ( currentBufferSizeInBytes + length  )
      // Add the bit set to the buffer
      super . add ( bitSetIn ) ;
      currentBufferSizeInBytes // update the buffer length
         = currentBufferSizeInBytes + length ;
      return true ;
   }  // end of public void add ( Utilities.JNWBitSet bitSet )


   /**Remove the first JNWBitSet or String and reset the currentBufferSizeInBytes.
    * Returns null if the buffer is empty.*/
   public  Utility.JNWBitSet  poll (  )
   {
      Utility.JNWBitSet bitSet = null ;
      if ( ( bitSet =  ( Utility.JNWBitSet ) super . poll ( )  )
      == null ) return bitSet ;
      currentBufferSizeInBytes // Update the buffer size
         = currentBufferSizeInBytes // This is need since the bit set may NOT be
         // an a multiple of 8 bits long.
         - ( bitSet .getJNWLength (  ) + 7 ) / 8  ;
      return bitSet ;
   }  // end of public void add ( Utilities.JNWBitSet bitSet )


   /**Returns the current buffer length in bytes*/
   public int getCurrentBufferSizeInBytes ( )
   {
      return currentBufferSizeInBytes ;
   }


   /**Decreases the amount of data stored in the buffer.  This is used to account
    *for TCP packets stored in the transmit and receive buffers while the node 
    *waits for an ACK message or to receive a complete message.  When the model 
    * creates an AttemptToSendTCPFrame Event, the TCP frame is held in the
    *Event.  In reality, the TCP frame would be held in the transmit buffer.  The
    *fragments of the received message would be held in the receive buffer.*/
   public void DecreaseCurrentBufferSize ( int decrease )
   {
      currentBufferSizeInBytes = currentBufferSizeInBytes - decrease ;
   }


   /**Increases the amount of data stroed in the buffer.  This is used to account
    *for TCP packets stored in the transmit and receive buffers while the node 
    *waits for an ACK message or to receive a complete message.  When the model 
    * creates a AttemptToSendTCPFrame Event, the TCP frame is held in the 
    *Event.  In reallity, the TCP frame would be held in the transmit buffer.  The
    *fragments of the received message would be held in the receive buffer.*/
   public void IncreaseCurrentBufferSize ( int increase )
   {
      currentBufferSizeInBytes = currentBufferSizeInBytes + increase ;
   }

   
   
   /**For testing only.*/
   private static class TESTER
   {
      JNWBuffer testbuffer 
         = new JNWBuffer ( 1000 ) ;
      
      public boolean add ( JNWBitSet bitSetIn )
      {
         return testbuffer. add ( bitSetIn ) ;
      }
      
   }
   
   /** For testing only*/
   public static void main_Testing_Only ( String [ ] args )
   {
      JNWBuffer testbuffer = new JNWBuffer ( 1000 ) ;
      
      JNWBitSet data = new JNWBitSet ( "00000001 11111111 00000000 11111111 01010101") ;
      testbuffer .add ( data ) ;
      
      ////       System.out.println("Data is a frame? = " + FrameOrNot ( data ) );
      System.exit ( 0 ) ;
      
      ModelData.PrintPermissions.printBufferOverflows = true ;
      String string = "0123456789" ;
      JNWBuffer buffer = new JNWBuffer ( 1000 ) ;
      JNWBitSet bitSet = new  JNWBitSet ( 800  ) ;
      TESTER tester = new TESTER ( ) ;
////        bitSet . setJNWBitSetToString ( string + string ) ;
      int i = 0 ;
      
      while ( i < 20)
      {
         tester. add ( bitSet . clone ( )) ;
////        buffer . add ( string ) ;
         System.out.println ("Buffer length after adding bit set = " + tester . testbuffer . getCurrentBufferSizeInBytes ( )
         + " bytes" ) ;
         ////       buffer . add ( bitSet ) ;
////        System.out.println("Buffer after adding BitSet = " + buffer . getCurrentBufferSizeInBytes ( )
////        + " bytes" ) ;
////        bitSet. append ( bitSet ) ;
         i = i + 1 ;
      }
      System.out.println ( "\n\nWRITING BUFFER ") ;
      while ( buffer.size ( ) > 0 )
      {
         Object object = buffer . poll ( ) ;
         System.out.println ("Buffer length = " + buffer . getCurrentBufferSizeInBytes ( )
         + " object = " + object .toString ( ) ) ;
      }
   }  // end of main
}  // end of public class JNWBuffer



//////20071205//    public boolean add ( Object objectIn ) {
//////20071205        Object object = bitSetIn ;
//////20071205        object = objectIn . clone ( ) ;
////System.out.println ( "IN JNWBuffer at beginning "
////+"\n Buffer size currently " + this . size( ) ) ;
////Utility.JNWBitSet bitSet = ( Utility.JNWBitSet ) object ;
////System.out.println ( "IN JNWBuffer before adding "
////+"input packet =  " + bitSet.writeBitSet ( ) ) ;
////        int length = 0 ;/*= ( bitSet . getJNWLength (  ) + 7 ) / 8 */;
//////20071205/        if ( object instanceof  String  ) {
//////20071205            length = ( ( String ) object ) .length ( ) ;
//////20071205        }
//////20071205        else if ( object instanceof  Utility.JNWBitSet  ) {
//////20071205            int length = ( ( ( Utility.JNWBitSet ) object )
//////20071205/               . getJNWLength (  ) + 7 ) / 8 ;


//////20071205/           if ( FrameOrNot ( ( Utility.JNWBitSet ) object ) == false ) {
//////20071205              int fromIndex = ( ( Utility.JNWBitSet ) object ) . getJNWLength ( ) - 8 - 3 ;
//////20071205              int fromIndex = ( ( Utility.JNWBitSet ) object ) . getJNWLength ( ) - 8 - 3 ;
//////20071205               JNWBitSet precedenceBits
//////20071205                   = ( ( Utility.JNWBitSet ) object ). getSubset( fromIndex , fromIndex + 3  ) ;
//////20071205        }  // end of else if ( object instanceof  Utility.JNWBitSet  ) {
//////20071205         else {
//////20071205            Utility.Out.err ("ERROR IN Utility.JNWBuffer.add: object "
//////20071205 /               + object .toString ( ) + " neither"
//////20071205 /               + " a String nor a Utility.JNWBitSet" + carriageReturn ) ;
//////20071205 /        }

////           System.out.println ( " this = " + this.toString ( ) ) ;
////          System.out.println ( " this . getClass()  = " + this . getClass() .toString ( ) ) ;
////           System.out.println ( " this . getClass() . getDeclaringClass()  = " + this . getClass() . getEnclosingClass() .toString ( ) ) ;
////           System.out.println ( " this . getClass() . getDeclaringClass() . getName ( )  = " + this . getClass() . getDeclaringClass() . getName ( ) .toString ( ) ) ;
////System.out.println ("ADDING OBJECT TO BUFFER")   ;
//////20071205             super . addFirst ( object ) ;  // the head of the buffer
////System.out.println ("Adding object first: object = " + ( ( Utility.JNWBitSet ) object ) .writeBitSet() ) ;
//////20071205            super . add( object ) ;
////System.out.println ("Adding object last: object = " + ( ( Utility.JNWBitSet ) object ) .writeBitSet() ) ;






////      if ( object == null ) { return null ; }
////        Utility.JNWBitSet bitSet = ( Utility.JNWBitSet ) super . poll ( ) ;
////System.out.println ("Buffer size " + this .size()  ) ;
////        object = /*( Utility.JNWBitSet ) */ super . poll ( ) ;
////        if ( bitSet == null ) { return null ; }
////System.out.println ("IN JNWBiffer: object is of class" + object.getClass().getName() ) ;
////      if ( object instanceof Utility.JNWBitSet  ) {
////System.out.println ("IN JNWBiffer: A bit set of size " + ( ( Utility.JNWBitSet ) object ) .getJNWLength (  ) ) ;

////       }
////      else {
////System.out.println ("IN JNWBiffer: A String of size " + ( ( String ) object ) . length (  ) ) ;
////         currentBufferSizeInBytes // Update the buffer size
////            = currentBufferSizeInBytes -  ( ( String ) object ) . length (  )   ;
////       }
////      return  ( String ) object ;