/*
 * JNWBuffer.java
 *
 * Created on June 1, 2007, 1:26 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package Utility;

/** Creates a FIFO buffer for use in storing frames, packets, etc. in the
 *interface or elsewhere.  The buffer has a finite size in bytes.  JNWBitSet
 * arriving will cause the buffer to compare the remaining memory to the
 * size of the arriving JNWBitSet.  If the new data cannot fit in the
 * buffer, it is discarded, and the data is lost.  The buffer will
 *accept ONLY String or JNEBitSets; any other class will cause an error message
 *
 *<p>Since the buffer size is saved as an integer, the maximum buffer size is
 *268 magabytes.
 *
 *Version 1.001 adds a test for a null value from the poll with
 * <li>if ( ( object =  super . poll ( )  )  == null ) return object ;</li>
 * *<p><b>Versions</b>
 *<li><b>Version 1.001</b> adds a test for a null value from the poll with
 * <li>if ( ( object =  super . poll ( )  )  == null ) return object ;</li></li>
 *<li><b>Version 1.002</b> of 20071118 adds the boolean addFirst, which if set
 * will add the Object to the head of the buffer rather than the tail.
 * <p>It also contains the method FrameOrNot ( JNWBitSet objectIn ) which
 * tests whether the bit set is an HDLC frame or an Ethernet frame.</li>
 *<li><b>Version 1.100</b> of 20071205 removes the storage capability for
 * Strings.  This capabilty is never used.</li>
 * @author dixonw
 * @version 1.100
 */
public class JNWBufferwithPriority extends JNWBuffer
   implements GUIConstants . General
{
   /**A boolean to tell the add method to add the Object as the
    *head of the list rather than the last element in the list.
    *This is used for moving network control messages to the front
    *of the queue.  This is to be used wheN the IP packet has the
    *Network Control precedence set in the type of service field of
    * the IP header.
    * <p>There could an error introduced here.  The JNWBiffer does
    * not know if it is storing an IPPacket or a frame.  But the three
    * bits are NOT set in the Ethernet frame since the preamble is
    * alternating "01010101" for seven bytes.  It is possible that these
    * three bits will be set in an HDLC frame.  */
   boolean addFirst ;
   
   
   /** Creates a new instance of JNWBuffer with a specified maximum
    * buffer size specified in bytes.*/
   public JNWBufferwithPriority ( int maximumBufferSizeInKilobytes )
   {
      super ( maximumBufferSizeInKilobytes ) ;
   }  // end of public JNWBuffer( int maximumBufferSizeInBits )
   
   
   /**Adds either a JNWBitSet or a String to the buffer is possible,
    *otherwise, the JNWBitSet is discarded.  Bit lengths for a JNWBitSet
    *are rounded up to the next largest byte; that is, 7 bits will
    *occupy one byte.  The buffer will accept ONLY a JNEBitSet;
    *any other class will cause an error message*/
   public boolean add ( JNWBitSet bitSetIn )
   {
      addFirst = false ;
      int length = ( bitSetIn . getJNWLength ( ) + 7 ) / 8 ;
      // Search the object for the presence of the network
      // control bits
      // First test whether this is a frame
      // The logic is kinda weak here.  I know of no way to be 
      // ABSOLUTELY sure that the bit set is a frame or not
      if ( FrameOrNot ( bitSetIn ) == false )
      {  // This is an IP packet
         // The location of the precedence bits in the IP header
         int fromIndex =  bitSetIn . getJNWLength ( ) - 8 - 3 ;
         JNWBitSet precedenceBits
            = bitSetIn . getSubset ( fromIndex , fromIndex + 3  ) ;
         // test the precedence bits of the IP header
         if ( precedenceBits . cardinality ( ) == 3 )
         {// All three bist are set
            addFirst = true ;
         }  // end of if ( precedenceBits . cardinality ( ) == 3 )
      }  // end of if ( FrameOrNot ( ( Utility.JNWBitSet ) object ) == false )
      // is the BitSet too long?
      if ( ( currentBufferSizeInBytes + length  )
      > maximumBufferSizeInBytes )  // Is there an overflow?
      {  // Yes.  Should I print an error message?
         if ( ModelData.PrintPermissions.printBufferOverflows == true )
         {
            Utility.err.println ("ERROR IN Utility.JNWBuffer.add: buffer overflow "
               + "for buffer instantiated in class "
               + this . getClass () . getName ( )  ) ;
         }   // end of if ( ModelDatas.PrintPermissions.printBufferOverflows == true )
         return false ; }  // end of if ( ( currentBufferSizeInBytes + length  )
      if ( addFirst == true )
      { // if addFirst == true add the object to
         super . addFirst ( bitSetIn ) ;  // the head of the buffer
      }
      else
      { // else add it to the end
         super . add ( bitSetIn ) ;
      }  // update the buffer length
      currentBufferSizeInBytes // update the buffer length
         = currentBufferSizeInBytes + length ;
      return true ;
   }  // end of public void add ( Utilities.JNWBitSet bitSet )
    
 
   /**Returns a boolean telling whether the JNWBitSet is a frame (HDLC or Ethernet)
    * or not.  The method tests for the first 8 bytes of an Ethernet frame and the
    first and last bytes of an HDLC frame.*/
   public static boolean FrameOrNot ( JNWBitSet bitSetIn )
   {
      JNWBitSet HDLCHeader = new JNWBitSet ( "01111110" ) ;
      // Does the bit set start and end with an HDLC header
      if ( ( bitSetIn . beginsWith ( HDLCHeader ) & bitSetIn . endsWith ( HDLCHeader ) )
         | bitSetIn . beginsWith ( ModelDataLink.EthernetUtilities.ethernetPreamble )  )
      {
         return true ;
      }
      return false ;
   }  // end of FrameOrNot
}  // end of public class JNWBufferwithPriority

   
   
   /**For testing only.*/
/*   private static class TESTER
   {
      JNWBuffer testbuffer = new JNWBuffer ( 1000 ) ;
      
      public boolean add ( JNWBitSet bitSetIn )
      {
         return testbuffer. add ( bitSetIn ) ;
      }
   }
*/   
   /* For testing only
   public static void main_Testing_Only ( String [ ] args )
   {
      JNWBuffer testbuffer = new JNWBuffer ( 1000 ) ;
      
      JNWBitSet data = new JNWBitSet ( "00000001 11111111 00000000 11111111 01010101") ;
      testbuffer .add ( data ) ;
      
      ////       System.out.println("Data is a frame? = " + FrameOrNot ( data ) );
      System.exit ( 0 ) ;
      
      ModelData.PrintPermissions.printBufferOverflows = true ;
      String string = "0123456789" ;
      JNWBuffer buffer = new JNWBuffer ( 1000 ) ;
      JNWBitSet bitSet = new  JNWBitSet ( 800  ) ;
      TESTER tester = new TESTER ( ) ;
////        bitSet . setJNWBitSetToString ( string + string ) ;
      int i = 0 ;
      
      while ( i < 20)
      {
         tester. add ( bitSet . clone ( )) ;
////        buffer . add ( string ) ;
         System.out.println ("Buffer length after adding bit set = " + tester . testbuffer . getCurrentBufferSizeInBytes ( )
         + " bytes" ) ;
         ////       buffer . add ( bitSet ) ;
////        System.out.println("Buffer after adding BitSet = " + buffer . getCurrentBufferSizeInBytes ( )
////        + " bytes" ) ;
////        bitSet. append ( bitSet ) ;
         i = i + 1 ;
      }
      System.out.println ( "\n\nWRITING BUFFER ") ;
      while ( buffer.size ( ) > 0 )
      {
         Object object = buffer . poll ( ) ;
         System.out.println ("Buffer length = " + buffer . getCurrentBufferSizeInBytes ( )
         + " object = " + object .toString ( ) ) ;
      }
   }  // end of main
*/