package Utility;
/*
 * JNWRandom.java
 *
 * Created on January 6, 2007, 10:02 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

/**JNWRandom extends java.util.Random by adding a method, nextPoisson, to compute a
 *Poisson distributed variable and nextExponential, an exponentially distributed
 *varaible.  The class in implemented as a singleton.  The class is instantiated as:
 *<p><li> Utility.JNWRandom randomNumberGernerator = Utility.JNWRandom .getInstance ( ) ; </li>
 * 
 * @author dixonw
 * @version 1.100
 *@see <a href="http://en.wikipedia.org/wiki/Exponential_distribution">Exponential Distribution</a>
 *@see <a href="http://en.wikipedia.org/wiki/Poisson_distribution">Poisson Distribution</a>
 */
public class JNWRandom extends java.util.Random {

/**This is the seed for the random number generator of class GMURandom
  *It can be changed in the Global Variables GUI.*/
public static long seedForJNWRandomNumberGenerator = 123456789 ;    
    
/**The singleton of a java.util.JNWRandom*/
 private static final JNWRandom randomNumberGenerator 
    = new JNWRandom ( ) ;
    
    /**
     * Creates a new instance of JNWRandom and initializes the random number
     * generator to seedForJNWRandomNumberGenerator.
     */
    private JNWRandom() {
        super ( ) ;
        // Initialize the random number generator to 
        // seedForJNWRandomNumberGenerator
        this . setSeed ( seedForJNWRandomNumberGenerator ) ;
    }
    
    
    /** Returns the single instance of a JNWRandom.*/    
     public static JNWRandom getInstance( )  {
         return randomNumberGenerator ;
     }
     
     
     /** Method to return exponentially distributed number with mean 1 / mu.
     *@param mu the inverse of the mean value of an exponential distribution.
      *@see <a href="http://en.wikipedia.org/wiki/Exponential_distribution">Exponential Distribution</a>*/
    public double nextExponential ( double mu ) {
       return  - java.lang.Math.log ( randomNumberGenerator . nextDouble( ) ) / mu ;
    }  // end of nextExponential
  
    
    /** Method to return a Poisson distributed number with mean lambda.
     *@param lambda the mean value of an Poisson distribution.
     *@see <a href="http://en.wikipedia.org/wiki/Poisson_distribution">Poisson Distribution</a>*/
    public int nextPoisson ( double lambda ) {
       int k = 0 ;
       double p = 1.0 ; 
       double L = java.lang.Math . exp( - lambda ) ; 
       do {
          k = k + 1 ;
          p = p * randomNumberGenerator . nextDouble() ;
       } while ( p >= L ) ;
       return  k  - 1 ;
    }  // end of nextPoisson
   
    
   /**For testing only.*/
    public static void main_Testing_Only ( String[ ] args ) {
        JNWRandom randomNG = JNWRandom.getInstance ( ) ;
        for ( int i = 0 ; i < 10 ; i++ ) {
            System.out.println ("Exponential distribition = " +  randomNG .nextExponential( 1.0 /1500.0 ));
        }  // end of for 
        for ( int i = 0 ; i < 10 ; i++ ) {
            System.out.println ("Poisson distribition = " +  randomNG .nextPoisson( 500.0  ));
        }  // end of for 
    }  // end of main
    
}  // end of class JNWRandom
