package Utility;
/*
 * JNWRandom.java
 *
 * Created on January 6, 2007, 10:02 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

/**JNWRandom_NonSingleton provides a non-singleton version of Utility.JNWRandom for
 * use in computing random numbers for each link.  JNWRandom_NonSingleton extends
 * java.util.Random by adding a method, nextPoisson, to compute a
 *Poisson distributed variable and nextExponential, an exponentially distributed
 *varaible.  This class in <b>NOT</b> implemented as a singleton.  Instantiate it
 * in the normal way.  
 * 
 * @author dixonw
 * @version 1.100
 *@see <a href="http://en.wikipedia.org/wiki/Exponential_distribution">Exponential Distribution</a>
 *@see <a href="http://en.wikipedia.org/wiki/Poisson_distribution">Poisson Distribution</a>
 */
public class JNWRandom_NonSingleton extends java.util.Random {   
    
    /**Creates a new instance of JNWRandom and initializes the random number
     * number generator to seedForJNWRandomNumberGenerator.*/
    public JNWRandom_NonSingleton() {
       super ( ) ;
    }
    
     /** Method to return exponentially distributed number with mean 1 / mu.
     *@param mu the inverse of the mean value of an exponential distribution.
      *@see <a href="http://en.wikipedia.org/wiki/Exponential_distribution">Exponential Distribution</a>*/
    public double nextExponential ( double mu ) {
       return  - java.lang.Math.log ( this . nextDouble( ) ) / mu ;
    }  // end of nextExponential
  
    
    /** Method to return a Poisson distributed number with mean lambda.
     *@param lambda the mean value of an Poisson distribution.
     *@see <a href="http://en.wikipedia.org/wiki/Poisson_distribution">Poisson Distribution</a>*/
    public int nextPoisson ( double lambda ) {
       int k = 0 ;
       double p = 1.0 ; 
       double L = java.lang.Math . exp( - lambda ) ; 
       do {
          k = k + 1 ;
          p = p * this . nextDouble() ;
       } while ( p >= L ) ;
       return  k  - 1 ;
    }  // end of nextPoisson
   
    
   /**For testing only.*/
    public static void main_Testing_Only ( String[ ] args ) {
        JNWRandom_NonSingleton randomNG = new JNWRandom_NonSingleton ( ) ;
        for ( int i = 0 ; i < 10 ; i++ ) {
            System.out.println ("Exponential distribition = " 
               +  randomNG .nextExponential( 1.0 /1500.0 ));
        }  // end of for 
        for ( int i = 0 ; i < 10 ; i++ ) {
            System.out.println ("Poisson distribition = " 
               +  randomNG .nextPoisson( 500.0  ));
        }  // end of for 
    }  // end of main
    
}  // end of class JNWRandom
